#ident "$Id: palo_mem.c 46 2016-12-03 23:17:24Z buildsa $"
/*
 * Copyright (C) 2005 MIPS Technologies, Inc.  All rights reserved.
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */
#ident "$Id: palo_mem.c 46 2016-12-03 23:17:24Z buildsa $"
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/bootmem.h>
#include <linux/pfn.h>
#include <linux/proc_fs.h>

#include <asm/bootinfo.h>
#include <asm/processor.h>
#include <asm/r4kcache.h>
#include <asm/page.h>
#include <asm/sections.h>

#include <palo/include/palomem.h>
#include <palo/fw/mcpu/include/palo_conf.h>
#include <palo/fw/mcpu/include/board_conf.h>
#include <palo/include/misc_mregs.h>
#include "prom.h"
#include "regio.h"

__BUILD_SET_C0(ecc)

#define C0CONFIG_DSP    (1 << 23)           // Data Scratchpad Present
#define C0ECC_SPR       (1 << 28)           // Scratchpad RAM test
#define SPTAG0_E        (1 << 7)            // Scratchpad RAM enable

#define MCPU_DSPRAM_START 0x10000000

static int __init configure_dspram(u_int32_t basepa, u_int32_t *lenp)
{
	u_int32_t c0config, sptag1;
	u_int32_t len;

	c0config = read_c0_config();
	if (!(c0config & C0CONFIG_DSP)) {
		printk("Data SPRAM not found %08x\n", c0config);
		return -1;
	}

	/* Enable SPRAM configuration access */
	set_c0_ecc(C0ECC_SPR);
	ehb();

	/* Enable SPRAM at the supplied address */
	write_c0_dtaglo(MCPU_DSPRAM_START | SPTAG0_E);
	ehb();
	cache_op(Index_Store_Tag_D, 0);
	ehb();

	/* Read the SPRAM length */
	cache_op(Index_Load_Tag_D, 8);
	ehb();
	sptag1 = read_c0_dtaglo();
	ehb();

	/* Disable SPRAM configuration access */
	clear_c0_ecc(C0ECC_SPR);

	len = sptag1 & -4096;

	printk("Data SPRAM: 0x%08x-0x%08x\n", basepa, basepa + len - 1);

	*lenp = len;

	return 0;
}

static char memregions_buf[1024];
static int memregions_buflen;

static void __init add_procfs_region(char *name,
	phys_t start, phys_t size, long type)
{
	char *buf = &memregions_buf[memregions_buflen];
	size_t buflen = sizeof(memregions_buf) - memregions_buflen;
	int r;

	// First time through, print header
	if (memregions_buflen == 0) {
		r = snprintf(buf, buflen, "%-10s %-10s %-10s %-10s\n",
			"NAME", "START", "END", "SIZE");
		if (r < 0 || r > buflen)
			r = buflen;
		memregions_buflen += r;
		buf += r;
		buflen -= r;
	}

	printk("memregion %-10s 0x%08lx 0x%08lx 0x%08lx\n",
		name, (long) start, (long) (start + size), (long) size);
	r = snprintf(buf, buflen, "%-10s 0x%08lx 0x%08lx 0x%08lx\n",
		name, (long) start, (long) (start + size), (long) size);
	if (r < 0 || r > buflen)
		r = buflen;
	memregions_buflen += r;
}

/*
 * Memory regions are installed according to the
 * memory map described in palomem.h.
 */
struct memregion {
	char *name;
	phys_t base;
	phys_t size;
	u32 flags;
#define RF_MEM        0x1 /* add with add_memory_region */
#define RF_PROCFS     0x2 /* include in procfs memregions */
	long type;
};
#define MEMREGION(REG, FLAGS, TYPE) { \
        .name = PALOMEM_## REG ##_NAME, \
        .base = PALOMEM_## REG ##_BASE, \
        .size = PALOMEM_## REG ##_SIZE, \
        .flags = FLAGS, \
        .type = TYPE, \
    }

static void __init add_memregions(struct memregion *mr)
{
	for (; mr && mr->name; mr++) {
		if (mr->flags & RF_MEM)
			add_memory_region(mr->base, mr->size, mr->type);
		if (mr->flags & RF_PROCFS)
			add_procfs_region(mr->name, mr->base, mr->size,
				mr->type);
	}
}

#ifdef CONFIG_BASE_PALO

static struct memregion palo_memregions[] = {
	MEMREGION(ECTX,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(LINUX,   RF_MEM | RF_PROCFS, BOOT_MEM_RAM),
	MEMREGION(MEMTUN,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(HWIO,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(PMPCIE,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(DIAG,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(WQCACHE, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(RQCACHE, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(EXTBL, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU1, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU2, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	{
		.name = NULL,}
};

static void __init init_memregions(void)
{
	add_memregions(palo_memregions);
}
#endif /* CONFIG_BASE_PALO */

#ifdef CONFIG_BASE_SERENO

static struct memregion sereno_memregions[] = {
	MEMREGION(ECTX,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(DIAG,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(MEMTUN,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(LINUX,   RF_MEM | RF_PROCFS, BOOT_MEM_RAM),
	MEMREGION(EXTBL,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU1,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(PMPCIE,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	{ .name = NULL, }
};
static struct memregion sereno_memregions128[] = {
	MEMREGION(HWIO128, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECC128, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	{.name = NULL,}
};
static struct memregion sereno_memregions256[] = {
	MEMREGION(ECPU2, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(HWIO256, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(GENIO, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECC256, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	{.name = NULL,}
};

/*
 * Initialize memory that was mistakenly not initialized by an early
 * development version of u-boot. This is temporary code that can be
 * removed in any post-Capitola release.
 */
static void __init ecc_init_sereno(void)
{
        u_int32_t *start = (u_int32_t *)KSEG1ADDR(0x0dff0000);
        u_int32_t size = 0x0e000000 - 0x0dff0000;
        u_int32_t *end = (u_int32_t *)((u_int8_t *)start + size);

        for ( ; start < end; start++)
                *start = 0;
}

static void __init init_memregions(void)
{
	u_int memsize;

	add_memregions(sereno_memregions);

	memsize = readreg(MISC_MREGS_BE_CFG_TRANSLATE_REGION0_SIZE);
	switch (memsize) {
	case 1:
		add_memregions(sereno_memregions128);
		break;
	case 2:
		add_memregions(sereno_memregions256);
		ecc_init_sereno();
		break;
	case 3:
		printk("only using 256MB of 512MB\n");
		add_memregions(sereno_memregions256);
		break;
	default:
		printk("unknown memory size, using 128MB\n");
		add_memregions(sereno_memregions128);
		break;
	}
}
#endif /* CONFIG_BASE_SERENO */

#ifdef CONFIG_BASE_CRUZ

static struct memregion cruz_memregions[] = {
	MEMREGION(ECTX, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(DIAG, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(MEMTUN, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(LINUX, RF_MEM | RF_PROCFS, BOOT_MEM_RAM),
	MEMREGION(LINUXHI, RF_MEM | RF_PROCFS, BOOT_MEM_RAM),
	MEMREGION(EXTBL,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU1,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU2,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU3,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(ECPU4,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(HWIO,    RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(PMPCIE,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(FWDTAB,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(VIFTBL,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(RDMAFT,  RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(USNICFT, RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	MEMREGION(GENIO,   RF_MEM | RF_PROCFS, BOOT_MEM_RESERVED),
	{ .name = NULL, }
};

static void __init init_memregions(void)
{
	add_memregions(cruz_memregions);
}
#endif /* CONFIG_BASE_CRUZ */

void __init prom_meminit(void)
{
	u_int32_t len;

	init_memregions();
	configure_dspram(MCPU_DSPRAM_START, &len);
}

void __init prom_free_prom_memory(void)
{
	/* Nothing to do! */
}

/* override of arch/mips/mm/cache.c: __uncached_access */
int __uncached_access(struct file *file, unsigned long addr)
{
	if (file->f_flags & O_SYNC)
		return 1;

	// Linux useable ram is cacheable
	if (addr >= PALOMEM_LINUX_BASE &&
		addr < PALOMEM_LINUX_BASE + PALOMEM_LINUX_SIZE)
		return 0;

	// anything else should be uncached
	return 1;
}

#ifdef CONFIG_PROC_FS

static int memregions_read(char *page, char **start, off_t off,
	int count, int *eof, void *data)
{
	int srcleft, copylen = 0;

	*start = page;
	srcleft = memregions_buflen - off;
	if (srcleft > 0) {
		copylen = srcleft < count ? srcleft : count;
		memcpy(page, &memregions_buf[off], copylen);
	}
	*eof = copylen >= srcleft;
	return copylen;
}

static int __init memregions_install_procfs(void)
{
	create_proc_read_entry("memregions", 0, NULL, memregions_read, NULL);
	return 0;
}

static void __exit memregions_remove_procfs(void)
{
	remove_proc_entry("memregions", NULL);
}

module_init(memregions_install_procfs)
module_exit(memregions_remove_procfs)
#endif
