/*
 * Copyright 2006 Nuova Systems, Inc.  All rights reserved.
 * Copyright 2009 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_net.c 46 2016-12-03 23:17:24Z buildsa $"

#include <linux/param.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/addrspace.h>
#include <palo/include/palo_defs.h>
#include <palo/fw/common/include/ecpu_res.h>
#include <palo/include/wq_regs.h>
#include <palo/include/rq_regs.h>
#include <palo/include/cq_regs.h>
#include <palo/include/intr_regs.h>
#include <palo/include/intrc_regs.h>
#include <palo/include/net_regs.h>
#include <palo/include/regs_macros.h>
#include <common/include/hwecpu_addr.h>
#include <drivers/palo/mnic/vnic_resource.h>
#include <drivers/palo/mnic/vnic_enet.h>

#include <palo/fw/mcpu/include/palo_conf.h>
#include <palo/fw/mcpu/include/board_conf.h>

#define uintptr_t u_int32_t

#define NWQ MNIC_NWQ
#define NRQ MNIC_NRQ
#define NCQ MNIC_NCQ
#define NINTR MNIC_NINTR

#define mnic_wq_base(idx)   MNIC_WQ_BASE(palo_conf_wq_count(), idx)
#define mnic_rq_base(idx)   MNIC_RQ_BASE(palo_conf_rq_count(), idx)
#define mnic_cq_base(idx)   MNIC_CQ_BASE(palo_conf_cq_count(), idx)
#define mnic_intr_base(idx) \
    MNIC_INTR_BASE(palo_conf_intr_count() - palo_conf_intr_grp_count(), idx)

enum bar_type {
	BAR_BAR0,
	BAR_WQ,
	BAR_RQ,
	BAR_CQ,
	BAR_INTR,
	BAR_INTR_TABLE,
	BAR_INTR_PBA,
	BAR_DEV,
	BAR_MAX,
};

struct bar0_data {
	struct vnic_resource_header header;
	struct vnic_resource resources[RES_TYPE_MAX];
};

struct resource_info {
	u_int32_t wq_base;
	u_int32_t wq_size;
	u_int32_t rq_base;
	u_int32_t rq_size;
	u_int32_t cq_base;
	u_int32_t cq_size;
	u_int32_t in_base;
	u_int32_t in_size;
	u_int32_t lif_id;
	u_int32_t vnic_id;
};

struct enic_device {
	u_int8_t uif;
	u_int8_t is_loop;
	u_int16_t loop_tag;
	struct bar0_data bar0;
	struct vnic_enet_config enet_config;
	struct resource_info resource_info;
	struct resource resources[BAR_MAX];
	struct platform_device pdev;
};

//
// Template for bar0 data
//
struct bar0_data bar0 = {
	.header =
	{ VNIC_RES_MAGIC, VNIC_RES_VERSION,},
	.resources =
	{
		// .bar_offset = 0 for all resources[]
		{ .type = RES_TYPE_WQ, .bar = BAR_WQ, .count = NWQ,},
		{ .type = RES_TYPE_RQ, .bar = BAR_RQ, .count = NRQ,},
		{ .type = RES_TYPE_CQ, .bar = BAR_CQ, .count = NCQ,},
		{ .type = RES_TYPE_INTR_CTRL, .bar = BAR_INTR, .count = NINTR,},
		{ .type = RES_TYPE_INTR_TABLE, .bar = BAR_INTR_TABLE,
			.count = NINTR,},
		{ .type = RES_TYPE_INTR_PBA, .bar = BAR_INTR_PBA,
			.count = NINTR,},
		{ .type = RES_TYPE_DEV, .bar = BAR_DEV,
			.count = sizeof(struct vnic_enet_config),},
		{ .type = RES_TYPE_EOL,} // EOL must be last
	},
};

//
// Template for Ethernet device-specific region
//
struct vnic_enet_config enet_config = {
	.flags = VENETF_RXCSUM | VENETF_TXCSUM | VENETF_TSO | VENETF_LRO,
	.wq_desc_count = 128,
	.rq_desc_count = 128,
	.mtu = 1500,
	.intr_timer = 83,
	.intr_timer_type = 0,
	.intr_mode = 0,
	.intr_timer_usec = 125,
};

//
// Template for sub-port Ethernet device-specific region
//
struct vnic_enet_config enet_config_subport = {
	.flags = VENETF_RXCSUM | VENETF_TXCSUM | VENETF_TSO | VENETF_LRO,
	.wq_desc_count = 32,
	.rq_desc_count = 32,
	.mtu = 1500,
	.intr_timer = 83,
	.intr_timer_type = 0,
	.intr_mode = 0,
	.intr_timer_usec = 125,
};

static void __init init_enic_bar0(struct enic_device *edev)
{
	edev->bar0 = bar0;

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((ulong) & edev->bar0, sizeof(edev->bar0));
#endif
}

static void __init init_enic_config(struct enic_device *edev)
{
	char devname[16];

	// Some entries have a predefined devname
	memcpy(devname, edev->enet_config.devname, sizeof(devname));
	if (strnlen(devname, sizeof(devname)) == 0) {
		edev->enet_config = enet_config_subport;
	} else {
		edev->enet_config = enet_config;
	}
	memcpy(edev->enet_config.devname, devname,
		sizeof(edev->enet_config.devname));

	if (edev->is_loop) {
		/*
		 * Configure loopback frames.
		 * Turn off rx csum.
		 */
		edev->enet_config.flags |= VENETF_LOOP;
		edev->enet_config.loop_tag = edev->loop_tag;
		edev->enet_config.flags &= ~VENETF_RXCSUM;
	}

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((ulong)&edev->enet_config,
		sizeof(edev->enet_config));
#endif
}

static void __init init_enic_resource_info(struct enic_device *edev)
{
	const int idx = edev->pdev.id;
	struct resource_info *ri = &edev->resource_info;

	ri->wq_base = mnic_wq_base(idx);
	ri->wq_size = NWQ;
	ri->rq_base = mnic_rq_base(idx);
	ri->rq_size = NRQ;
	ri->cq_base = mnic_cq_base(idx);
	ri->cq_size = NCQ;
	ri->in_base = mnic_intr_base(idx);
	ri->in_size = NINTR;

	ri->lif_id = MNIC_LIFID(idx);
	ri->vnic_id = MNIC_VNICID(idx);
}

static void __init init_enic_resources(struct enic_device *edev)
{
	struct resource_info *ri = &edev->resource_info;
	struct resource *res = edev->resources;
	u_int32_t start;

	res = &edev->resources[BAR_BAR0];
	res->start = CPHYSADDR((u_int32_t) & edev->bar0);
	res->end = res->start + sizeof(edev->bar0) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_WQ];
	res->start = WQ_HREGS_ADDR(ri->wq_base);
	res->end = WQ_HREGS_ADDR(ri->wq_base + ri->wq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_RQ];
	res->start = RQ_HREGS_ADDR(ri->rq_base);
	res->end = RQ_HREGS_ADDR(ri->rq_base + ri->rq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_CQ];
	res->start = CQ_HREGS_ADDR(ri->cq_base);
	res->end = CQ_HREGS_ADDR(ri->cq_base + ri->cq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR];
	res->start = INTRCTRL_HREGS_ADDR(ri->in_base);
	res->end = INTRCTRL_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_TABLE];
	res->start = INTRMSI_HREGS_ADDR(ri->in_base);
	res->end = INTRMSI_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_PBA];
	intrpba_enc(&start, ri->in_base, ri->in_size, NOR2C);
	res->start = start;
	res->end = (start + roundup(ri->in_size, 32) / 8) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_DEV];
	res->start = CPHYSADDR((u_int32_t) & edev->enet_config);
	res->end = res->start + sizeof(edev->enet_config) - 1;
	res->flags = IORESOURCE_MEM;
}

static void __init init_enic(struct enic_device *edev)
{
	init_enic_bar0(edev);
	init_enic_config(edev);
	init_enic_resource_info(edev);
	init_enic_resources(edev);

	printk(KERN_INFO "Palo %s%d uif %d\n",
		edev->pdev.name, edev->pdev.id, edev->uif);
}


static struct enic_device enic_devices[] = {
	{
		.uif = MNIC_UIF(0),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 0,
			.resource = enic_devices[0].resources,
			.num_resources = ARRAY_SIZE(enic_devices[0].resources)
		},
	},
	{
		.uif = MNIC_UIF(1),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 1,
			.resource = enic_devices[1].resources,
			.num_resources = ARRAY_SIZE(enic_devices[1].resources)
		},
	},
	{
		.uif = MNIC_UIF(2),
		.is_loop = 1,
		.loop_tag = 0, // init in palo_net_setup
		.pdev =
		{
			.name = "enic",
			.id = 2,
			.resource = enic_devices[2].resources,
			.num_resources = ARRAY_SIZE(enic_devices[2].resources)
		},
	},
#if defined CONFIG_BASE_SERENO || defined CONFIG_BASE_CRUZ
	{
		.uif = MNIC_UIF(3),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 3,
			.resource = enic_devices[3].resources,
			.num_resources = ARRAY_SIZE(enic_devices[3].resources)
		},
		.enet_config =
		{
			.devname = "eth0-%d",
		},
	},
	{
		.uif = MNIC_UIF(4),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 4,
			.resource = enic_devices[4].resources,
			.num_resources = ARRAY_SIZE(enic_devices[4].resources)
		},
		.enet_config =
		{
			.devname = "eth0-%d",
		},
	},
	{
		.uif = MNIC_UIF(5),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 5,
			.resource = enic_devices[5].resources,
			.num_resources = ARRAY_SIZE(enic_devices[5].resources)
		},
		.enet_config =
		{
			.devname = "eth0-%d",
		},
	},
	{
		.uif = MNIC_UIF(6),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 6,
			.resource = enic_devices[6].resources,
			.num_resources = ARRAY_SIZE(enic_devices[6].resources)
		},
		.enet_config =
		{
			.devname = "eth0-%d",
		},
	},
	{
		.uif = MNIC_UIF(7),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 7,
			.resource = enic_devices[7].resources,
			.num_resources = ARRAY_SIZE(enic_devices[7].resources)
		},
		.enet_config =
		{
			.devname = "eth1-%d",
		},
	},
	{
		.uif = MNIC_UIF(8),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 8,
			.resource = enic_devices[8].resources,
			.num_resources = ARRAY_SIZE(enic_devices[8].resources)
		},
		.enet_config =
		{
			.devname = "eth1-%d",
		},
	},
	{
		.uif = MNIC_UIF(9),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 9,
			.resource = enic_devices[9].resources,
			.num_resources = ARRAY_SIZE(enic_devices[9].resources)
		},
		.enet_config =
		{
			.devname = "eth1-%d",
		},
	},
	{
		.uif = MNIC_UIF(10),
		.is_loop = 0,
		.pdev =
		{
			.name = "enic",
			.id = 10,
			.resource = enic_devices[10].resources,
			.num_resources = ARRAY_SIZE(enic_devices[10].resources)
		},
		.enet_config =
		{
			.devname = "eth1-%d",
		},
	},
#endif /* SERENO || CRUZ */
};


/*
 * Array of platform_device objects to add to the platform bus.
 */
static struct platform_device *devices[] = {
	&enic_devices[0].pdev,
	&enic_devices[1].pdev,
	&enic_devices[2].pdev,
	&enic_devices[3].pdev,
#if defined CONFIG_BASE_SERENO || defined CONFIG_BASE_CRUZ
	&enic_devices[4].pdev,
	&enic_devices[5].pdev,
	&enic_devices[6].pdev,
	&enic_devices[7].pdev,
	&enic_devices[8].pdev,
	&enic_devices[9].pdev,
	&enic_devices[10].pdev,
#endif /* SERENO || CRUZ */
};

void __init palo_net_setup(void)
{
	const int nenics = ARRAY_SIZE(enic_devices);
	int i;

	BUG_ON(nenics > ARRAY_SIZE(devices));

	/* initialize all enic devices
	 * For PALO and SERENO - lif[15:8], rq[7:0]
	 * For CRUZ            - lif[14:4], rq[3:0]
	 */
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
	enic_devices[MNIC_BMC].loop_tag = (board_conf_bmc_lifid() << 8);
#elif defined CONFIG_BASE_CRUZ
	enic_devices[MNIC_BMC].loop_tag = (board_conf_bmc_lifid() << 4);
#endif /* PALO || SERENO */
	for (i = 0; i < nenics; i++) {
		init_enic(&enic_devices[i]);
	}

	/* register all platform devices */
	platform_add_devices(devices, nenics);
}
