/*
 * Copyright 2010 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_rq2wq.c 43 2016-12-03 22:59:19Z buildsa $"

#include <linux/param.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/platform_device.h>
#include <asm/io.h>
#include <asm/addrspace.h>
#include <palo/include/palo_defs.h>
#include <palo/fw/common/include/ecpu_res.h>
#include <palo/include/misc_regs_defs.h>
#include <palo/include/wq_regs.h>
#include <palo/include/rq_regs.h>
#include <palo/include/cq_regs.h>
#include <palo/include/intr_regs.h>
#include <palo/include/intrc_regs.h>
#include <palo/include/net_regs.h>
#include <palo/include/regs_macros.h>
#include <common/include/hwecpu_addr.h>
#include <drivers/palo/rq2wq/vnic_resource.h>
#include <drivers/palo/rq2wq/vnic_enet.h>

#include <palo/fw/mcpu/include/palo_conf.h>
#include <palo/fw/mcpu/include/board_conf.h>

#define uintptr_t u_int32_t

#define NWQ BMC_SLOW_NWQ
#define NRQ BMC_SLOW_NRQ
#define NCQ BMC_SLOW_NCQ
#define NINTR BMC_SLOW_NINTR

#define bmc_slow_wq_base()   BMC_SLOW_WQ_BASE(palo_conf_wq_count())
#define bmc_slow_rq_base()   BMC_SLOW_RQ_BASE(palo_conf_rq_count())
#define bmc_slow_cq_base()   BMC_SLOW_CQ_BASE(palo_conf_cq_count())
#define bmc_slow_intr_base() BMC_SLOW_INTR_BASE(palo_conf_intr_count())

enum bar_type {
	BAR_BAR0,
	BAR_WQ,
	BAR_RQ,
	BAR_CQ,
	BAR_INTR,
	BAR_INTR_TABLE,
	BAR_INTR_PBA,
	BAR_DEV,
	BAR_MAX,
};

struct bar0_data {
	struct vnicky_resource_header header;
	struct vnicky_resource resources[RES_TYPE_MAX];
};

struct resource_info {
	u_int32_t wq_base;
	u_int32_t wq_size;
	u_int32_t rq_base;
	u_int32_t rq_size;
	u_int32_t cq_base;
	u_int32_t cq_size;
	u_int32_t in_base;
	u_int32_t in_size;
	u_int32_t lif_id;
};

struct rq2wq_device {
	struct bar0_data bar0;
	struct vnicky_enet_config enet_config;
	struct resource_info resource_info;
	struct resource resources[BAR_MAX];
	struct platform_device pdev;
};

//
// Template for bar0 data
//
static struct bar0_data bar0 = {
	.header =
	{ VNIC_RES_MAGIC, VNIC_RES_VERSION,},
	.resources =
	{
		// .bar_offset = 0 for all resources[]
		{ .type = RES_TYPE_WQ, .bar = BAR_WQ, .count = NWQ,},
		{ .type = RES_TYPE_RQ, .bar = BAR_RQ, .count = NRQ,},
		{ .type = RES_TYPE_CQ, .bar = BAR_CQ, .count = NCQ,},
		{ .type = RES_TYPE_INTR_CTRL, .bar = BAR_INTR, .count = NINTR,},
		{ .type = RES_TYPE_INTR_TABLE, .bar = BAR_INTR_TABLE,
			.count = NINTR,},
		{ .type = RES_TYPE_INTR_PBA, .bar = BAR_INTR_PBA,
			.count = NINTR,},
		{ .type = RES_TYPE_DEV, .bar = BAR_DEV,
			.count = sizeof(struct vnicky_enet_config),},
		{ .type = RES_TYPE_EOL,}, // EOL must be last
	},
};

//
// Template for Ethernet device-specific region
//
static struct vnicky_enet_config rq2wq_config = {
	.flags = VENETF_RXCSUM | VENETF_TXCSUM | VENETF_TSO | VENETF_LRO,
	.wq_desc_count = 512,
	.rq_desc_count = 512,
	.mtu = 1500,
	.intr_timer = 83,
	.intr_timer_type = 0,
	.intr_mode = 0,
	.intr_timer_usec = 125,
};

static void __init init_rq2wq_bar0(struct rq2wq_device *edev)
{
	edev->bar0 = bar0;

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((ulong) & edev->bar0, sizeof(edev->bar0));
#endif
}

static void __init init_rq2wq_config(struct rq2wq_device *edev)
{
	edev->enet_config = rq2wq_config;

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((ulong) & edev->enet_config,
		sizeof(edev->enet_config));
#endif
}

static void __init init_rq2wq_resource_info(struct rq2wq_device *edev)
{
	struct resource_info *ri = &edev->resource_info;

	ri->wq_base = bmc_slow_wq_base();
	ri->wq_size = NWQ;
	ri->rq_base = bmc_slow_rq_base();
	ri->rq_size = NRQ;
	ri->cq_base = bmc_slow_cq_base();
	ri->cq_size = NCQ;
	ri->in_base = bmc_slow_intr_base();
	ri->in_size = NINTR;

	ri->lif_id = board_conf_bmc_lifid();
}

static void __init init_rq2wq_resources(struct rq2wq_device *edev)
{
	struct resource_info *ri = &edev->resource_info;
	struct resource *res = edev->resources;
	u_int32_t start;

	res = &edev->resources[BAR_BAR0];
	res->start = CPHYSADDR((u_int32_t) & edev->bar0);
	res->end = res->start + sizeof(edev->bar0) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_WQ];
	res->start = WQ_HREGS_ADDR(ri->wq_base);
	res->end = WQ_HREGS_ADDR(ri->wq_base + ri->wq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_RQ];
	res->start = RQ_HREGS_ADDR(ri->rq_base);
	res->end = RQ_HREGS_ADDR(ri->rq_base + ri->rq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_CQ];
	res->start = CQ_HREGS_ADDR(ri->cq_base);
	res->end = CQ_HREGS_ADDR(ri->cq_base + ri->cq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR];
	res->start = INTRCTRL_HREGS_ADDR(ri->in_base);
	res->end = INTRCTRL_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_TABLE];
	res->start = INTRMSI_HREGS_ADDR(ri->in_base);
	res->end = INTRMSI_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_PBA];
	intrpba_enc(&start, ri->in_base, ri->in_size, NOR2C);
	res->start = start;
	res->end = (start + roundup(ri->in_size, 32) / 8) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_DEV];
	res->start = CPHYSADDR((u_int32_t) & edev->enet_config);
	res->end = res->start + sizeof(edev->enet_config) - 1;
	res->flags = IORESOURCE_MEM;
}

static void __init init_rq2wq(struct rq2wq_device *edev)
{
	init_rq2wq_bar0(edev);
	init_rq2wq_config(edev);
	init_rq2wq_resource_info(edev);
	init_rq2wq_resources(edev);

	printk(KERN_INFO "Palo %s%d\n", edev->pdev.name, edev->pdev.id);
}

static struct rq2wq_device rq2wq_devices[] = {
	{
		.pdev =
		{
			.name = "rq2wq",
			.id = 0,
			.resource = rq2wq_devices[0].resources,
			.num_resources = ARRAY_SIZE(rq2wq_devices[0].resources)
		},
	}
};

static struct platform_device *devices[] = {
	&rq2wq_devices[0].pdev,
};

void __init palo_rq2wq_setup(void)
{
	init_rq2wq(&rq2wq_devices[0]);
	platform_add_devices(devices, 1);
}
