
/*
 * Copyright 2013 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: aclmgt-rc.c 43 2016-12-03 22:59:19Z buildsa $"

#include <linux/bitmap.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/types.h>

#include <asm/io.h>
#include <asm/mach-palo/rc.h>
#include <palo/include/cruz/rc_mregs.h>
#include "rc.h"

//
// These are like the standard Linux kernel versions of similar name,
// but remove the power-of-two region restriction they impose.
//
#define NUMT            RC_MREGS_DHS_UP_ACL_RAM_MAXCNT
#define ACL_RAM_NWORDS  RC_MREGS_DHS_UP_ACL_RAM_NWORDS

static DECLARE_BITMAP(acl_bitmap, NUMT);
static DEFINE_SPINLOCK(acl_map_lock);

static int bitmap_allocate_region_n(unsigned long *bitmap, int pos, int nbits)
{
	int i;

	for (i = pos; i < pos + nbits; i++) {
		set_bit(i, bitmap);
	}
	return 0;
}

static int bitmap_find_free_region_n(unsigned long *bitmap, int bits, int nbits)
{
	int start = 0;

	while (start + nbits <= bits) {
		int next_set = find_next_bit(bitmap, bits, start);

		if (next_set >= start + nbits) {
			bitmap_allocate_region_n(bitmap, start, nbits);
			return start;
		}

		start = find_next_zero_bit(bitmap, bits, next_set);
	}
	return -ENOMEM;
}

static void bitmap_release_region_n(unsigned long *bitmap, int pos, int nbits)
{
	int i;

	for (i = pos; i < pos + nbits; i++) {
		clear_bit(i, bitmap);
	}
}

static int cam_alloc(void)
{
	int cam_bit;
	cam_bit = bitmap_find_free_region_n(acl_bitmap, NUMT, 1);
	return cam_bit < NUMT ? cam_bit : -1;
}

static void cam_free(int cam_ind)
{
	bitmap_release_region_n(acl_bitmap, cam_ind, 1);
}

#define TCAM_HOST_INDEX_SHIFT	(7)
#define TCAM_ADDR_SPACE_SHIFT	(11)
#define TCAM_UPS_ADDR_SHIFT		(12)
#define VALID_SHIFT             (63)

int rc_acl_update(struct acl_tcam tcam, struct acl_ram ram)
{
	int index  = cam_alloc();
	int *base_addr, *tcam_xaddr, *tcam_yaddr;
	u64 tcam_reg, tcam_x, tcam_y;
	unsigned long flags;

	if (index < 0 || index >= NUMT)
		return -1;

	spin_lock_irqsave(&acl_map_lock, flags);
	// RAM setup
	base_addr = (int *)CKSEG1ADDR(RC_MREGS_DHS_UP_ACL_RAM_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index));
	tcam_xaddr = (int *)CKSEG1ADDR(RC_MREGS_DHS_TCAM_X_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index));
	tcam_yaddr = (int *)CKSEG1ADDR(RC_MREGS_DHS_TCAM_Y_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index));

	rc_mregs_dhs_up_acl_ram_make(base_addr,
			1,
			ram.action,
			ram.vnic_ovr,
			ram.vnic_ovr_en,
			ram.addr_ovr,
			ram.addr_ovr_en,
			ram.addr_space,
			ram.addr_size,
			ram.ecc);

	// TCAM setup
	tcam_reg = (tcam.vnic_offset |
			   (tcam.hostid << TCAM_HOST_INDEX_SHIFT) |
			   (tcam.addr_space << TCAM_ADDR_SPACE_SHIFT) |
			   ((tcam.upstream_addr >> 2) << TCAM_UPS_ADDR_SHIFT) |
			   ((u64)1 << VALID_SHIFT));

	tcam_x = tcam_reg & (~(tcam.vnic_offset |
			(tcam.hostid << TCAM_HOST_INDEX_SHIFT) |
			(tcam.addr_space << TCAM_ADDR_SPACE_SHIFT)));

	tcam_y = ~tcam_reg & (~(tcam.vnic_offset |
			(tcam.hostid << TCAM_HOST_INDEX_SHIFT) |
			(tcam.addr_space << TCAM_ADDR_SPACE_SHIFT)));

	rc_mregs_dhs_tcam_x_make(tcam_xaddr, tcam_x, 1);
	rc_mregs_dhs_tcam_y_make(tcam_yaddr, tcam_y, 1);

	spin_unlock_irqrestore(&acl_map_lock, flags);
	return index;
}
EXPORT_SYMBOL(rc_acl_update);

void rc_acl_clear(int index)
{
	int base_addr, tcam_xaddr, tcam_yaddr;
	unsigned long flags;

	if (index >= NUMT)
		return;
	base_addr = (int)RC_MREGS_DHS_UP_ACL_RAM_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index);
	tcam_xaddr = (int)RC_MREGS_DHS_TCAM_X_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index);
	tcam_yaddr = (int)RC_MREGS_DHS_TCAM_Y_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index);
	spin_lock_irqsave(&acl_map_lock, flags);
	writeregs(tcam_xaddr, 0, ACL_RAM_NWORDS);
	writeregs(tcam_yaddr, 0, ACL_RAM_NWORDS);
	writeregs(base_addr, 0, ACL_RAM_NWORDS);
	cam_free(index);
	spin_unlock_irqrestore(&acl_map_lock, flags);
}
EXPORT_SYMBOL(rc_acl_clear);

char *decode_size[32] = {
	"Invalid",
	"Invalid",
	"Invalid",
	"Invalid",
	"Invalid",
	"32B",
	"64B",
	"128B",
	"256B",
	"512B",
	"1KB",
	"2KB",
	"4KB",
	"8KB",
	"16KB",
	"32KB",
	"64KB",
	"128KB",
	"256KB",
	"512KB",
	"1MB",
	"2MB",
	"4MB",
	"8MB",
	"16MB",
	"32MB",
	"64MB",
	"128MB",
	"256MB",
	"512MB",
	"1GB",
	"Invalid",
};

static int parse_tcam(int index, int *base_addr,
		int *tcam_xaddr, int *tcam_yaddr, char *buf)
{
	char *buf_ret = buf;
	u8 action, vnic_ovr, vnic_ovr_en, addr_ovr_en;
	u8 addr_space, addr_size, ecc;
	u8 valid;
	u64 addr_ovr;

	buf_ret += sprintf(buf_ret, "Entry: %03d\n", index);
	buf_ret += sprintf(buf_ret, "TCAM X:\n");
	buf_ret += sprintf(buf_ret,
			"vnic_off hid addrspace upstream_addr valid\n");
	buf_ret += sprintf(buf_ret, "%08d %03d %9s 0x%16llx %d\n",
		tcam_xaddr[0] & 0x7F,
		(tcam_xaddr[0] >> TCAM_HOST_INDEX_SHIFT) & 0xF,
		((tcam_xaddr[0] >> TCAM_ADDR_SPACE_SHIFT) & 1) ?
		"local" : "host",
		((u64)((u64)(tcam_xaddr[1] )  << 20) |
			tcam_xaddr[0] >> TCAM_UPS_ADDR_SHIFT), ((tcam_xaddr[1] >> 31) & 0x1) );

	buf_ret += sprintf(buf_ret, "TCAM Y:\n");
	buf_ret += sprintf(buf_ret,
			"vnic_off hid addrspace upstream_addr valid \n");
	buf_ret += sprintf(buf_ret, "%08d %03d %9s 0x%16llx %d\n",
		tcam_yaddr[0] & 0x7F,
		(tcam_yaddr[0] >> TCAM_HOST_INDEX_SHIFT) & 0xF,
		((tcam_yaddr[0] >> TCAM_ADDR_SPACE_SHIFT) & 1) ?
		"local" : "host",
		((u64)((u64)(tcam_yaddr[1])  << 20) |
			tcam_yaddr[0] >> TCAM_UPS_ADDR_SHIFT),((tcam_yaddr[1] >> 31)&0x1));

	rc_mregs_dhs_up_acl_ram_get(base_addr,
			&valid,
			&action,
			&vnic_ovr,
			&vnic_ovr_en,
			&addr_ovr,
			&addr_ovr_en,
			&addr_space,
			&addr_size,
			&ecc);

	buf_ret += sprintf(buf_ret, "ACL_RAM:\n");
	buf_ret += sprintf(buf_ret,
		"valid act vnc_ovr vnc_ovr_en addrspace addr_size"
		" ecc addr_ovr_en addr_ovr\n");
	buf_ret += sprintf(buf_ret,
		"%05d %03d %07d %10s %9s %9s %03d %11s 0x%016llx\n",
		valid, action, vnic_ovr,
		vnic_ovr_en ? "enable" : "disable",
		addr_space ? "local" : "host",
		decode_size[addr_size],
		ecc,
		addr_ovr_en ? "enable" : "disable",
		addr_ovr);

	return buf_ret - buf;
}

int acl_info_show(char *buf)
{
	// Display the TCAM and RAM contents
	int *base_addr, *tcam_xaddr, *tcam_yaddr;
	char *buf_ret = buf;
	int index;

	for (index = 0; index < NUMT; index++) {
		base_addr = (int *)CKSEG1ADDR(
				RC_MREGS_DHS_UP_ACL_RAM_ADDR_COMPUTE(
					RC_MREGS_BASE_ADDR, index));
		tcam_xaddr = (int *)CKSEG1ADDR(RC_MREGS_DHS_TCAM_X_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index));
		tcam_yaddr = (int *)CKSEG1ADDR(RC_MREGS_DHS_TCAM_Y_ADDR_COMPUTE(
				RC_MREGS_BASE_ADDR, index));
		if ((base_addr[0] & 1)) {
			buf_ret += sprintf(buf_ret, "--------\n");
			buf_ret += parse_tcam(index, base_addr,
					tcam_xaddr, tcam_yaddr, buf_ret);
		}
	}
	return buf_ret - buf;
}


static int __init acl_init(void)
{
	spin_lock_init(&acl_map_lock);
	writereg(RC_MREGS_CFG_TCAM_GRST, 1); // Reset TCAM
	writereg(RC_MREGS_CFG_TCAM_GRST, 0); // Unreset TCAM
	return 0;
}
arch_initcall(acl_init);
