/*
 * Copyright 2013 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: evt-rc.c 43 2016-12-03 22:59:19Z buildsa $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/pci.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <asm/mach-palo/rc.h>

#include "rc.h"
#include <palo/fw/mcpu/include/rc_dev.h>

static void rc_add_event(struct rc_event *ev);

int rc_have_events(void)
{
	struct rc_info *rcdev = &rc_dev_info;
	int have_events = 0;
	unsigned long flags;

	spin_lock_irqsave(&rcdev->events_lock, flags);
	have_events = !TAILQ_EMPTY(&rcdev->events);
	spin_unlock_irqrestore(&rcdev->events_lock, flags);
	return have_events;
}

static int rc_evt_limit = 1000;

void rc_evt_ind(int bdf, char cause)
{
	struct rc_event *ev;

	ev = kzalloc(sizeof(struct rc_event), GFP_ATOMIC);
	if (!ev) {
		printk("No free memory for event:%x %02x:%02x.%02x\n",
			cause, (bdf >> 8) & 0xFF, (bdf >> 3) & 0x1F, bdf & 0x7);
		return;
	}
	ev->type = RC_EVT_IND;
	ev->ind_cause = cause;
	ev->bdf = bdf;

	rc_add_event(ev);
}

void rc_evt_ecc(int ecc_cause, int ecc_status)
{
	struct rc_event *ev;

	ev = kzalloc(sizeof(struct rc_event), GFP_ATOMIC);
	if (!ev) {
		printk("No free memory for ecc event allocation\n");
		return;
	}
	ev->type = RC_EVT_ECC;
	ev->ecc_cause = ecc_cause;
	ev->ecc_status = ecc_status;

	rc_add_event(ev);
}

static void rc_add_event(struct rc_event *ev)
{
	struct rc_info *rcdev = &rc_dev_info;
	unsigned long flags;
	static int warned;

	if (!rc_evt_limit || atomic_read(&rcdev->nevents) < rc_evt_limit) {
		spin_lock_irqsave(&rcdev->events_lock, flags);
		TAILQ_INSERT_TAIL(&rcdev->events, ev, list);
		atomic_inc(&rcdev->nevents);
		spin_unlock_irqrestore(&rcdev->events_lock, flags);
		wake_up_interruptible(&rcdev->waitq);
	} else {
		if (!warned) {
			printk(KERN_WARNING PFX
			       "add_event limit %d exceeded\n",
			       rc_evt_limit);
			warned = 1;
		}
		kfree(ev);
	}
}


static struct rc_event *rc_get_event(void)
{
	struct rc_info *rcdev = &rc_dev_info;
	struct rc_event *ev = NULL;
	unsigned long flags;

	spin_lock_irqsave(&rcdev->events_lock, flags);
	if (!TAILQ_EMPTY(&rcdev->events)) {
		ev = TAILQ_FIRST(&rcdev->events);
		TAILQ_REMOVE(&rcdev->events, ev, list);
		atomic_dec(&rcdev->nevents);
	}
	spin_unlock_irqrestore(&rcdev->events_lock, flags);
	return ev;
}

int rc_copy_events_to_user(struct rc_event __user *uev)
{

	int r, nev;

	if (!access_ok(VERIFY_WRITE, uev,
		       sizeof (struct rc_event) * RC_GET_EVENTS_MAX))
		return -EFAULT;

	for (nev = 0; nev < RC_GET_EVENTS_MAX; nev++, uev++) {
		struct rc_event *ev = rc_get_event();
		if (!ev)
			break;
		r = copy_to_user(uev, ev, sizeof(struct rc_event));
		BUG_ON(r);
		kfree(ev);
	}
	return nev;
}

void __devinit rc_evt_init(void)
{
	struct rc_info *rcdev = &rc_dev_info;

	TAILQ_INIT(&rcdev->events);
	spin_lock_init(&rcdev->events_lock);
	init_waitqueue_head(&rcdev->waitq);
}
