/*
 * Copyright 2012 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: msi-rc.c 43 2016-12-03 22:59:19Z buildsa $"
#include <linux/msi.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <asm/mach-palo/paloint.h>

#include <palo/include/cruz/intrc_mregs.h>
#include <palo/include/intrc_regs.h>
#include "ecpu_res.h"
#include "rc.h"

#define INTRC_STATUS(lvl)       CKSEG1ADDR(INTRC_MREGS_STATUS(0, lvl))
#define INTRC_MASKSET(lvl)      CKSEG1ADDR(INTRC_MREGS_MASK_SET(0, lvl))
#define INTRC_MASKCLEAR(lvl)    CKSEG1ADDR(INTRC_MREGS_MASK_CLEAR(0, lvl))

static DEFINE_SPINLOCK(rc_msi_lock);
#define MAX_MSI_IRQ RC_DEVICE_INT_SRC_COUNT

static void rc_msix_enable_irq(struct irq_data *d)
{
	int lvl = 0, src = 0;
	int irq = d->irq - MIPS_CPU_IRQ_BASE;

	palo_int_from_irq(irq, &lvl, &src);
	writel(1 << src, (uint32_t *)INTRC_MASKCLEAR(lvl));
	(void) readl((void *)INTRC_MASKCLEAR(lvl));
	unmask_msi_irq(d);
}

static void rc_msix_disable_irq(struct irq_data *d)
{
	int lvl = 0, src = 0;
	int irq = d->irq - MIPS_CPU_IRQ_BASE;

	palo_int_from_irq(irq, &lvl, &src);
	writel(1 << src, (uint32_t *)INTRC_MASKSET(lvl));
	(void) readl((void *)INTRC_MASKSET(lvl));
	mask_msi_irq(d);
}

static void rc_msix_ack_irq(struct irq_data *d)
{
	int lvl = 0, src = 0;
	int irq = d->irq - MIPS_CPU_IRQ_BASE;

	palo_int_from_irq(irq, &lvl, &src);
	writel(1 << src, (uint32_t *)INTRC_MASKSET(lvl));
	(void) readl((void *)INTRC_MASKSET(lvl));

	writel(1 << src, (uint32_t *)INTRC_STATUS(lvl));
	(void) readl((void *)INTRC_STATUS(lvl));
}

static void rc_msix_end_irq(struct irq_data *d)
{
	int lvl = 0, src = 0;
	int irq = d->irq - MIPS_CPU_IRQ_BASE;

	if (!(d->state_use_accessors &
                (IRQD_IRQ_INPROGRESS | IRQD_IRQ_DISABLED))) {
		palo_int_from_irq(irq, &lvl, &src);
	 	writel(1 << src, (uint32_t *)INTRC_MASKCLEAR(lvl));
		(void) readl((void *)INTRC_MASKCLEAR(lvl));
	}
}



static struct irq_chip msix_chip = {
	.irq_unmask  = rc_msix_enable_irq,
	.irq_mask    = rc_msix_disable_irq,
	.irq_enable  = rc_msix_enable_irq,
	.irq_disable = rc_msix_disable_irq,
	.irq_ack     = rc_msix_ack_irq,
	.irq_eoi     = rc_msix_end_irq,
	.name        = "rc-msi",
};


static char irq_free[MAX_MSI_IRQ];

static int rc_irq(int *index)
{
	/* Allocate an unused irq */
	int irq = -1, i;

	spin_lock(&rc_msi_lock);
	for (i = 0; i < MAX_MSI_IRQ; i++) {
		if (!irq_free[i]) {
			irq_free[i] = 1;
			break;
		}
	}
	if (i < MAX_MSI_IRQ) {
		irq = palo_int_to_irq(RC_DEVICE_INT_LVL,
            RC_DEVICE_INT_SRC_BASE + i);
		*index = i;
	}
	spin_unlock(&rc_msi_lock);
	return irq;
}

static int setup_msi_msg_address(struct pci_dev *dev,
		struct msi_msg *msg, int index, int msi_index)
{
	if (dev->msi_address[msi_index] != 0) {
		msg->address_lo = dev->msi_address[msi_index];
	} else {
		msg->address_lo =
		INTRC_MREGS_SET_INT_LEVEL3_CPU0 + (index * sizeof (int));
	}
	msg->address_hi = 0x00000000;
	msg->data = 0x1;
	return 0;
}

int rc_setup_msi_irq(struct pci_dev *dev, struct msi_desc *desc, int type,
		int msi_index)
{
	struct msi_msg msg;
	int irq = 0, ret, irq_index = 0;

	irq = rc_irq(&irq_index);
	if (irq < 0)
		return -1;

	setup_msi_msg_address(dev, &msg, irq_index, msi_index);
	ret = irq_set_msi_desc(irq, desc);
	if (ret < 0)
		printk("irq_set_msi_desc failed\n");
	write_msi_msg(irq, &msg);
	irq_set_chip_and_handler_name(irq, &msix_chip,
			handle_level_irq,
			(type == PCI_CAP_ID_MSI) ? "MSI" : "MSIX");
	return 0;
}

int rc_setup_msi_irqs(struct pci_dev *dev, int nvec, int type)
{
	struct msi_desc *entry;
	int ret, index = 0;

	if (type == PCI_CAP_ID_MSI && nvec > 1)
		return 1;

	list_for_each_entry(entry, &dev->msi_list, list) {
		ret = rc_setup_msi_irq(dev, entry, type, index);
		if (ret < 0)
			return ret;
		if (ret > 0)
			return -ENOSPC;
		index++;
	}
	return 0;
}

void rc_teardown_msi_irq(unsigned int irq)
{
	int lvl, src;

	palo_int_from_irq(irq, &lvl, &src);
	irq_free[src] = 0;
}

void rc_teardown_msi_irqs(struct pci_dev *dev)
{
	struct msi_desc *entry;

	list_for_each_entry(entry, &dev->msi_list, list) {
		int i, nvec;
		if (entry->irq == 0)
			continue;
		nvec = 1 << entry->msi_attrib.multiple;
		for (i = 0; i < nvec; i++) {
			rc_teardown_msi_irq(entry->irq + i);
			dynamic_irq_cleanup(entry->irq + i);
		}
	}
}
