/*
 * Copyright 2012 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: remap-rc.c 43 2016-12-03 22:59:19Z buildsa $"

#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>

#include <asm/io.h>
#include <palo/include/pcie_mregs.h>
#include <palo/include/cruz/rc_mregs.h>
#include <palo/include/cruz/proc_mregs.h>

#include "rc.h"

#define WINDOW_COUNT	4
#define WINDOW_SIZE		0x400000
#define RC_BASE_0		0x40000000
#define RC_BASE_2		0x60000000
#define DNS_TBL_OFFSET  12

typedef struct {
	int base_address;
	char valid;
	int start_offset;
	char ref_count;
} remap_window_t;

typedef struct {
	int window_base;
	int remap_base;
} window_reg_t;

static DEFINE_SPINLOCK(remap_lock);
static remap_window_t rwindow[WINDOW_COUNT];

static window_reg_t  remap_reg[WINDOW_COUNT] = {
	{ PROC_MREGS_BIG_WINDOW_BASE1, PROC_MREGS_BIG_REMAP_WINDOW1 },
	{ PROC_MREGS_BIG_WINDOW_BASE2, PROC_MREGS_BIG_REMAP_WINDOW2 },
	{ PROC_MREGS_BIG_WINDOW_BASE3, PROC_MREGS_BIG_REMAP_WINDOW3 },
	{ PROC_MREGS_BIG_WINDOW_BASE0, PROC_MREGS_BIG_REMAP_WINDOW0 - 0x3000 },
};

// Ensure that nobody other than PCI drivers will call this.
// Since we are allocating 128M, it may crash the system.

static char offset_in_range(phys_t offset, char *port)
{
	if (offset >= PORT_0_MEM_START && offset < PORT_0_MEM_END) {
		*port = 0;
		return 0;
	} else if (offset >= PORT_2_MEM_START && offset < PORT_2_MEM_END) {
		*port = 2;
		return 0;
	}
	return -1;
}

static char usable_window(phys_t offset, unsigned long size, int *address,
		int dev_index, int bar_index)
{
	int i;
	dev_mem_t *devmem;
	if ((dev_index == -1) || (bar_index == -1)) {
		return -1;
	}
	devmem = &rc_vnic[dev_index].mem_mapped[bar_index];

	for (i = 0; i < WINDOW_COUNT; i++) {
		if (rwindow[i].valid && devmem->valid) {
			if (((rwindow[i].base_address + WINDOW_SIZE) >=
				(offset + size)) &&
				(rwindow[i].base_address <= offset)) {
				spin_lock(&remap_lock);
				*address = devmem->remap_base +
					offset - devmem->dev_base;
				rwindow[i].ref_count++;
				spin_unlock(&remap_lock);
				return 0;
			}
		}
	}
	return -1;
}

static void program_window(int base_address, int offset, int window_id, int
        *address, char port, int dev_index, int bar_index)
{
	int base_next = 0x04090001;
	int dest_address;
	dev_mem_t *devmem = &rc_vnic[dev_index].mem_mapped[bar_index];

	if (port)
		dest_address = RC_BASE_2 + base_address;
	else
		dest_address = RC_BASE_0 + base_address;

	writereg(remap_reg[window_id].window_base, dest_address);
	readreg(remap_reg[window_id].window_base);
	writereg(remap_reg[window_id].window_base + 4, base_next);
	readreg(remap_reg[window_id].window_base + 4);
	writereg(PROC_MREGS_WINDOW_ATTRIBUTE_CACHED, 0x2);
	readreg(PROC_MREGS_WINDOW_ATTRIBUTE_CACHED);

	*address = remap_reg[window_id].remap_base + offset;
	devmem->window = window_id;
	devmem->dev_base = base_address + offset;
	devmem->remap_base = remap_reg[window_id].remap_base + offset;
	devmem->valid = 1;
	devmem->vdev_base = (int)CKSEG1ADDR(devmem->dev_base);
}

static char allocate_window(phys_t offset, unsigned long size, int *address,
        char port, int dev_index, int bar_index)
{
	int i = 0;
	int found = 0;

	if ((dev_index == -1) || (bar_index == -1)) {
		return -1;
	}
	for (i = 0; i < WINDOW_COUNT; i++) {
		if (rwindow[i].valid)
			continue;
		else {
			found = 1;
			break;
		}
	}
	if (found) {
		spin_lock(&remap_lock);
		rwindow[i].start_offset = offset % WINDOW_SIZE;
		rwindow[i].base_address =  offset -  rwindow[i].start_offset;
		program_window(rwindow[i].base_address,
				rwindow[i].start_offset, i,
				address, port, dev_index,
				bar_index);
		rwindow[i].valid = 1;
		rwindow[i].ref_count++;
		spin_unlock(&remap_lock);
		return 0;
	} else {
		return -1;
	}
}

static int dev_index_by_offset(phys_t offset, unsigned long size,
		int *dev_index, int *bar_index)
{
	unsigned char index, i, found = 0;
	struct pci_dev *dev;
	struct resource *res;

	for (index = 0; index < MAX_VNIC_COUNT; index++) {
		if (rc_vnic[index].valid) {
			dev = rc_vnic[index].pdev;
			for (i = 0; i < BAR_COUNT; i++) {
				res = &dev->resource[i];
				if ((res->start <= offset) &&
					((res->end + 1) >= (offset + size))) {
					found = 1;
					goto down;
				}
			}
		}
	}
	return -1;
down:
	printk(KERN_DEBUG "found entry: %x at bar: %x\n", index, i);
	*dev_index = index;
	*bar_index = i;
	return 0;
}

int remap_address(phys_t offset, unsigned long size, int *address)
{
	unsigned char port = 0;
	int dev_index = -1, bar_index = -1;

	if (!offset_in_range(offset, &port)) {
		dev_index_by_offset(offset, size, &dev_index, &bar_index);
		if (usable_window(offset, size, address,
                    dev_index, bar_index)) {
			if (!allocate_window(offset, size,
					address, port, dev_index,
					bar_index))
				return 0;
			return 1;
		} else {
			return 0;
		}
	}
	return -1;
}
EXPORT_SYMBOL(remap_address);


void unmap_address(struct pci_dev *pdev, int address)
{
// Need to free the mem regions appropriately
// Currently the addresses in the window are not monitored. Once they are
// provided, we don't handle the unloading of the driver. This needs to be
// changed.
	int busno = pdev->bus->number;
	int dev = PCI_SLOT(pdev->devfn);
	int func = PCI_FUNC(pdev->devfn);
	int bdf = make_bdf(busno, dev, func);
	int i, j, window_id = 0, found = 0;
	dev_mem_t *devmem;
	for (i = 0; i < key_index; i++) {
		if (rc_vnic[i].valid && (rc_vnic[i].bdf == bdf)) {
			for (j = 0; j < BAR_COUNT; j++) {
				devmem = &rc_vnic[i].mem_mapped[j];
				if (devmem->valid) {
					if ((address >= devmem->vdev_base) &&
							address < WINDOW_SIZE) {
						window_id = devmem->window;
						found = 1;
						goto down;
					}
				}
			}
		}
	}
down:
	if (found) {
		// Assume that unmap address is called
		// only when the device
		// is not going to be used.
		spin_lock(&remap_lock);
		rwindow[window_id].ref_count--;
		if (rwindow[window_id].ref_count == 0) {
			memset(&rwindow[window_id],
					0,
					sizeof(remap_window_t));
			memset(&rc_vnic[i].mem_mapped[0],
					0,
					sizeof(dev_mem_t) * BAR_COUNT);
		}
		spin_unlock(&remap_lock);
	}
}
EXPORT_SYMBOL(unmap_address);

void *map_cfg(struct pci_dev *pdev)
{
	int busno = pdev->bus->number;
	int dev = PCI_SLOT(pdev->devfn);
	int func = PCI_FUNC(pdev->devfn);
	int bdf = make_bdf(busno, dev, func);
	int i, remap_window[4], found = 0;

	for (i = 0; i < key_index; i++) {
		if (bdf == rc_vnic[i].bdf) {
			found = 1;
			break;
		}
	}
	if (found) {
		remap_window[0] = 0x40000000 | (i << DNS_TBL_OFFSET);
		remap_window[1] = 0x00090009;
		remap_window[2] = 0x4;
		remap_window[3] = 0x3;

		writereg(PROC_MREGS_WINDOW_BASE, remap_window[0]);
		writereg(PROC_MREGS_WINDOW_BASE + 4, remap_window[1]);
		writereg(PROC_MREGS_WINDOW_BASE + 8, remap_window[2]);
		writereg(PROC_MREGS_WINDOW_BASE + 18, remap_window[3]);
		return (void *)CKSEG1ADDR(PROC_MREGS_REMAP_WINDOW);
	}
	return NULL;
}

static int __init remap_init(void)
{
	spin_lock_init(&remap_lock);
	return 0;
}
arch_initcall(remap_init);
