/*
 * Copyright (c) 2007 Nuova Systems, Inc.  All rights reserved.
 */

#ifndef _KCOMPAT_MIPS_H_
#define _KCOMPAT_MIPS_H_

/*
 * MIPS Linux doesn't include PCI, so this provides a fake PCI API
 * that lets the driver mainline continue to use the PCI API.  Rather
 * than a PCI device, the driver instantiates a platform device that
 * pretends to be a PCI device.
 */

#include <linux/platform_device.h>
#include <paloint.h>

#include <palo/fw/common/include/ecpu_res.h>

#define CONFIG_MNIC         1
#define VNIC_PADDR_TARGET	0x0001000000000000ULL

#undef readq
#define readq kcompat_readq
static inline u64 readq(void __iomem *addr)
{
	return ioread32(addr) | (((u64)ioread32(addr + 4)) << 32);
}

#undef writeq 
#define writeq kcompat_writeq
static inline void writeq(u64 val, void __iomem *addr)
{
	iowrite32((u32)val, addr);
	iowrite32((u32)(val >> 32), addr + 4);
}

struct mnic_msix_entry {
	u16 vector;	/* kernel uses to write allocated vector */
	u16 entry;	/* driver uses to specify entry, OS writes */
};

#undef PCI_ROM_RESOURCE
#define PCI_ROM_RESOURCE	16

#define pci_enable_device kcompat_pci_enable_device
static inline int pci_enable_device(struct pci_dev *dev) {return 0;}

#define pci_set_master kcompat_pci_set_master
static inline void pci_set_master(struct pci_dev *dev) {}

#define pci_set_dma_mask kcompat_pci_set_dma_mask
static inline int pci_set_dma_mask(struct pci_dev *dev, u64 mask) {return 0;}

#define pci_find_capability kcompat_pci_find_capability
static inline int pci_find_capability(struct pci_dev *dev, int cap)
	{return cap == PCI_CAP_ID_MSI;}

#define pci_request_regions kcompat_pci_request_regions
static inline int pci_request_regions(struct pci_dev *pdev,
	const char *res_name) {return 0;}

#define pci_release_regions kcompat_pci_release_regions
static inline void pci_release_regions(struct pci_dev *pdev) {}

#define pci_set_consistent_dma_mask kcompat_pci_set_consistent_dma_mask
static inline int kcompat_pci_set_consistent_dma_mask(struct pci_dev *dev,
	u64 mask) {return 0;}

static inline int mnic_pci_enable_device_mem(struct pci_dev* dev) {return 0;}

static inline int mnic_pci_enable_msix(struct pci_dev* dev,
	struct mnic_msix_entry *entries, int nvec) {return -1;}

static inline void mnic_pci_disable_msix(struct pci_dev* dev) {}

static inline int mnic_pci_enable_msi(struct pci_dev *dev) {return 0;}

static inline void mnic_pci_disable_msi(struct pci_dev *dev) {}

#define pci_alloc_consistent kcompat_pci_alloc_consistent
static inline void *kcompat_pci_alloc_consistent(struct pci_dev *pdev,
	size_t size, dma_addr_t *dma_handle)
{
    return dma_alloc_coherent(&pdev->dev, size,
            dma_handle, GFP_KERNEL);
}

#define pci_free_consistent kcompat_pci_free_consistent
static inline void kcompat_pci_free_consistent(struct pci_dev *pdev,
	size_t size, void *cpu_addr, dma_addr_t dma_addr)
{
	dma_free_coherent(&pdev->dev, size, cpu_addr, dma_addr);
}

#define pci_iomap kcompat_pci_iomap
static inline void __iomem *pci_iomap(struct pci_dev *dev, int bar,
	unsigned long maxlen)
{
	unsigned long start = pci_resource_start(dev, bar);
	unsigned long len = pci_resource_len(dev, bar);
	unsigned long flags = pci_resource_flags(dev, bar);

	if (!len || !start)
		return NULL;
	if (maxlen && len > maxlen)
		len = maxlen;
	if (flags & IORESOURCE_IO)
		return NULL; /* Sorry, no support for IO */
	if (flags & IORESOURCE_MEM) {
		if (flags & IORESOURCE_CACHEABLE)
			return ioremap(start, len);
		return ioremap_nocache(start, len);
	}
	/* What? */
	return NULL;
}

static int kcompat_probe_device(struct device *dev)
{
	struct platform_device *pd = to_platform_device(dev);
	struct pci_driver *pci_drv = to_pci_driver(dev->driver);
	struct pci_dev *pdev = kzalloc(sizeof(struct pci_dev), GFP_KERNEL);
	int int_lvl, int_src;

	if (pd->flags & PLATFORM_DEV_SHARE_INTR) {
		int_lvl = PALO_ENIC_INT_LVL_SHARED;
		int_src = PALO_ENIC_INT_SRC_SHARED;
	} else {
		int_lvl = PALO_ENIC_INT_LVL(pd->id);
		int_src = PALO_ENIC_INT_SRC(pd->id);
	}

	if (!pdev)
		return -ENOMEM;

	dev->platform_data = pdev;
	pdev->dev = pd->dev;
	printk(KERN_INFO "enic idx = %d, int_lvl = %d, int_src = %d\n",
			pd->id, int_lvl, int_src);
	pdev->irq = palo_int_to_irq(int_lvl, int_src);

	memcpy(pdev->resource, pd->resource,
			pd->num_resources * sizeof (struct resource));

	return pci_drv->probe(pdev, NULL);
}

static int kcompat_remove_device(struct device *dev)
{
	struct pci_driver *pci_drv = to_pci_driver(dev->driver);
    struct pci_dev *pdev = dev->platform_data;

	pci_drv->remove(pdev);

	kfree(pdev);
	
	return 0;
}

#define mnic_pci_register_driver kcompat_pci_register_driver
static inline int mnic_pci_register_driver(struct pci_driver *pci_drv)
{
	struct device_driver *driver = &pci_drv->driver;

	driver->name = pci_drv->name;
	driver->bus = &platform_bus_type;
	driver->owner = THIS_MODULE;
	driver->probe = kcompat_probe_device;
	driver->remove = kcompat_remove_device;

	return driver_register(driver);
}

#endif /* _KCOMPAT_MIPS_H_ */
