
/*
 * Copyright 2013 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: ejtag.c 59 2017-01-21 21:37:10Z buildsa $"

#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/mipsregs.h>
#include <asm/addrspace.h>
#include <asm/cacheops.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <palo/fw/mcpu/include/ecpuctx.h>
#include <palo/fw/mcpu/include/paloecpu.h>
#include "palo_ecpu.h"

/*
 * GO - Continue executing the target.
 */
static u_int32_t
cmd_go(paloecpu_dbpage_t *pg)
{
	u_int32_t val;

	/*
	 * Turn off the single-step bit in the CP0 Debug Register.
	 */
	write_c0_debug(read_c0_debug() & ~DBG_SST);

	/*
	 * Enable interrupts in the DRSEG Debug Control Register.
	 */
	val = *(u_int32_t *)0xff300000;
	val |= (1 << 4);
	*(u_int32_t *)0xff300000 = val;

	/*
	 * Terminate the command loop.
	 */
	pg->stopped = 0;

	return PALOECPU_DBG_STAT(OK);
}

/*
 * STEP - Single step.
 */
static u_int32_t
cmd_step(paloecpu_dbpage_t *pg)
{
	u_int32_t val;

	/*
	 * Set the single-step bit in the CP0 Debug Register.
	 */
	write_c0_debug(read_c0_debug() | DBG_SST);

	/*
	 * Disable interrupts in the DRSEG Debug Control Register.  Disabling
	 * interrupts here supercedes the target's own setting in the CP0
	 * Status register, so we won't single-step into (for example) the
	 * next timer interrupt.
	 */
	val = *(u_int32_t *)0xff300000;
	val &= ~(1 << 4);
	*(u_int32_t *)0xff300000 = val;

	/*
	 * Terminate the command loop.
	 */
	pg->stopped = 0;

	return PALOECPU_DBG_STAT(OK);
}

/*
 * READ_MEM - Read target memory.
 * We don't check accessibility of the address.  If we take an exception
 * then we're re-enter the debugger and report the failure.
 */
static u_int32_t
cmd_read_mem(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = pg->arg[0];
	u_int32_t size = pg->arg[1];
	u_int32_t count = pg->arg[2];
	u_int32_t i;

	switch (size) {
	case 1: {
		const u_int8_t *sp = (u_int8_t *)addr;
		u_int8_t *dp = (u_int8_t *)pg->data;

		if (count > sizeof (pg->data)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	case 2: {
		const u_int16_t *sp = (u_int16_t *)addr;
		u_int16_t *dp = (u_int16_t *)pg->data;

		if ((addr & 0x1) || (count > sizeof (pg->data) / 2)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	case 4: {
		const u_int32_t *sp = (u_int32_t *)addr;
		u_int32_t *dp = (u_int32_t *)pg->data;

		if ((addr & 0x3) || (count > sizeof (pg->data) / 4)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	default:
		return -1;
	}
	return PALOECPU_DBG_STAT(OK);
}

/*
 * WRITE_MEM - Write target memory.
 * We don't check accessibility of the address.  If we take an exception
 * then we're re-enter the debugger and report the failure.
 */
static u_int32_t
cmd_write_mem(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = pg->arg[0];
	u_int32_t size = pg->arg[1];
	u_int32_t count = pg->arg[2];
	u_int32_t i;

	switch (size) {
	case 1: {
		const u_int8_t *sp = (u_int8_t *)pg->data;
		u_int8_t *dp = (u_int8_t *)addr;

		if (count > sizeof (pg->data)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	case 2: {
		const u_int16_t *sp = (u_int16_t *)pg->data;
		u_int16_t *dp = (u_int16_t *)addr;

		if ((addr & 0x1) || (count > sizeof (pg->data) / 2)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	case 4: {
		const u_int32_t *sp = (u_int32_t *)pg->data;
		u_int32_t *dp = (u_int32_t *)addr;

		if ((addr & 0x3) || (count > sizeof (pg->data) / 4)) {
			return PALOECPU_DBG_STAT(INVAL);
		}
		for (i = 0; i < count; i++) {
			*dp++ = *sp++;
		}
		break;
	}
	default:
		return PALOECPU_DBG_STAT(INVAL);
	}
	return PALOECPU_DBG_STAT(OK);
}

/*
 * READ_CP0 - Read a CP0 Register.
 */
static u_int32_t
cmd_read_cp0(paloecpu_dbpage_t *pg)
{
	u_int32_t reg = pg->arg[0];
	u_int32_t sel = pg->arg[1];
	u_int32_t code[3];

	/*
	 * Synthesize and invoke the following:
	 *	mfc0  v0, reg, sel
	 *	jr	 ra
	 *	nop
	 */
	code[0] = 0x40020000 | ((reg & 0x1f) << 11) | (sel & 0x7);
	code[1] = 0x03e00008;
	code[2] = 0x00000000;
	asm volatile("sync" ::: "memory");
	pg->data[0] = ((u_int32_t (*)(void))code)();

	return PALOECPU_DBG_STAT(OK);
}

/*
 * WRITE_CP0 - Write a CP0 Register.
 */
static u_int32_t
cmd_write_cp0(paloecpu_dbpage_t *pg)
{
	u_int32_t reg = pg->arg[0];
	u_int32_t sel = pg->arg[1];
	u_int32_t val = pg->arg[2];
	u_int32_t code[3];

	/*
	 * Synthesize and invoke the following:
	 *	mtc0  a0, reg, sel
	 *	jr.hb ra
	 *	nop
	 */
	code[0] = 0x40840000 | ((reg & 0x1f) << 11) | (sel & 0x7);
	code[1] = 0x03e00408;
	code[2] = 0x00000000;
	asm volatile("sync" ::: "memory");
	((void (*)(u_int32_t))code)(val);

	return PALOECPU_DBG_STAT(OK);
}

/*
 * CACHE_FLUSH - Flush the D or I cache.
 */
static u_int32_t
cmd_cache_flush(paloecpu_dbpage_t *pg)
{
	u_int32_t id = pg->arg[0];
	u_int32_t addr, sz;

	if (id == PALOECPU_DBG_CACHE_D) {
		sz = 32 * 4 * (64 << ((read_c0_config1() >> 13) & 0x7));
		for (addr = KSEG0ADDR(0); addr < KSEG0ADDR(sz); addr += 32) {
			cache_op(Index_Writeback_Inv_D, addr);
		}
	} else {
		sz = 32 * 4 * (64 << ((read_c0_config1() >> 22) & 0x7));
		for (addr = KSEG0ADDR(0); addr < KSEG0ADDR(sz); addr += 32) {
			cache_op(Index_Invalidate_I, addr);
		}
	}
	return PALOECPU_DBG_STAT(OK);
}

/*
 * WRITE_INSTR - Write an opcode to memory.
 */
static u_int32_t
cmd_write_instr(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = pg->arg[0];
	u_int32_t opcode = pg->arg[1];

	if (addr < KSEG0 || addr >= KSEG2) {
		// cannot yet handle non KSEG memory
		return PALOECPU_DBG_STAT(INVAL);
	}

	/*
	 * Push out via uncached space and, if the address as in KSEG0,
	 * invalidate the I-Cache.
	 */
	*(u_int32_t *)KSEG1ADDR(addr) = opcode;
	if (addr < KSEG1) {
		cache_op(Hit_Invalidate_I, addr);
	}
	return PALOECPU_DBG_STAT(OK);
}

/*
 * Save and restore TLB register state.
 */
typedef struct {
	u_int32_t index;
	u_int32_t entryhi;
	u_int32_t entrylo0;
	u_int32_t entrylo1;
	u_int32_t pagemask;
} tlbstate_t;

static void
tlbop_begin(tlbstate_t *p)
{
	p->index    = read_c0_index();
	p->entryhi  = read_c0_entryhi();
	p->entrylo0 = read_c0_entrylo0();
	p->entrylo1 = read_c0_entrylo1();
	p->pagemask = read_c0_pagemask();
}

static void
tlbop_end(tlbstate_t *p)
{
	write_c0_index(p->index);
	write_c0_entryhi(p->entryhi);
	write_c0_entrylo0(p->entrylo0);
	write_c0_entrylo1(p->entrylo1);
	write_c0_pagemask(p->pagemask);
}

/*
 * READ_TLB - Read a TLB entry.
 */
static u_int32_t
cmd_read_tlb(paloecpu_dbpage_t *pg)
{
	u_int32_t index = pg->arg[0];
	tlbstate_t state;

	tlbop_begin(&state);
	write_c0_index(index);
	asm volatile("ehb; tlbr; ehb");
	pg->data[0] = read_c0_entrylo0();
	pg->data[1] = read_c0_entrylo1();
	pg->data[2] = read_c0_entryhi();
	pg->data[3] = read_c0_pagemask();
	tlbop_end(&state);

	return PALOECPU_DBG_STAT(OK);
}

/*
 * WRITE_TLB - Write a TLB entry.
 */
static u_int32_t
cmd_write_tlb(paloecpu_dbpage_t *pg)
{
	u_int32_t index = pg->arg[0];
	tlbstate_t state;

	tlbop_begin(&state);
	write_c0_index(index);
	write_c0_entrylo0(pg->data[0]);
	write_c0_entrylo1(pg->data[1]);
	write_c0_entryhi(pg->data[2]);
	write_c0_pagemask(pg->data[3]);
	asm volatile("ehb; tlbwi; ehb");
	tlbop_end(&state);

	return PALOECPU_DBG_STAT(OK);
}

/*
 * Save and restore cache state.
 */
typedef struct {
	u_int32_t   ecc;
	u_int32_t   taglo;
	u_int32_t   datalo;
	u_int32_t   dtaglo;
	u_int32_t   ddatalo;
} cachestate_t;

static void
cacheop_begin(cachestate_t *p)
{
	p->ecc = read_c0_ecc();
	p->taglo = read_c0_taglo();
	p->datalo = read_c0_datalo();
	p->dtaglo = read_c0_dtaglo();
	p->ddatalo = read_c0_ddatalo();
	write_c0_ecc(p->ecc & 0x80000000); /* keep Parity-Enable */
	asm volatile("ehb");
}

static void
cacheop_end(cachestate_t *p)
{
	write_c0_ecc(p->ecc);
	write_c0_taglo(p->taglo);
	write_c0_datalo(p->datalo);
	write_c0_dtaglo(p->dtaglo);
	write_c0_ddatalo(p->ddatalo);
}

/*
 * READ_DCACHET - Read D-Cache tags.
 */
static u_int32_t
cmd_read_dcachet(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = KSEG0 | (pg->arg[0] << 5);
	cachestate_t state;

	cacheop_begin(&state);
	cache_op(Index_Load_Tag_D, addr);
	asm volatile("ehb");
	pg->data[0] = read_c0_dtaglo();
	pg->data[1] = 0;
	cacheop_end(&state);
	return PALOECPU_DBG_STAT(OK);
}

/*
 * READ_ICACHET - Read I-Cache tags.
 */
static u_int32_t
cmd_read_icachet(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = KSEG0 | (pg->arg[0] << 5);
	cachestate_t state;

	cacheop_begin(&state);
	cache_op(Index_Load_Tag_I, addr);
	asm volatile("ehb");
	pg->data[0] = read_c0_taglo();
	pg->data[1] = 0;
	cacheop_end(&state);
	return PALOECPU_DBG_STAT(OK);
}

/*
 * READ_DCACHE - Read D-Cache data.
 */
static u_int32_t
cmd_read_dcache(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = KSEG0 | pg->arg[0];
	cachestate_t state;
	int i;

	cacheop_begin(&state);
	for (i = 0; i < 8; i++) {
		cache_op(Index_Load_Tag_D, addr);
		asm volatile("ehb");
		pg->data[i] = read_c0_ddatalo();
		addr += 4;
	}
	cacheop_end(&state);
	return PALOECPU_DBG_STAT(OK);
}

/*
 * READ_ICACHE - Read I-Cache data.
 */
static u_int32_t
cmd_read_icache(paloecpu_dbpage_t *pg)
{
	u_int32_t addr = KSEG0 | pg->arg[0];
	cachestate_t state;
	int i;

	cacheop_begin(&state);
	for (i = 0; i < 8; i++) {
		cache_op(Index_Load_Tag_I, addr);
		asm volatile("ehb");
		pg->data[i] = read_c0_datalo();
		addr += 4;
	}
	cacheop_end(&state);
	return PALOECPU_DBG_STAT(OK);
}

/*
 * Command dispatch table.
 */
typedef struct {
	u_int32_t minargs;
	u_int32_t (*func)(paloecpu_dbpage_t *pg);
} cmdentry_t;

static cmdentry_t ejcmdtab[PALOECPU_DBG_CMD_NUM] = {
	[PALOECPU_DBG_CMD_GO]		= { 0, cmd_go			},
	[PALOECPU_DBG_CMD_STEP]		= { 0, cmd_step			},
	[PALOECPU_DBG_CMD_READ_MEM]	= { 3, cmd_read_mem		},
	[PALOECPU_DBG_CMD_WRITE_MEM]	= { 3, cmd_write_mem		},
	[PALOECPU_DBG_CMD_READ_CP0]	= { 2, cmd_read_cp0		},
	[PALOECPU_DBG_CMD_WRITE_CP0]	= { 3, cmd_write_cp0		},
	[PALOECPU_DBG_CMD_CACHE_FLUSH]	= { 1, cmd_cache_flush		},
	[PALOECPU_DBG_CMD_WRITE_INSTR]	= { 2, cmd_write_instr		},
	[PALOECPU_DBG_CMD_READ_TLB]	= { 1, cmd_read_tlb		},
	[PALOECPU_DBG_CMD_WRITE_TLB]	= { 1, cmd_write_tlb		},
	[PALOECPU_DBG_CMD_READ_DCACHET]	= { 1, cmd_read_dcachet		},
	[PALOECPU_DBG_CMD_READ_ICACHET]	= { 1, cmd_read_icachet		},
	[PALOECPU_DBG_CMD_READ_DCACHE]	= { 1, cmd_read_dcache		},
	[PALOECPU_DBG_CMD_READ_ICACHE]	= { 1, cmd_read_icache		},
};

static u_int32_t
command_handle(paloecpu_dbpage_t *pg)
{
	u_int32_t cmd = pg->cmd_stat;
	cmdentry_t *cp = &ejcmdtab[cmd];

	if (cmd >= PALOECPU_DBG_CMD_NUM || cp->func == NULL) {
		return PALOECPU_DBG_STAT(CMDUNKNOWN);
	}
	if (pg->argc < cp->minargs) {
		return PALOECPU_DBG_STAT(ARGCNT);
	}
	return (*cp->func)(pg);
}

static void
command_wait(paloecpu_dbpage_t *pg)
{
	while (pg->cmdseqno == pg->cmdackno) {
		rmb(); /* wait */
	}
}

static void
command_reply(paloecpu_dbpage_t *pg, u_int32_t status)
{
	pg->cmd_stat = status;
	pg->cmdackno = pg->cmdseqno;
	wmb();
}

/*
 * eCPU EJTAG (SDBBP) Handler 
 *
 * This is the debugger front-end.  Although part of the mcpu Linux kernel,
 * this code is actually executed by the other ecpus - as vectored via their
 * ecpuctx.dbgvec vector.  Having this code as part of the Linux kernel
 * removes any requirement on the ecpu target code.
 *
 * The ecpu sits in the do/while loop, waiting for a command to show up
 * in the shared memory page, taking action based on the command (via the
 * ejcmdtab table above), and replying.
 *
 * If a command clears the 'stopped' flag (i.e. GO and STEP commands) then
 * the do/while loop terminates the execution continues in the ecpu's own code.
 *
 * On entry:
 *  ctx points to the ecpuctx in KSEG1.
 *  regs points to the stacked machine registers (in KSEG1).
 *
 * On exit:
 *  The machine registers are restored from regs.
 */
void
ejtag_handler(ecpuctx_t *ctx, paloecpu_xcptregs_t *regs)
{
	paloecpu_dbpage_t *pg = (paloecpu_dbpage_t *)ctx->dbpage;
	u_int32_t ret;

	/*
	 * Prevent the c0_counter from counting whilst in debug mode.
	 */
	write_c0_debug(read_c0_debug() & ~DBG_COUNTDM);

	/*
	 * Determine how we got here.
	 */
	if (regs->debug & DBG_INTBITS) {
		/*
		 * Normal entry into the debugger.
		 */
		++pg->stats.entry_count;

		/*
		 * External DINT is spurious.  Count these and return.
		 */
		if (regs->debug & DBG_DINT) {
			++pg->stats.dint_count;
			return;
		}

		/*
		 * Return if not armed.
		 */
		if (!pg->armed) {
			++pg->stats.disarmed_count;
			return;
		}

		/*
		 * Copy the machine state from regs into the published
		 * register state in the debug page.
		 */
		memcpy((void *)&pg->regs, regs, sizeof (*regs));
		pg->stopped = 1;
	} else {
		/*
		 * Re-entry into the debugger (e.g. fault on previous command)
		 * Don't copy the stacked registers as they describe the
		 * internal exception rather than the desired original ecpu 
		 * exception.
		 * The debugger application is waiting for the response to
		 * its command.  Report that we took an exception.
		 */
		++pg->stats.reentry_count;
		pg->data[0] = (read_c0_debug() >> 10) & 0x1f;
		command_reply(pg, PALOECPU_DBG_STAT(EXCEPTION));
	}

	/*
	 * Command processing loop.
	 */
	do {
		command_wait(pg);
		ret = command_handle(pg);
		command_reply(pg, ret);
	} while (pg->stopped);

	/*
	 * Continue with the new state.
	 */
	memcpy(regs, (void *)&pg->regs, sizeof (*regs));
}
