/*
 * Copyright 2009 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_ecpu.c 59 2017-01-21 21:37:10Z buildsa $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/mtd/mtd.h>

#include "regio.h"
#include "paloint.h"

#include <palo/include/palo_defs.h>
#include <palo/include/intrc_regs.h>
#include <palo/fw/mcpu/include/ecpuctx.h>
#include <palo/fw/mcpu/include/paloecpu.h>
#include <palo/fw/common/include/ecpu_res.h>

#define DRV_NAME "paloecpu"
#define DRV_DESCRIPTION "Driver for Palo eCPUs"
#define DRV_VERSION "1.0"
#define DRV_COPYRIGHT "Copyright 2009 Cisco Systems"
#define PFX DRV_NAME ": "

MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

static int paloecpu_major;

static u_int32_t ecpu_hold_reg[] = {
	0,
	INTRC_MREGS_RESET_HOLD_CPU1,
	INTRC_MREGS_RESET_HOLD_CPU2,
#ifdef CONFIG_BASE_CRUZ
	INTRC_MREGS_RESET_HOLD_CPU3,
	INTRC_MREGS_RESET_HOLD_CPU4,
#endif
};

static u_int32_t ecpu_nmi_reg[] = {
	0,
	INTRC_MREGS_FORCE_NMI_CPU1,
	INTRC_MREGS_FORCE_NMI_CPU2,
#ifdef CONFIG_BASE_CRUZ
	INTRC_MREGS_FORCE_NMI_CPU3,
	INTRC_MREGS_FORCE_NMI_CPU4,
#endif
};

typedef struct {
	int		ecpuid;		// ecpu id (0..N)
	char		name[16];	// e.g. "ecpu1"
	int		irq;		// virtual console interrupt
	paloecpu_vcons_t *vcons;	// virtual console memory
	dma_addr_t	vcons_pa;	// virtual console phys addr
	paloecpu_dbpage_t *dbpage;	// debug message page
	dma_addr_t	dbpage_pa;	// debug message page phys addr
	struct platform_device *pdev;	// platform_device
	struct mutex	mutex;		// paloecpu_dev_t mutex
	wait_queue_head_t rxwaitq;	// virtual console rx wait queue
	wait_queue_head_t txwaitq;	// virtual console tx wait queue
	u_int32_t	*remote_intr;	// virtual console ecpu interrupt
} paloecpu_dev_t;

static paloecpu_dev_t paloecpu[PALO_MAX_ECPU];

typedef struct {
	paloecpu_dev_t	*dev;
	int		is_datapath;
} paloecpu_open_file_t;

static irqreturn_t paloecpu_intr(int irq, void *arg)
{
	paloecpu_dev_t *v = arg;
	paloecpu_vcons_fifo_t *rx = &v->vcons->tx;	// our rx is their tx
	paloecpu_vcons_fifo_t *tx = &v->vcons->rx;	// our tx is their rx

	if (!paloecpu_vcons_fifo_isempty(rx)) {
		wake_up_interruptible(&v->rxwaitq);
	}
	if (!paloecpu_vcons_fifo_isfull(tx)) {
		wake_up_interruptible(&v->txwaitq);
	}
	return IRQ_HANDLED;
}

static void
paloecpu_ring_doorbell(paloecpu_dev_t *v)
{
	if (v->remote_intr != NULL) {
		writel(1, v->remote_intr);
	}
}

static ssize_t paloecpu_read(struct file *filp, char __user *buf,
		size_t count, loff_t *pos)
{
	paloecpu_open_file_t *op = filp->private_data;
	paloecpu_dev_t *v = op->dev;
	paloecpu_vcons_fifo_t *rx = &v->vcons->tx;	// our rx is their tx
	char mbuf[PALOECPU_VCONS_BUFSZ];

	/*
	 * Return zero bytes for mcpu or control file.
	 */
	if (!op->is_datapath || v->ecpuid == 0) {
		return 0;
	}

	if (mutex_lock_interruptible(&v->mutex)) {
		return -ERESTARTSYS;
	}
	while (paloecpu_vcons_fifo_isempty(rx)) {
		mutex_unlock(&v->mutex);
		if (filp->f_flags & O_NONBLOCK) {
			return -EAGAIN;
		}
		if (wait_event_interruptible(v->rxwaitq,
			!paloecpu_vcons_fifo_isempty(rx))) {
			return -ERESTARTSYS;
		}
		if (mutex_lock_interruptible(&v->mutex)) {
			return -ERESTARTSYS;
		}
	}
	if (count > sizeof (buf)) {
		count = sizeof (buf);
	}
	count = paloecpu_vcons_fifo_read(rx, mbuf, count);
	if (copy_to_user(buf, mbuf, count)) {
		mutex_unlock(&v->mutex);
		return -EFAULT;
	}
	paloecpu_ring_doorbell(v);
	mutex_unlock(&v->mutex);
	return count;
}

static ssize_t paloecpu_write(struct file *filp, const char __user *buf,
		size_t count, loff_t *pos)
{
	paloecpu_open_file_t *op = filp->private_data;
	paloecpu_dev_t *v = op->dev;
	paloecpu_vcons_fifo_t *tx = &v->vcons->rx;	// our tx is their rx
	char mbuf[PALOECPU_VCONS_BUFSZ];

	/*
	 * Swallow data written to control or mcpu.
	 */
	if (!op->is_datapath || v->ecpuid == 0) {
		return count;
	}

	if (mutex_lock_interruptible(&v->mutex)) {
		return -ERESTARTSYS;
	}
	while (paloecpu_vcons_fifo_isfull(tx)) {
		mutex_unlock(&v->mutex);
		if (filp->f_flags & O_NONBLOCK) {
			return -EAGAIN;
		}
		if (wait_event_interruptible(v->txwaitq,
			!paloecpu_vcons_fifo_isfull(tx))) {
			return -ERESTARTSYS;
		}
		if (mutex_lock_interruptible(&v->mutex)) {
			return -ERESTARTSYS;
		}
	}
	if (count > sizeof (mbuf)) {
		count = sizeof (mbuf);
	}
	if (copy_from_user(mbuf, buf, count)) {
		mutex_unlock(&v->mutex);
		return -EFAULT;
	}
	count = paloecpu_vcons_fifo_write(tx, mbuf, count);
	paloecpu_ring_doorbell(v);
	mutex_unlock(&v->mutex);
	return count;
}

static unsigned int paloecpu_poll(struct file *filp,
		struct poll_table_struct *wait)
{
	paloecpu_open_file_t *op = filp->private_data;
	paloecpu_dev_t *v = op->dev;
	paloecpu_vcons_fifo_t *rx = &v->vcons->tx;	// our rx is their tx
	unsigned int mask = 0;

	/*
	 * No poll for control or mcpu.
	 */
	if (!op->is_datapath || v->ecpuid == 0) {
		return 0;
	}

	mutex_lock(&v->mutex);
	poll_wait(filp, &v->rxwaitq, wait);
	if (!paloecpu_vcons_fifo_isempty(rx)) {
		mask = POLLIN | POLLRDNORM;
	}
	mutex_unlock(&v->mutex);

	return mask;
}

static int paloecpu_run_ecpu(paloecpu_dev_t *v)
{
	extern void paloecpu_nmi_entry(void);
	extern void paloecpu_ejtag_entry(void);
	int polls, r;
	u32 holdreg;
	ecpupage_t *ep;
	volatile ecpuctx_t *ctx;
	extern struct mtd_info *palo_mtd;

	// cannot run the mcpu or out-of-bounds
	if (v->ecpuid == 0 || v->ecpuid >= ARRAY_SIZE(ecpu_hold_reg))
		return -EINVAL;

	/*
	 * First, acquire the flash to be sure the flash part
	 * is in array mode.  The ecpu reset vectors are in flash and
	 * and we need the flash in array mode before letting them go.
	 * This call may sleep waiting for a write or an erase to complete
	 * or something, but once we get back we know the flash part is
	 * ready for ecpu instruction fetches.
	 *
	 * Note that acquire will return with a spin_lock held, so
	 * we should take care to be quick and not sleep/fault or
	 * do anything else that might take a "long" time.
	 */
	if (palo_mtd->acquire) {
		r = palo_mtd->acquire(palo_mtd);
		if (r)
			return r;
	}

	/*
	 * Determine the ecpu context.
	 */
	ep = (ecpupage_t *)CKSEG1ADDR(ECPUCTX_MEMBASE);
	ctx = &ep->ctx[v->ecpuid];

	/*
	 * Set up the virtual ecpu console address.
	 */
	ctx->vcons = v->vcons_pa;

	/*
	 * Set up the debugger and its vectors.
	 * NMI and EJTAG entry-points are in KSEG1 (uncached), as are the
	 * data pointers.  This avoids all use of the L1 cache to facilitate
	 * debugging before (and during) L1 cache initialization.
	 */
	ctx->dbpage = KSEG1ADDR(v->dbpage_pa);
	ctx->nmivec = KSEG1ADDR((u_int32_t)paloecpu_nmi_entry);
	ctx->nmimagic = ECPUCTX_NMIMAGIC;
	ctx->dbgvec = KSEG1ADDR((u_int32_t)paloecpu_ejtag_entry);
	ctx->dbgmagic = ECPUCTX_DBGMAGIC;
	wmb();
	(void)ctx->dbgmagic; /* read to flush */
	rmb();

	/*
	 * Release the eCPU from reset.
	 */
	holdreg = ecpu_hold_reg[v->ecpuid];
	writereg(holdreg, 0);
	(void)readreg(holdreg); /* read back to flush */

	/*
	 * Wait for ecpu to start.  It'll clear magic when started.
	 * ecpu starts are normally immediate, but sometimes
	 * slow starts with an fs2 attached.
	 */
	polls = 10000;
	do {
		udelay(100);
	} while (ctx->magic != 0 && --polls > 0);

	if (palo_mtd->release)
		palo_mtd->release(palo_mtd);

	/* Did it start up? */
	if (ctx->magic != 0)
		return -ETIMEDOUT;

	return 0;
}

static int paloecpu_stop_ecpu(paloecpu_dev_t *v)
{
	u32 holdreg;

	// cannot reset the mcpu or out-of-bounds
	if (v->ecpuid == 0 || v->ecpuid >= ARRAY_SIZE(ecpu_hold_reg))
		return -EINVAL;
	holdreg = ecpu_hold_reg[v->ecpuid];
	writereg(holdreg, 1);
	(void)readreg(holdreg); /* read back to flush */

	return 0;
}

static int paloecpu_nmi_ecpu(paloecpu_dev_t *v)
{
	extern struct mtd_info *palo_mtd;
	u32 nmireg;
	int r;

	// cannot nmi the mcpu or out-of-bounds
	if (v->ecpuid == 0 || v->ecpuid >= ARRAY_SIZE(ecpu_hold_reg))
		return -EINVAL;

	/*
	 * Acquire the flash.
	 */
	if (palo_mtd->acquire) {
		r = palo_mtd->acquire(palo_mtd);
		if (r)
			return r;
	}

	nmireg = ecpu_nmi_reg[v->ecpuid];
	writereg(nmireg, 1);
	(void)readreg(nmireg); /* read back to flush */

	/*
	 * Wait a while.  The NMI code should only be a few cycles.
	 */
	udelay(100);

	if (palo_mtd->release)
		palo_mtd->release(palo_mtd);

	return 0;
}

static int paloecpu_run_mcpu(void)
{
	ecpupage_t *ep;
	volatile ecpuctx_t *ctx;
	u_int32_t ebase, status, cause;

	/*
	 * The mcpu is running Linux so we are not going to reset it here.
	 * Instead, this API allows us to invoke ecpu-like test code on the
	 * mcpu.  The same ecpuctx block is expected to be prepared with
	 * the appropriate vectors.  Instead of reseting mcpu we simply call
	 * the target function with interrupts disabled.
	 * When the test code returns (which could be after an arbitrarily
	 * long time) we re-enable interrupts and continue.
	 */
	ep = (ecpupage_t *)CKSEG1ADDR(ECPUCTX_MEMBASE);
	ctx = &ep->ctx[0];
	if (ctx->magic != ECPUCTX_MAGIC) {
		return -1;
	}
	if (ctx->entry == 0) {
		return -1;
	}
	ctx->magic = 0;
	local_irq_disable();
	ebase = read_c0_ebase();
	cause = read_c0_cause();
	status = read_c0_status();
	(*(void (*)(u_int32_t arg))(ctx->entry))(ctx->arg);
	write_c0_status(status);
	write_c0_cause(cause);
	write_c0_ebase(ebase);
	asm volatile("ehb");
	local_irq_enable();

	return 0;
}

static long paloecpu_ioctl(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	paloecpu_open_file_t *op = filp->private_data;
	paloecpu_dev_t *v = op->dev;
	int r = -ENOTTY;
	u_int32_t addr;

	if (op->is_datapath) {
		return r;
	}
	if (mutex_lock_interruptible(&v->mutex)) {
		return -ERESTARTSYS;
	}
	switch (cmd) {
	case PALOECPU_IOC_RUN:
		if (v->ecpuid == 0) {
			r = paloecpu_run_mcpu();
		} else {
			r = paloecpu_run_ecpu(v);
		}
		break;

	case PALOECPU_IOC_STOP:
		r = paloecpu_stop_ecpu(v);
		break;

	case PALOECPU_IOC_NMI:
		r = paloecpu_nmi_ecpu(v);
		break;

	case PALOECPU_IOC_GETDBPAGE:
		if (sizeof (paloecpu_xcptregs_t) != PALOECPU_XCPTREGS_SIZE) {
			printk(KERN_ERR "PALOECPU_XCPTREGS_SIZE != "
				"sizeof (paloecpu_xcptregs_t)\n");
			r = -EIO;
		} else {
			addr = v->dbpage_pa;
			r = copy_to_user(argp, &addr, sizeof (u_int32_t));
			if (r) {
				r = -EFAULT;
			}
		}
		break;

	default:
		break;
	}
	mutex_unlock(&v->mutex);
	return r;
}

static void paloecpu_flush_rx(paloecpu_dev_t *v)
{
	paloecpu_vcons_fifo_t *rx = &v->vcons->tx;	// our rx is their tx
	char mbuf[PALOECPU_VCONS_BUFSZ];

	paloecpu_vcons_fifo_read(rx, mbuf, sizeof (mbuf));
	paloecpu_ring_doorbell(v);
}

static int paloecpu_open(struct inode *inode, struct file *filp)
{
	int ecpuid = iminor(inode) & 0xf;
	int is_datapath = (iminor(inode) > 0xf);
	paloecpu_dev_t *v = &paloecpu[ecpuid];
	paloecpu_open_file_t *op;

	if (ecpuid < 0 || ecpuid >= PALO_MAX_ECPU) {
		return -ENODEV;
	}
	if (mutex_lock_interruptible(&v->mutex)) {
		return -ERESTARTSYS;
	}
	op = kmalloc(sizeof (*op), GFP_KERNEL);
	if (!op) {
		mutex_unlock(&v->mutex);
		return -ENOMEM;
	}
	op->dev = v;
	op->is_datapath = is_datapath;
	filp->private_data = op;
	if (v->vcons && is_datapath) {
		if (++v->vcons->connected == 1) {
			/*
			 * Flush garbage from the read fifo on the initial
			 * connection.
			 */
			paloecpu_flush_rx(v);
		}
	}
	mutex_unlock(&v->mutex);
	return 0;
}

static int paloecpu_release(struct inode *inode, struct file *filp)
{
	paloecpu_open_file_t *op = filp->private_data;
	paloecpu_dev_t *v = op->dev;

	if (mutex_lock_interruptible(&v->mutex)) {
		return -ERESTARTSYS;
	}
	if (v->vcons && op->is_datapath) {
		--v->vcons->connected;
		paloecpu_ring_doorbell(v);
	}
	filp->private_data = NULL;
	kfree(op);
	mutex_unlock(&v->mutex);

	return 0;
}

static struct file_operations paloecpu_fops = {
	.owner	 = THIS_MODULE,
	.open	 = paloecpu_open,
	.read	 = paloecpu_read,
	.write	 = paloecpu_write,
	.unlocked_ioctl	 = paloecpu_ioctl,
	.poll	 = paloecpu_poll,
	.release = paloecpu_release,
};

static int __devinit paloecpu_probe(struct platform_device *pdev)
{
	int ecpuid = pdev->id;
	paloecpu_dev_t *v = &paloecpu[ecpuid];
	int r, lvl, src, irq;
	u_int32_t intr_pa;

	if (ecpuid < 0 || ecpuid >= PALO_MAX_ECPU) {
		return -EINVAL;
	}
	v->ecpuid = ecpuid;
	v->pdev = pdev;
	snprintf(v->name, sizeof (v->name), "%s%d", DRV_NAME, ecpuid);

	/*
	 * For ecpus (not mcpu) allocate the debug and virtual console
	 * resources.
	 */
	if (ecpuid != 0) {
		v->dbpage = dma_alloc_coherent(&pdev->dev, PALOECPU_DBPAGESIZE,
				&v->dbpage_pa, GFP_KERNEL);
		if (v->dbpage == NULL) {
			return -ENOMEM;
		}
		v->vcons = dma_alloc_coherent(&pdev->dev, sizeof (*v->vcons),
				&v->vcons_pa, GFP_KERNEL);
		if (v->vcons == NULL) {
			dma_free_coherent(&pdev->dev, PALOECPU_DBPAGESIZE,
				v->dbpage, v->dbpage_pa);
			return -ENOMEM;
		}
		init_waitqueue_head(&v->rxwaitq);
		init_waitqueue_head(&v->txwaitq);

		/*
		 * Request the interrupt for the mcpu doorbell.  This will be
		 * poked by the ecpu to indicate vcons activity to us.
		 */
		lvl = VCONS_INT_LVL;
		src = VCONS1_INT_SRC + ecpuid - 1;
		irq = palo_int_to_irq(lvl, src);
		r = request_irq(irq, paloecpu_intr, IRQF_DISABLED, v->name, v);
		if (r) {
			printk(KERN_ERR "%s unable to request irq\n", v->name);
			dma_free_coherent(&pdev->dev, sizeof (*v->vcons),
				v->vcons, v->vcons_pa);
			dma_free_coherent(&pdev->dev, PALOECPU_DBPAGESIZE,
				v->dbpage, v->dbpage_pa);
			return r;
		}
		v->irq = irq;

		/*
		 * Calculate the remote ecpu's doorbell interrupt address that
		 * we will use to signal vcons activity to them.
		 * The ecpu firmware uses irq 0 (lvl=0, src=0).
		 */
		lvl = 0;
		src = 0;
		intrc_set_int_level_addr_enc(&intr_pa,
			INTRC_MREGS_BASE_ADDR, ecpuid, lvl, src);
		v->remote_intr = (u_int32_t *)CKSEG1ADDR(intr_pa);
	}
	mutex_init(&v->mutex);
	platform_set_drvdata(pdev, v);
	printk(KERN_INFO PFX "device %d initialized\n", ecpuid);
	return 0;
}

static int paloecpu_remove(struct platform_device *pdev)
{
	paloecpu_dev_t *v = &paloecpu[pdev->id];

	if (v->irq) {
		free_irq(v->irq, v);
	}
	if (v->vcons != NULL) {
		dma_free_coherent(&pdev->dev, sizeof (*v->vcons),
			v->vcons, v->vcons_pa);
	}
	if (v->dbpage != NULL) {
		dma_free_coherent(&pdev->dev, PALOECPU_DBPAGESIZE,
			v->dbpage, v->dbpage_pa);
	}
	return 0;
}

static struct platform_driver paloecpu_driver = {
	.probe	= paloecpu_probe,
	.remove = paloecpu_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name  = DRV_NAME
	}
};

static int __init paloecpu_init(void)
{
	int r;

	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO PFX "%s\n", DRV_COPYRIGHT);

	r = register_chrdev(0, DRV_NAME, &paloecpu_fops);
	if (r < 0) {
		printk(KERN_INFO PFX "register_chrdev return %d\n", r);
		return r;
	}
	paloecpu_major = r;
	printk(KERN_INFO PFX "major=%d\n", paloecpu_major);
	r = platform_driver_register(&paloecpu_driver);
	if (r) {
		printk(KERN_INFO PFX "platform_driver_register failed\n");
		unregister_chrdev(paloecpu_major, DRV_NAME);
	}
	return r;
}

static void __exit paloecpu_exit(void)
{
	platform_driver_unregister(&paloecpu_driver);
	unregister_chrdev(paloecpu_major, DRV_NAME);
}

module_init(paloecpu_init);
module_exit(paloecpu_exit);
