/*
 * Copyright 2008 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_i2c.c 21 2016-12-02 21:25:59Z buildsa $"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/i2c.h>
#include <linux/jiffies.h>
#include <linux/delay.h>
#include <linux/timex.h>
#include <linux/cdev.h>
#include <linux/kdev_t.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/timex.h>
//#include <asm/semaphore.h>
#include <linux/slab.h>
#include <asm/io.h>

#include <linux/platform_device.h>
#include "palo/include/synopsis/dw_i2c.h"
#include "palo/include/misc_mregs.h"
#include "palo/include/palo_i2ctiming.h"
#include "palo/fw/mcpu/include/palo_i2c_defs.h"
#include "palo_i2c.h"

MODULE_LICENSE("TBD");
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION(DRV_DESCRIPTION);

static int
is_timedout(struct palo_i2c *i2c)
{
	if (time_after(jiffies, i2c->xfer_timeout))
		return (1);
	else
		return (0);
}

static int
palo_i2c_master_initialize(struct palo_i2c *i2c, u_int16_t target)
{
	int port = i2c->port;

	if (readreg(IC_ENABLE(port)) != 0) {
		/*
		 * Ensure that we don't disable the I2C controller mid-operation.
		 */
		while (!is_timedout(i2c)) {
			if (!(readreg(IC_STATUS(port)) &
				IC_STATUS_MST_ACTIVITY)) {
				break;
			}
		}
		if (is_timedout(i2c)) {
			dev_info(i2c->dev, "init timeout\n");
			return -1;
		}
		writereg(IC_ENABLE(port), 0);
	}

	writereg(IC_CON(port),
		IC_CON_SLAVE_DISABLE | IC_CON_RESTART_EN |
		IC_CON_SPEED_STANDARD | IC_CON_MASTER_MODE);

	writereg(IC_INTR_MASK(port), 0);

	writereg(IC_SS_HCNT(port), SS_SCL_HCNT);
	writereg(IC_SS_LCNT(port), SS_SCL_LCNT);

	i2c->target = target;
	writereg(IC_TAR(port), target);

	return 0;
}

static void
palo_i2c_controller_reset(void)
{
	u_int32_t val;

	/*
	 * Reset the I2C controller by toggling the I2C bit in the
	 * misc_sw_hard_rst register.
	 */
	val = readreg(MISC_MREGS_SW_HARD_RST);
	val |= MISC_MREGS_SW_HARD_RST_I2C;
	writereg(MISC_MREGS_SW_HARD_RST, val);
	udelay(5);
	val &= ~MISC_MREGS_SW_HARD_RST_I2C;
	writereg(MISC_MREGS_SW_HARD_RST, val);
	udelay(5);
}

static int
palo_i2c_do_master_read_write(struct palo_i2c *i2c, int is_read,
	u_int8_t *adbuf, int adlen,
	u_int8_t *buf, int len)
{
	int adcnt, txcnt, rxcnt, rxlen, nstarts, maxstarts, port;
	unsigned long flags;
	u_int32_t intr;

	port = i2c->port;

	writereg(IC_ENABLE(port), 1);

	adcnt = txcnt = rxcnt = 0;
	rxlen = is_read ? len : 0;
	maxstarts = is_read ? 2 : 1;
	nstarts = 0;

	writereg(IC_TX_TL(port), i2c->txfsz - 1);
	writereg(IC_RX_TL(port), 0);

	/*
	 * Fill the TX FIFO with interrupts disabled.  As soon as we write the
	 * first byte the controller will start.  If an interrupt takes us away
	 * long enough for the TX FIFO to drain, a STOP will be generated and
	 * we'll have a split transaction.
	 */
	local_irq_save(flags);
	for (adcnt = 0; adcnt < adlen; adcnt++) {
		writereg(IC_DATA_CMD(port), IC_DATA_CMD_WRITE | adbuf[adcnt]);
	}
	while (adlen + txcnt < i2c->txfsz && txcnt < len) {
		if (is_read)
			writereg(IC_DATA_CMD(port), IC_DATA_CMD_READ);
		else
			writereg(IC_DATA_CMD(port),
				IC_DATA_CMD_WRITE | buf[txcnt]);
		++txcnt;
	}
	local_irq_restore(flags);

	while (!is_timedout(i2c)) {
		intr = readreg(IC_RAW_INTR_STAT(port));
		if (intr & IC_INTR_RX_FULL) {
			while (rxcnt < rxlen &&
				(readreg(IC_STATUS(port)) & IC_STATUS_RFNE))
				buf[rxcnt++] = readreg(IC_DATA_CMD(port));
		}
		if (intr & IC_INTR_RX_OVER)
			return (PALO_I2C_ERR_RX_OVER);
		if (intr & IC_INTR_TX_ABRT)
			return (PALO_I2C_ERR_TX_ABORT);
		if (intr & IC_INTR_START_DET) {
			if (++nstarts > maxstarts)
				return (PALO_I2C_ERR_TX_SPLIT_START);
			readreg(IC_CLR_START_DET(port));
		}
		if (intr & IC_INTR_STOP_DET) {
			if (txcnt != len || rxcnt != rxlen)
				return (PALO_I2C_ERR_TX_SPLIT_STOP);
		}
		if (intr & IC_INTR_TX_EMPTY) {
			if (txcnt < len && readreg(IC_TXFLR(port)) < i2c->txfsz)
			{
				if (is_read)
					writereg(IC_DATA_CMD(port),
						IC_DATA_CMD_READ);
				else
					writereg(IC_DATA_CMD(port),
						IC_DATA_CMD_WRITE | buf[txcnt]);
				++txcnt;
			}
		}
		if (txcnt == len && rxcnt == rxlen) {
			if (intr & IC_INTR_ACTIVITY)
				return (0);
		}

		if (i2c->break_armed && --i2c->break_cnt == 0) {
			if (i2c->break_armed == 1) {
				writereg(IC_ENABLE(port), 0);
			} else {
				palo_i2c_controller_reset();
			}
			i2c->break_armed = 0;
			return (PALO_I2C_ERR_TIMEOUT);
		}
		++i2c->break_total;
	}
	return (PALO_I2C_ERR_TIMEOUT);
}

static int
palo_i2c_master_read_write(struct palo_i2c *i2c, struct i2c_msg *msg, int num)
{
	int adlen, len, port, is_read, ret;
	u_int8_t *adbuf, *buf;

	i2c->last_jiffies = jiffies_64;
	i2c->last_cycles = get_cycles();

	i2c->xfer_timeout = jiffies + i2c->adapter.timeout;

	if (msg[0].addr != i2c->target || i2c->error) {
		if (palo_i2c_master_initialize(i2c, msg[0].addr) < 0) {
			i2c->error = 1;
			return (PALO_I2C_ERR_TIMEOUT);
		}
	}
	i2c->error = 0;

	if (num == 1) {
		adbuf = NULL;
		adlen = 0;
		buf = msg[0].buf;
		len = msg[0].len;
		is_read = msg[0].flags & I2C_M_RD;
	} else {
		adbuf = msg[0].buf;
		adlen = msg[0].len;
		buf = msg[1].buf;
		len = msg[1].len;
		is_read = msg[1].flags & I2C_M_RD;
	}

	port = i2c->port;
	if (adlen > i2c->txfsz) {
		printk(KERN_ERR PFX "Invalid adlen: %d > %d\n",
			adlen, i2c->txfsz);
		return -EINVAL;
	}

	ret = palo_i2c_do_master_read_write(i2c, is_read, adbuf, adlen,
		buf, len);
	if (ret != PALO_I2C_ERR_TIMEOUT) {
		while (!is_timedout(i2c)) {
			if (!(readreg(IC_STATUS(port)) &
				IC_STATUS_MST_ACTIVITY)) {
				writereg(IC_ENABLE(port), 0);
				return (ret);
			}
		}
		dev_info(i2c->dev, "term timeout\n");
	}
	i2c->error = 1;
	return (PALO_I2C_ERR_TIMEOUT);
}

static int
palo_i2c_doxfer(struct palo_i2c *i2c, struct i2c_msg *msg, int num)
{
	const char *err = NULL;
	int ret;

	ret = palo_i2c_master_read_write(i2c, msg, num);
	palo_i2c_log_xfer(i2c, msg, num, ret);
	switch (ret) {
	case PALO_I2C_ERR_RX_OVER:
		err = "rx-overrun";
		ret = -EAGAIN;
		break;

	case PALO_I2C_ERR_TX_ABORT:
		(void) readreg(IC_CLR_TX_ABRT(i2c->port));
		ret = -ENXIO;
		break;

	case PALO_I2C_ERR_TX_SPLIT_START:
		err = "split-transaction-start";
		ret = -EAGAIN;
		break;

	case PALO_I2C_ERR_TX_SPLIT_STOP:
		err = "split-transaction-stop";
		ret = -EAGAIN;
		break;

	case PALO_I2C_ERR_TIMEOUT:
		err = "timeout";
		ret = -EIO;
		break;

	default:
		break;
	}

	if (err != NULL)
		dev_info(i2c->dev, "target 0x%x: %s error\n", i2c->target, err);

	return ret;
}

static int
palo_i2c_xfer(struct i2c_adapter *adapter, struct i2c_msg *msg, int num)
{
	struct palo_i2c *i2c = adapter->algo_data;
	u_int32_t flags;

	if (num != 1 && num != 2)
		return -EINVAL;
	/* if (msg[0].len == 0)
	    return -EINVAL;
	 */

	flags = msg[0].flags;
	if (num == 2) {
		/* 
		 * Two part transactions must be to the same slave,
		 * must begin with a 1 - 4 byte write.
		 */
		if (msg[1].addr != msg[0].addr || (msg[0].flags & I2C_M_RD) ||
			msg[1].len == 0 || msg[0].len > 4) {
			return -EINVAL;
		}
		flags |= msg[1].flags;
	}

	/*
	 * Only the I2C_M_RD flag is supported.
	 */
	if (flags & ~I2C_M_RD)
		return -EINVAL;

	return palo_i2c_doxfer(i2c, msg, num);
}

static u32
palo_i2c_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C;
}

static struct i2c_algorithm palo_i2c_algorithm = {
	.master_xfer = palo_i2c_xfer,
	.functionality = palo_i2c_func
};

static ssize_t i2c_attr_break_arm_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	return sprintf(buf, "%d\n", i2c->break_armed);
}

static ssize_t i2c_attr_break_arm_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	sscanf(buf, "%d", &i2c->break_armed);
	return count;
}

static ssize_t i2c_attr_break_cnt_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	return sprintf(buf, "%d\n", i2c->break_cnt);
}

static ssize_t i2c_attr_break_cnt_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	sscanf(buf, "%d", &i2c->break_cnt);
	return count;
}

static ssize_t i2c_attr_break_total_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	return sprintf(buf, "%d\n", i2c->break_total);
}

static ssize_t i2c_attr_break_total_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct palo_i2c *i2c = dev_to_palo_i2c(dev);

	sscanf(buf, "%d", &i2c->break_total);
	return count;
}

static DEVICE_ATTR(i2c_break_arm, S_IRUGO | S_IWUSR,
	i2c_attr_break_arm_show,
	i2c_attr_break_arm_store);
static DEVICE_ATTR(i2c_break_cnt, S_IRUGO | S_IWUSR,
	i2c_attr_break_cnt_show,
	i2c_attr_break_cnt_store);
static DEVICE_ATTR(i2c_break_total, S_IRUGO | S_IWUSR,
	i2c_attr_break_total_show,
	i2c_attr_break_total_store);

static int
palo_i2c_probe(struct platform_device *pdev)
{
	struct palo_i2c *i2c;
	struct i2c_adapter *adapter;
	int ret;

	i2c = kzalloc(sizeof(struct palo_i2c), GFP_KERNEL);
	if (i2c == NULL)
		return -ENOMEM;
	i2c->port = pdev->id;
	i2c->target = -1;
	i2c->dev = &pdev->dev;
	i2c->txfsz = ((readreg(IC_COMP_PARAM_1(i2c->port)) >> 16) & 0xff) + 1;

	mutex_init(&i2c->mut);
	spin_lock_init(&i2c->spinlock);
	ret = palo_i2c_log_init(i2c);
	if (ret)
		return ret;

	adapter = &i2c->adapter;
	adapter->algo = &palo_i2c_algorithm;
	strlcpy(adapter->name, pdev->name, sizeof(adapter->name));
	adapter->retries = 3; /* not used */
	adapter->timeout = msecs_to_jiffies(PALO_I2C_TIMEOUT_MS);
	adapter->algo_data = i2c;
	adapter->class = I2C_CLASS_HWMON;
	adapter->nr = pdev->id;
	adapter->dev.parent = &pdev->dev;

	ret = i2c_add_numbered_adapter(adapter);
	if (ret) {
		dev_err(&pdev->dev, "adapter %s registration failed\n",
			adapter->name);
		kfree(i2c);
		return ret;
	}

	platform_set_drvdata(pdev, i2c);

	ret = device_create_file(&pdev->dev, &dev_attr_i2c_break_arm);
	if (ret != 0) {
		dev_err(&pdev->dev, "create_file failed\n");
	}
	ret = device_create_file(&pdev->dev, &dev_attr_i2c_break_cnt);
	if (ret != 0) {
		dev_err(&pdev->dev, "create_file failed\n");
	}
	ret = device_create_file(&pdev->dev, &dev_attr_i2c_break_total);
	if (ret != 0) {
		dev_err(&pdev->dev, "create_file failed\n");
	}
	dev_info(&pdev->dev, "installed\n");

	return 0;
}

static __exit int
palo_i2c_remove(struct platform_device *pdev)
{
	struct palo_i2c *i2c = platform_get_drvdata(pdev);

	platform_set_drvdata(pdev, NULL);
	i2c_del_adapter(&i2c->adapter);
	kfree(i2c);

	return 0;
}

static struct platform_driver palo_i2c_driver = {
	.probe = palo_i2c_probe,
	.remove = __devexit_p(palo_i2c_remove),
	.driver =
	{
		.name = "paloi2c",
		.owner = THIS_MODULE,
	},
};

static int __init
palo_i2c_init(void)
{
	dev_t dev;
	int ret;

	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO PFX "%s\n", DRV_COPYRIGHT);

	if (palo_i2c_logmajor != 0) {
		dev = MKDEV(palo_i2c_logmajor, 0);
		ret = register_chrdev_region(dev, PALO_NUM_I2C, DRV_NAME);
	} else {
		ret = alloc_chrdev_region(&dev, 0, PALO_NUM_I2C, DRV_NAME);
		palo_i2c_logmajor = MAJOR(dev);
	}
	if (ret) {
		printk(KERN_ERR DRV_NAME "log cannot acquire major number\n");
		return ret;
	}
	printk(KERN_INFO DRV_NAME "log  major=%d\n", MAJOR(dev));

	return platform_driver_register(&palo_i2c_driver);
}

static void __exit
palo_i2c_exit(void)
{
	platform_driver_unregister(&palo_i2c_driver);
}

module_init(palo_i2c_init);
module_exit(palo_i2c_exit);
