/*
 * Copyright 2010 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_nic.c 43 2016-12-03 22:59:19Z buildsa $"

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/init.h>
#include <linux/fs.h>

#include <linux/param.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/addrspace.h>

#include <palo/fw/common/include/ecpu_res.h>

#include <palo/fw/mcpu/include/palonic.h>
#include <palo/include/regs_macros.h>
#include <palo/include/wq_regs.h>
#include <palo/include/rq_regs.h>
#include <palo/include/cq_regs.h>
#include <palo/include/intr_regs.h>
#include <palo/include/intrc_regs.h>
#include <palo/include/net_regs.h>
#include <common/include/hwecpu_addr.h>
#include <drivers/palo/mnic/vnic_resource.h>
#include <drivers/palo/mnic/vnic_enet.h>
#include <palo/fw/mcpu/include/palo_conf.h>


#include "intrc_mregs.h"
#include "palonic.h"

#define DRV_NAME "palonic"
#define DRV_DESCRIPTION "Driver for Palo NIC Controller"
#define DRV_VERSION "1.0"
#define DRV_COPYRIGHT "Copyright 2010 Cisco Systems"
#define PFX DRV_NAME ": "

MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_LICENSE("TBD");
MODULE_VERSION(DRV_VERSION);

static int palonic_major;

static inline u32
readreg(const u32 pa)
{
	return readl((void *) CKSEG1ADDR(pa));
}

static void inline
writereg(const u32 pa, const u32 val)
{
	writel(val, (void *) CKSEG1ADDR(pa));
}

enum bar_type {
	BAR_BAR0,
	BAR_WQ,
	BAR_RQ,
	BAR_CQ,
	BAR_INTR,
	BAR_INTR_TABLE,
	BAR_INTR_PBA,
	BAR_DEV,
	BAR_MAX,
};

struct bar0_data {
	struct vnic_resource_header header;
	struct vnic_resource resources[RES_TYPE_MAX];
};

//
// Template for bar0 data
//
static struct bar0_data palo_nic_bar0 = {
	.header = { VNIC_RES_MAGIC, VNIC_RES_VERSION, },
	.resources = {
		// .bar_offset = 0 for all resources[]
		{ .type = RES_TYPE_WQ,        .bar = BAR_WQ, .count = 0,},
		{ .type = RES_TYPE_RQ,        .bar = BAR_RQ, .count = 0,},
		{ .type = RES_TYPE_CQ,        .bar = BAR_CQ, .count = 0,},
		{ .type = RES_TYPE_INTR_CTRL, .bar = BAR_INTR, .count = 0,},
		{ .type = RES_TYPE_INTR_TABLE, .bar = BAR_INTR_TABLE,
			.count = 1,},
		{ .type = RES_TYPE_INTR_PBA,  .bar = BAR_INTR_PBA, .count = 1,},
		{ .type = RES_TYPE_DEV,       .bar = BAR_DEV,
			.count = sizeof(struct vnic_enet_config),},
		{ .type = RES_TYPE_EOL,}, // EOL must be last
	},
};

//
// Template for Ethernet device-specific region
//
static struct vnic_enet_config palo_nic_enet_config = {
	.flags = VENETF_RXCSUM | VENETF_TXCSUM | VENETF_TSO | VENETF_LRO,
	.wq_desc_count = 64,
	.rq_desc_count = 64,
	.mtu = 1500,
	//    .intr_timer = 83,
	.intr_timer_type = 0,
	.intr_mode = 0,
	.intr_timer_usec = 125,
};

struct enic_device {
	struct bar0_data bar0;
	struct vnic_enet_config enet_config;
	struct resource resources[BAR_MAX];
	struct platform_device pdev;
};

static void
init_enic_bar0(struct enic_device *edev,
	struct palo_nic *nic_info)
{
	struct resource_info *ri = &nic_info->resources;
	edev->bar0 = palo_nic_bar0;
	edev->bar0.resources[0].count = ri->wq_size;
	edev->bar0.resources[1].count = ri->rq_size;
	edev->bar0.resources[2].count = ri->cq_size;
	edev->bar0.resources[3].count = ri->in_size;

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((unsigned long) &edev->bar0, sizeof(edev->bar0));
#endif
}

static void
init_enic_config(struct enic_device *edev,
	struct palo_nic *nic_info)
{
	edev->enet_config = palo_nic_enet_config;
	strncpy(edev->enet_config.devname, nic_info->dev_name, DEVICE_NAME_LEN);
	edev->enet_config.mtu = nic_info->mtu;
	if (nic_info->flags & PALONIC_F_SHARE_INTR) {
		edev->enet_config.flags |= VENETF_SHARE_INTR;
	}
	printk(KERN_INFO "dev_name %s mtu %d flags %d\n", nic_info->dev_name,
			nic_info->mtu, edev->enet_config.flags);

#ifdef CONFIG_DMA_NONCOHERENT
	dma_cache_wback_inv((unsigned long) &edev->enet_config,
			sizeof(edev->enet_config));
#endif
}

static void
init_enic_resources(struct enic_device *edev,
	struct palo_nic *nic_info)
{
	struct resource *res = edev->resources;
	struct resource_info *ri = &nic_info->resources;
	u_int32_t start;

	res = &edev->resources[BAR_BAR0];
	res->start = CPHYSADDR((u_int32_t) & edev->bar0);
	res->end = res->start + sizeof(edev->bar0) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_WQ];
	res->start = WQ_HREGS_ADDR(ri->wq_base);
	res->end = WQ_HREGS_ADDR(ri->wq_base + ri->wq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_RQ];
	res->start = RQ_HREGS_ADDR(ri->rq_base);
	res->end = RQ_HREGS_ADDR(ri->rq_base + ri->rq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_CQ];
	res->start = CQ_HREGS_ADDR(ri->cq_base);
	res->end = CQ_HREGS_ADDR(ri->cq_base + ri->cq_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR];
	res->start = INTRCTRL_HREGS_ADDR(ri->in_base);
	res->end = INTRCTRL_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_TABLE];
	res->start = INTRMSI_HREGS_ADDR(ri->in_base);
	res->end = INTRMSI_HREGS_ADDR(ri->in_base + ri->in_size) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_INTR_PBA];
	intrpba_enc(&start, ri->in_base, ri->in_size, NOR2C);
	res->start = start;
	res->end = (start + roundup(ri->in_size, 32) / 8) - 1;
	res->flags = IORESOURCE_MEM;

	res = &edev->resources[BAR_DEV];
	res->start = CPHYSADDR((u_int32_t) & edev->enet_config);
	res->end = res->start + sizeof(edev->enet_config) - 1;
	res->flags = IORESOURCE_MEM;
}

static void
init_enic(struct enic_device *edev, struct palo_nic *nic_info)
{
	init_enic_bar0(edev, nic_info);
	init_enic_config(edev, nic_info);
	init_enic_resources(edev, nic_info);
	printk(KERN_INFO "Palo %s%d\n",
		edev->pdev.name, edev->pdev.id);
}

static struct enic_device *enic_platform_devices[PALONIC_MAX_COUNT];

static void
palonic_device_release(struct device *dev)
{
}

static int
palonic_get_next_index(void)
{
	int i;

	for (i = 0; i < PALONIC_MAX_COUNT; i++) {
		if (enic_platform_devices[i] == NULL) {
			return i;
		}
	}
	return -1;
}

static int
palonic_add_nic(struct palo_nic *req)
{
	int nic_id, ret;
	struct enic_device *user_device;
	nic_id = palonic_get_next_index();

	if (nic_id < 0) {
		return -EBUSY;
	}
	user_device = kzalloc(sizeof(struct enic_device), GFP_KERNEL);
	if (user_device == NULL) {
		return -ENOMEM;
	}
	user_device->pdev.id = PALONIC_PLATFORM_DEVICE_ID(nic_id);
	user_device->pdev.name = "enic";
	user_device->pdev.num_resources = ARRAY_SIZE(user_device->resources);
	user_device->pdev.resource = user_device->resources;
	user_device->pdev.dev.release = palonic_device_release;

	init_enic(user_device, req);
	if (user_device->enet_config.flags & VENETF_SHARE_INTR) {
		user_device->pdev.flags |= PLATFORM_DEV_SHARE_INTR;
	}

	printk(KERN_INFO "registering platform device %p \n",
		&(user_device->pdev));
	ret = platform_device_register(&user_device->pdev);
	if (ret) {
		kfree(user_device);
		return -EBUSY;
	}
	enic_platform_devices[nic_id] = user_device;

	return user_device->pdev.id;
}

static int
palonic_remove_nic(u_int8_t platform_device_id)
{
	int i;
	struct enic_device *user_device = NULL;

	printk(KERN_INFO "Removing palonic_with platform id %d\n",
		platform_device_id);
	for (i = 0; i < PALONIC_MAX_COUNT; i++) {
		if (enic_platform_devices[i] != NULL &&
			enic_platform_devices[i]->pdev.id == platform_device_id)
		{
			user_device = enic_platform_devices[i];
			break;
		}
	}
	if (user_device == NULL) {
		printk(KERN_INFO "ENIC platform device %d does not exist\n",
			platform_device_id);
		return -ENXIO;
	}
	platform_device_unregister(&(user_device->pdev));
	printk(KERN_INFO "platform device unregister completed\n");
	enic_platform_devices[i] = NULL;
	kfree(user_device);
	return 0;
}

static long
palonic_ioctl(struct file *filp,
              unsigned int cmd, unsigned long arg)
{
	int r;
	void __user *uarg;
	struct palo_nic nic_req;

	switch (cmd) {
	case PALONIC_IOC_ADD:
		uarg = (void __user *) arg;
		r = copy_from_user(&nic_req, uarg, sizeof(struct palo_nic));
		if (r == 0) {
			r = palonic_add_nic(&nic_req);
		} else {
			r = -EFAULT;
		}
		break;
	case PALONIC_IOC_REMOVE:
		r = palonic_remove_nic((u_int8_t) arg);
		break;
	default:
		r = -ENOTTY;
		break;
	}
	return r;
}

static int
palonic_open(struct inode *inode, struct file *filep)
{
	return 0;
}

static int
palonic_release(struct inode *inode, struct file *filep)
{
	return 0;
}

static struct file_operations palonic_fops = {
	.owner = THIS_MODULE,
	.open = palonic_open,
	.release = palonic_release,
	.unlocked_ioctl = palonic_ioctl,
};

static int __devinit
palonic_probe(struct device *dev)
{
	int r = register_chrdev(palonic_major, DRV_NAME, &palonic_fops);

	if (r < 0)
		return r;
	palonic_major = r;
	printk(KERN_INFO PFX "palonic major %d\n", palonic_major);
	return 0;
}

static int __devinit
palonic_remove(struct device *dev)
{
	return 0;
}

static struct device_driver palonic_driver = {
	.name = DRV_NAME,
	.bus = &platform_bus_type,
	.owner = THIS_MODULE,
	.probe = palonic_probe,
	.remove = palonic_remove,
};

static int __init
palonic_init(void)
{
	int i;

	for (i = 0; i < PALONIC_MAX_COUNT; i++) {
		enic_platform_devices[i] = NULL;
	}
	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO PFX "%s\n", DRV_COPYRIGHT);

	return driver_register(&palonic_driver);
}

static void __exit palonic_exit(void)
{
	driver_unregister(&palonic_driver);
}

module_init(palonic_init);
module_exit(palonic_exit);
