/*
 * Copyright 2007 Nuova Systems, Inc.  All rights reserved.
 * Copyright 2009 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: ppcie_evt.c 118590 2012-11-28 23:09:17Z brasmit2 $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/sched.h>

#include <palo/include/vnic_devcmd.h>

#include "pciemgr.h"
#include "pciesvc.h"

#include "ppcie.h"
#include "ppcie_evt.h"


int ppcie_have_events(void)
{
    struct ppcie_device_info *pdi = &ppcie_device_info;
    int have_events;
    unsigned long flags;

    spin_lock_irqsave(&pdi->events_lock, flags);
    have_events = !TAILQ_EMPTY(&pdi->events);
    spin_unlock_irqrestore(&pdi->events_lock, flags);
    return (have_events);
}

/*
 * 1000 is arbitrary.  We want to have enough that we don't
 * overflow before mcp starts up and host events are going to
 * be limited before mcp is up.  We don't want so many that we
 * use gobs of memory if mcp has died or something blocked
 * it for some reason.
 */
static int ppcie_evt_limit = 1000;

void ppcie_new_add_event(void *iev, int len, enum ppcie_event_type type,
                     u_int16_t id, u_int16_t bdf)
{
    struct ppcie_device_info *pdi = &ppcie_device_info;
    unsigned long flags;
    static int warned;
    struct ppcie_event *ev;

    if (!ppcie_evt_limit || atomic_read(&pdi->nevents) < ppcie_evt_limit) {
        ev = kzalloc(sizeof (*ev) + len, GFP_ATOMIC);
        if (ev) {
            ev->type = type;
            ev->id = id;
            ev->bdf = bdf;
            ev->len = len;
            if (len) memcpy(&ev->u, iev, len);

            spin_lock_irqsave(&pdi->events_lock, flags);
            TAILQ_INSERT_TAIL(&pdi->events, ev, list);
            atomic_inc(&pdi->nevents);
            spin_unlock_irqrestore(&pdi->events_lock, flags);
            wake_up_interruptible(&pdi->waitq);
        } else {
            printk(KERN_WARNING PFX "no mem for event type %d\n", type);
        }
    } else {
        if (!warned) {
            printk(KERN_WARNING PFX "add_event limit %d exceeded\n",
                    ppcie_evt_limit);
            warned = 1;
        }
    }
}

static struct ppcie_event *ppcie_get_event(void)
{
    struct ppcie_device_info *pdi = &ppcie_device_info;
    struct ppcie_event *ev = NULL;
    unsigned long flags;

    spin_lock_irqsave(&pdi->events_lock, flags);
    if (!TAILQ_EMPTY(&pdi->events)) {
        ev = TAILQ_FIRST(&pdi->events);
        TAILQ_REMOVE(&pdi->events, ev, list);
        atomic_dec(&pdi->nevents);
    }
    spin_unlock_irqrestore(&pdi->events_lock, flags);
    return (ev);
}


int ppcie_copy_events_to_user(struct ppcie_event __user *uev)
{

    int r, nev;

    if (!access_ok(VERIFY_WRITE, uev,
               sizeof (struct ppcie_event) * PPCIE_GET_EVENTS_MAX)) {
            return (-EFAULT);
    }

    for (nev = 0; nev < PPCIE_GET_EVENTS_MAX; nev++, uev++) {
        struct ppcie_event *ev = ppcie_get_event();
        if (!ev) {
            break;
        }
        r = copy_to_user(uev, ev, sizeof (struct ppcie_event) -
                             sizeof(ev->list));
        BUG_ON(r);
        kfree(ev);
    }
    return (nev);
}

void ppcie_evt_bus_chg(hostid_t hostid, u_int8_t busno)
{
    struct ppcie_event_bus_chg bus_chg;

    bus_chg.busno = busno;

    ppcie_new_add_event(&bus_chg, sizeof (bus_chg), PPCIE_EVT_BUS_CHG,
                        hostid, 0);
}

static void ppcie_evt_idevt(const u_int8_t id, const int evtype, u_int32_t iter)
{
    struct ppcie_event_host_iter host_iter;
    host_iter.host_iter = iter;
    ppcie_new_add_event(&host_iter, sizeof (host_iter), evtype, id, 0);
}

void ppcie_evt_hostup(const hostid_t hostid, u_int32_t iter)
{
    ppcie_evt_idevt(hostid, PPCIE_EVT_HOSTUP, iter);
}

void ppcie_evt_hostdn(const hostid_t hostid, u_int32_t iter)
{
    ppcie_evt_idevt(hostid, PPCIE_EVT_HOSTDN, iter);
}

void ppcie_evt_linkerr(const hostid_t hostid,
                       const u8 errtype, const u8 errcode)
{
    struct ppcie_event_linkerr linkerr;

    memset(&linkerr, 0, sizeof (linkerr));
    linkerr.errtype = errtype;
    linkerr.errcode = errcode;
    ppcie_new_add_event(&linkerr, sizeof (linkerr), PPCIE_EVT_LINKERR,
                        hostid, 0);
}

void __devinit ppcie_evt_init(void)
{
    struct ppcie_device_info *pdi = &ppcie_device_info;

    TAILQ_INIT(&pdi->events);
    spin_lock_init(&pdi->events_lock);
    init_waitqueue_head(&pdi->waitq);
}

int ppcie_delete_devcmds(hostid_t hostid, u16 bdf)
{
    struct ppcie_device_info *pdi = &ppcie_device_info;
    struct ppcie_event *ev;
    struct ppcie_event *tmp;
    unsigned long flags;
    int nfound = 0;

    spin_lock_irqsave(&pdi->events_lock, flags);
    TAILQ_FOREACH_SAFE(ev, &pdi->events, list, tmp) {
        if (ev->type == PPCIE_EVT_DEVCMD &&
            ev->id == hostid &&
            ev->bdf == bdf) {
            TAILQ_REMOVE(&pdi->events, ev, list);
            atomic_dec(&pdi->nevents);
            nfound++;
        }
    }
    spin_unlock_irqrestore(&pdi->events_lock, flags);
    return nfound;
}
