/*
 * Copyright 2012 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: ppcie_host.c 146652 2013-10-09 22:50:27Z sborikar $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/fs.h>

#include <palo/include/palo_defs.h>
#include <palo/include/pcie_mregs.h>
#include <palo/fw/mcpu/include/palo_conf.h>
#include <palo/fw/mcpu/include/board_conf.h>
#include "ppcie.h"
#include "ppcie_host.h"

#define NHOSTS PALO_MAX_HOSTS
#define NHIDS  PALO_MAX_HIDS
#define NSHIDS 2

// A hidmap entry describes which hid interface is active
// and how many lanes are connected.
typedef struct hidmap_s {
	hid_t    hid;
	u_int32_t width;
	u_int32_t flags;
	u_int8_t valid;
} hidmap_t;

// A hostmap entry maps to a {hid,shid} pair.
typedef struct hostmap_s {
	hid_t    hid;
	shid_t   shid;
	u_int8_t valid;
} hostmap_t;

typedef struct hostinfo_s {
	u_int32_t nhids;
	u_int32_t nhosts;
	hidmap_t hidmap[NHIDS];
	hostmap_t hostmap[NHOSTS];
	hostid_t hostrevmap[NHIDS][NSHIDS];
} hostinfo_t;

static hostinfo_t hostinfo;

////////////////////////////////////////////////////////////////
// Module APIs
////////////////////////////////////////////////////////////////

hostid_t ppcie_hostid_max(void)
{
	return (hostinfo.nhosts);
}

u_int8_t ppcie_hostid_hid(const hostid_t hostid)
{
	BUG_ON(hostid >= hostinfo.nhosts);
	return (hostinfo.hostmap[hostid].hid);
}

u_int8_t ppcie_hostid_shid(const hostid_t hostid)
{
	BUG_ON(hostid >= hostinfo.nhosts);
	return (hostinfo.hostmap[hostid].shid);
}

u_int8_t ppcie_hid_pp(const hid_t hid)
{
	return ((hid >> 2) & 0x3);
}

u_int8_t ppcie_hid_pps(const hid_t hid)
{
	return (hid & 0x3);
}

u_int8_t ppcie_hostid_pp(const hostid_t hostid)
{
	return (ppcie_hid_pp(ppcie_hostid_hid(hostid)));
}

u_int8_t ppcie_hostid_pps(const hostid_t hostid)
{
	return (ppcie_hid_pps(ppcie_hostid_hid(hostid)));
}

hostid_t ppcie_hostid_make(const hid_t hid, const shid_t shid)
{
	BUG_ON(hid >= NHIDS);
	BUG_ON(shid >= NSHIDS);
	return (hostinfo.hostrevmap[hid][shid]);
}

void ppcie_foreach_hid(ppcie_hostcb_t cb, void *cbarg)
{
	hostinfo_t *hi = &hostinfo;
	int i;

	for (i = 0; i < hi->nhids; i++) {
		cb(hi->hidmap[i].hid, cbarg);
	}
}

//
// reqidx - Indirect transaction request index.  Cruz has 16 indirect
//     transaction request info buffers and it is up to software how
//     to use them to service the 16 hosts.  For now, use request index
//     N for physical "hid" N.
//
u_int8_t ppcie_hostid_reqidx(const hostid_t hostid)
{
	return (ppcie_hostid_hid(hostid));
}

hostid_t ppcie_reqidx_hostid(const u_int8_t reqidx)
{
	return (ppcie_hostid_make(reqidx, 0));
}

void ppcie_get_linkparams(const hid_t hid, u32 *width, u32 *flags)
{
	int i;

	BUG_ON(hid >= NHIDS);
	for (i = 0; i < hostinfo.nhids; i++) {
		hidmap_t *hm = &hostinfo.hidmap[i];
		if (hm->hid == hid) {
			if (width) *width = hm->width;
			if (flags) *flags = hm->flags;
			return;
		}
	}
	// inactive link
	if (width) *width = 0;
	if (flags) *flags = 0;

}

void ppcie_set_linkparams(const hid_t hid, const u32 width, const u32 flags)
{
	int i;

	BUG_ON(hid >= NHIDS);
	for (i = 0; i < hostinfo.nhids; i++) {
		hidmap_t *hm = &hostinfo.hidmap[i];
		if (hm->hid == hid) {
			hm->width = width;
			hm->flags = flags;
			return;
		}
	}
}

int ppcie_hostid_get_hostinfo(struct ppcie_hostinfo *ppciehi)
{
	hostinfo_t *hi = &hostinfo;
	int i;

	memset(ppciehi, 0, sizeof (struct ppcie_hostinfo));
	ppciehi->nhids = hi->nhids;
	for (i = 0; i < hi->nhids; i++) {
		ppciehi->hidmap[i].hid = hi->hidmap[i].hid;
		ppciehi->hidmap[i].width = hi->hidmap[i].width;
	}
	ppciehi->nhosts = hi->nhosts;
	for (i = 0; i < hi->nhosts; i++) {
		ppciehi->hostmap[i].hid = hi->hostmap[i].hid;
		ppciehi->hostmap[i].shid = hi->hostmap[i].shid;
	}
	return (0);
}

////////////////////////////////////////////////////////////////
// Initialization routines
////////////////////////////////////////////////////////////////

//
// hostid:hid mapped 1:1.  Initialized at runtime instead of static
// initializers to handle different PALO_MAX_HOSTS.  This is the
// case used for single-host platforms, e.g. Palo/Sereno.
//
static __devinit void hostinfo_init_default(hostinfo_t *hi)
{
	hid_t h;

	memset(hi, 0, sizeof (hostinfo_t));
	hi->nhids = PALO_MAX_HIDS;
	for (h = 0; h < PALO_MAX_HIDS; h++) {
		hi->hidmap[h].hid = h;
	}

	hi->nhosts = PALO_MAX_HOSTS;
	for (h = 0; h < PALO_MAX_HOSTS; h++) {
		hi->hostmap[h].hid = h;
	}
}

#if defined CONFIG_BASE_CRUZ
static __devinit int
hostinfo_init_hidmap(hidmap_t *hm, hid_t hid, int width, char *s)
{
	if (width < 0 || width > 8) {
		printk(KERN_WARNING PFX "invalid link width %d\n", width);
		return (-1);
	}
	hm->width = width;
	hm->hid = hid;
	for (; s && *s; s++) {
		switch (*s) {
		case 'T': hm->flags |= PCIE_LINKF_TX_FLIP; break;
		case 'R': hm->flags |= PCIE_LINKF_RX_FLIP; break;
		default:
			printk(KERN_WARNING PFX
			       "invalid linkwidths char %c\n", *s);
			return (-1);
		}
	}
	hm->valid++;
	return (0);
}

//
// <width>[TR][=hostid[:hostid]...]
//
static __devinit int hostinfo_init_hidspec(hostinfo_t *hi, hid_t hid, char *s)
{
	int hmidx = hi->nhids;
	hidmap_t *hm = &hi->hidmap[hmidx];
	char *equal = strchr(s, '=');
	char *trailing;
	int width, hostid;

	// First process the width param up to the '=',
	// including any trailing chars saved in "trailing".
	if (s == NULL || *s == '\0') {
		return (0);
	}
	width = simple_strtoul(s, &trailing, 0);
	if (equal != NULL) {
		 *equal = '\0';
	}
	if (hostinfo_init_hidmap(hm, hid, width, trailing) < 0) {
		return -1;
	}
	hi->nhids++;

	// If we have '=' process the "=hostid[:hostid]..." params
	// else assume next available hostid.
	if (equal) {
		char *hstr;
		shid_t shid = 0;

		s = equal + 1;
		while ((hstr = strsep(&s, ":")) != NULL && shid < NSHIDS) {
			hostid = simple_strtoul(hstr, NULL, 0);
			hi->hostmap[hostid].hid = hid;
			hi->hostmap[hostid].shid = shid;
			hi->hostmap[hostid].valid++;
			shid++;
			if (hi->nhosts < hostid + 1)
				hi->nhosts = hostid + 1;
		}
	} else {
		hostid = hmidx; // next available
		hi->hostmap[hostid].hid = hid;
		hi->hostmap[hostid].shid = 0;
		hi->hostmap[hostid].valid++;
		if (hi->nhosts < hostid + 1)
			hi->nhosts = hostid + 1;
	}
	return (0);
}

//
// Handle comma-separated tokens of this form:
//     <hid>[TR][=hostid[:hostid]...]
// e.g.
//     4=0,0,4=1,0,4=2,0,4=3,0,4=4,0,4=5,0,4=6,0,4=7,0
//     4=0,,4=1,,4=2,,4=3,,4=4,,4=5,,4=6,,4=7,
//     8=0:1,,,8=2:3
//
static __devinit int hostinfo_init_string_normal(hostinfo_t *hi, char *str)
{
	hid_t hid;
	int   r = 0;
	char localstr[128], *lstr = localstr;

	memset(hi, 0, sizeof (hostinfo_t));
	// save a local copy of str for strsep() to step on.
	strncpy(lstr, str, sizeof (localstr));
	for (hid = 0; hid < NHIDS && lstr && r == 0; hid++) {
		char *hidstr = strsep(&lstr, ",");
		r = hostinfo_init_hidspec(hi, hid, hidstr);
	}
	return (r);
}

//
// Handle simple linkwidth strings such as:
//     80008000         -- 2 x8 links
//     4040404040404040 -- 8 x4 links
//     2222222222222222 -- 16 x2 links
//
static __devinit int hostinfo_init_string_basic(hostinfo_t *hi, char *str)
{
	char  buf[8];
	hid_t hid;
	int r = 0;

	memset(hi, 0, sizeof (hostinfo_t));
	for (hid = 0; hid < NHIDS && *str && r == 0; str++, hid++) {
		buf[0] = *str;
		buf[1] = '\0';
		r = hostinfo_init_hidspec(hi, hid, buf);
	}
	return (r);
}

//
// Validate the resulting hostinfo from str is correct.
// We expect each element to be mentioned once and only once.
//
static __devinit int hostinfo_validate(hostinfo_t *hi, char *str)
{
	int h;

	for (h = 0; h < hi->nhids; h++) {
		hidmap_t *hidm = &hi->hidmap[h];
		if (hidm->valid == 1) continue; // expect valid == 1

		// valid is 0 means the spec string skipped it
		// valid > 1 means the spec string mentioned it twice
		printk(KERN_ERR PFX "bad pcie_linkwidths %s\n", str);
		printk(KERN_ERR PFX "  hidmap%d: hid %d valid %d\n",
		       h, hidm->hid, hidm->valid);
		return (-1);
	}
	for (h = 0; h < hi->nhosts; h++) {
		hostmap_t *hostm = &hi->hostmap[h];
		if (hostm->valid == 1) continue; // expect valid == 1

		// valid is 0 means the spec string skipped it
		// valid > 1 means the spec string mentioned it twice
		printk(KERN_ERR PFX "bad pcie_linkwidths %s\n", str);
		printk(KERN_ERR PFX "  hostmap%d: hid %d shid %d valid %d\n",
		       h, hostm->hid, hostm->shid, hostm->valid);
		return (-1);
	}
	return (0);
}

static __devinit void hostinfo_display(hostinfo_t *hi, char *str)
{
	hid_t hid;
        int   h;

	printk(KERN_INFO PFX "pcie_linkwidths = %s\n", str);
	for (h = 0; h < hi->nhids; h++) {
		hidmap_t *hidmap = &hi->hidmap[h];
		hid = hidmap->hid;
		printk(KERN_INFO PFX "%d: hid %d pp %d pps %d width x%d%s%s\n",
		       h, hid, ppcie_hid_pp(hid), ppcie_hid_pps(hid),
		       hidmap->width,
		       hidmap->flags & PCIE_LINKF_TX_FLIP ? " tx_flip" : "",
		       hidmap->flags & PCIE_LINKF_RX_FLIP ? " rx_flip" : "");
	}
	for (h = 0; h < hi->nhosts; h++) {
		printk(KERN_INFO PFX "host %d: hid %d shid %d\n",
		       h, hi->hostmap[h].hid, hi->hostmap[h].shid);
	}
}

static __devinit int hostinfo_init_string(hostinfo_t *hi, char *str)
{
	int r = 0;

	if (strchr(str, ',') != NULL || strchr(str, '=') != NULL) {
		r = hostinfo_init_string_normal(hi, str);
	} else {
		r = hostinfo_init_string_basic(hi, str);
	}
	if (r == 0) {
		// validate what we parsed from the string
		r = hostinfo_validate(hi, str);
	}
	if (r == 0) {
		// display what we were told
		hostinfo_display(hi, str);
	}
	return (r);
}
#endif // CRUZ

static __devinit void hostinfo_init(hostinfo_t *hi)
{
#if defined CONFIG_BASE_CRUZ
	char *val;
	extern char *prom_getenv(const char *s);

        // Cruz asic is flexible, allow custom widths via envar.
	if (palo_conf_is_asic()) {
		val = prom_getenv("pcie_linkwidths");
		if (val && hostinfo_init_string(hi, val) == 0)
			return;
	}

	val = board_conf_pcie_linkwidths();
	if (val && val[0] != '\0' && hostinfo_init_string(hi, val) == 0)
		return;
#endif // CRUZ

	hostinfo_init_default(hi);
}

// build reverse map
static __devinit void hostrevmap_init(hostinfo_t *hi)
{
	hostid_t hostid;

	for (hostid = 0; hostid < ppcie_hostid_max(); hostid++) {
		const u8 hid = ppcie_hostid_hid(hostid);
		const u8 shid = ppcie_hostid_shid(hostid);
		hi->hostrevmap[hid][shid] = hostid;
	}
}

void __devinit ppcie_host_init(void)
{
	hostinfo_t *hi = &hostinfo;

	hostinfo_init(hi);
	hostrevmap_init(hi);
}
