/*
 * Copyright 2007 Nuova Systems, Inc.  All rights reserved.
 * Copyright 2009 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: ppcie_ioctl.c 122077 2013-01-25 22:19:10Z punsheno $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>

#include <palo/fw/common/include/ecpu_res.h>
#include <palo/include/intrc_regs.h>
#include <palo/include/pcie_mregs.h>
#include <palo/include/pcie_regs_defs.h>
#include <palo/fw/mcpu/include/palo_conf.h>

#include "pciemgr.h"
#include "pciesvc.h"
#include "pm_private.h"
#include "ps_sys.h"
#include "ppcie.h"
#include "ppcie_pcie.h"
#include "ppcie_evt.h"
#include "ppcie_cfg.h"
#include "ppcie_bar.h"
#include "ppcie_trn.h"
#include "ppcie_host.h"
#include "ppcie_bdft.h"
#include "ps_stats.h"
#include "ppcie_vstate.h"
// #include "ppcie_event.h"
#if defined CONFIG_BASE_CRUZ
#include "ppcie_umt.h"
#include "ppcie_urt.h"
#endif // CONFIG_BASE_CRUZ

/*
 * for  ppcie_rc_cfg_mode()
 */
#include <palo/include/cruz/rc_mregs.h>

int ppcie_vnic_load(pmvnich_t pmvnich)
{
    int r;
    pmvnic_t *pmvnic;


    pmvnic = pmvnich_to_pmvnic(pmvnich);
    if (pmvnic == NULL) {
        return (-1);
    }

    r = ppcie_cfg_init_pmvnic(pmvnic);
    if (r) {
        goto error_out;
    }

    r = ppcie_bars_load(pmvnic);
    if (r) {
        goto error_out;
    }

    ps_vnic_add_hlink(pmvnich, pmvnic);
    return (0);

error_out:
    printk("%s: vnich %d loaded failed with err 0x%x\n", __func__, pmvnich, r);
    return (r);
}

int ppcie_vnic_unload(pmvnich_t pmvnich)
{
    pmvnic_t *pmvnic;

    pmvnic = pmvnich_to_pmvnic(pmvnich);
    if (pmvnic == NULL) {
        return (-1);
    }

    ppcie_cfg_unload(pmcfgh_to_pmcfg(pmvnic->pmcfgh));
    ppcie_bars_unload(pmvnic, 1);
    ppcie_bdft_unload(pmvnic->vnicid);
    ps_vnic_del_hlink(pmvnich, pmvnic);
    return (0);
}

static int ppcie_ioctl_ready(void *uarg)
{
    struct ppcie_host_ready host_ready;
    u8 pp;
    u8 pps;
    u_int32_t hostid;

    if (copy_from_user(&host_ready, uarg, sizeof (host_ready))) {
        return -EFAULT;
    }
    hostid = host_ready.hostid;
    if (hostid >= ppcie_hostid_max()) {
        printk(KERN_ERR "hostid %d >= hostid_max %d\n", hostid,
                                                        ppcie_hostid_max());
        return -EINVAL;
    }

    pp  = ppcie_hostid_pp(hostid);
    pps = ppcie_hostid_pps(hostid);
    ppcie_pcie_vnics_ready(hostid, pp, pps, host_ready.hostdn_iter);
    return (0);
}

static int ppcie_ioctl_chip_reset(void)
{
    ppcie_pcie_chip_reset();
    return (0);
}

static int ppcie_ioctl_get_events(struct ppcie_event __user *uev)
{
    return (ppcie_copy_events_to_user(uev));
}

static int ppcie_ioctl_get_hostinfo(void *uarg)
{
    struct ppcie_hostinfo ppciehi;
    int r;

    r = ppcie_hostid_get_hostinfo(&ppciehi);
    if (r == 0) {
        r = copy_to_user(uarg, &ppciehi, sizeof (struct ppcie_hostinfo));
    }
    return (r);
}

static int ppcie_ioctl_active_standby(unsigned long arg)
{
    int r = 0;
#ifdef CONFIG_BASE_CRUZ
    pmvnic_t *pmvnic;
    pmvnich_t pmvnich;
    int do_active;
    extern int actstandby;

    if ((arg >> 16) & 1 )
        do_active = 1;
    else
        do_active = 0;
    pmvnich = (pmvnich_t)(arg & 0xffff);

    pmvnic = pmvnich_to_pmvnic(pmvnich);
    if (pmvnic == NULL) {
        return (-ENODEV);
    }

    actstandby = 1;

    if (do_active)
        pmvnic->dev_flags &= ~PMVNIC_DEVF_PCISTANDBY;
    else
        pmvnic->dev_flags |= PMVNIC_DEVF_PCISTANDBY;

    ppcie_bdft_load(pmvnic->vnicid, pmvnic->hid, pmvnic->shid,
                    bdf_to_bus(pmvnic->bdf),
                    bdf_to_dev(pmvnic->bdf),
                    bdf_to_fnc(pmvnic->bdf),
                    do_active);

    ppcie_umt_cmt_load(pmcfgh_to_pmcfg(pmvnic->pmcfgh), pmvnic->pmcfgh);
    ppcie_active_standby_bars_reload(pmvnic);

    actstandby = 0;
#endif
    return (r);
}


static long ppcie_ioctl_crt_accmode_reload(unsigned long arg)
{
    pmcfgh_t pmcfgh = (pmcfgh_t)(arg & 0xFFFF);
    return ppcie_cfg_crt_load(pmcfgh_to_pmcfg(pmcfgh), pmcfgh);
}

static int ppcie_ioctl_delete_devcmds(u32 arg)
{
    hostid_t hostid = arg >> 16;
    u16 bdf = arg & 0xffff;

    if (hostid >= ppcie_hostid_max()) {
        printk(KERN_ERR "hostid %d not in range of 0 to %d\n",
                        hostid, ppcie_hostid_max() -1);
        return -EINVAL;
    }
    return ppcie_delete_devcmds(hostid, bdf);
}

long ppcie_ioctl(struct file *filp,
                unsigned int cmd, unsigned long arg)
{
    void __user *uarg = (void __user *)arg;
    int r;

    ppcie_stats.ioctl++;

    switch (cmd) {

    case PPCIE_IOC_LOAD:
        r = ppcie_vnic_load((pmvnich_t)arg);
        break;

    case PPCIE_IOC_UNLOAD:
        r = ppcie_vnic_unload((pmvnich_t)arg);
        break;

    case PPCIE_IOC_READY:
        r = ppcie_ioctl_ready(uarg);
        break;

    case PPCIE_IOC_CHIP_RESET:
        r = ppcie_ioctl_chip_reset();
        break;

    case PPCIE_IOC_GET_EVENTS:
        r = ppcie_ioctl_get_events(uarg);
        break;

    case PPCIE_IOC_GET_HOSTINFO:
        r = ppcie_ioctl_get_hostinfo(uarg);
        break;

    case PPCIE_IOC_ACTIVE_STANDBY:
        r = ppcie_ioctl_active_standby(arg);
        break;

    case PPCIE_IOC_CRT_RELOAD:
        r = ppcie_ioctl_crt_accmode_reload(arg);
        break;

    case PPCIE_IOC_DELETE_DEVCMDS:
        r = ppcie_ioctl_delete_devcmds(arg);
        break;

    case PPCIE_IOC_VNIC_STATE_CFG:
        r = ppcie_ioctl_vnic_st_cfg(uarg);
        break;

    default:
        printk("UNKNOWN ioctl cmd 0x%x, arg 0x%lx\n", cmd,  arg);
        r = -ENOTTY;
        break;
    }

    return (r);
}
