/*
 * Copyright 2007 Nuova Systems, Inc.  All rights reserved.
 * Copyright 2009-2015 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: ppcie_pcie.c 166654 2014-03-29 00:30:12Z brasmit2 $"

#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/notifier.h>
#include <linux/pci_regs.h>

#include <paloint.h>
#include <palo/fw/mcpu/include/palo_conf.h>
#include <palo/fw/mcpu/include/board_conf.h>

#include <palo/fw/common/include/ecpu_res.h>
#include <palo/include/pcie_mregs.h>
#include <palo/include/misc_mregs.h>
#include <palo/include/net_mregs.h>
#include <palo/include/synopsis/dw_wdt.h>

#include "pciemgr.h"
#include "pciesvc.h"
#include "pmimpl.h"
#include "ps_sys.h"
#include "ppcie_evt.h"
#include "ps_stats.h"
#include "ppcie.h"
#include "ppcie_pcie.h"
#include "ppcie_host.h"
#include "ppcie_bdft.h"
#include "ppcie_evt.h"

#define memory_barrier()    do { asm volatile("sync" ::: "memory"); } while (0)

//
// This module implements a collection of utilities to
// manage the pcie link.
//
// Some interesting register bits:
//
//     LTSSM_ENABLE - Link Training Something State Machine
//         When set, host sees pcie link up.  When clear, no pcie link to host.
//     FORCE_RESET_OK - When set, pcie bus resets cause Palo reset.
//         When clear, pcie resets close the barrier but don't reset Palo.
//     RETRY_EN - Config retry enable.  When set, all host config cycles
//         generate a CRS (Config Retry Status) response.  When clear,
//         host config cycles are allowed through to be processed by Palo.
//     BARRIER - The BARRIER between the sub-block of Palo in the pcie
//         clock domain, and the other sub-blocks in the Palo clock domain.
//         When the BARRIER is closed, cpu accesses to any registers in the
//         pcie clock domain (those with the PCIEP_MREGS_ prefix) return
//         failure which the cpu interprets as a DBE (Data Bus Error)
//         exception.  The BARRIER closes automatically when a pcie bus reset
//         arrives that does NOT reset all of Palo.
//
// Sequence of events:
//     power on/reset - At reset
//         LTSSM_ENABLE = 0;		// no pcie link to host
//         RETRY_EN = 1;		// host config retries (CRS)
//         FORCE_RESET_OK = 1;		// we will reset if host resets
//         BARRIER = 0;			// barrier closed
//     uboot brings link up (LTSSM_ENABLE)
//         BARRIER = 1;			// bring up barrier to access LTSSM
//         LTSSM_ENABLE = 1;		// bring up pcie link to host
//
//         allows BIOS to detect pcie link for our slot
//         otherwise it'll skip our slot.
//         (also gives us snake-eyes link lights on fpga boards)
//     linux kernel boots, ppcie driver initializes, registers for the
//         pcie mac intr which fires when the pcie mac resets.
//         FORCE_RESET_OK = 0;		// ignore host resets
//     when pcie_mac_reset() {
//         BARRIER = 1;			// bring up barrier to access LTSSM etc
//         LTSSM_ENABLE = 1;		// bring up pcie link to host
//         if (vnics_are_ready)
//             pcie_create_vnics();     // restore vnic regs
//     }
//
//     mcp loads the vnic configuration, then
//         vnics_are_ready = 1;
//         RETRY_EN = 0;                // config retries off, host scans bus
//
//     host sets the Palo pcie bridge secondary bus number during scan,
//     which fires the SEC_BUS_NUM_CHG intr, then:
//         FORCE_RESET_OK = 1;		// we will reset if host resets
//
// Clearing PCIE_FORCE_RESET_OK means we will continue to run
// when the host issues a PCIe reset.  The barrier will close
// on PCIe reset.  Be sure to open barrier on reset.
// Setting PCIE_FORCE_RESET_OK causes us to reset when the host resets.
//
// Secondary bus number change indicates the host has begun scanning the
// pcie bus. This is a good indication that we are done with the reset flurry
// at startup, so arrange to reset with the host now.
// Read type 1 config space word that contains secondary bus. Secondary bus
// is byte 2
//

typedef enum pciest_e {
    PCIEST_MACRST,	// MAC is in reset
    PCIEST_DOWN,	// MAC is up reporting link down
    PCIEST_LINKUP,	// Link is up
    PCIEST_UP,		// Link is active, secondary bus set
    PCIEST_FAULT,	// Link had a fault condition

    PCIEST_MAX
} pciest_t;

static const char *pciest_names[PCIEST_MAX] = {
#define PCIEST_NAME(ST) \
    [PCIEST_##ST] = #ST
    PCIEST_NAME(MACRST),
    PCIEST_NAME(DOWN),
    PCIEST_NAME(LINKUP),
    PCIEST_NAME(UP),
    PCIEST_NAME(FAULT),
};

typedef enum pcieev_e {
    PCIEEV_MACDN,	// MAC went down
    PCIEEV_MACUP,	// MAC came up
    PCIEEV_LINKDN,	// Link went down
    PCIEEV_LINKUP,	// Link came up
    PCIEEV_BUS_CHG,	// Secondary bus got set
    PCIEEV_SET_CFG,	// Request to change config

    PCIEEV_MAX
} pcieev_t;

static const char *pcieev_names[PCIEEV_MAX] = {
#define PCIEEV_NAME(EV) \
    [PCIEEV_##EV] = #EV
    PCIEEV_NAME(MACDN),
    PCIEEV_NAME(MACUP),
    PCIEEV_NAME(LINKDN),
    PCIEEV_NAME(LINKUP),
    PCIEEV_NAME(BUS_CHG),
    PCIEEV_NAME(SET_CFG),
};

typedef struct pcie_hid_info_s {
    spinlock_t lock;            // fsm lock
    char name[8];               // "hidX"
    char *fault_reason;         // fault reason string
    char *last_fault_reason;    // last fault reason string
    u32 nfaults;                // count of faults
    u32 maxpllpolls;            // largest pll polls
    u32 ltssm_st_changed;       // "recent" ltssm_st_changed intrs
    u32 ltssm_ms;               // time (in ms) of last ltssm_st_changed intr
    hid_t hid;                  // hid for us
    u_int32_t flags;            // PCIEF_* flags
    u_int32_t hostdn_iter;      // hostdn iter
    pciest_t state;             // current state PCIEST_*
    pcieev_t event;             // last event PCIEEV_*
    struct kobject kobj;        // for sysfs
    trace_log_t log;            // tracing
    pcie_hid_stats_t tstats;    // total hid stats
    pcie_hid_int_err_stats_t int_err_stats; // int err stats
    pcie_hid_int_err_stats_t tint_err_stats; // total int err stats
} pcie_hid_info_t;

#define PCIEF_RESET		0x01	// pcie interface in reset
#define PCIEF_CFG_CRS           0x02    // configured for Config Retries
#define PCIEF_LOOPBACK          0x04    // link is in loopback state
#define PCIEF_FAULT             0x08    // link had a fault

#define tracep(p, ev, fmt, ...) \
    cmn_trace(&(p)->log, ev, fmt, ##__VA_ARGS__)

#define PCIETR_FSM      0x01
#define PCIETR_INFO     0x02

// UCSM discovery setup, decommission and recommission scenario
// pll does not lock at first place, retry of pll check after
// serdes reset, locks the pll.
#define PPS_PLL_CHECK_RETRY_COUNT	1

static pcie_hid_info_t pcie_hid_info[PALO_MAX_HIDS];
static u32 pcie_ntrace = 1024;
static u32 hostid_max;
static spinlock_t pcie_lock;
// ppcie_stats_t ppcie_stats;
static u8 pcie_reset_on_macup;
static u8 pcie_sw_reset_with_host;
static u32 pcie_ltssm_ms = 1;           // ltssm_st_changed intr time
static u32 pcie_ltssm_count = 100;      // ltssm_st_changed intr threshold
static u32 pcie_set_linkwidth_while_up; // allow set_linkwidth even if link up

static void
pciep_debug_port_disable(void)
{
#if defined CONFIG_BASE_SERENO
    u_int32_t val = readreg(PCIEP_MREGS_CFG_P_DEBUG_PORT);

    val &= ~0x1;     // bit 0 is enable
    writereg(PCIEP_MREGS_CFG_P_DEBUG_PORT, val);
#endif
}

static void
pcie_asic_reset(void)
{
    writereg(DW_WDT_CR(0), DW_WDT_CR_WDT_EN); /* Enable, RESET (without NMI) */
    writereg(DW_WDT_TORR(0), 0); /* Timeout asap */
    writereg(DW_WDT_CRR(0), DW_WDT_CRR_MAGIC); /* kick */
    for (;;) {
        // Wait for the chip reset
    }
}

/*
 * Caller is requesting a reset on the next host reboot
 * (probably after a firmware update on a rack server).
 * We'll request the link fsm perform a reset on the next
 * macup event.
 */
void
ppcie_pcie_chip_reset(void)
{
    pcie_reset_on_macup = 1;
}

static inline void
pcie_reset_with_host(u8 on)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    writereg(MISC_MREGS_PCIE_FORCE_RESET_OK, on);
#endif /* PALO || SERENO */

    // If hardware doesn't support FORCE_RESET_OK our link fsm
    // will reset on macdn.
	pcie_sw_reset_with_host = on;
}

static void
pcie_log_reset_state(void)
{
	char *en = "enabled";
	char *dis = "disabled";
#if defined CONFIG_BASE_SERENO
	u8 force;

	force = readreg(MISC_MREGS_PCIE_FORCE_RESET_OK);
	printk(KERN_DEBUG PFX "hardware reset with host %s\n", force ? en : dis);
#endif /* SERENO */
	printk(KERN_DEBUG PFX "software reset with host %s\n",
		pcie_sw_reset_with_host ? en : dis);
	printk(KERN_DEBUG PFX "reset on mac up %s, "
		"card mode = %d, power level is %s\n",
		pcie_reset_on_macup ? en : dis,
		board_conf_cisco_card_mode(),
		board_conf_is_lowpower() ? "low" : "full");
}

//
// Open the pcie barrier.  Barrier must be open before accessing
// any registers in the PCIEP_MREGS_ domain, otherwise we will
// get a data bus error (DBE) response from those registers while
// barrier is closed.
//
static inline void
pcie_barrier_open(const u8 pp, const u8 pps)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    // write to _OPEN reg opens pcie barrier
    writereg(PCIEC_MREGS_WRO_BARRIER_OPEN, 0x1);
    // read to be sure write above has landed...
    //     (void)readreg(PCIEC_MREGS_WRO_BARRIER_OPEN);
    // ...but better to read STAtus register rather than
    // the WRite-Only reg we just wrote.
    (void)readreg(PCIEC_MREGS_STA_BARRIER);
#endif /* PALO || SERENO */
}

static inline void
pcie_barrier_gate_open(const u8 pp, const u8 pps)
{
#if defined CONFIG_BASE_CRUZ
    writereg(PPSC_MREGS_(WRO_BARRIER_OPEN, pp, pps), 1);
    (void)readreg(PPSC_MREGS_(STA_BARRIER, pp, pps));
#endif /* CRUZ */
}

static inline void
pcie_barrier_gate_close(const u8 pp, const u8 pps)
{
#if defined CONFIG_BASE_CRUZ
    writereg(PPSC_MREGS_(WRO_BARRIER_CLOSE, pp, pps), 1);
    (void)readreg(PPSC_MREGS_(STA_BARRIER, pp, pps));
#endif /* CRUZ */
}

static void
pcie_close_and_idle(const u8 pp, const u8 pps)
{
    pcie_barrier_gate_close(pp, pps);
    // XXX check STA_BUSY and BARRIER_MARKER
}

//
// Bring up the pcie link.  Should do this soon after mac is
// out of reset, and after pcie barrier is open.  Once the
// link is up, the host will get Config Retry Status (CRS)
// for any config space accesses.
// (CRS will stop once RETRY_EN is cleared.)
//
static inline void
pcie_link_up(const u8 pp, const u8 pps)
{
#if defined CONFIG_BASE_PALO
    u_int32_t reg = readreg(PCIEP_MREGS_CFG_MAC);
    reg |= PCIEP_MREGS_CFG_MAC_APP_LTSSM_ENABLE;
    writereg(PCIEP_MREGS_CFG_MAC, reg);
#endif /* PALO */
#if defined CONFIG_BASE_SERENO
    u_int32_t reg = readreg(PCIEP_MREGS_CFG_MAC);
    u_int32_t em = readreg(MISC_MREGS_EMULATION);
    if (em == MISC_EMULATION_SERENO) {
        reg |= PCIEP_MREGS_CFG_MAC_RX_LANE_FLIP_EN;
        reg |= PCIEP_MREGS_CFG_MAC_TX_LANE_FLIP_EN;
        writereg(PCIEP_MREGS_CFG_MAC, reg);
    }
    if (board_conf_is_oakland() || board_conf_is_walnutcreek()) {
        reg &= ~(PCIEP_MREGS_CFG_MAC_RX_LANE_FLIP_EN |
                 PCIEP_MREGS_CFG_MAC_TX_LANE_FLIP_EN);
        writereg(PCIEP_MREGS_CFG_MAC, reg);
    }
    reg |= PCIEP_MREGS_CFG_MAC_APP_LTSSM_ENABLE;
    writereg(PCIEP_MREGS_CFG_MAC, reg);
#endif /* SERENO */
#if defined CONFIG_BASE_CRUZ
    u32 reg = readreg(PPSP_MREGS_(CFG_MAC0, pp, pps));
    reg |= PPSP_MREGS_PPS0_CFG_MAC0_APP_LTSSM_ENABLE;
    writereg(PPSP_MREGS_(CFG_MAC0, pp, pps), reg);
#endif /* CRUZ */
}

static inline void
pcie_link_dn(const u8 pp, const u8 pps)
{
#if defined CONFIG_BASE_PALO
    u_int32_t reg = readreg(PCIEP_MREGS_CFG_MAC);
    reg &= ~PCIEP_MREGS_CFG_MAC_APP_LTSSM_ENABLE;
    writereg(PCIEP_MREGS_CFG_MAC, reg);
#endif /* PALO */
#if defined CONFIG_BASE_SERENO
    u_int32_t reg = readreg(PCIEP_MREGS_CFG_MAC);
    reg &= ~PCIEP_MREGS_CFG_MAC_APP_LTSSM_ENABLE;
    writereg(PCIEP_MREGS_CFG_MAC, reg);
#endif /* SERENO */
#if defined CONFIG_BASE_CRUZ
    u32 reg = readreg(PPSP_MREGS_(CFG_MAC0, pp, pps));
    reg &= ~PPSP_MREGS_PPS0_CFG_MAC0_APP_LTSSM_ENABLE;
    writereg(PPSP_MREGS_(CFG_MAC0, pp, pps), reg);
#endif /* CRUZ */
}

//
// Enable PCIe target main mode.
//
static inline void
pcie_tar_main_mode_en(void)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    u_int32_t reg = readreg(PCIEP_MREGS_CFG_TAR_MAIN_MODE);
    reg |= (PCIEP_MREGS_CFG_TAR_MAIN_MODE_EN |
            PCIEP_MREGS_CFG_TAR_MAIN_MODE_IND_ECRC_ERR |
            PCIEP_MREGS_CFG_TAR_MAIN_MODE_IND_UNS_TRN |
            PCIEP_MREGS_CFG_TAR_MAIN_MODE_IND_PSND_TRN |
            PCIEP_MREGS_CFG_TAR_MAIN_MODE_IND_PMV |
            PCIEP_MREGS_CFG_TAR_MAIN_MODE_BUS_ADJ);
#if !defined CONFIG_BASE_PALO
    // This was added in Sereno, but is likely to be in all future revisions
    reg |= PCIEP_MREGS_CFG_TAR_MAIN_MODE_CFG_UNALN_EN;
#endif /* !CONFIG_BASE_PALO */
    writereg(PCIEP_MREGS_CFG_TAR_MAIN_MODE, reg);
#endif /* PALO || SERENO */
}

static void
pcie_tar_main_mode_init(void)
{
#if defined CONFIG_BASE_CRUZ
    u_int8_t   en;
    u_int8_t   ind_dis;
    u_int8_t   ind_ecrc_err;
    u_int8_t   ind_uns_trn;
    u_int8_t   ind_pmv;
    u_int8_t   ind_psnd_trn;
    u_int8_t   ind_msg_trn;
    u_int8_t   ind_arq_ecc_err;
    u_int8_t   sgl_pnd;
    u_int8_t   fen_en;
    u_int8_t   cfg_unaln_en;
    u_int8_t   cpl_td_miss;
    u_int8_t   sub_host_id_sel;
    u_int16_t  ecrc_chk_en;
    u_int16_t  bus_adj;
    u_int8_t   ss2crz_access_en;
    u_int8_t   tcam_ser_check_en;
    u_int32_t regs[PC_MREGS_CFG_TAR_MAIN_MODE_NWORDS];

    readregs(PC_MREGS_CFG_TAR_MAIN_MODE, regs,
	     PC_MREGS_CFG_TAR_MAIN_MODE_NWORDS);
    pc_mregs_cfg_tar_main_mode_get(regs, &en, &ind_dis, &ind_ecrc_err,
				   &ind_uns_trn, &ind_pmv, &ind_psnd_trn,
				   &ind_msg_trn, &ind_arq_ecc_err, &sgl_pnd,
				   &fen_en, &cfg_unaln_en, &cpl_td_miss,
				   &sub_host_id_sel, &ecrc_chk_en, &bus_adj,
				   &ss2crz_access_en, &tcam_ser_check_en);
    en = 1;
    bus_adj |= 0xffff; // bus adjust all hosts
    cfg_unaln_en = 1;
    // ss2crz_access_en = board_conf_supports_seaside(); enable if have seaside
	ss2crz_access_en = 0; /* seaside not supported */
    tcam_ser_check_en = 0; // XXXCRUZ for now ignore tcam bmt crc check
    pc_mregs_cfg_tar_main_mode_make(regs, en, ind_dis, ind_ecrc_err,
				    ind_uns_trn, ind_pmv, ind_psnd_trn,
				    ind_msg_trn, ind_arq_ecc_err, sgl_pnd,
				    fen_en, cfg_unaln_en, cpl_td_miss,
				    sub_host_id_sel, ecrc_chk_en, bus_adj,
				    ss2crz_access_en, tcam_ser_check_en);
    writeregs(PC_MREGS_CFG_TAR_MAIN_MODE, regs,
	      PC_MREGS_CFG_TAR_MAIN_MODE_NWORDS);
#endif /* CRUZ */
}

//
// Release the host from endless config retries (CRS).
//
static inline void
pcie_set_crs(u8 pp, u8 pps, int on)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    u32 reg = readreg(PCIEP_MREGS_CFG_MAC);
    if (on) {
        reg |= PCIEP_MREGS_CFG_MAC_APP_REQ_RETRY_EN;
    } else {
        reg &= ~PCIEP_MREGS_CFG_MAC_APP_REQ_RETRY_EN;
    }
    writereg(PCIEP_MREGS_CFG_MAC, reg);
#endif /* PALO || SERENO */
#if defined CONFIG_BASE_CRUZ
    u32 reg = readreg(PPSP_MREGS_(CFG_MAC0, pp, pps));
    if (on) {
        reg |= PPSP_MREGS_PPS0_CFG_MAC0_APP_REQ_RETRY_EN;
    } else {
        reg &= ~PPSP_MREGS_PPS0_CFG_MAC0_APP_REQ_RETRY_EN;
    }
    writereg(PPSP_MREGS_(CFG_MAC0, pp, pps), reg);
#endif /* CRUZ */
}

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO

#define INT_MAC_REG(pp, pps)		PCIEC_MREGS_INT_MAC
#define INT_MAC_MSK_REG(pp, pps)	PCIEC_MREGS_INT_MAC_MSK
#define INT_MACF_CORE_RST_DN		PCIEC_MREGS_INT_MAC_CORE_RST_DN
#define INT_MACF_CORE_RST_DN2UP		PCIEC_MREGS_INT_MAC_CORE_RST_DN2UP
#define INT_MACF_CORE_RST_UP2DN		PCIEC_MREGS_INT_MAC_CORE_RST_UP2DN
#define INT_MACF_SEC_BUS_NUM_CHG	PCIEC_MREGS_INT_MAC_SEC_BUS_NUM_CHG
#define INT_MACF_XMLH_LINK_UP2DN	PCIEC_MREGS_INT_MAC_XMLH_LINK_UP2DN
#define INT_MACF_XMLH_LINK_DN2UP	PCIEC_MREGS_INT_MAC_XMLH_LINK_DN2UP
#define INT_MAC_MSKF_CORE_RST_DN	PCIEC_MREGS_INT_MAC_MSK_CORE_RST_DN
#define INT_MAC_MSKF_CORE_RST_DN2UP	PCIEC_MREGS_INT_MAC_MSK_CORE_RST_DN2UP
#define INT_MAC_MSKF_CORE_RST_UP2DN	PCIEC_MREGS_INT_MAC_MSK_CORE_RST_UP2DN
#define INT_MAC_MSKF_SEC_BUS_NUM_CHG	PCIEC_MREGS_INT_MAC_MSK_SEC_BUS_NUM_CHG
#define INT_MAC_MSKF_XMLH_LINK_UP2DN	PCIEC_MREGS_INT_MAC_MSK_XMLH_LINK_UP2DN
#define INT_MAC_MSKF_XMLH_LINK_DN2UP	PCIEC_MREGS_INT_MAC_MSK_XMLH_LINK_DN2UP
#define STA_RST_REG(pp, pps)		PCIEC_MREGS_STA_RST
#define STA_RST_MAC_CORE_RST_N		PCIEC_MREGS_STA_RST_MAC_CORE_RST_N
#define STA_RST_MAC_XMLH_LINK_UP	PCIEC_MREGS_STA_RST_MAC_XMLH_LINK_UP

#elif defined CONFIG_BASE_CRUZ

#define INT_MAC_REG(pp, pps)		PPSC_MREGS_(INT_MAC, pp, pps)
#define INT_MAC_MSK_REG(pp, pps)	PPSC_MREGS_(INT_MAC_MSK, pp, pps)
#define INT_MACF_CORE_RST_DN		PPSC_MREGS_PPS0_INT_MAC_CORE_RST_DN
#define INT_MACF_CORE_RST_DN2UP		PPSC_MREGS_PPS0_INT_MAC_CORE_RST_DN2UP
#define INT_MACF_CORE_RST_UP2DN		PPSC_MREGS_PPS0_INT_MAC_CORE_RST_UP2DN
#define INT_MACF_SEC_BUS_NUM_CHG	PPSC_MREGS_PPS0_INT_MAC_SEC_BUS_NUM_CHG
#define INT_MACF_XMLH_LINK_UP2DN        PPSC_MREGS_PPS0_INT_MAC_XMLH_LINK_UP2DN
#define INT_MACF_XMLH_LINK_DN2UP        PPSC_MREGS_PPS0_INT_MAC_XMLH_LINK_DN2UP
#define INT_MAC_MSKF_CORE_RST_DN	\
	PPSC_MREGS_PPS0_INT_MAC_MSK_CORE_RST_DN
#define INT_MAC_MSKF_CORE_RST_DN2UP	\
	PPSC_MREGS_PPS0_INT_MAC_MSK_CORE_RST_DN2UP
#define INT_MAC_MSKF_CORE_RST_UP2DN	\
	PPSC_MREGS_PPS0_INT_MAC_MSK_CORE_RST_UP2DN
#define INT_MAC_MSKF_SEC_BUS_NUM_CHG	\
	PPSC_MREGS_PPS0_INT_MAC_MSK_SEC_BUS_NUM_CHG
#define INT_MAC_MSKF_XMLH_LINK_UP2DN    \
	PPSC_MREGS_PPS0_INT_MAC_MSK_XMLH_LINK_UP2DN
#define INT_MAC_MSKF_XMLH_LINK_DN2UP    \
	PPSC_MREGS_PPS0_INT_MAC_MSK_XMLH_LINK_DN2UP
#define STA_RST_REG(pp, pps)   		PPSC_MREGS_(STA_RST, pp, pps)
#define STA_RST_MAC_CORE_RST_N		PPSC_MREGS_PPS0_STA_RST_MAC_CORE_RST_N
#define STA_RST_MAC_XMLH_LINK_UP	\
	PPSC_MREGS_PPS0_STA_RST_MAC_XMLH_LINK_UP

#endif /* CONFIG_BASE */

//
// NOTE: MSK reg is really an ENAble reg.  A 1 set in MSK allows
// an intr to come through.
//
static inline void
pcie_mask_mac_intr(const u8 pp, const u8 pps, const u32 src)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    u32 reg = readreg(PCIEC_MREGS_INT_MAC_MSK);
    reg &= ~src;
    writereg(PCIEC_MREGS_INT_MAC_MSK, reg);
    (void)readreg(PCIEC_MREGS_INT_MAC_MSK);
#endif /* PALO || SERENO */
#if defined CONFIG_BASE_CRUZ
    u32 reg = readreg(PPSC_MREGS_(INT_MAC_MSK, pp, pps));
    reg &= ~src;
    writereg(PPSC_MREGS_(INT_MAC_MSK, pp, pps), reg);
    (void)readreg(PPSC_MREGS_(INT_MAC_MSK, pp, pps));
#endif /* CRUZ */
}

static inline void
pcie_unmask_mac_intr(const u8 pp, const u8 pps, const u32 src)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    u32 reg = readreg(PCIEC_MREGS_INT_MAC_MSK);
    reg |= src;
    writereg(PCIEC_MREGS_INT_MAC_MSK, reg);
    (void)readreg(PCIEC_MREGS_INT_MAC_MSK);
#endif /* PALO || SERENO */
#if defined CONFIG_BASE_CRUZ
    u32 reg = readreg(PPSC_MREGS_(INT_MAC_MSK, pp, pps));
    reg |= src;
    writereg(PPSC_MREGS_(INT_MAC_MSK, pp, pps), reg);
    (void)readreg(PPSC_MREGS_(INT_MAC_MSK, pp, pps));
#endif /* CRUZ */
}

static void
pcie_unmask_mac_intr_hid(u_int32_t hid)
{
    const u8 pp  = ppcie_hid_pp(hid);
    const u8 pps = ppcie_hid_pps(hid);
    const u_int32_t mskf = (INT_MAC_MSKF_CORE_RST_DN |
                            INT_MAC_MSKF_CORE_RST_UP2DN |
                            INT_MAC_MSKF_CORE_RST_DN2UP |
                            INT_MAC_MSKF_SEC_BUS_NUM_CHG |
                            INT_MAC_MSKF_XMLH_LINK_UP2DN |
                            INT_MAC_MSKF_XMLH_LINK_DN2UP);

    pcie_mask_mac_intr(pp, pps, ~0);            // mask all
    pcie_unmask_mac_intr(pp, pps, mskf);        // unmask ours
}

static void __devinit
pcie_unmask_mac_intr_cb(u_int32_t hid, void *arg)
{
    pcie_unmask_mac_intr_hid(hid);
}

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
#define INT_ERR_REG(pp, pps)		PCIEP_MREGS_INT_ERR
#define INT_ERR_REG_(FIELD, pp, pps)    PCIEP_MREGS_INT_ERR_##FIELD
#define INT_ERR_MSK_REG(pp, pps)        PCIEP_MREGS_INT_ERR_MSK
#elif defined CONFIG_BASE_CRUZ
#define INT_ERR_REG(pp, pps)	        PPSP_MREGS_(INT_ERR, pp, pps)
#define INT_ERR_REG_(FIELD, pp, pps)    PPSP_MREGS_PPS0_INT_ERR_##FIELD
#define INT_ERR_MSK_REG(pp, pps)        PPSP_MREGS_(INT_ERR_MSK, pp, pps)
#endif /* CONFIG_BASE */

static inline void
pcie_mask_err_intr(const u8 pp, const u8 pps, const u32 src)
{
    u32 reg = readreg(INT_ERR_MSK_REG(pp, pps));
    reg &= ~src;
    writereg(INT_ERR_MSK_REG(pp, pps), reg);
    (void)readreg(INT_ERR_MSK_REG(pp, pps));
}

static inline void
pcie_unmask_err_intr(const u8 pp, const u8 pps, const u32 src)
{
    u32 reg = readreg(INT_ERR_MSK_REG(pp, pps));
    reg |= src;
    writereg(INT_ERR_MSK_REG(pp, pps), reg);
    (void)readreg(INT_ERR_MSK_REG(pp, pps));
}

static void
pcie_unmask_err_intr_hid(u_int32_t hid)
{
    const u8 pp  = ppcie_hid_pp(hid);
    const u8 pps = ppcie_hid_pps(hid);

    pcie_unmask_err_intr(pp, pps, ~0);        // unmask all
}

//
// Apply some minor fixups to the config space of the
// synopsys core pcie bridge.
//
static void pcie_fixup_pciecap(void)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
#define SW_CFG_PCIE_CAPOFF      0x70
    const u_int32_t cap = PCIEP_MREGS_DHS_PCIE_SW_CFG + SW_CFG_PCIE_CAPOFF;
    u_int32_t reg;

    // Turn on Extended Tags by default
    // Modern PCIe allows this.  See PCI-SIG doc
    // ECN_Extended_Tag_Enable_Default, Sep 5, 2008.
    reg = readreg(cap + PCI_EXP_DEVCTL);
    reg |= PCI_EXP_DEVCTL_EXT_TAG;
    writereg(cap + PCI_EXP_DEVCTL, reg);

#ifdef CONFIG_BASE_PALO
    // L0s and L1 should be 0 for functions other than endpoints.
    // Our Palo Synopsis core switch had these non-zero.
    reg = readreg(cap + PCI_EXP_DEVCAP);
    reg &= ~PCI_EXP_DEVCAP_L0S;
    reg &= ~PCI_EXP_DEVCAP_L1;
    writereg(cap + PCI_EXP_DEVCAP, reg);
#endif // CONFIG_BASE_PALO
#endif /* PALO || SERENO */
}

//
// PALO, SERENO - DESCRIPTION
// lm0, sw_rdb - Synopsys DWC_pcie_sw core replay data buffer
// lm1, sw_rsb - Synopsys DWC_pcie_sw core replay start of TLP
// lm2, (sw_rhq) - Synopsys DWC_pcie_sw core receive header queue (Palo-only)
// lm3, (sw_rdq) - Synopsys DWC_pcie_sw core receive data queue (Palo-only)
// lm4, tar_cmt - Target CMT
// lm5, tar_crt - Target CRT (swinit)
// lm6, tar_bmt - Target BMT
// lm7, tar_brt - Target BRT (swinit)
// lm8, ini_crf - Initiator command request FIFO
// lm9, ini_drf - Initiator data request FIFO
// lm10, ini_bdft - Initiator BDFT (swinit)
// lm11, ini_rrq - Initiator response request queue
//
static void pcie_unmask_int_errs(void)
{
#ifdef CONFIG_BASE_PALO
	writereg(PCIEP_MREGS_INT_LM0_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM1_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM2_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM3_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM4_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM5_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM6_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM7_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM8_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM9_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM10_ECC_ERROR_MSK, ~0);
	writereg(PCIEP_MREGS_INT_LM11_ECC_ERROR_MSK, ~0);
#endif // CONFIG_BASE_PALO
#ifdef CONFIG_BASE_SERENO
	writereg(PCIEP_MREGS_INT_ECC_ERROR_MSK, ~0);
#endif // CONFIG_BASE_SERENO
}

//
// Set PCIe initiator timeout.  If for some reason we do not get
// a completion from the host in the timeout_value time then a
// non-zero timeout_value will cause a pcie_p_int_err.ini_cpl_timeout.
// Such a timeout indicates a serious problem on the PCIe bus.
//
// pcie_p_cfg_ini_main_mode.timeout_ticks divides the PCIe clock
// to give a quanta of 50us.  Then timeout_value sets the number of
// units to elapse before a timeout is signaled.
//
static void pcie_ini_cpl_timeout(u32 timeout_us)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
	u_int32_t reg;
	u_int16_t timeout_tick;
	u_int16_t timeout_value;
	u_int8_t bus_adj;
	u_int8_t dca_en;
	u_int8_t addr_ext_en;
	u_int8_t rlx_ord_en;
	u_int8_t halt;
#ifndef CONFIG_BASE_PALO
	u_int8_t ecrc_chk_en;
#endif // !CONFIG_BASE_PALO

	reg = readreg(PCIEP_MREGS_CFG_INI_MAIN_MODE);
	pciep_mregs_cfg_ini_main_mode_get(&reg,
	  &timeout_tick, &timeout_value, &bus_adj, &dca_en,
	  &addr_ext_en, &rlx_ord_en, &halt
#ifndef CONFIG_BASE_PALO
	  , &ecrc_chk_en
#endif // !CONFIG_BASE_PALO
	  );

	timeout_value = timeout_us / 50;

	pciep_mregs_cfg_ini_main_mode_make(&reg,
	  timeout_tick, timeout_value, bus_adj, dca_en,
	  addr_ext_en, rlx_ord_en, halt
#ifndef CONFIG_BASE_PALO
	  , ecrc_chk_en
#endif // !CONFIG_BASE_PALO
	);
	writereg(PCIEP_MREGS_CFG_INI_MAIN_MODE, reg);
#endif // PALO || SERENO
}

static void pcie_ini_init(void)
{
	pcie_ini_cpl_timeout(50000); // 50ms
}

static void
pcie_set_fault(pcie_hid_info_t *p, char *reason, const u8 errcode)
{
	const int hostid = ppcie_hostid_make(p->hid, 0);

	tracep(p, PCIETR_INFO, "FAULT: %s", reason);
	printk(KERN_ERR PFX "%s FAULT: %s\n", p->name, reason);
	ppcie_evt_linkerr(hostid, PPCIE_ERRTYPE_HARD, errcode);
	p->state = PCIEST_FAULT;
        p->nfaults++;
	if ((p->flags & PCIEF_FAULT) == 0) {
		// Save first fault reason
		p->flags |= PCIEF_FAULT;
		p->fault_reason = reason;
	}
}

static void
pcie_clear_fault(pcie_hid_info_t *p)
{
	if (p->flags & PCIEF_FAULT) {
		p->flags &= ~PCIEF_FAULT;
		p->last_fault_reason = p->fault_reason;
		p->fault_reason = NULL;
	}
}

#if defined CONFIG_BASE_CRUZ
typedef struct oventry {
	u_int32_t off;
	u_int32_t val;
} oventry_t;

#define MAXOV 32
typedef struct ovinfo {
	u_int32_t nov;
	oventry_t ov[MAXOV];
} ovinfo_t;

#define MAXPP  PP_MREGS_MAXCNT
#define MAXPPS PPSP_MREGS_MAXCNT
typedef struct ovgroup {
	ovinfo_t gbl;
	ovinfo_t pps[MAXPP][MAXPPS];
} ovgroup_t;

static ovgroup_t serovg;
static ovgroup_t pcsovg;
static ovgroup_t cfgovg;
static ovgroup_t rstovg; // cfg_ini_barrier_req_stall_threshold

static void
ovinfo_init(ovinfo_t *ovinfo, char *str)
{
	int n;
	char *equal, *comma;
	char *off, *val;

	n = 0;
	while (str && n < MAXOV) {
		equal = strchr(str, '=');
		comma = strchr(str, ',');
		if (equal && (!comma || equal < comma)) {
			off = str;
			val = equal + 1;
		} else {
			off = "";
			val = str;
		}
		ovinfo->ov[n].off = simple_strtoul(off, NULL, 0);
		ovinfo->ov[n].val = simple_strtoul(val, NULL, 0);
		n++;

		// find next off=val
		str = comma;
		if (str) str++;
	}
	ovinfo->nov = n;
}

static void
ovgroup_init(ovgroup_t *ovg, char *str)
{
	extern char *prom_getenv(const char *s);
	char *val, buf[80];
	int pp, pps;

	// Get global settings
	val = prom_getenv(str);
	if (val) {
		printk("%s override = %s\n", str, val);
		ovinfo_init(&ovg->gbl, val);
	}
	// Get per-pps settings
	for (pp = 0; pp < MAXPP; pp++) {
		for (pps = 0; pps < MAXPPS; pps++) {
			snprintf(buf, sizeof (buf), "%s_%d_%d", str, pp, pps);
			val = prom_getenv(buf);
			if (val) {
				printk("%s override = %s\n", buf, val);
				ovinfo_init(&ovg->pps[pp][pps], val);
			}
		}
	}
}

static void
ovinfo_write(ovinfo_t *ovinfo, u_int32_t base)
{
    int i;

    for (i = 0; i < ovinfo->nov; i++) {
        oventry_t *ov = &ovinfo->ov[i];

        writereg(base + ov->off, ov->val);
    }
}

//
// For CruzB (LSI),
// serdes config regs are found in these register blocks:
//     PP<pp>_DHS_SER_<pps>
// pipe/pcs config regs are found in these register blocks:
//     PP<pp>_DHS_PIPE_<pps>
//
#define PP_MREGS_DHS_SER_STRIDE \
	(PP_MREGS_DHS_SER_1 - PP_MREGS_DHS_SER_0)
#define AHB_SER(pp, pps) \
	(PP_MREGS_BASE(pp) + \
	 PP_MREGS_DHS_SER_0_OFFSET + (pps) * PP_MREGS_DHS_SER_STRIDE)

#define PP_MREGS_DHS_PIPE_STRIDE \
	(PP_MREGS_DHS_PIPE_1 - PP_MREGS_DHS_PIPE_0)
#define AHB_PCS(pp, pps) \
	(PP_MREGS_BASE(pp) + \
	 PP_MREGS_DHS_PIPE_0_OFFSET + (pps) * PP_MREGS_DHS_PIPE_STRIDE)

static ovinfo_t pcsdef_cruz = {
	.nov = 0,
	.ov = {
	},
};

//
// OCP block is for CruzA (TI) only.
//
// OCP block is 64k words = 256k bytes
// The lower 32k words (128k bytes) is serdes register,
// upper 32k words (128k bytes) is pcs registers.
// 8k words (32k bytes) of serdes/pcs region per pps.
//
// 0x00000000 +-----------------+
//            |   serdes pps0   |
//            |   serdes pps1   |
//            |   serdes pps2   |
//            |   serdes pps3   |
// 0x00020000 +-----------------+
//            |      pcs pps0   |
//            |      pcs pps1   |
//            |      pcs pps2   |
//            |      pcs pps3   |
// 0x00040000 +-----------------+
//
#define PP_MREGS_DHS_OCP_OFFSET 0x0000
#define OCP_SER(pp, pps) \
	(PP_MREGS_(DHS_OCP, pp) | ((pps) << 15))
#define OCP_PCS(pp, pps) \
	(PP_MREGS_(DHS_OCP, pp) | ((pps) << 15) | 0x20000)

static ovinfo_t pcsdef_cruz_a = {
	.nov = 17,
	.ov = {
		 { 0x0020, 0x32 }, // Set PCS lane0-7 stopthreshold
		 { 0x0420, 0x32 }, // PCS register offset
		 { 0x0820, 0x32 }, // 0x8 x i100 = 0x32
		 { 0x0c20, 0x32 },
		 { 0x1020, 0x32 },
		 { 0x1420, 0x32 },
		 { 0x1820, 0x32 },
		 { 0x1c20, 0x32 },
		 { 0x001c, 0x26 }, // Set PCS lane0-7 Passthreshold
		 { 0x041c, 0x26 }, // PCS register offset
		 { 0x081c, 0x26 }, // 0x7 x i100 = 0x26
		 { 0x0c1c, 0x26 },
		 { 0x101c, 0x26 },
		 { 0x141c, 0x26 },
		 { 0x181c, 0x26 },
		 { 0x1c1c, 0x26 },
		 { 0x01bc, 0x01 }, // Set PCS GlobalAdapt = 1
	},
};

static void
pps_sta_mac(const int pp, const int pps,
	    int *ltssmstp, int *xmlhp, int *rdhlp)
{
	u_int8_t xmlh_ltssm_state, xmlh_link_up, rdlh_link_up;
	u_int8_t pm_curnt_state, pm_dstate, pm_pme_en, pm_status, aux_pm_en;
	u_int8_t pm_xtlh_block_tlp, pm_req_dwsp_turnoff;
	u_int8_t cfg_hw_auto_sp_dis;
	u_int32_t sta_mac;

	sta_mac = readreg(PPSP_MREGS_(STA_MAC, pp, pps));
	ppsp_mregs_pps0_sta_mac_get(&sta_mac,
				    &xmlh_ltssm_state,
				    &xmlh_link_up, &rdlh_link_up,
				    &pm_curnt_state, &pm_dstate,
				    &pm_pme_en, &pm_status, &aux_pm_en,
				    &pm_xtlh_block_tlp, &pm_req_dwsp_turnoff,
				    &cfg_hw_auto_sp_dis);
	if (ltssmstp) *ltssmstp = xmlh_ltssm_state;
	if (xmlhp) *xmlhp = xmlh_link_up;
	if (rdhlp) *rdhlp = rdlh_link_up;
}

static int
pps_in_compliance_state(const u8 pp, const u8 pps)
{
    int ltssmst;

    pps_sta_mac(pp, pps, &ltssmst, NULL, NULL);
    return (ltssmst == 0x3);
}

static void
pps_ser_init(const u8 pp, const u8 pps)
{
	u_int32_t ser;

	if (readreg(MISC_MREGS_VERSION) >= 0xb0) {
		ser = AHB_SER(pp, pps);
	} else {
		ser = OCP_SER(pp, pps);
	}

        // Apply any requested custom settings
        ovinfo_write(&serovg.gbl, ser);
        ovinfo_write(&serovg.pps[pp][pps], ser);
}

static void
pps_pcs_init(const u8 pp, const u8 pps)
{
	u_int32_t pcs;

	if (readreg(MISC_MREGS_VERSION) >= 0xb0) {
		pcs = AHB_PCS(pp, pps);
		ovinfo_write(&pcsdef_cruz, pcs);
	} else {
		pcs = OCP_PCS(pp, pps);
		ovinfo_write(&pcsdef_cruz_a, pcs);
	}

        // Apply any requested custom settings
        ovinfo_write(&pcsovg.gbl, pcs);
        ovinfo_write(&pcsovg.pps[pp][pps], pcs);
}

static int
pps_pll_poll(const u8 pp, const u8 ser)
{
#define STA_SER_PLL_STRIDE	(PP_MREGS_STA_SER1_PLL - PP_MREGS_STA_SER0_PLL)
#define STA_SER_PLL(pp, ser) \
	(PP_MREGS_(STA_SER0_PLL, pp) + ((ser) * STA_SER_PLL_STRIDE))
	const u32 sta_pll = STA_SER_PLL(pp, ser);
	int polls;

	for (polls = 0; polls < 10; polls++) {
		if (readreg(sta_pll) != 0) return polls;
		udelay(100);
	}
	return -1;
}

static int
pps_pll_locked(const u8 pp, const u8 pps, const u32 width)
{
	int r0, r1, r2, r3;

	r0 = r1 = r2 = r3 = 0;
	switch (width) {
	case 1:
	case 2:
		// pps0 owns serdes0, pps1 owns serdes1, etc
		r0 = pps_pll_poll(pp, pps);
		break;
	case 4:
		// pps0 owns serdes0/1, pps2 owns serdes2/3
		r0 = pps_pll_poll(pp, pps);
		r1 = pps_pll_poll(pp, pps + 1);
		break;
	case 8:
		// pps0 owns serdes0/1/2/3
		r0 = pps_pll_poll(pp, 0);
		r1 = pps_pll_poll(pp, 1);
		r2 = pps_pll_poll(pp, 2);
		r3 = pps_pll_poll(pp, 3);
		break;
	default:
		break;
	}
	// Did any of the polls return fail?
	if (r0 < 0 || r1 < 0 || r2 < 0 || r3 < 0)
		return -1;
	// Return sum of all polls.
	return r0 + r1 + r2 + r3;
}

#define PP_MREGS_PCSD_CFG_SER_RESET_STRIDE \
	(PP_MREGS_PCSD_CFG_1_SER_RESET - PP_MREGS_PCSD_CFG_0_SER_RESET)
#define PP_MREGS_PCSD_CFG_SER_RESET(pp, pps) \
	(PP_MREGS_BASE(pp) + \
	 PP_MREGS_PCSD_CFG_0_SER_RESET_OFFSET + \
	 (pps) * PP_MREGS_PCSD_CFG_SER_RESET_STRIDE)

static void
pps_set_ser_reset(const u8 pp, const u8 pps, const int on)
{
	const u32 flags = (PP_MREGS_PCSD_CFG_0_SER_RESET_AHB_DHS2AHB_RESET |
			   PP_MREGS_PCSD_CFG_0_SER_RESET_CORE_DHS2AHB_RESET);
	u32 reg = readreg(PP_MREGS_PCSD_CFG_SER_RESET(pp, pps));
	if (on) {
		reg |= flags;
	} else {
		reg &= ~flags;
	}
	writereg(PP_MREGS_PCSD_CFG_SER_RESET(pp, pps), reg);
	(void)readreg(PP_MREGS_PCSD_CFG_SER_RESET(pp, pps));
}

#define PP_MREGS_PCSD_CFG_PIPE_STRIDE \
	(PP_MREGS_PCSD_CFG_1_PIPE - PP_MREGS_PCSD_CFG_0_PIPE)
#define PP_MREGS_PCSD_CFG_PIPE(pp, pps) \
	(PP_MREGS_BASE(pp) + \
	 PP_MREGS_PCSD_CFG_0_PIPE_OFFSET + \
	 (pps) * PP_MREGS_PCSD_CFG_PIPE_STRIDE)

static void
pps_set_pipe_reset(const u8 pp, const u8 pps, const int on)
{
	const u32 flags = (PP_MREGS_PCSD_CFG_0_PIPE_PIPE_RESET |
			   PP_MREGS_PCSD_CFG_0_PIPE_TX_CFG_SERDES_RESET |
			   PP_MREGS_PCSD_CFG_0_PIPE_SERDES_RESET |
			   PP_MREGS_PCSD_CFG_0_PIPE_DHS2AHB_RESET |
			   PP_MREGS_PCSD_CFG_0_PIPE_SERDES_AHB_RESET);
	u32 reg = readreg(PP_MREGS_PCSD_CFG_PIPE(pp, pps));
	if (on) {
		reg |= flags;
	} else {
		reg &= ~flags;
	}
	writereg(PP_MREGS_PCSD_CFG_PIPE(pp, pps), reg);
	(void)readreg(PP_MREGS_PCSD_CFG_PIPE(pp, pps));
}

//
// Reset the pcie serdes.  We want both the PPx_PCSD_CFG_x_SER_RESET
// and the PPx_PCSD_CFG_x_PIPE register bits to go together through
// the sequence so both serdes and pipe reset together.
//
// Writing '1' to the reset bits is not enough to guarantee the blocks
// get reset.  To put the blocks into reset, we want an off->on edge
// transition on the reset bits in each register.
//
static void
pps_set_serdes_reset(const u8 pp, const u8 pps, const int on)
{
	// Only on the CruzB ASIC do we have the LSI serdes/pipe.
	if (palo_conf_is_asic() &&
	    readreg(MISC_MREGS_VERSION) >= 0xb0) {
		if (on) {
			// Transition from "off" to "on" causes the reset,
			pps_set_ser_reset(pp, pps, 0);
			pps_set_pipe_reset(pp, pps, 0);
			udelay(100);
			pps_set_ser_reset(pp, pps, 1);
			pps_set_pipe_reset(pp, pps, 1);
			udelay(100);
		} else {
			pps_set_ser_reset(pp, pps, 0);
			pps_set_pipe_reset(pp, pps, 0);
			udelay(100);
		}
	}
}

static void
pps_serdes_reset(const u8 pp, const u8 pps)
{
	pps_set_serdes_reset(pp, pps, 1);
	pps_set_serdes_reset(pp, pps, 0);
}

static void
pcie_set_serdes_reset(pcie_hid_info_t *p, const int on)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	unsigned long flags;

	spin_lock_irqsave(&p->lock, flags);
	pps_set_serdes_reset(pp, pps, on);
	spin_unlock_irqrestore(&p->lock, flags);
}

static int act_tar2naxi_idwidth;

static void
pps_tar_init(const u8 pp, const u8 pps)
{
	u_int8_t   idwidth;
	u_int8_t   halt;
	u_int8_t   mwr_gt16_en;
	u_int8_t   steer_cpl_act_sb;
	u_int8_t   steer_cpl_sub_host_id;
	u_int8_t   posted_wrack_en;
	u_int32_t reg;

	reg = readreg(PPSP_MREGS_(CFG_TAR_MODE, pp, pps));
	ppsp_mregs_pps0_cfg_tar_mode_get(&reg,
					 &idwidth,
					 &halt,
					 &mwr_gt16_en,
					 &steer_cpl_act_sb,
					 &steer_cpl_sub_host_id,
					 &posted_wrack_en);
	idwidth = act_tar2naxi_idwidth;
	mwr_gt16_en = 1;
    // if (board_conf_supports_seaside()) {
		// Steer the active / standby to TLP Hdr Byte1[7]
		// steer_cpl_act_sb = 1;
	// }
	ppsp_mregs_pps0_cfg_tar_mode_make(&reg,
					  idwidth,
					  halt,
					  mwr_gt16_en,
					  steer_cpl_act_sb,
					  steer_cpl_sub_host_id,
					  posted_wrack_en);
	writereg(PPSP_MREGS_(CFG_TAR_MODE, pp, pps), reg);
}

static void
pps_set_steer_mode(u_int8_t pp, u_int8_t pps)
{
	u_int8_t steer_ini_sub_host;
	u_int8_t non_ext_tag_limit16;
	u_int8_t steer_ini_act_sb;
	u_int32_t reg;
	u_int8_t mode = 0x0; // don't enable (also hw default)

	reg = readreg(PPSP_MREGS_(CFG_INI_SS_STEER_MODE, pp, pps));
	ppsp_mregs_pps0_cfg_ini_ss_steer_mode_get(&reg,
			&steer_ini_act_sb,
			&steer_ini_sub_host,
			&non_ext_tag_limit16);

	steer_ini_act_sb = mode;

	ppsp_mregs_pps0_cfg_ini_ss_steer_mode_make(&reg,
			steer_ini_act_sb,
			steer_ini_sub_host,
			non_ext_tag_limit16);

	writereg(PPSP_MREGS_(CFG_INI_SS_STEER_MODE, pp, pps), reg);
}

static void
pps_ini_init(const u8 pp, const u8 pps)
{
	u_int16_t timeout_tick;
	u_int16_t timeout_value;
	u_int8_t bus_adj;
	u_int8_t obff_en;
	u_int8_t addr_ext_en;
	u_int8_t obff_thres;
	u_int8_t halt;
	u_int8_t ecrc_chk_en;
	u_int8_t bypass_xadm_cdts_chk;
	u_int32_t reg;
	const u_int32_t timeout_us = 50000; // 50ms

	reg = readreg(PPSP_MREGS_(CFG_INI_MAIN_MODE, pp, pps));
	ppsp_mregs_pps0_cfg_ini_main_mode_get(&reg,
					      &timeout_tick,
					      &timeout_value,
					      &bus_adj,
					      &obff_en,
					      &addr_ext_en,
					      &obff_thres,
					      &halt,
					      &ecrc_chk_en,
					      &bypass_xadm_cdts_chk);
	//
	// Set PCIe initiator timeout.  If for some reason we do not get
	// a completion from the host in the timeout_value time then a
	// non-zero timeout_value will cause a int_err.ini_cpl_timeout.
	//
	// cfg_ini_main_mode.timeout_ticks divides the PCIe clock
	// to give a quanta of 50us.  Then timeout_value sets the number of
	// units to elapse before a timeout is signaled.
	//
	timeout_value = timeout_us / 50;

	ppsp_mregs_pps0_cfg_ini_main_mode_make(&reg,
					       timeout_tick,
					       timeout_value,
					       bus_adj,
					       obff_en,
					       addr_ext_en,
					       obff_thres,
					       halt,
					       ecrc_chk_en,
					       bypass_xadm_cdts_chk);
	writereg(PPSP_MREGS_(CFG_INI_MAIN_MODE, pp, pps), reg);

}

static int gen3_eq_ctrl = 0x20;

static void
pps_cfg_init(const u8 pp, const u8 pps, const u32 width)
{
	u_int32_t cfg = PP_MREGS_PPS_(DHS_SW_CFG, pp, pps);
	u_int32_t reg;

	if (palo_conf_is_asic()) {
		if (readreg(MISC_MREGS_VERSION) <= 0xa0) {
			// GEN3_EQ_CONTROL_REG_8A8_DBI_548 = 0x400
			// (EIEOS_CNT_RST_DIS)
			writereg(cfg + 0x890, 0x400);
			// Set DWC_SW0_GEN3_EQ_CONTROL_REG_8A8_DBI_554 = 0x100
			writereg(cfg + 0x8a8, gen3_eq_ctrl);
			// Set DWC_SW0_GEN3_EQ_CONTROL_REG_8AC_DBI_555 = 0x20
			writereg(cfg + 0x8ac, 0x0020);
		} else {
			// #minimum 4 ms in eq state with
			// converge window depth 8 #0x24
			writereg(cfg + 0x80c, 0x00020864);
			// preset 5
			writereg(cfg + 0x8ac, 0x00000044);
			writereg(cfg + 0x8a8, 0x00002020);
			writereg(cfg + 0x890, 0x00000001);
		}

		// fix pcie cap link cap register
		reg = readreg(cfg + 0x7c);
                // fix link width
		switch (width) {
		case 1: reg &= ~0x3f0; reg |= 1 << 4; break;
		case 2: reg &= ~0x3f0; reg |= 2 << 4; break;
		case 4: reg &= ~0x3f0; reg |= 4 << 4; break;
		case 8: reg &= ~0x3f0; reg |= 8 << 4; break;
		}
		// Disable ASPM support on Cruz because ASPM causes too many
		// interoperability issues.  PCIe forums recommend disabling
		// ASPM to work around these.
		reg &= ~0x0c00;
		writereg(cfg + 0x7c, reg);

		// update subsys_id
		if (readreg(MISC_MREGS_VERSION) > 0xa0) {
			reg = readreg(cfg + 0xf4);
			reg &= ~0xffff0000;
			reg |= board_conf_subsys_id() << 16;
			writereg(cfg + 0xf4, reg);
		}

		// Apply any requested custom settings
		ovinfo_write(&cfgovg.gbl, cfg);
		ovinfo_write(&cfgovg.pps[pp][pps], cfg);
	}
}

static void
pps_cfg_mac1(const u8 pp, const u8 pps, const u32 width, const u32 flags)
{
	u_int32_t reg;
	u_int8_t auto_ltssm_enable;
	u_int8_t app_link_cap_mask;
	u_int8_t rx_lane_flip_en;
	u_int8_t tx_lane_flip_en;
	u_int8_t halt_on_retry_en;
	u_int8_t msg_drop;

	reg = readreg(PPSC_MREGS_(CFG_MAC1, pp, pps));
	ppsc_mregs_pps0_cfg_mac1_get(&reg,
				     &auto_ltssm_enable, &app_link_cap_mask,
				     &rx_lane_flip_en, &tx_lane_flip_en,
				     &halt_on_retry_en, &msg_drop);

	// map linkwidth spec to cap_mask
	switch (width) {
	case 1: app_link_cap_mask = 0x1; break;
	case 2: app_link_cap_mask = 0x3; break;
	case 4: app_link_cap_mask = 0x7; break;
	case 8: app_link_cap_mask = 0xf; break;
	default: break; // no change
	}

	// apply lane flip requests
	tx_lane_flip_en = (flags & PCIE_LINKF_TX_FLIP) != 0;
	rx_lane_flip_en = (flags & PCIE_LINKF_RX_FLIP) != 0;

	ppsc_mregs_pps0_cfg_mac1_make(&reg,
				      auto_ltssm_enable, app_link_cap_mask,
				      rx_lane_flip_en, tx_lane_flip_en,
				      halt_on_retry_en, msg_drop);
	writereg(PPSC_MREGS_(CFG_MAC1, pp, pps), reg);
}

static void
pps_cfg_guardband(const u8 pp, const u8 pps)
{
	u_int8_t nph_ro, nph_ord, ph_ord, pd_ord;
	u_int32_t reg;

	// guardband is CruzB only (and ok for CruzitaB)
	if (readreg(MISC_MREGS_VERSION) < 0xb0)
		return;

	reg = readreg(PPSP_MREGS_(CFG_INI_FC_CDTS_GUARDBAND, pp, pps));
	ppsp_mregs_pps0_cfg_ini_fc_cdts_guardband_get(&reg,
						      &nph_ro,
						      &nph_ord,
						      &ph_ord,
						      &pd_ord);
	nph_ro = 0x2;
	nph_ord = 0x2;
	ph_ord = 0x2;
	pd_ord = 0x10;

	ppsp_mregs_pps0_cfg_ini_fc_cdts_guardband_make(&reg,
						       nph_ro,
						       nph_ord,
						       ph_ord,
						       pd_ord);
	writereg(PPSP_MREGS_(CFG_INI_FC_CDTS_GUARDBAND, pp, pps), reg);
}

static void
pps_cfg_req_stall_threshold(const u8 pp, const u8 pps)
{
	// req_stall_threshold was added in CruzB.
	if (readreg(MISC_MREGS_VERSION) >= 0xb0) {
		const u32 req_stall_threshold =
		  PPSC_MREGS_(CFG_INI_BARRIER_REQ_STALL_THRESHOLD, pp, pps);

		ovinfo_write(&rstovg.gbl, req_stall_threshold);
		ovinfo_write(&rstovg.pps[pp][pps], req_stall_threshold);
	}
}

static int
pps_init(pcie_hid_info_t *p)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	u32 width, flags;
	int retry = 0;

	ppcie_get_linkparams(p->hid, &width, &flags);

	// Poll for locked PLLs (asic-only).
        if (palo_conf_is_asic()) {
		int polls = pps_pll_locked(pp, pps, width);
		while (polls < 0) {
			if (++retry > PPS_PLL_CHECK_RETRY_COUNT) {
				pcie_set_fault(p, "pll check failed",
					       PPCIE_ERRCODE_PPS_PLL);
				return -1;
			}
			pps_serdes_reset(pp, pps);
			polls = pps_pll_locked(pp, pps, width);
		}
		if (polls > p->maxpllpolls) {
			tracep(p, PCIETR_INFO, "pll polls %d", polls);
			p->maxpllpolls = polls;
		}
	}

	pps_tar_init(pp, pps);
	pps_set_steer_mode(pp, pps);
	pps_ini_init(pp, pps);
	pps_cfg_init(pp, pps, width);
	pps_cfg_mac1(pp, pps, width, flags);
	pps_ser_init(pp, pps);
	pps_pcs_init(pp, pps);
	pps_cfg_guardband(pp, pps);
	pps_cfg_req_stall_threshold(pp, pps);
	return 0;
}
#endif // CRUZ

//
// Initialize the registers in the PCIE_P_* domain.
// Note that the PCIE_P_* domain registers are reset
// each time a pciep reset happens, so this might be
// called several times in one "boot session".
//
static void pciep_initialize(void)
{
	pcie_fixup_pciecap();
	pcie_unmask_int_errs();
	pcie_ini_init();
}

static void
pcie_serdes_pll_disable(void)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
	u32 i, addr, val;

	for (i = 0; i < 4; i++) {
		addr = PCIEC_MREGS_CFG_SER0_PLL +
	               ((PCIEC_MREGS_CFG_SER1_PLL - PCIEC_MREGS_CFG_SER0_PLL)
			* i);
		val = readreg(addr);
		val |= PCIEC_MREGS_CFG_SER0_PLL_ENABLE_MAN;
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER0_PLL_ENABLE_MAN;
		writereg(addr, val);
	}
	for (i = 0; i < 15; i++) {
		addr = PCIEC_MREGS_CFG_SER_RX0 +
 	               ((PCIEC_MREGS_CFG_SER_RX1 - PCIEC_MREGS_CFG_SER_RX0)
			* i);
		val = readreg(addr);
		val |= PCIEC_MREGS_CFG_SER_RX0_ENRX_MAN;
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER_RX0_ENRX_MAN;
		writereg(addr, val);

		addr = PCIEC_MREGS_CFG_SER_TX0 +
                       ((PCIEC_MREGS_CFG_SER_TX1 - PCIEC_MREGS_CFG_SER_TX0)
			* i);
		val = readreg(addr);
		val |= PCIEC_MREGS_CFG_SER_TX1_ENTX_MAN;
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER_TX0_ENTX_MAN;
		writereg(addr, val);
	}
#endif // CONFIG_BASE_PALO || CONFIG_BASE_SERENO
}

static void
pcie_serdes_pll_enable(void)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
	u32 i, addr, val;

	for (i = 0; i < 4; i++) {
		addr = PCIEC_MREGS_CFG_SER0_PLL +
	               ((PCIEC_MREGS_CFG_SER1_PLL - PCIEC_MREGS_CFG_SER0_PLL)
			* i);
		val = readreg(addr);
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER0_PLL_ENABLE_MAN;
		writereg(addr, val);
	}
	for (i = 0; i < 15; i++) {
		addr = PCIEC_MREGS_CFG_SER_RX0 +
                       ((PCIEC_MREGS_CFG_SER_RX1 - PCIEC_MREGS_CFG_SER_RX0)
			* i);
		val = readreg(addr);
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER_RX0_ENRX_MAN;
		writereg(addr, val);
		addr = PCIEC_MREGS_CFG_SER_TX0 +
                       ((PCIEC_MREGS_CFG_SER_TX1 - PCIEC_MREGS_CFG_SER_TX0)
			* i);
		val = readreg(addr);
		val &= ~(u_int32_t)PCIEC_MREGS_CFG_SER_TX0_ENTX_MAN;
		writereg(addr, val);
	}
#endif // CONFIG_BASE_PALO || CONFIG_BASE_SERENO
}

#if defined CONFIG_BASE_CRUZ
// CruzA Octavo Reset bits
#define PP_MREGS_SW_SOFT_RST_OCP_OCTAVO0                    (1UL << 5)
#define PP_MREGS_SW_SOFT_RST_OCP_OCTAVO1                    (1UL << 6)
#define PP_MREGS_SW_SOFT_RST_OCP_OCTAVO2                    (1UL << 7)
#define PP_MREGS_SW_SOFT_RST_OCP_OCTAVO3                    (1UL << 8)

static u_int32_t
get_serdesbits(const int pp, const int pps, const u32 width)
{
	u_int32_t serdesbits;

	// Determine serdes bits based on link width.
	switch (width) {
	case 1:
	case 2:
		// pps0 owns serdes0, pps1 owns serdes1, etc
		serdesbits = PP_MREGS_SW_SOFT_RST_OCP_OCTAVO0 << pps;
		break;
	case 4:
		// pps0 owns serdes0/1, pps2 owns serdes2/3
		serdesbits = (PP_MREGS_SW_SOFT_RST_OCP_OCTAVO0 |
			      PP_MREGS_SW_SOFT_RST_OCP_OCTAVO1) << pps;
		break;
	case 8:
		// pps0 owns serdes0/1/2/3
		serdesbits = (PP_MREGS_SW_SOFT_RST_OCP_OCTAVO0 |
			      PP_MREGS_SW_SOFT_RST_OCP_OCTAVO1 |
			      PP_MREGS_SW_SOFT_RST_OCP_OCTAVO2 |
			      PP_MREGS_SW_SOFT_RST_OCP_OCTAVO3);
		break;
	default:
		serdesbits = 0;
		break;
	}
	return serdesbits;
}

static void
pcie_set_softrst(pcie_hid_info_t *p, const int on)
{
// CruzA #defines
#define PP_MREGS_SW_SOFT_RST_OCP_MASTER                     (1UL << 4)
#define PP_MREGS_SW_SOFT_RST_OCP_PCS0123                    (1UL << 9)
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	const u32 ppsbit = 1 << pps;
	u32 serdesbits = 0;
	unsigned long flags;
	u32 reg;

	// CruzA had serdes bits that needed to be reset.
	if (readreg(MISC_MREGS_VERSION) <= 0xa0) {
		u32 width;
		ppcie_get_linkparams(p->hid, &width, NULL);
		if (width < 2) width = 2;
		serdesbits = get_serdesbits(pp, pps, width);
		if (!on) {
			// Also bringing up the octavo master and pcs0123.
			serdesbits |= (PP_MREGS_SW_SOFT_RST_OCP_MASTER |
				       PP_MREGS_SW_SOFT_RST_OCP_PCS0123);
		}
	}

	spin_lock_irqsave(&p->lock, flags);
	reg = readreg(PP_MREGS_(SW_SOFT_RST, pp));
	if (on) {
		reg |= (ppsbit | serdesbits);
	} else {
		reg &= ~(ppsbit | serdesbits);
	}
	writereg(PP_MREGS_(SW_SOFT_RST, pp), reg);
	(void)readreg(PP_MREGS_(SW_SOFT_RST, pp));
	spin_unlock_irqrestore(&p->lock, flags);
}

static void
pcie_set_hardrst(pcie_hid_info_t *p, const int on)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	const u32 ppsbit = 1 << pps;
	unsigned long flags;
	u32 reg;

	spin_lock_irqsave(&p->lock, flags);
	reg = readreg(PP_MREGS_(SW_HARD_RST, pp));
	if (on) {
		reg |= ppsbit;
		p->flags |= PCIEF_RESET;
	} else {
		reg &= ~ppsbit;
		p->flags &= ~PCIEF_RESET;
	}
	writereg(PP_MREGS_(SW_HARD_RST, pp), reg);
	(void)readreg(PP_MREGS_(SW_HARD_RST, pp));
	spin_unlock_irqrestore(&p->lock, flags);
}

static void
pcie_set_reset(pcie_hid_info_t *p, const int on)
{
	if (on) {
		// serdes reset
		pcie_set_serdes_reset(p, 1);

		// in hard reset...
		pcie_set_hardrst(p, 1);
		// ...then in soft reset.
		pcie_set_softrst(p, 1);
	} else {
		// serdes unreset
		pcie_set_serdes_reset(p, 0);

		// bring out of soft reset first...
		pcie_set_softrst(p, 0);
		// ...then bring out of hard reset.
		pcie_set_hardrst(p, 0);
	}
}

static void
pcie_reset(pcie_hid_info_t *p)
{
	tracep(p, PCIETR_INFO, "link reset");
	pcie_set_reset(p, 1);
	pcie_set_reset(p, 0);
	// hard reset will reset intr mask, so re-init
	pcie_unmask_mac_intr_hid(p->hid);
}

static int __devinit
pcie_pll_cfg1(void)
{
	int polls;

	// val0:pll_en high,load high,lresetn=0
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x03f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);
	// val1:restn=1, rest are the same
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x07f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);
	// val2:lrestn=0
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x03f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);
	// val3:lrestn=1
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x07f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);
	// same as val3
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x07f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);
	// same as above to give enough cycles for load to be stable.
	writereg(MISC_MREGS_MIPL_CLK_HPLL, 0x07f92708);
	readreg (MISC_MREGS_MIPL_CLK_HPLL);

	// now poll pll lock
	for (polls = 0; polls < 10; polls++) {
		if (readreg(MISC_MREGS_PISD_PLL_500_LOCK)) {
			return polls;
		}
		udelay(1000);
	}
	return -1;
}

static int __devinit
pcie_pll_cfg(void)
{
	int i, polls;

	for (i = 0; i < 10; i++) {
		if ((polls = pcie_pll_cfg1()) >= 0) {
			printk(KERN_INFO PFX "pcie_pll_cfg locked %d:%d\n",
			       i, polls);
			return 0;
		}
	}
	printk(KERN_ERR PFX "pcie_pll_cfg failed to lock\n");
	return -1;
}

// ================================================================
//
// Each PCIe PP module has 4 PPS "slices", each PPS has 4 dedicated
// serdes blocks, each serdes block has 2 pcie gen3 lanes.  So each PP
// has 8 pcie lanes that can be configured in x2, x4, or x8 configs
// to connect to up to 4 hosts.  These are configs provided by hardware:
//
// ================================================
// x8 - single pps x8, only 1 host interface
// ------------------------------------------------
//       /- serdes0
// pps0 /-- serdes1
//      \-- serdes2
//       \- serdes3
// ================================================
// x4 - two pps x4, 2 host interfaces
// ------------------------------------------------
// pps0 /-- serdes0
//      \-- serdes1
// pps2 /-- serdes2
//      \-- serdes3
// ================================================
// x2 - four pps x2, 4 host interfaces
// ------------------------------------------------
// pps0 <-- serdes0
// pps1 <-- serdes1
// pps2 <-- serdes2
// pps3 <-- serdes3
//
// Configuring this is a PP-global operation since it can affect the
// config of all the pps's in the group.
//
// For reference, see Cruz TOO, Table 13: Valid combinations of PP linkwidths.
//

static pps_linkwidth_t
pps_linkwidth(const u32 width)
{
	pps_linkwidth_t pps_width;

	switch (width) {
	default:
	case 2: pps_width = PPS_LINKWIDTH_x2; break;
	case 4: pps_width = PPS_LINKWIDTH_x4; break;
	case 8: pps_width = PPS_LINKWIDTH_x8; break;
	}
	return pps_width;
}

//
// Configure pps0.  pps0 can support any width x2, x4, x8
// but larger widths span the serdes of the adjacent pps's
// so they must be held in reset.
//
static void
pps0_cfg_linkwidth(const u32 pps_width, u8 *pps0, u8 *pps1, u8 *pps2, u8 *pps3)
{
	// pps0 supports any width x2, x4, x8
	switch (pps_width) {
	case PPS_LINKWIDTH_x2:
		// x2 forces pps1 to x2, pps2/3 to x4 or x2,x2
		*pps0 = pps_width;
		*pps1 = PPS_LINKWIDTH_x2;
		if (*pps2 == 0) { // pps0 was x8?
			*pps2 = PPS_LINKWIDTH_x4;
		}
		break;
	case PPS_LINKWIDTH_x4:
		// x4 means pps1 must be unused
		*pps0 = pps_width;
		*pps1 = 0;
		if (*pps2 == 0) { // pps0 was x8?
			*pps2 = PPS_LINKWIDTH_x4;
		}
		break;
	case PPS_LINKWIDTH_x8:
		// x8 means pps1,pps2,pps3 must be unused
		*pps0 = pps_width;
		*pps1 = *pps2 = *pps3 = 0;
		break;
	}
}

//
// Configure pps1.  pps1 can support only x2, and if configured
// then pps0 must not be x4 or x8 since those wider settings would
// span the serdes of our pps1 link.
//
static void
pps1_cfg_linkwidth(const u32 pps_width, u8 *pps0, u8 *pps1, u8 *pps2, u8 *pps3)
{
	// pps can only be configured x2
	if (pps_width != PPS_LINKWIDTH_x2) return;

	*pps1 = pps_width;
	*pps0 = PPS_LINKWIDTH_x2;
	if (*pps2 == 0) {
		*pps2 = PPS_LINKWIDTH_x4;
	}
}

//
// Configure pps2.  pps2 can support x2 or x4.
//
static void
pps2_cfg_linkwidth(const u32 pps_width, u8 *pps0, u8 *pps1, u8 *pps2, u8 *pps3)
{
	// pps2 supports width x2, x4
	switch (pps_width) {
	case PPS_LINKWIDTH_x2:
		// configuring pps2 means pps3 is x2, pps0 must not be x8
		*pps2 = pps_width;
		*pps3 = PPS_LINKWIDTH_x2;
		if (*pps0 == PPS_LINKWIDTH_x8) {
			*pps0 = PPS_LINKWIDTH_x4;
		}
		break;
	case PPS_LINKWIDTH_x4:
		// configuring pps2 x4 means pps0 must not be x8, and
		// pps3 must be unused
		*pps2 = pps_width;
		*pps3 = 0;
		if (*pps0 == PPS_LINKWIDTH_x8) {
			*pps0 = PPS_LINKWIDTH_x4;
		}
		break;
	}
}

//
// Configure pps3.  Like pps1, pps3 can support only x2, and if configured
// then pps2 must not be x4 and pps0 must not be x8 since those wider settings
// would span the serdes of our pps3 link.
//
static void
pps3_cfg_linkwidth(const u32 pps_width, u8 *pps0, u8 *pps1, u8 *pps2, u8 *pps3)
{
	// pps can only be configured x2
	if (pps_width != PPS_LINKWIDTH_x2) return;

	*pps3 = pps_width;
	*pps2 = PPS_LINKWIDTH_x2;
	if (*pps0 == PPS_LINKWIDTH_x8) {
		*pps0 = PPS_LINKWIDTH_x4;
	}
}

static void
pps_cfg_linkwidth(const u8 pp, const u8 pps, const u32 width)
{
	u_int32_t cfg_linkwidth_reg = PP_MREGS_(CFG_LINKWIDTH, pp);
	u_int32_t cfg;
	u_int8_t pps0, pps1, pps2, pps3;
	pps_linkwidth_t pps_width;

	if (readreg(MISC_MREGS_VERSION) <= 0xa0) {
#define PP_MREGS_CFG_LINKWIDTH_CRUZA_OFFSET 0x6a530
		cfg_linkwidth_reg = PP_MREGS_(CFG_LINKWIDTH_CRUZA, pp);
	}

	// If requesting x0(reset), config x2 here, link stays in reset
	// If requesting x1, config x2 and we'll mask 1 lane in cfg_mac1.
	if (width <= 1) {
		pps_width = pps_linkwidth(2);
	} else {
		pps_width = pps_linkwidth(width);
	}

	//
	// Read the current lane config, adjust for the new setting
	// including validating any adjacent links that share serdes
	// lanes, then replace the config.
	//
	cfg = readreg(cfg_linkwidth_reg);
	pp_mregs_cfg_linkwidth_get(&cfg, &pps0, &pps1, &pps2, &pps3);

	switch (pps) {
	case 0:
		pps0_cfg_linkwidth(pps_width, &pps0, &pps1, &pps2, &pps3);
		break;
	case 1:
		pps1_cfg_linkwidth(pps_width, &pps0, &pps1, &pps2, &pps3);
		break;
	case 2:
		pps2_cfg_linkwidth(pps_width, &pps0, &pps1, &pps2, &pps3);
		break;
	case 3:
		pps3_cfg_linkwidth(pps_width, &pps0, &pps1, &pps2, &pps3);
		break;
	}
	pp_mregs_cfg_linkwidth_make(&cfg, pps0, pps1, pps2, pps3);
	writereg(cfg_linkwidth_reg, cfg);
	(void)readreg(cfg_linkwidth_reg);
}
#endif /* CRUZ */

static void
pcie_bus_chg(pcie_hid_info_t *p)
{
#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    // read config space word that contains secondary bus
    const u_int32_t busdw = readreg(PCIEP_MREGS_DHS_PCIE_SW_CFG + 0x18);
    const u_int32_t pcie_cap = PCIEP_MREGS_DHS_PCIE_SW_CFG + 0x70;
#elif defined CONFIG_BASE_CRUZ
    // pp,pps only used for Cruz
    const u8 pp  __attribute__ ((unused)) = ppcie_hid_pp(p->hid);
    const u8 pps __attribute__ ((unused)) = ppcie_hid_pps(p->hid);
    const u_int32_t busdw = readreg(PP_MREGS_PPS_(DHS_SW_CFG, pp, pps) + 0x18);
    const u_int32_t pcie_cap = PP_MREGS_PPS_(DHS_SW_CFG, pp, pps) + 0x70;
#endif
    const int hostid = ppcie_hostid_make(p->hid, 0);
    const u_int8_t secbus = busdw >> 8; // secondary bus is byte 2
    const u_int32_t lnksta = readreg(pcie_cap + (PCI_EXP_LNKSTA & ~0x3)) >> 16;
    const u_int32_t neg_gen = lnksta & 0xf;
    const u_int32_t neg_wid = (lnksta >> 4) & 0x3f;
    static u8 reset_with_host;

    tracep(p, PCIETR_INFO,
           "host%d: gen%d x%d pcie secondary bus changed to %d",
           hostid, neg_gen, neg_wid, secbus);

#if defined CONFIG_BASE_CRUZ
    pcie_unmask_mac_intr(pp, pps,
			 PPSC_MREGS_PPS0_INT_MAC_INI_BARRIER_REQ_EXCESS_STALL |
			 PPSC_MREGS_PPS0_INT_MAC_LTSSM_ST_CHANGED);
#endif

    //
    // Secondary bus number change indicates
    // the host has begun scanning the pcie bus.
    // This is a good indication that we are done
    // with the reset flurry at startup, so arrange
    // to reset with the host now.
    //
	if (board_conf_reset_with_host() && !reset_with_host) {
		printk(KERN_DEBUG PFX "host%d: reset with host\n", hostid);
		pcie_reset_with_host(1);
		reset_with_host = 1;
	}

    // Generate this event *after* we're resetting-with-host,
    // because mcp may read PCIEP_* registers when it gets this event.
    ppcie_evt_bus_chg(hostid, secbus);
}

static void
pcie_set_cfg(pcie_hid_info_t *p)
{
    const u8 pp  = ppcie_hid_pp(p->hid);
    const u8 pps = ppcie_hid_pps(p->hid);
    const int crs = (p->flags & PCIEF_CFG_CRS) != 0;

    tracep(p, PCIETR_INFO, "pcie_set_cfg hid %d crs %d", p->hid, crs);
    // if !crs, then vnics are ready
    if (!crs) {
        // const int hostid = ppcie_hostid_make(p->hid, 0);
        ppcie_dev_link_reset(p->hid, 0);
        pcie_tar_main_mode_en();
    }
    pcie_set_crs(pp, pps, crs);
}

// r = a + b
static void
pcie_hid_stats_add(pcie_hid_stats_t *r,
                   pcie_hid_stats_t *a, pcie_hid_stats_t *b)
{
    u_int64_t *rp, *ap, *bp;
    int i, n;

    rp = (u_int64_t *)r;
    ap = (u_int64_t *)a;
    bp = (u_int64_t *)b;

    n = sizeof (pcie_hid_stats_t) / sizeof (u_int64_t);
    for (i = 0; i < n; i++, rp++, ap++, bp++) {
        *rp = *ap + *bp;
    }
}

//
// Sum the hid stats across all ecpus.
//
static void
pcie_hid_stats_sum(hid_t hid, pcie_hid_stats_t *hstats)
{
    int i;
    extern pm_impl_t *pmimpl_vaddr;

    for (i = 0; i < PALO_MAX_ECPU; i++) {
        pcie_hid_stats_t *estats = &pmimpl_vaddr->pcie_hid_stats[i][hid];
        pcie_hid_stats_add(hstats, hstats, estats);
    }
}

//
// Sum the hid stats across all ecpus,
// and reset ecpu stats.
//
static void
pcie_hid_stats_reset(pcie_hid_info_t *p)
{
    int i;
    pcie_hid_stats_t *tstats = &p->tstats;
    extern pm_impl_t *pmimpl_vaddr;

    for (i = 0; i < PALO_MAX_ECPU; i++) {
        pcie_hid_stats_t *estats = &pmimpl_vaddr->pcie_hid_stats[i][p->hid];
        pcie_hid_stats_add(tstats, tstats, estats);
        memset(estats, 0, sizeof (pcie_hid_stats_t));
    }
}

static void
pcie_err_stats_reset(pcie_hid_info_t *p)
{
    pcie_hid_int_err_stats_t *estats = &p->int_err_stats;
    pcie_hid_int_err_stats_t *tstats = &p->tint_err_stats;
    u_int64_t *tp, *ep;
    int i, n;

    tp = (u_int64_t *)tstats;
    ep = (u_int64_t *)estats;

    n = sizeof (pcie_hid_int_err_stats_t) / sizeof (u_int64_t);
    for (i = 0; i < n; i++, tp++, ep++) {
        *tp += *ep;
    }

    memset(estats, 0, sizeof (pcie_hid_int_err_stats_t));
}

static void
pcie_stats_reset(pcie_hid_info_t *p)
{
    pcie_hid_stats_reset(p);
    pcie_err_stats_reset(p);
}

static void
pcie_hostup(pcie_hid_info_t *p)
{
    const int hostid = ppcie_hostid_make(p->hid, 0);

    tracep(p, PCIETR_INFO, "hostup %d hid %d iter %d", hostid, p->hid,
                                                       p->hostdn_iter);
    ppcie_evt_hostup(hostid, p->hostdn_iter);
}

static void
pcie_hostdn(pcie_hid_info_t *p)
{
    const int hostid = ppcie_hostid_make(p->hid, 0);
    const u8 pp  = ppcie_hid_pp(p->hid);
    const u8 pps = ppcie_hid_pps(p->hid);

    // Don't close barrier if we are in loopback.
    if ((p->flags & PCIEF_LOOPBACK) == 0) {
        pcie_close_and_idle(pp, pps);
    }
    ppcie_bdft_host_reset(p->hid, 0);

    tracep(p, PCIETR_INFO, "hostdn %d hid %d iter %d", hostid, p->hid,
                                                       p->hostdn_iter);
    ppcie_dev_hid_reset(p->hid, 0, 0);

    pcie_barrier_gate_open(pp, pps);
    pcie_set_crs(pp, pps, 1);
    p->flags |= PCIEF_CFG_CRS;
    ppcie_evt_hostdn(hostid, p->hostdn_iter);
    p->hostdn_iter++;
}

static int
pcie_macup(pcie_hid_info_t *p)
{
    const u8 pp  = ppcie_hid_pp(p->hid);
    const u8 pps = ppcie_hid_pps(p->hid);

    if (pcie_reset_on_macup) {
        pcie_asic_reset();
        /* NOTREACHED */
    }

#if defined CONFIG_BASE_CRUZ
    // Do this before LTSSM_ENABLE
    if (pps_init(p) < 0) {
        printk("pps_iint failed\n");
	    return -1;
    }
#endif // CRUZ
    pcie_stats_reset(p);
    pcie_barrier_open(pp, pps);
    tracep(p, PCIETR_INFO, "ltssm enabled");
    pcie_link_up(pp, pps);
    pciep_initialize();

    pcie_barrier_gate_open(pp, pps);
    pciep_debug_port_disable();
    pcie_unmask_err_intr_hid(p->hid);
    return 0;
}

static void
pcie_macdn(pcie_hid_info_t *p)
{
    if (p->flags & PCIEF_LOOPBACK) {
        p->flags &= ~PCIEF_LOOPBACK;
        tracep(p, PCIETR_INFO, "exit loopback");
    }
	pcie_log_reset_state();
	if (pcie_sw_reset_with_host && !board_conf_cisco_card_mode()) {
		/*
		 * Emulating reset-with-host function in software.
		 * macdn event triggers reset to ASIC if not in cisco_card_mode.
		 * That is what the hardware would do if the PCIE_FORCE_RESET_OK
		 * register is set.
		 */
        pcie_asic_reset();
        /* NOTREACHED */
    }
    pcie_clear_fault(p); // reset any fault conditions
}

static void
pcie_linkup(pcie_hid_info_t *p)
{
    pcie_hostup(p);
    pcie_set_cfg(p);
}

static void
pcie_linkdn(pcie_hid_info_t *p)
{
#if defined CONFIG_BASE_CRUZ
    const u8 pp  = ppcie_hid_pp(p->hid);
    const u8 pps = ppcie_hid_pps(p->hid);
    const u32 sta_barrier = readreg(PPSC_MREGS_(STA_BARRIER, pp, pps));
    const u32 sta_mac = readreg(PPSP_MREGS_(STA_MAC, pp, pps));
    const u32 sta_mac_diag = readreg(PPSP_MREGS_(STA_MAC_DIAG, pp, pps));
    const u32 cfg_mac0 = readreg(PPSP_MREGS_(CFG_MAC0, pp, pps));
    const u32 cfg_mac1 = readreg(PPSC_MREGS_(CFG_MAC1, pp, pps));

    tracep(p, PCIETR_INFO,
           "linkdn sta_barrier %d sta_mac 0x%x sta_mac_diag 0x%x",
           sta_barrier, sta_mac, sta_mac_diag);
    tracep(p, PCIETR_INFO,
           "  cfg_mac0 0x%x cfg_mac1 0x%x",
           cfg_mac0, cfg_mac1);

    pcie_mask_mac_intr(pp, pps,
		       PPSC_MREGS_PPS0_INT_MAC_INI_BARRIER_REQ_EXCESS_STALL |
		       PPSC_MREGS_PPS0_INT_MAC_LTSSM_ST_CHANGED);
#endif // CRUZ
    pcie_hostdn(p);
}

////////////////////////////////////////////////////////////////
// FSM routines
////////////////////////////////////////////////////////////////

// Invalid transition.  Don't expect this to happen.
static void
pcie_fsm_inv(pcie_hid_info_t *p)
{
    tracep(p, PCIETR_FSM, "pcie_fsm_inv: %s + %s => INVALID_TRANSITION",
           pciest_names[p->state], pcieev_names[p->event]);
    pcie_set_fault(p, "invalid transition", PPCIE_ERRCODE_FSM_FAULT);
}

// NOP transition, nothing to do
static void
pcie_fsm_nop(pcie_hid_info_t *p)
{
    // This space intentionally left blank.
}

// MAC is up.
static void
pcie_fsm_macup(pcie_hid_info_t *p)
{
    if (pcie_macup(p) == 0)
	    p->state = PCIEST_DOWN;
    else
	    p->state = PCIEST_FAULT;
}

// Low-level MAC went down.
static void
pcie_fsm_macdn(pcie_hid_info_t *p)
{
    p->state = PCIEST_MACRST;
    pcie_macdn(p);
}

// Link came up.
static void
pcie_fsm_linkup(pcie_hid_info_t *p)
{
    p->state = PCIEST_LINKUP;
    pcie_linkup(p);
}

// Link went down.
static void
pcie_fsm_linkdn(pcie_hid_info_t *p)
{
    p->state = PCIEST_DOWN;
    pcie_linkdn(p);
}

//
// We have been signalled that the host has written the bridge secondary
// bus number register.  This is a clue that our pcie bus is discovered,
// and downstream device enumeration has begun.
//
static void
pcie_fsm_up(pcie_hid_info_t *p)
{
    p->state = PCIEST_UP;
    pcie_bus_chg(p);
}

//
// We have been up and our bus was scanned (secondary bus number
// register written), but then the mac went down.
//
static void
pcie_fsm_up_macdn(pcie_hid_info_t *p)
{
    p->state = PCIEST_MACRST;
    pcie_linkdn(p);
    pcie_macdn(p);
}

//
// We were up then maybe someone did a "hard reset" which caused
// the mac to suddenly go to reset then out again.
//
static void
pcie_fsm_up_macup(pcie_hid_info_t *p)
{
    p->state = PCIEST_DOWN;
    pcie_linkdn(p);
    pcie_macdn(p);
    pcie_macup(p);
}

//
// We have been up and our bus was scanned (secondary bus number
// register written), but then the link went down.
//
static void
pcie_fsm_up_linkdn(pcie_hid_info_t *p)
{
    p->state = PCIEST_DOWN;
    pcie_linkdn(p);
}

static void
pcie_fsm_bus_chg(pcie_hid_info_t *p)
{
    pcie_bus_chg(p);
}

static void
pcie_fsm_set_cfg(pcie_hid_info_t *p)
{
    pcie_set_cfg(p);
}

#define INV pcie_fsm_inv
#define NOP pcie_fsm_nop
#define MUP pcie_fsm_macup
#define MDN pcie_fsm_macdn
#define LUP pcie_fsm_linkup
#define LDN pcie_fsm_linkdn
#define UP_ pcie_fsm_up
#define UMD pcie_fsm_up_macdn
#define UMU pcie_fsm_up_macup
#define ULD pcie_fsm_up_linkdn
#define BUS pcie_fsm_bus_chg
#define CFG pcie_fsm_set_cfg
typedef void (*pcie_fsm_func_t)(pcie_hid_info_t *p);
static pcie_fsm_func_t pcie_fsm_transition[PCIEST_MAX][PCIEEV_MAX] = {
    // [STATE]       + EV:
    //                   MACDN
    //                   |    MACUP
    //                   |    |    LINKDN
    //                   |    |    |    LINKUP
    //                   |    |    |    |    BUS_CHG
    //                   |    |    |    |    |    SET_CFG
    [PCIEST_MACRST]  = { NOP, MUP, INV, INV, INV, NOP },
    [PCIEST_DOWN]    = { MDN, NOP, INV, LUP, INV, NOP },
    [PCIEST_LINKUP]  = { MDN, UMU, LDN, INV, UP_, CFG },
    [PCIEST_UP]      = { UMD, UMU, ULD, INV, BUS, CFG },
    [PCIEST_FAULT]   = { MDN, NOP, NOP, NOP, NOP, NOP },
};

static void
pcie_fsm(pcie_hid_info_t *p, pcieev_t ev)
{
    pciest_t st;
    unsigned long flags;

    spin_lock_irqsave(&p->lock, flags);
    st = p->state;      // get current state
    p->event = ev;      // set current event

    pcie_fsm_transition[st][ev](p);

    tracep(p, PCIETR_FSM, "pcie_fsm: %s + %s => %s",
           pciest_names[st], pcieev_names[ev], pciest_names[p->state]);

    spin_unlock_irqrestore(&p->lock, flags);
}

void
ppcie_pcie_vnics_ready(const int hostid, const u8 pp, const u8 pps,
                       u_int32_t hostready_iter)
{
    const hid_t hid = ppcie_hostid_hid(hostid);
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    unsigned long flags;

    spin_lock_irqsave(&p->lock, flags);
    if (hostready_iter == p->hostdn_iter) {
        p->flags &= ~PCIEF_CFG_CRS;
    }
    spin_unlock_irqrestore(&p->lock, flags);
    tracep(p, PCIETR_INFO, "pcie_vnics_ready hid %d iter(%d, %d)",
              hid, hostready_iter, p->hostdn_iter);
    pcie_fsm(p, PCIEEV_SET_CFG);
}

static void
_pcie_mac_intr_hid(u_int32_t hid, void *arg)
{
    pcie_hid_stats_t *hstats;
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    // pp,pps only used for Cruz
    const u8 pp  __attribute__ ((unused)) = ppcie_hid_pp(hid);
    const u8 pps __attribute__ ((unused)) = ppcie_hid_pps(hid);
    const u32 unmasked = readreg(INT_MAC_MSK_REG(pp, pps));
    const u32 int_mac = readreg(INT_MAC_REG(pp, pps)) & unmasked;
    u32 sta_rst;
    irqreturn_t *irqretp = arg;

    // ack intrs first, before processing
    writereg(INT_MAC_REG(pp, pps), int_mac);
    (void)readreg(INT_MAC_REG(pp, pps)); // read to flush
    memory_barrier();

    sta_rst = readreg(STA_RST_REG(pp, pps));

    if (int_mac) {
        tracep(p, PCIETR_INFO, "int_mac 0x%x sta_rst 0x%x", int_mac, sta_rst);
    }

    hstats = ppcie_get_pcie_hid_stats(p->hid);
#if defined CONFIG_BASE_CRUZ
#define is_pcie_loopback(st) ((st) >= 0x1a && (st) <= 0x1d)

    if (int_mac & PPSC_MREGS_PPS0_INT_MAC_INI_BARRIER_REQ_EXCESS_STALL) {
	tracep(p, PCIETR_INFO, "int_mac ini_barrier_req_excess_stall");
	hstats->ini_barrier_req_excess_stall++;
    }
    if (int_mac & PPSC_MREGS_PPS0_INT_MAC_LTSSM_ST_CHANGED) {
	unsigned int now, delta;
	int ltssmst;

	pps_sta_mac(pp, pps, &ltssmst, NULL, NULL);
	tracep(p, PCIETR_INFO, "int_mac ltssm_st_changed 0x%02x", ltssmst);
	hstats->ltssm_st_changed++;
        if (is_pcie_loopback(ltssmst)) {
            if ((p->flags & PCIEF_LOOPBACK) == 0) {
                p->flags |= PCIEF_LOOPBACK;
                tracep(p, PCIETR_INFO, "enter loopback");
            }
        } else {
            if (p->flags & PCIEF_LOOPBACK) {
                p->flags &= ~PCIEF_LOOPBACK;
                tracep(p, PCIETR_INFO, "exit loopback");
            }
        }

	// Links with low quality might send "many" ltssm_st_changed
	// interrupts.  If we get too many interrupts in a small amount
        // of time, disable the int and send a notice.
	now = jiffies_to_msecs(get_jiffies_64());
        delta = now - p->ltssm_ms;
	if (delta < pcie_ltssm_ms) {
            if (++p->ltssm_st_changed >= pcie_ltssm_count) {
                const int hostid = ppcie_hostid_make(p->hid, 0);
                pcie_mask_mac_intr(pp, pps,
                                   PPSC_MREGS_PPS0_INT_MAC_LTSSM_ST_CHANGED);
                ppcie_evt_linkerr(hostid,
                                  PPCIE_ERRTYPE_SOFT,
                                  PPCIE_ERRCODE_LTSSM_THRESHOLD);
                // We took action above, so reset for next time.
                p->ltssm_ms = now;
                p->ltssm_st_changed = 0;
            }
        } else {
            // This is the first ltssm_st_changed of a new burst.
            p->ltssm_ms = now;
            p->ltssm_st_changed = 1;
        }
    }
#endif // CONFIG_BASE_CRUZ
    if (int_mac & INT_MACF_XMLH_LINK_UP2DN) {
        hstats->link_dn++;
        pcie_fsm(p, PCIEEV_LINKDN);
    }
    if (int_mac & INT_MACF_CORE_RST_DN2UP) {
        hstats->mac_dn++;
        pcie_fsm(p, PCIEEV_MACDN);
    }
    if (sta_rst & STA_RST_MAC_CORE_RST_N) {
        if (int_mac & INT_MACF_CORE_RST_UP2DN ||
            int_mac & INT_MACF_CORE_RST_DN) {
            hstats->mac_up++;
            pcie_fsm(p, PCIEEV_MACUP);
        }
        if (int_mac & INT_MACF_XMLH_LINK_DN2UP) {
            hstats->link_up++;
            pcie_fsm(p, PCIEEV_LINKUP);
        }
        if (int_mac & INT_MACF_SEC_BUS_NUM_CHG) {
            hstats->bus_chg++;
            pcie_fsm(p, PCIEEV_BUS_CHG);
        }
    }
    // Only need this to get us a MACUP event at power up.
    // Once up, _UP2DN will interrupt us.
    if (int_mac & INT_MACF_CORE_RST_DN) {
        pcie_mask_mac_intr(pp, pps, INT_MAC_MSKF_CORE_RST_DN);
    }
    if (int_mac) {
        *irqretp = IRQ_HANDLED;
    }
}

static void
log_int_err(pcie_hid_info_t *p, u32 int_err)
{
    pcie_hid_int_err_stats_t *estats = &p->int_err_stats;

#define CHK(int_err, FIELD, stat) \
    if ((int_err) & INT_ERR_REG_(FIELD, pp, pps)) { \
        tracep(p, PCIETR_INFO, "int_err: " #FIELD); \
        estats->stat++; \
    }

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    CHK(int_err, TAR_RSP_UNEXP, tar_rsp_unexp);
    CHK(int_err, TAR_RSP_FEN_AERR, tar_rsp_fen_aerr);
    CHK(int_err, TAR_RSP_DIR_AERR, tar_rsp_dir_aerr);
    CHK(int_err, TAR_RSP_DIR_DERR, tar_rsp_dir_derr);
    CHK(int_err, TAR_RSP_IND_INFO_AERR, tar_rsp_ind_info_aerr);
    CHK(int_err, INI_REQ_CRF_ERR_UNC, ini_req_crf_err_unc);
    CHK(int_err, INI_REQ_BDFT_ERR_UNC, ini_req_bdft_err_unc);
#elif defined CONFIG_BASE_CRUZ
    CHK(int_err, INI_REQ_ORD_CRF_ERR_UNC, ini_req_ord_crf_err_unc);
    CHK(int_err, INI_REQ_RO_CRF_ERR_UNC, ini_req_ro_crf_err_unc);
    CHK(int_err,
        INI_REQ_STALL_TIMER_EXCESS_HALT, ini_req_stall_timer_excess_halt);
    CHK(int_err,
        TAR_REQ_STALL_TIMER_EXCESS_HALT, tar_req_stall_timer_excess_halt);
    CHK(int_err,
        TAR_CPL_STALL_TIMER_EXCESS_HALT, tar_cpl_stall_timer_excess_halt);
#endif /* CONFIG_BASE */
    CHK(int_err, INI_REQ_BUS_MASTER_DIS, ini_req_bus_master_dis);
    CHK(int_err, INI_REQ_DRF_ERR_UNC, ini_req_drf_err_unc);
    CHK(int_err, INI_REQ_UNS, ini_req_uns);
    CHK(int_err, INI_REQ_BDFT_MISS, ini_req_bdft_miss);
    CHK(int_err, INI_CPL_UNEXP, ini_cpl_unexp);
    CHK(int_err, INI_CPL_DLLP_ABORT, ini_cpl_dllp_abort);
    CHK(int_err, INI_CPL_TLP_ABORT, ini_cpl_tlp_abort);
    CHK(int_err, INI_CPL_TLP_POISONED, ini_cpl_tlp_poisoned);
    CHK(int_err, INI_CPL_ECRC_ERR, ini_cpl_ecrc_err);
    CHK(int_err, INI_CPL_STAT_ERR, ini_cpl_stat_err);
    CHK(int_err, INI_CPL_TIMEOUT, ini_cpl_timeout);
    CHK(int_err, INI_RSP_AERR, ini_rsp_aerr);
    CHK(int_err, INI_RSP_DERR, ini_rsp_derr);
}

static void
_pcie_err_intr_hid(u_int32_t hid, void *arg)
{
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    // pp,pps only used for Cruz
    const u8 pp  __attribute__ ((unused)) = ppcie_hid_pp(hid);
    const u8 pps __attribute__ ((unused)) = ppcie_hid_pps(hid);
    u32 int_err, unmasked;
    irqreturn_t *irqretp = arg;

    // If in MACRST state then we won't be receiving int_err interrupts,
    // and we will generate a bus error (discarded by bus error handler)
    // if we access the int_err register.  This tests saves us from the
    // log noise of discarding a bunch of bus error notices.
    if (p->state == PCIEST_MACRST) return;

    unmasked = readreg(INT_ERR_MSK_REG(pp, pps));
    int_err = readreg(INT_ERR_REG(pp, pps)) & unmasked;
    if (int_err) {
        // ack intrs first, before processing
        writereg(INT_ERR_REG(pp, pps), int_err);
        (void)readreg(INT_ERR_REG(pp, pps));
        memory_barrier();

        log_int_err(p, int_err);
        *irqretp = IRQ_HANDLED;
    }
}

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
static irqreturn_t
pcie_mac_intr_handler(int irq, void *arg)
{
    irqreturn_t irqret = IRQ_NONE;

    _pcie_mac_intr_hid(0, &irqret);
    return (irqret);
}
static irqreturn_t
pcie_err_intr_handler(int irq, void *arg)
{
    irqreturn_t irqret = IRQ_NONE;

    _pcie_err_intr_hid(0, &irqret);
    return (irqret);
}
#elif defined CONFIG_BASE_CRUZ
static void
_pcie_intr_handler(u_int32_t hid, void *arg)
{
    irqreturn_t *irqretp = arg;

    if (*irqretp == IRQ_NONE) _pcie_mac_intr_hid(hid, irqretp);
    if (*irqretp == IRQ_NONE) _pcie_err_intr_hid(hid, irqretp);
}

static irqreturn_t
pcie_intr_handler(int irq, void *arg)
{
    irqreturn_t irqret = IRQ_NONE;

    ppcie_foreach_hid(_pcie_intr_handler, &irqret);
    return (irqret);
}
#endif /* CONFIG_BASE */

static irqreturn_t
pwr_intr(int irq, void *data)
{
    unsigned long flags;
    int main_power;

    spin_lock_irqsave(&pcie_lock, flags);
    main_power = board_conf_pwr_intr_chg_pol();
    spin_unlock_irqrestore(&pcie_lock, flags);

    if (main_power) {
        pcie_serdes_pll_enable();       // enable in main power mode
    } else {
        pcie_serdes_pll_disable();      // disable in standby power mode
    }
    return IRQ_HANDLED;
}

static int
pwr_intr_init(struct platform_device *dev)
 {
     int x = board_conf_pwr_intr_init();
     if (x != -1) {
         int pwr_irq = palo_int_to_irq(0, 22 + x);
         int ret = request_irq(pwr_irq, pwr_intr, 0, DRV_NAME "-pwr", dev);
         if (ret != 0) {
             printk(KERN_ERR PFX "Unable to request pwr irq %d\n", pwr_irq);
             return ret;
         }
     }
     return 0;
}

#if defined CONFIG_BASE_CRUZ
static struct timer_list pcie_link_poll_timer;
static int pcie_link_poll_timer_ms = 10;

//
// The pcie link sometimes gets stuck in compliance state when
// our pcie link comes up and the peer is there but not up.
// The link hardware state machine automatically dives into
// compliance state in that case, and the only way out is to reset
// our side.  We poll for this state occasionally and if we find
// it issue a reset.
//
static void
pcie_link_poll_hid(u_int32_t hid, void *arg)
{
	pcie_hid_info_t *p = &pcie_hid_info[hid];
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);

	//
	// We only need to check for compliance state if we are
	// in DOWN state.  If we are still in reset then we can't
	// be in compliance state and we can't poll the sta_mac reg
	// because the pcie pp/pps block might be held in reset.
	// If we are UP then we have made it past the compliance state
	// in the hardware state machine.
	//
	if (p->state == PCIEST_DOWN && pps_in_compliance_state(pp, pps)) {
		tracep(p, PCIETR_INFO, "pcie compliance state reset");
		// The reset effectively is a manual MACDN.
		pcie_fsm(p, PCIEEV_MACDN);
		pcie_reset(p);
	}
}

static void
pcie_link_poll(unsigned long arg)
{
	ppcie_foreach_hid(pcie_link_poll_hid, NULL);
	mod_timer(&pcie_link_poll_timer,
		  jiffies + HZ / 1000 * pcie_link_poll_timer_ms);
}

static void
pcie_link_poll_init(void)
{
	init_timer(&pcie_link_poll_timer);
	pcie_link_poll_timer.function = pcie_link_poll;
	pcie_link_poll_timer.expires = (jiffies +
					(HZ / 1000 * pcie_link_poll_timer_ms));
	add_timer(&pcie_link_poll_timer);
}

//
// Validate the new width is valid for this interface.  We have to validate
// that the width is 0,1,2,4,8, that this width is supported by this pps,
// and that no other pps's are using pcie lanes that would be needed by
// this pps.
//
// For example, consider pps0, pps1, pps2, pps3.  If we want to configure
// pps2 in x4 mode, we must verify that pps3 is not up in x2 mode, and
// pps0 must not be up in x8 mode.  If either of those are true, then
// there is a conflict with the pcie lanes in use by another pps in
// the same pp.
//
// The valid_linkwidth[] table encodes the valid link width
// combinations as described in the Cruz TOO, Section 2.6.5,
// Table 13: "Valid combinations of PP linkwidths".
//
static int
pcie_valid_linkwidth(pcie_hid_info_t *p, const u32 width)
{
	const u8 pps = ppcie_hid_pps(p->hid);
	u32 ow[4], nw[4], nwx[4];
	int i;
#define x0      0
#define x2      1
#define x4      2
#define x8      3
#define xINV	0xff
	static const u8 map_xX[9] = {
		x0, x2, x2, xINV, x4, xINV, xINV, xINV, x8
	};
	// See Cruz TOO, Section 2.6.5,
	// Table 13: "Valid combinations of PP linkwidths".
	static const u8 valid_linkwidth[4][4][4][4] = {
		[x0][x0][x0][x0] = 1, // 0 links
		[x0][x0][x0][x2] = 1, // x2 combos
		[x0][x0][x2][x0] = 1,
		[x0][x0][x2][x2] = 1,
		[x0][x2][x0][x0] = 1,
		[x0][x2][x0][x2] = 1,
		[x0][x2][x2][x0] = 1,
		[x0][x2][x2][x2] = 1,
		[x2][x0][x0][x0] = 1,
		[x2][x0][x0][x2] = 1,
		[x2][x0][x2][x0] = 1,
		[x2][x0][x2][x2] = 1,
		[x2][x2][x0][x0] = 1,
		[x2][x2][x0][x2] = 1,
		[x2][x2][x2][x0] = 1,
		[x2][x2][x2][x2] = 1,
		[x0][x0][x4][x0] = 1, // 1 x4 link
		[x0][x2][x4][x0] = 1,
		[x2][x0][x4][x0] = 1,
		[x2][x2][x4][x0] = 1,
		[x4][x0][x0][x0] = 1,
		[x4][x0][x0][x2] = 1,
		[x4][x0][x2][x0] = 1,
		[x4][x0][x2][x2] = 1,
		[x4][x0][x4][x0] = 1, // 2 x4 links
		[x8][x0][x0][x0] = 1, // 1 x8 link
	};

	// Check if this is a valid width to consider (0,1,2,4,8).
	if (width > 8 || (width & (width - 1)) != 0) {
		printk(KERN_ERR PFX
		       "pcie_valid_linkwidth: %s invalid width %d\n",
		       p->name, width);
		tracep(p, PCIETR_INFO,
		       "pcie_valid_linkwidth: %s invalid width %d",
		       p->name, width);
		return (-EINVAL);
	}

	// Fetch the original link widths of all pps's in our group.
	for (i = 0; i < 4; i++) {
		ppcie_get_linkparams((p->hid & ~0x3) + i, &ow[i], NULL);
		nw[i] = ow[i];
	}
	// Replace our desired new width.
	nw[pps] = width;

	// Map widths to xX index [0-3] into valid_linkwidth[] table.
	for (i = 0; i < 4; i++) {
		nwx[i] = map_xX[ nw[i] ];
	}

	tracep(p, PCIETR_INFO,
	       "pcie_valid_linkwidth: [%d %d %d %d] -> [%d %d %d %d]",
	       ow[0], ow[1], ow[2], ow[3],
	       nw[0], nw[1], nw[2], nw[3]);

	// Test if this table row is valid.
	if (!valid_linkwidth[ nwx[0] ][ nwx[1] ][ nwx[2] ][ nwx[3] ]) {
		printk(KERN_ERR PFX
		       "pcie_valid_linkwidth: %s bad width combo "
		       "[%d %d %d %d]\n",
		       p->name, nw[0], nw[1], nw[2], nw[3]);
		tracep(p, PCIETR_INFO,
		       "pcie_valid_linkwidth: %s bad width combo "
		       "[%d %d %d %d]",
		       p->name, nw[0], nw[1], nw[2], nw[3]);
		return (-EIO);
	}

	return (0);
}

static int
pcie_set_linkwidth(pcie_hid_info_t *p, const u32 width, const u32 flags)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	int i, num_affected, ret;

	// Only allow linkwidth changes while link down.
	if (p->state > PCIEST_DOWN && !pcie_set_linkwidth_while_up) {
		printk(KERN_ERR PFX
		       "pcie_set_linkwidth: %s width %d flags 0x%x"
			" when link is active, state %s\n",
		       p->name, width, flags, pciest_names[p->state]);
		tracep(p, PCIETR_INFO,
		       "pcie_set_linkwidth: width %d flags 0x%x"
			" when link is active, state %s",
		       width, flags, pciest_names[p->state]);
		return (-EBUSY);
	}

	// Check for a valid link config.
	ret = pcie_valid_linkwidth(p, width);
	if (ret < 0) {
		return (ret);
	}

	printk(KERN_INFO PFX
               "pcie_set_linkwidth: %s width %d flags 0x%x\n",
               p->name, width, flags);
	tracep(p, PCIETR_INFO,
               "pcie_set_linkwidth: width %d flags 0x%x", width, flags);

	switch (width) {
	case 0: // 0 = Disable the link, leave pps in reset.
	case 1:
	case 2: num_affected = 1; break;
	case 4: num_affected = 2; break;
	case 8: num_affected = 4; break;
	default: num_affected = 0; break;
	}

	// put affected pps's in reset
	for (i = 0; i < num_affected; i++) {
		pcie_hid_info_t *p1 = &pcie_hid_info[p->hid + i];

		pcie_fsm(p1, PCIEEV_MACDN);
		pcie_set_reset(p1, 1);
		ppcie_set_linkparams(p1->hid, 0, 0);
	}

	// set new linkparams width/flags
	ppcie_set_linkparams(p->hid, width, flags);
	// config new width
	pps_cfg_linkwidth(pp, pps, width);

	// If width != 0 reset link to reset serdes and make operational,
	// else skip reset and leave pps held in reset from above.
	if (width != 0) {
		pcie_reset(p);
	}
	return (0);
}
#endif // CRUZ

static void __devinit
pcie_check_env(void)
{
#if defined CONFIG_BASE_CRUZ
    extern char *prom_getenv(const char *s);
    char *val;

    val = prom_getenv("act_tar2naxi_idwidth");
    if (val) {
	act_tar2naxi_idwidth = simple_strtoul(val, NULL, 0);
	printk("act_tar2naxi_idwidth override = 0x%x\n", act_tar2naxi_idwidth);
    }
    val = prom_getenv("gen3_eq_ctrl");
    if (val) {
	gen3_eq_ctrl = simple_strtoul(val, NULL, 0);
	printk("gen3_eq_ctrl override = 0x%x\n", gen3_eq_ctrl);
    }
    val = prom_getenv("pcie_link_poll_timer_ms");
    if (val) {
	pcie_link_poll_timer_ms = simple_strtoul(val, NULL, 0);
	printk("pcie_link_poll_timer_ms override = %d\n",
	       pcie_link_poll_timer_ms);
    }
    val = prom_getenv("pcie_ntrace");
    if (val) {
	pcie_ntrace = simple_strtoul(val, NULL, 0);
	printk("pcie_ntrace override = %d\n", pcie_ntrace);
    }
    ovgroup_init(&pcsovg, "pcie_pcs");
    ovgroup_init(&serovg, "pcie_ser");
    ovgroup_init(&cfgovg, "pcie_cfg");
    ovgroup_init(&rstovg, "pcie_req_stall_threshold");
#endif // CRUZ
}

////////////////////////////////////////////////////////////////
// sysfs stuff
////////////////////////////////////////////////////////////////

// host related sysfs

typedef ssize_t (*pcie_sysfs_hid_showf_t)(pcie_hid_info_t *p, char *buf);
typedef ssize_t (*pcie_sysfs_hid_storef_t)(pcie_hid_info_t *p,
                                           const char *buf,
                                           const size_t buflen);

typedef struct pcie_sysfs_hid_attr {
	struct attribute attr;
	pcie_sysfs_hid_showf_t show;
	pcie_sysfs_hid_storef_t store;
} pcie_sysfs_hid_attr_t;

#define PPCIE_HID_ATTR(name, mode, showf, storef) \
static pcie_sysfs_hid_attr_t pcie_sysfs_hid_attr_##name = \
	__ATTR(name, mode, showf, storef)

#define attr_to_hidattr(a) container_of(a, pcie_sysfs_hid_attr_t, attr)
#define kobj_to_hidinfo(k) container_of(k, pcie_hid_info_t, kobj)

static ssize_t
pcie_sysfs_show(struct kobject *kobj, struct attribute *attr, char *data)
{
	pcie_sysfs_hid_attr_t *hidattr = attr_to_hidattr(attr);
        pcie_hid_info_t *p = kobj_to_hidinfo(kobj);
        ssize_t sz = 0;

	if (hidattr->show) {
		sz = hidattr->show(p, data);
	}
        return sz;
}

static ssize_t
pcie_sysfs_store(struct kobject *kobj, struct attribute *attr,
                 const char *data, size_t len)
{
	pcie_sysfs_hid_attr_t *hidattr = attr_to_hidattr(attr);
        pcie_hid_info_t *p = kobj_to_hidinfo(kobj);
        ssize_t sz = len;

	if (hidattr->store) {
		sz = hidattr->store(p, data, len);
	}
        return sz;
}

static struct sysfs_ops pcie_sysfs_ops = {
	.show = pcie_sysfs_show,
	.store = pcie_sysfs_store,
};

static ssize_t
pcie_sysfs_hid_attr_debug_show(pcie_hid_info_t *p, char *buf)
{
	char *b = buf;
	u32 width __attribute__ ((unused)) ;
	u32 flags __attribute__ ((unused));

	b += sprintf(b, "state: %s\n", pciest_names[p->state]);
	b += sprintf(b, "event: %s\n", pcieev_names[p->event]);

#if defined CONFIG_BASE_CRUZ
	ppcie_get_linkparams(p->hid, &width, &flags);

	b += sprintf(b, "linkwidth: x%d%s%s\n",
		     width,
		     flags & PCIE_LINKF_TX_FLIP ? " tx_flip" : "",
		     flags & PCIE_LINKF_RX_FLIP ? " rx_flip" : "");
#endif

#define PF(p, FLAG) ((p)->flags & PCIEF_##FLAG ? " " #FLAG : "")
	b += sprintf(b, "flags:%s%s%s%s\n",
		     PF(p, RESET), PF(p, CFG_CRS),
		     PF(p, LOOPBACK), PF(p, FAULT));
	if (p->fault_reason)
		b += sprintf(b, "fault reason: %s\n", p->fault_reason);
	if (p->last_fault_reason)
		b += sprintf(b, "last fault reason: %s\n",
			     p->last_fault_reason);
        b += sprintf(b, "nfaults: %d\n", p->nfaults);
        b += sprintf(b, "maxpllpolls: %d\n", p->maxpllpolls);
        b += sprintf(b, "ltssm_st_changed: %d\n", p->ltssm_st_changed);
	return b - buf;
}

#define print_stat(b, s, t, var)              \
    do { if (s->var || t->var) \
             b += sprintf(b, "%32s : %20llu %20llu\n",  \
                          #var, s->var, s->var + t->var); \
    } while (0)

static ssize_t
pcie_sysfs_hid_attr_stats_show(pcie_hid_info_t *p, char *buf)
{
    char *b = buf;
    pcie_hid_stats_t stats;
    pcie_hid_stats_t *h, *t;

    h = &stats;
    memset(h, 0, sizeof (pcie_hid_stats_t));
    pcie_hid_stats_sum(p->hid, h);
    t = &p->tstats;

    b += sprintf(b, "%32s : %20s %20s\n", "STATISTIC", "CURRENT", "TOTAL");

    print_stat(b, h, t, cfg_rd);
    print_stat(b, h, t, cfg_wr);
    print_stat(b, h, t, cfg_unaligned_rd);
    print_stat(b, h, t, cfg_unaligned_wr);
    print_stat(b, h, t, bar_rd);
    print_stat(b, h, t, bar_wr);
    print_stat(b, h, t, bmt_miss);
    print_stat(b, h, t, bmt_miss_retries);
    print_stat(b, h, t, bmt_oor);
    print_stat(b, h, t, pmv);
    print_stat(b, h, t, unhandled_cause);
    print_stat(b, h, t, unexpected_reqidx);
    print_stat(b, h, t, ini_barrier_req_excess_stall);
    print_stat(b, h, t, ltssm_st_changed);
    print_stat(b, h, t, mac_up);
    print_stat(b, h, t, mac_dn);
    print_stat(b, h, t, link_up);
    print_stat(b, h, t, link_dn);
    print_stat(b, h, t, bus_chg);

    return b - buf;
}

static ssize_t
pcie_sysfs_hid_attr_int_err_stats_show(pcie_hid_info_t *p, char *buf)
{
    char *b = buf;
    pcie_hid_int_err_stats_t *e = &p->int_err_stats;
    pcie_hid_int_err_stats_t *t = &p->tint_err_stats;

    b += sprintf(b, "%32s : %20s %20s\n", "STATISTIC", "CURRENT", "TOTAL");

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    print_stat(b, e, t, tar_rsp_unexp);
    print_stat(b, e, t, tar_rsp_fen_aerr);
    print_stat(b, e, t, tar_rsp_dir_aerr);
    print_stat(b, e, t, tar_rsp_dir_derr);
    print_stat(b, e, t, tar_rsp_ind_info_aerr);
    print_stat(b, e, t, ini_req_crf_err_unc);
    print_stat(b, e, t, ini_req_bdft_err_unc);
#elif defined CONFIG_BASE_CRUZ
    print_stat(b, e, t, ini_req_ord_crf_err_unc);
    print_stat(b, e, t, ini_req_ro_crf_err_unc);
    print_stat(b, e, t, ini_req_stall_timer_excess_halt);
    print_stat(b, e, t, tar_req_stall_timer_excess_halt);
    print_stat(b, e, t, tar_cpl_stall_timer_excess_halt);
#endif /* CONFIG_BASE */
    print_stat(b, e, t, ini_req_bus_master_dis);
    print_stat(b, e, t, ini_req_drf_err_unc);
    print_stat(b, e, t, ini_req_uns);
    print_stat(b, e, t, ini_req_bdft_miss);
    print_stat(b, e, t, ini_cpl_unexp);
    print_stat(b, e, t, ini_cpl_dllp_abort);
    print_stat(b, e, t, ini_cpl_tlp_abort);
    print_stat(b, e, t, ini_cpl_tlp_poisoned);
    print_stat(b, e, t, ini_cpl_ecrc_err);
    print_stat(b, e, t, ini_cpl_stat_err);
    print_stat(b, e, t, ini_cpl_timeout);
    print_stat(b, e, t, ini_rsp_aerr);
    print_stat(b, e, t, ini_rsp_derr);

    return b - buf;
}

static ssize_t
pcie_sysfs_hid_attr_width_show(pcie_hid_info_t *p, char *buf)
{
	u32 width, flags;

	ppcie_get_linkparams(p->hid, &width, &flags);
	return sprintf(buf, "%d%s%s\n", width,
		       flags & PCIE_LINKF_TX_FLIP ? "T" : "",
		       flags & PCIE_LINKF_RX_FLIP ? "R" : "");
}

static ssize_t
pcie_sysfs_hid_attr_width_store(pcie_hid_info_t *p,
				const char *buf, const size_t buflen)
{
#if defined CONFIG_BASE_CRUZ
	char *trailing, *s;
	u32 width, flags;
	ssize_t ret;

	width = simple_strtoul(buf, &trailing, 0);
	flags = 0;
	for (s = trailing; s && *s; s++) {
		switch (*s) {
		case 'T': flags |= PCIE_LINKF_TX_FLIP; break;
		case 'R': flags |= PCIE_LINKF_RX_FLIP; break;
		case ' ': case '\t': case '\n': break;
		default:
			printk(KERN_WARNING PFX
			       "invalid width char %c ignored\n", *s);
			break;
		}
	}
	ret = pcie_set_linkwidth(p, width, flags);
	if (ret < 0)
		return ret;

	return buflen;
#else
	return -EIO;
#endif
}

static ssize_t
pcie_sysfs_hid_attr_reset_show(pcie_hid_info_t *p, char *buf)
{
	unsigned long flags;
	int on;

	spin_lock_irqsave(&p->lock, flags);
	on = (p->flags & PCIEF_RESET) != 0;
	spin_unlock_irqrestore(&p->lock, flags);

	return sprintf(buf, "%d\n", on);
}

static ssize_t
pcie_sysfs_hid_attr_reset_store(pcie_hid_info_t *p,
                                const char *buf, const size_t buflen)
{
#if defined CONFIG_BASE_CRUZ
	unsigned long on = simple_strtoul(buf, NULL, 0);

	tracep(p, PCIETR_INFO, "sysfs reset %s", on ? "on" : "off");
	if (on) {
		pcie_fsm(p, PCIEEV_MACDN);
		pcie_set_reset(p, 1);
	} else {
		pcie_reset(p);
	}
	return buflen;
#else
	return -EIO;
#endif
}

static ssize_t
pcie_sysfs_hid_attr_trace_evmask_show(pcie_hid_info_t *p, char *buf)
{
	return sprintf(buf, "0x%x\n", trace_get_evmask(&p->log));
}

static ssize_t
pcie_sysfs_hid_attr_trace_evmask_store(pcie_hid_info_t *p,
                                       const char *buf, const size_t buflen)
{
	u_int32_t evmask = simple_strtoul(buf, NULL, 0);
	trace_set_evmask(&p->log, evmask);
	return buflen;
}

static ssize_t
pcie_sysfs_hid_attr_trace_log_show(pcie_hid_info_t *p, char *buf)
{
	return sprintf(buf, "0x%p\n", &p->log);
}

#if defined CONFIG_BASE_CRUZ
static ssize_t
pcie_sysfs_hid_attr_ovinfo_show(ovinfo_t *ovinfo, char *buf)
{
	char *b = buf;
	int i;

	for (i = 0; i < ovinfo->nov; i++) {
		oventry_t *ov = &ovinfo->ov[i];
		b += sprintf(b, "0x%x=0x%x\n", ov->off, ov->val);
	}
	return b - buf;
}

static ssize_t
pcie_sysfs_hid_attr_pcie_ser_show(pcie_hid_info_t *p, char *buf)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	ovinfo_t *ovinfo = &serovg.pps[pp][pps];
	return pcie_sysfs_hid_attr_ovinfo_show(ovinfo, buf);
}

static ssize_t
pcie_sysfs_hid_attr_pcie_pcs_show(pcie_hid_info_t *p, char *buf)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	ovinfo_t *ovinfo = &pcsovg.pps[pp][pps];
	return pcie_sysfs_hid_attr_ovinfo_show(ovinfo, buf);
}

static ssize_t
pcie_sysfs_hid_attr_pcie_cfg_show(pcie_hid_info_t *p, char *buf)
{
	const u8 pp  = ppcie_hid_pp(p->hid);
	const u8 pps = ppcie_hid_pps(p->hid);
	ovinfo_t *ovinfo = &cfgovg.pps[pp][pps];
	return pcie_sysfs_hid_attr_ovinfo_show(ovinfo, buf);
}
#endif /* CRUZ */

PPCIE_HID_ATTR(debug,        0444, pcie_sysfs_hid_attr_debug_show, NULL);
PPCIE_HID_ATTR(stats,        0444, pcie_sysfs_hid_attr_stats_show, NULL);
PPCIE_HID_ATTR(int_err_stats,0444, pcie_sysfs_hid_attr_int_err_stats_show, NULL);
PPCIE_HID_ATTR(width,        0666, pcie_sysfs_hid_attr_width_show,
                                   pcie_sysfs_hid_attr_width_store);
PPCIE_HID_ATTR(reset,        0666, pcie_sysfs_hid_attr_reset_show,
                                   pcie_sysfs_hid_attr_reset_store);
PPCIE_HID_ATTR(trace_log,    0444, pcie_sysfs_hid_attr_trace_log_show, NULL);
PPCIE_HID_ATTR(trace_evmask, 0666, pcie_sysfs_hid_attr_trace_evmask_show,
                                   pcie_sysfs_hid_attr_trace_evmask_store);
#if defined CONFIG_BASE_CRUZ
PPCIE_HID_ATTR(pcie_ser,     0444, pcie_sysfs_hid_attr_pcie_ser_show, NULL);
PPCIE_HID_ATTR(pcie_pcs,     0444, pcie_sysfs_hid_attr_pcie_pcs_show, NULL);
PPCIE_HID_ATTR(pcie_cfg,     0444, pcie_sysfs_hid_attr_pcie_cfg_show, NULL);
#endif /* CRUZ */

static struct attribute *pcie_sysfs_default_hid_attrs[] = {
	&pcie_sysfs_hid_attr_debug.attr,
	&pcie_sysfs_hid_attr_stats.attr,
	&pcie_sysfs_hid_attr_int_err_stats.attr,
	&pcie_sysfs_hid_attr_width.attr,
	&pcie_sysfs_hid_attr_reset.attr,
	&pcie_sysfs_hid_attr_trace_log.attr,
	&pcie_sysfs_hid_attr_trace_evmask.attr,
	NULL
};

static struct kobj_type pcie_sysfs_hid_ktype = {
	.release = NULL,
	.sysfs_ops = &pcie_sysfs_ops,
	.default_attrs = pcie_sysfs_default_hid_attrs,
};

static void __devinit
pcie_add_hid_sysfs(struct platform_device *dev, pcie_hid_info_t *p)
{
	const u8 pp  __attribute__ ((unused)) = ppcie_hid_pp(p->hid);
	const u8 pps __attribute__ ((unused)) = ppcie_hid_pps(p->hid);
	struct kobject *kobj = &p->kobj;
	int r;

    r = kobject_init_and_add(kobj, &pcie_sysfs_hid_ktype,
                                 &dev->dev.kobj, "hid%d", p->hid);
	if (r) {
            printk(KERN_WARNING PFX "kobject_register hid%d failed %d\n",
                   p->hid, r);
	}
#if defined CONFIG_BASE_CRUZ
	if (serovg.pps[pp][pps].nov) {
		r = sysfs_create_file(kobj,
				      &pcie_sysfs_hid_attr_pcie_ser.attr);
		if (r) printk(KERN_WARNING PFX "pcie_ser attr failed %d\n", r);
	}
	if (pcsovg.pps[pp][pps].nov) {
		r = sysfs_create_file(kobj,
				      &pcie_sysfs_hid_attr_pcie_pcs.attr);
		if (r) printk(KERN_WARNING PFX "pcie_pcs attr failed %d\n", r);
	}
	if (cfgovg.pps[pp][pps].nov) {
		r = sysfs_create_file(kobj,
				      &pcie_sysfs_hid_attr_pcie_cfg.attr);
		if (r) printk(KERN_WARNING PFX "pcie_cfg attr failed %d\n", r);
	}
#endif /* CRUZ */
}

// Generic PCIe sysfs

static ssize_t
ppcie_sysfs_pcie_show_stats(struct device *dev,
                            struct device_attribute *attr, char *buf)
{
    char *b = buf;
#define print_pcie_stat(b, var) \
    if (ppcie_stats.var) \
        b += sprintf(b, "%15s: %llu\n", #var, ppcie_stats.var);

    print_pcie_stat(b, bar_miss);
    print_pcie_stat(b, bar_missfail);
    print_pcie_stat(b, bars_evicted);
    print_pcie_stat(b, bars_loaded);
    print_pcie_stat(b, bars_loadov);
    print_pcie_stat(b, bars_reset);
    print_pcie_stat(b, ioctl);

    return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_show_sw_reset_with_host(struct device *dev,
                                  struct device_attribute *attr,
                                  char *buf)
{
	char *b = buf;

	b += sprintf(b, "%d\n", pcie_sw_reset_with_host);
	return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_store_sw_reset_with_host(struct device *dev,
                                   struct device_attribute *attr,
                                   const char *buf, size_t buflen)
{
	pcie_sw_reset_with_host = simple_strtoul(buf, NULL, 0);
	return buflen;
}

static ssize_t
ppcie_sysfs_pcie_show_reset_on_macup(struct device *dev,
				     struct device_attribute *attr, char *buf)
{
    char *b = buf;

    b += sprintf(b, "%d\n", pcie_reset_on_macup);
    return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_store_reset_on_macup(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t buflen)
{
    pcie_reset_on_macup = simple_strtoul(buf, NULL, 0);
    return buflen;
}

static ssize_t
ppcie_sysfs_pcie_show_ltssm_count(struct device *dev,
                                  struct device_attribute *attr, char *buf)
{
    char *b = buf;

    b += sprintf(b, "%d\n", pcie_ltssm_count);
    return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_store_ltssm_count(struct device *dev,
                                   struct device_attribute *attr,
                                   const char *buf, size_t buflen)
{
    pcie_ltssm_count = simple_strtoul(buf, NULL, 0);
    return buflen;
}

static ssize_t
ppcie_sysfs_pcie_show_ltssm_ms(struct device *dev,
                               struct device_attribute *attr, char *buf)
{
    char *b = buf;

    b += sprintf(b, "%d\n", pcie_ltssm_ms);
    return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_store_ltssm_ms(struct device *dev,
                                struct device_attribute *attr,
                                const char *buf, size_t buflen)
{
    pcie_ltssm_ms = simple_strtoul(buf, NULL, 0);
    return buflen;
}

static ssize_t
ppcie_sysfs_pcie_show_set_linkwidth_while_up(struct device *dev,
					     struct device_attribute *attr,
					     char *buf)
{
    char *b = buf;

    b += sprintf(b, "%d\n", pcie_set_linkwidth_while_up);
    return b - buf;
}

static ssize_t
ppcie_sysfs_pcie_store_set_linkwidth_while_up(struct device *dev,
					      struct device_attribute *attr,
					      const char *buf, size_t buflen)
{
    pcie_set_linkwidth_while_up = simple_strtoul(buf, NULL, 0);
    return buflen;
}

#define ppcie_sysfs_attr_ro(name) \
static DEVICE_ATTR(name, 0x444, ppcie_sysfs_pcie_show_##name, NULL);
#define ppcie_sysfs_attr_rw(name) \
static DEVICE_ATTR(name, 0x666, ppcie_sysfs_pcie_show_##name, \
				ppcie_sysfs_pcie_store_##name);

ppcie_sysfs_attr_ro(stats);
ppcie_sysfs_attr_rw(sw_reset_with_host);
ppcie_sysfs_attr_rw(reset_on_macup);
ppcie_sysfs_attr_rw(ltssm_count);
ppcie_sysfs_attr_rw(ltssm_ms);
ppcie_sysfs_attr_rw(set_linkwidth_while_up);

////////////////////////////////////////////////////////////////
// initialization routines
////////////////////////////////////////////////////////////////

static void __devinit
pcie_hid_info_init(void)
{
    u32 hid;

    for (hid = 0; hid < PALO_MAX_HIDS; hid++) {
        pcie_hid_info_t *p = &pcie_hid_info[hid];
        p->hid = hid;
        sprintf(p->name, "hid%d", p->hid);
        p->state = PCIEST_MACRST;
        p->flags |= PCIEF_RESET | PCIEF_CFG_CRS;
        spin_lock_init(&p->lock);
    }
}

static void __devinit
pcie_hid_init(u_int32_t hid, void *arg)
{
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    struct platform_device *dev = arg;
    char *buf;

    buf = kzalloc(pcie_ntrace * sizeof(trace_entry_t), GFP_ATOMIC);
    if (buf == NULL) {
        printk("No mem for trace\n");
    }
    cmn_trace_setup(&p->log, p->name, pcie_ntrace, buf);
    pcie_add_hid_sysfs(dev, p);
}

static u_int64_t
timestamp_to_us(trace_log_t *log, u_int64_t ts)
{
    do_div(ts, log->ts_div);
    return (ts);
}

static int
trace_entry_to_buf(trace_log_t *log, trace_entry_t *p, char *name,
                   char *buf, size_t buflen)
{
    u_int32_t word[TRACE_NWORDS + 1];
    int my_align, p_align, adj, n;
    u_int64_t s, us;

    /*
     * Recreate the original va_list alignment.
     */
    my_align = (((u_int32_t)word & 0x7) != 0);
    p_align = ((p->align & 0x7) != 0);
    adj = my_align ^ p_align;
    memcpy(&word[adj], p->word, TRACE_NWORDS * sizeof (u_int32_t));

    /*
     * Display the trace entry with the timestamp converted to seconds.
     */
    us = timestamp_to_us(log, p->timestamp);
    s = us;
    us = do_div(s, 1000000);

    n = snprintf(buf, buflen,
                 "%llu.%llu [%s:%2d] ", s, us, name, p->ev);

    if (n < buflen) {
        n += vsnprintf(buf + n, buflen - n, p->fmt, (va_list)(word + adj));
    }
    return (n);
}

//
// Seek nentries from end (head) of trace buffer ring,
// Expect nentries < 0.  Return index of selected
// entry in the trace ring.
//
static int
pcie_trace_seek(trace_log_t *log, int nentries)
{
    int head, tail, npos;

    head = log->head;
    tail = log->tail;
    if (head < tail) {
        head += log->nentries;
    }
    if (nentries > 0) {
        nentries = 0; // don't seek after end
    }
    npos = head + nentries;

    // Clamp new pos within [tail, head].
    if (npos > head) {
        npos = head;
    }
    if (npos < tail) {
        npos = tail;
    }
    // Wrap new pos, if necessary.
    if (npos >= log->nentries) {
        npos -= log->nentries;
    }
    return (npos);
}

static void
pcie_trace_dump_hid(u_int32_t hid, void *arg)
{
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    trace_log_t *log = &p->log;
    char buf[128];
    int pos;

    printk(KERN_INFO PFX "==== %s ====\n", p->name);
    if (log->buf == NULL || log->nentries == 0) {
        return;
    }

    pos = pcie_trace_seek(log, -10);
    while (pos != log->head) {
        trace_entry_to_buf(log, &log->buf[pos], p->name, buf, sizeof (buf));
        printk(KERN_INFO PFX "%s\n", buf);

        if (++pos == log->nentries) {
            pos = 0;
        }
    }
}

static void
pcie_trace_dump(void)
{
    printk(KERN_INFO PFX "panic pcie trace dump\n");
    ppcie_foreach_hid(pcie_trace_dump_hid, NULL);
}

//
// If we panic, arrange to reset with host again.
// Likely we are already following host unless we panic
// at startup before vnics have been created and the host
// released (via ppcie_pcie_vnics_ready()).
//
static int ppcie_pcie_panic_event(struct notifier_block *this,
				  unsigned long event, void *ptr)
{
    //printk(KERN_WARNING PFX "pcie panic so reset_with_host\n");
    pcie_reset_with_host(1);
    pcie_trace_dump();
    return NOTIFY_DONE;
}

static struct notifier_block ppcie_pcie_panic_block = {
    .notifier_call      = ppcie_pcie_panic_event,
    .priority           = INT_MAX,
};

static void
my_trace_ppcie_exp_intr(hid_t hid,  u_int32_t cmti)
{
    pcie_hid_info_t *p = &pcie_hid_info[hid];
    tracep(p, PCIETR_INFO, "trace_ppcie_exp_intr: hid %d cmti %d ",
                      hid, cmti);

}
extern void (*trace_ppcie_exp_intr)(hid_t hid,  u_int32_t cmti);

#if defined CONFIG_BASE_CRUZ
struct initargs {
	char *global_fault;
	int errcode;
};
static void __devinit
pp_init1(u_int32_t hid, void *arg)
{
	pcie_hid_info_t *p = &pcie_hid_info[hid];
	struct initargs *args = arg;
	char *global_fault = args->global_fault;
	int errcode = args->errcode;
	u32 width, flags;

	ppcie_get_linkparams(hid, &width, &flags);
	pcie_set_linkwidth(p, width, flags);

	tracep(p, PCIETR_INFO, "link initialized, hw_timer %lld core clocks",
	       readreg64(NET_MREGS_DHS_HW_TIMER));

	// If there is a global_fault, mark each hid that comes
	// through with the global fault reason.
	if (global_fault) {
		pcie_set_fault(p, global_fault, errcode);
	}
}

//
// Note: the SER_PLL_OFFSETs are NOT a constant stride.
// We use this function to find the correct offset.
//
static u_int32_t __devinit
pp_mregs_pcsd_cfg_ser_pll(const u8 pp, const u8 pps)
{
    u_int32_t base = PP_MREGS_BASE(pp);
    u_int32_t offset = 0;

    switch (pps) {
    default: // unknown pps gets pps0
    case 0: offset = PP_MREGS_PCSD_CFG_0_SER_PLL_OFFSET; break;
    case 1: offset = PP_MREGS_PCSD_CFG_1_SER_PLL_OFFSET; break;
    case 2: offset = PP_MREGS_PCSD_CFG_2_SER_PLL_OFFSET; break;
    case 3: offset = PP_MREGS_PCSD_CFG_3_SER_PLL_OFFSET; break;
    }
    return (base + offset);
}

static void __devinit
pps_ser_cfg(const u8 pp, const u8 pps)
{
	u_int32_t ser_pll = pp_mregs_pcsd_cfg_ser_pll(pp, pps);

	writereg(ser_pll,     0x84485926);
	writereg(ser_pll + 4, 0x00300020);
}

static void __devinit
pp_init(struct initargs *args)
{
	u8 pp, pps;

	// Put all interfaces in reset (this is initial power up state).
	for (pp = 0; pp < 4; pp++) {
		writereg(PP_MREGS_(SW_HARD_RST, pp), 0xf);
		writereg(PP_MREGS_(SW_SOFT_RST, pp), 0xf);
		pps_cfg_linkwidth(pp, 0, 8);
	}

	if (readreg(MISC_MREGS_VERSION) >= 0xb0) {
		// confg serdes pll
		for (pp = 0; pp < 4; pp++) {
			for (pps = 0; pps < 4; pps++) {
				pps_ser_cfg(pp, pps);
			}
		}
		// reset serdes
		for (pp = 0; pp < 4; pp++) {
			for (pps = 0; pps < 4; pps++) {
				pps_serdes_reset(pp, pps);
			}
		}
	}

	// Configure the active interfaces.
	ppcie_foreach_hid(pp_init1, args);
}
#endif // CONFIG_BASE_CRUZ

int __devinit
ppcie_pcie_init(struct platform_device *dev)
{
#if defined CONFIG_BASE_CRUZ
    struct initargs args;
#endif // CONFIG_BASE_CRUZ
    int irq;

    spin_lock_init(&pcie_lock);
    pcie_check_env();
    hostid_max = ppcie_hostid_max();
    trace_ppcie_exp_intr = my_trace_ppcie_exp_intr;

    pcie_hid_info_init();
    ppcie_foreach_hid(pcie_hid_init, dev);
    pcie_tar_main_mode_init();
#if defined CONFIG_BASE_CRUZ
    memset(&args, 0, sizeof (args));
    if (palo_conf_is_asic() && readreg(MISC_MREGS_VERSION) >= 0xb0) {
        if (pcie_pll_cfg() < 0) {
            args.global_fault = "pcie_pll_cfg failed to lock";
            args.errcode = PPCIE_ERRCODE_PLL_LOCK;
        }
    }
    pp_init(&args);
#endif // CONFIG_BASE_CRUZ

#define DEVICE_CREATE_FILE(devp, attr) \
    if (device_create_file(devp, &dev_attr_##attr)) { \
        printk(KERN_WARNING PFX "pcie " #attr " file create failed\n"); \
    }
    DEVICE_CREATE_FILE(&dev->dev, stats);
    DEVICE_CREATE_FILE(&dev->dev, sw_reset_with_host);
    DEVICE_CREATE_FILE(&dev->dev, reset_on_macup);
    DEVICE_CREATE_FILE(&dev->dev, ltssm_count);
    DEVICE_CREATE_FILE(&dev->dev, ltssm_ms);
    DEVICE_CREATE_FILE(&dev->dev, set_linkwidth_while_up);

#if defined CONFIG_BASE_PALO || defined CONFIG_BASE_SERENO
    irq = palo_int_to_irq(RNAXI_PCIEC_INT_LVL, RNAXI_PCIEC_INT_SRC);
    if (request_irq(irq, pcie_mac_intr_handler,
		    IRQF_DISABLED, DRV_NAME "-mac", dev)) {
        printk(KERN_ERR PFX "Unable to request pciec irq %d.\n", irq);
    }
    irq = palo_int_to_irq(RNAXI_PCIEP_INT_LVL, RNAXI_PCIEP_INT_SRC);
    if (request_irq(irq, pcie_err_intr_handler,
		    IRQF_DISABLED, DRV_NAME "-err", dev)) {
        printk(KERN_ERR PFX "Unable to request pciep irq %d.\n", irq);
    }
#endif /* PALO || SERENO */
#if defined CONFIG_BASE_CRUZ
    irq = palo_int_to_irq(RNAXI_PCIE_INT_LVL, RNAXI_PCIE_INT_SRC);
    if (request_irq(irq, pcie_intr_handler,
		    IRQF_DISABLED, DRV_NAME "-pcie", dev)) {
        printk(KERN_ERR PFX "Unable to request pcie irq %d.\n", irq);
    }
#endif /* CRUZ */

    if (board_conf_cisco_card_mode()) {
		/*
		 * In cisco_card_mode, setting this to 1, gives full control to
		 * BMC to reset the adapter. Sereno ASIC has an internal logic
		 * to AND FORCE_RESET_OK, SPARE-3 pin and PCIE-reset coming
		 * from the host to reset the ASIC. On Cruz, this is
		 * implemented in software by ANDing pcie_reset_on_macdn and 
		 *  SPARE-3 pin on PCIe mac down event.
		 */
		printk(KERN_DEBUG PFX "enabling reset with host\n");
		pcie_reset_with_host(1);
	} else {
		/*
		 * At init time, prevent rebooting the asic on pci resets
		 * until secondary bus number is assigned.
		 */
		printk(KERN_DEBUG PFX "disabling reset with host\n");
		pcie_reset_with_host(0);
    }
    if (board_conf_is_lowpower()) {
        pcie_serdes_pll_disable();
    } else {
        pciep_debug_port_disable();
    }
    pwr_intr_init(dev);

    // notify us if we panic
    atomic_notifier_chain_register(&panic_notifier_list,
                                   &ppcie_pcie_panic_block);
#if defined CONFIG_BASE_CRUZ
    if (board_conf_pcie_poll_compliance() && pcie_link_poll_timer_ms != 0) {
        pcie_link_poll_init();
    }
#endif
    ppcie_foreach_hid(pcie_unmask_mac_intr_cb, NULL);
    return 0;
}
