#include <asm/mipsregs.h>
#include <linux/jiffies.h>
#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/mach-palo/regio.h>

#include <palo/include/wq_mregs.h>
#include <palo/include/regs_macros.h>
#include <palo/include/palo_defs.h>
#include <palo/include/wq_regs.h>
#include <palo/include/wq_regs_ops.h>
#include <palo/fw/mcpu/include/vstate.h>

#include "ppcie_vstate.h"

#define NWQ_WORDS       (PALO_MAX_WQ / 32)

static u_int32_t wq_bitmap[NWQ_WORDS];
static u_int32_t max_lat_ticks;

//
// Palo and Sereno vnic_state uses {wq_base, wq_top} check for WQ range.
// WQs between [wq_base, wq_top] (inclusive) are scheduled for a vnic.
//
// Cruz vnic_state contains {wq_base, wq_count} and checks the WQ range
// between [wq_base, wq_base+wq_count). From the ASIC's POV, wq_count=0
// indicates 1024. From this API's POV, wq_count=0 is invalid,
// you must pass wq_count=1024 to configure 1024 wqs for vnic.
//
static void vnic_state_make(u_int32_t *vnic_state, int wq_base, int wq_cnt)
{
	int last_grant = 0;
#ifdef CONFIG_BASE_SERENO
	int wq_top = wq_base + wq_cnt - 1;
#elif defined CONFIG_BASE_CRUZ
	int wq_top = wq_cnt < 1024 ? wq_cnt : 0;
#endif

	wq_mregs_sched_dhs_vnic_state_make(vnic_state,
			wq_base, wq_top, last_grant
#ifdef CONFIG_BASE_SERENO
			, /*rsvd*/0
#endif // PALO || SERENO
	);
}

static void do_set_vnic_state(struct vnic_st_info *vstate, int uif)
{
	int vnic_state[WQ_MREGS_SCHED_DHS_VNIC_STATE_NWORDS];

	vnic_state_make(vnic_state, vstate->wq_base, vstate->wq_cnt);
	writeregs(WQ_MREGS_SCHED_VNIC_STATE(vstate->vnicid, uif, vstate->cos),
			vnic_state, WQ_MREGS_SCHED_DHS_VNIC_STATE_NWORDS);
}

static void do_clear_vnic_state(struct vnic_st_info *vstate, int uif)
{
	int vnic_state[WQ_MREGS_SCHED_DHS_VNIC_STATE_NWORDS];
	int cos;

	vnic_state_make(vnic_state, vstate->clear_wq, 1);
	for (cos = 0; cos < 8; cos++) {
		writeregs(WQ_MREGS_SCHED_VNIC_STATE(vstate->vnicid, uif, cos),
			    vnic_state, WQ_MREGS_SCHED_DHS_VNIC_STATE_NWORDS);
	}
}

int ppcie_ioctl_vnic_st_cfg(void __user *vnic_state)
{
	struct vnic_st_info vstate;
	unsigned long timeout;
	u_int32_t t0, dt;
	u_int8_t uif;
	int nrunning;
	int r;

	t0 = read_c0_count();
	r = copy_from_user(&vstate, vnic_state, sizeof(struct vnic_st_info));
	if (r)
		return r;

	r = copy_from_user(wq_bitmap,
		((struct vnic_st_info *)vnic_state)->wq_bitmap,
		sizeof (wq_bitmap));
	if (r)
		return r;

	// Clear mregs.master_enable on all allocated WQs
	wq_master_disable_bitmap(wq_bitmap, NWQ_WORDS);

	// Wait for running bit to clear everwhere
	timeout = jiffies + msecs_to_jiffies(100);    // 100ms
	do {
		wq_check_running_bitmap(wq_bitmap, NWQ_WORDS, &nrunning);
	} while (nrunning && !time_after(jiffies, timeout));
	if (nrunning) {
		printk(KERN_INFO "%s: %d WQs still running\n",
				__FUNCTION__, nrunning);
	}

	for (uif = 0; uif < PALO_MAX_UIF; ++uif) {
		if ((vstate.uif_mask & (1 << uif)) == 0) {
			continue;
		}
		if (vstate.op_flags & CLEAR_VNIC_STATE) {
			do_clear_vnic_state(&vstate, uif);
		}
		if (vstate.op_flags & SET_VNIC_STATE) {
			do_set_vnic_state(&vstate, uif);
		}
	}

	// Enable WQs
	wq_master_enable_bitmap(wq_bitmap, NWQ_WORDS);

	dt = read_c0_count() - t0;
	if (dt > max_lat_ticks) {
		max_lat_ticks = dt;
		printk(KERN_INFO "%s: max_lat_ticks = %u\n",
		__FUNCTION__, max_lat_ticks);
	}
	return 0;
}
