/*
 * Copyright 2010 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: palo_wdt.c 21 2016-12-02 21:25:59Z buildsa $"

#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/watchdog.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "palo/include/synopsis/dw_wdt.h"
#include "palo_wdt.h"

static int wdt_running;

static void palo_wdt_kick(void)
{
	writereg(DW_WDT_CRR(0), DW_WDT_CRR_MAGIC);
}

static int palo_wdt_start(int torr)
{
	if (wdt_running) {
		return -EINVAL;
	}
	if (torr < 0 || torr > 15) {
		return -EINVAL;
	}
	printk(KERN_INFO "palo_wdt started torr=%u\n", torr);
	wdt_running = 1;

	/*
	 * 786us between enable and first timeout - so enable and kick
	 * without the possibility of a delay.
	 */
	local_irq_disable();

	writereg(DW_WDT_TORR(0), torr);
	writereg(DW_WDT_CR(0), DW_WDT_CR_RMOD | DW_WDT_CR_WDT_EN);
	palo_wdt_kick();

	local_irq_enable();

	return 0;
}

static ssize_t palo_wdt_write(struct file *file,
		const char __user *data, size_t len, loff_t *ppos)
{
	if (len) {
		palo_wdt_kick();
	}
	return len;
}

static long palo_wdt_ioctl(struct file *file,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	int ret = -EOPNOTSUPP;
	int torr;

	switch (cmd) {
	case WDIOC_KEEPALIVE:
		palo_wdt_kick();
		return 0;

	case WDIOC_SETTIMEOUT:
		if (get_user(torr, p)) {
			return -EFAULT;
		}
		ret = palo_wdt_start(torr);
		break;

	default:
		break;
	}
	return ret;
}

static int palo_wdt_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static struct file_operations palo_wdt_fops = {
	.owner   = THIS_MODULE,
	.llseek  = no_llseek,
	.write	 = palo_wdt_write,
	.unlocked_ioctl   = palo_wdt_ioctl,
	.open    = palo_wdt_open,
};

static struct miscdevice palo_wdt_dev = {
	.minor   = WATCHDOG_MINOR,
	.name    = DRV_NAME,
	.fops    = &palo_wdt_fops
};

static int __init palo_wdt_init(void)
{
	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO PFX "%s\n", DRV_COPYRIGHT);
	return misc_register(&palo_wdt_dev);
}

static void __exit palo_wdt_exit(void)
{
	misc_deregister(&palo_wdt_dev);
}

module_init(palo_wdt_init);
module_exit(palo_wdt_exit);

MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
