/*
 * Copyright 2010 Cisco Systems, Inc.  All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include "kcompat.h"
#include "wq_enet_desc.h"
#include "rq_enet_desc.h"
#include "cq_enet_desc.h"
#include "vnic_enet.h"
#include "vnic_dev.h"
#include "vnic_wq.h"
#include "vnic_rq.h"
#include "vnic_cq.h"
#include "vnic_intr.h"
#include "vnic_stats.h"
#include "vnic_nic.h"

#define PCI_DEVICE_ID_CISCO_VIC_ENET	0x0043  /* ethernet vnicky */

#define DRV_INSTANCE_MAX	1
#define DRV_NAME		"rq2wq"
#define DRV_DESCRIPTION		"Cisco RQ-to-WQ Driver"
#define DRV_VERSION		"0.0.1.42396.ET.davidc"
#define DRV_COPYRIGHT		"Copyright 2010 Cisco Systems, Inc"
#define PFX			DRV_NAME ": "

#define RQ2WQ_MIN_WQ_DESCS	64
#define RQ2WQ_MAX_WQ_DESCS	4096
#define RQ2WQ_MIN_RQ_DESCS	64
#define RQ2WQ_MAX_RQ_DESCS	4096

#define RQ2WQ_MIN_MTU		576  /* minimum for IPv4 */
#define RQ2WQ_MAX_MTU		9000

#define RQ2WQ_SETTING(enic, f) ((rq2wq->config.flags & VENETF_##f) ? 1 : 0)

#define RQ2WQ_BARS_MAX		RES_TYPE_MAX

struct rq2wq {
	struct pci_dev *pdev;
	struct vnicky_dev *vdev;
	struct cdev cdev;
	struct vnicky_dev_bar bar[RQ2WQ_BARS_MAX];
	struct vnicky_enet_config config;
	struct vnicky_wq wq;
	struct vnicky_rq rq;
	struct vnicky_cq cq;
	struct vnicky_intr intr;
    spinlock_t rq2wq_lock;
};

static void rq2wq_wq_queue(struct vnicky_wq *wq, struct page *page,
	dma_addr_t dma_addr, unsigned int len)
{
	struct wq_enet_desc *desc = vnicky_wq_next_desc(wq);

	wq_enet_desc_enc(desc,
		(u64)dma_addr | VNIC_PADDR_TARGET,
		(u16)len,
		0, /* mss_or_csum_offset */
		0, /* hdr_len */
		(u8)WQ_ENET_OFFLOAD_MODE_CSUM,
		1, /* eop */
		1, /* cq_entry */
		0, /* fcoe_encap */
		0, /* vlan_tag_insert */
		0, /* vlan_tag */
		0  /* loopback */); 

	vnicky_wq_post(wq, page, dma_addr, len, 1 /* sop */, 1 /* eop */);
}

static void rq2wq_rq_queue(struct vnicky_rq *rq, struct page *page,
	dma_addr_t dma_addr)
{
	struct rq2wq *rq2wq = vnicky_dev_priv(rq->vdev);
	struct rq_enet_desc *desc = vnicky_rq_next_desc(rq);
	u8 type = RQ_ENET_TYPE_ONLY_SOP;
	unsigned int len = rq2wq->config.mtu + VLAN_ETH_HLEN;

	if (vnicky_rq_desc_avail(rq) == 0) {
		/*
		 * There is no space in RQ ring. This condition should
		 * never occur. In case we end up here, either we might have
		 * or might be posting same buffer twice
		 */
		printk(KERN_ERR PFX
			"RQ descriptor ring is full, post_index %x", rq->to_use->index);
		return;
	}
	rq_enet_desc_enc(desc,
		(u64)dma_addr | VNIC_PADDR_TARGET,
		type, (u16)len);

	vnicky_rq_post(rq, page, 0, dma_addr, len);
}

static void rq2wq_wq_to_rq(struct vnicky_wq *wq,
	struct cq_desc *cq_desc, struct vnicky_wq_buf *buf, void *opaque)
{      
	struct rq2wq *rq2wq = vnicky_dev_priv(wq->vdev);
	struct page *page = buf->os_buf;
	dma_addr_t dma_addr = buf->dma_addr;

	rq2wq_rq_queue(&rq2wq->rq, page, dma_addr);
}                                       

static void rq2wq_rq_to_wq(struct vnicky_rq *rq,
	struct cq_desc *cq_desc, struct vnicky_rq_buf *buf,
	int skipped, void *opaque)
{
	struct rq2wq *rq2wq = vnicky_dev_priv(rq->vdev);
	struct page *page = buf->os_buf;
	dma_addr_t dma_addr = buf->dma_addr;

	u8 type, color, eop, sop, ingress_port, vlan_stripped;
	u8 fcoe, fcoe_sof, fcoe_fc_crc_ok, fcoe_enc_error, fcoe_eof;
	u8 tcp_udp_csum_ok, udp, tcp, ipv4_csum_ok;
	u8 ipv6, ipv4, ipv4_fragment, fcs_ok, rss_type, csum_not_calc;
	u8 packet_error;
	u16 q_number, completed_index, bytes_written, vlan, checksum;
	u32 rss_hash;

	BUG_ON(skipped);

	cq_enet_rq_desc_dec((struct cq_enet_rq_desc *)cq_desc,
		&type, &color, &q_number, &completed_index,
		&ingress_port, &fcoe, &eop, &sop, &rss_type,
		&csum_not_calc, &rss_hash, &bytes_written,
		&packet_error, &vlan_stripped, &vlan, &checksum,
		&fcoe_sof, &fcoe_fc_crc_ok, &fcoe_enc_error,
		&fcoe_eof, &tcp_udp_csum_ok, &udp, &tcp,
		&ipv4_csum_ok, &ipv6, &ipv4, &ipv4_fragment,
		&fcs_ok);

	if (packet_error) {
		// Bad Packet, post the buffer back to RQ
		rq2wq_rq_queue(&rq2wq->rq, page, dma_addr);
		return;
	}

	if (eop && bytes_written > 0) {

		/* Good receive, queue it on the WQ
		*/

		rq2wq_wq_queue(&rq2wq->wq, page, dma_addr, bytes_written);

	} else {
		// XXX buffer overflow...log err somehow
		rq2wq_rq_queue(&rq2wq->rq, page, dma_addr);
	}
}

static int rq2wq_service(struct vnicky_dev *vdev, struct cq_desc *cq_desc,
	u8 type, u16 q_number, u16 completed_index, void *opaque)
{
	struct rq2wq *rq2wq = vnicky_dev_priv(vdev);

	switch (type) {
	case CQ_DESC_TYPE_WQ_ENET:
		vnicky_wq_service(&rq2wq->wq, cq_desc,
			completed_index, rq2wq_wq_to_rq,
			opaque);
		break;
	case CQ_DESC_TYPE_RQ_ENET:
		vnicky_rq_service(&rq2wq->rq, cq_desc,
			completed_index, VNIC_RQ_RETURN_DESC,
			rq2wq_rq_to_wq, opaque);
		break;
	default:
		BUG();
	}

	return 0;
}

static irqreturn_t rq2wq_isr_msi(int irq, void *data)
{
	struct rq2wq *rq2wq = data;
	unsigned int work_to_do = -1;  /* no limit */
	unsigned int work_done;

	work_done = vnicky_cq_service(&rq2wq->cq,
		work_to_do, rq2wq_service, NULL);
	
	vnicky_intr_return_credits(&rq2wq->intr,
		work_done,
		1 /* unmask intr */,
		1 /* reset intr timer */);

	return IRQ_HANDLED;
}

static int rq2wq_dev_wait(struct vnicky_dev *vdev,
	int (*start)(struct vnicky_dev *, int),
	int (*finished)(struct vnicky_dev *, int *),
	int arg)
{
	unsigned long time;
	int done;
	int err;

	BUG_ON(in_interrupt());

	err = start(vdev, arg);
	if (err)
		return err;

	/* Wait for func to complete...2 seconds max
	 */

	time = jiffies + (HZ * 2);
	do {

		err = finished(vdev, &done);
		if (err)
			return err;

		if (done)
			return 0;

		schedule_timeout_uninterruptible(HZ / 10);

	} while (time_after(time, jiffies));

	return -ETIMEDOUT;
}

static int rq2wq_dev_open(struct rq2wq *rq2wq)
{
	int err;

	err = rq2wq_dev_wait(rq2wq->vdev, vnicky_dev_open,
		vnicky_dev_open_done, 0);
	if (err)
		printk(KERN_ERR PFX
			"vNIC device open failed, err %d.\n", err);

	return err;
}

static int rq2wq_get_vnicky_config(struct rq2wq *rq2wq)
{
	struct vnicky_enet_config *c = &rq2wq->config;
	int err;

#define GET_CONFIG(m) \
	do { \
		err = vnicky_dev_spec(rq2wq->vdev, \
			offsetof(struct vnicky_enet_config, m), \
			sizeof(c->m), &c->m); \
		if (err) { \
			printk(KERN_ERR PFX \
				"Error getting %s, %d\n", #m, err); \
			return err; \
		} \
	} while (0)

	GET_CONFIG(flags);
	GET_CONFIG(wq_desc_count);
	GET_CONFIG(rq_desc_count);
	GET_CONFIG(mtu);
	GET_CONFIG(intr_timer_type);
	GET_CONFIG(intr_mode);
	GET_CONFIG(intr_timer_usec);

	c->wq_desc_count =
		min_t(u32, RQ2WQ_MAX_WQ_DESCS,
		max_t(u32, RQ2WQ_MIN_WQ_DESCS,
		c->wq_desc_count));
	c->wq_desc_count &= 0xffffffe0; /* must be aligned to groups of 32 */

	c->rq_desc_count =
		min_t(u32, RQ2WQ_MAX_RQ_DESCS,
		max_t(u32, RQ2WQ_MIN_RQ_DESCS,
		c->rq_desc_count));
	c->rq_desc_count &= 0xffffffe0; /* must be aligned to groups of 32 */

	if (c->mtu == 0)
		c->mtu = 1500;
	c->mtu = min_t(u16, RQ2WQ_MAX_MTU,
		max_t(u16, RQ2WQ_MIN_MTU,
		c->mtu));

	c->intr_timer_usec = min_t(u32,
		INTR_COALESCE_HW_TO_USEC(VNIC_INTR_TIMER_MAX),
		c->intr_timer_usec);

	printk(KERN_INFO PFX "vNIC wq/rq %d/%d\n",
		c->wq_desc_count, c->rq_desc_count);
	printk(KERN_INFO PFX "vNIC mtu %d csum tx/rx %d/%d tso/lro %d/%d "
		"intr timer %d usec\n",
		c->mtu, RQ2WQ_SETTING(rq2wq, TXCSUM),
		RQ2WQ_SETTING(rq2wq, RXCSUM), RQ2WQ_SETTING(rq2wq, TSO),
		RQ2WQ_SETTING(rq2wq, LRO), c->intr_timer_usec);

	return 0;
}

static int rq2wq_get_res_counts(struct rq2wq *rq2wq)
{
	unsigned int wq_count;
	unsigned int rq_count;
	unsigned int cq_count;
	unsigned int intr_count;

	wq_count = vnicky_dev_get_res_count(rq2wq->vdev, RES_TYPE_WQ);
	rq_count = vnicky_dev_get_res_count(rq2wq->vdev, RES_TYPE_RQ);
	cq_count = vnicky_dev_get_res_count(rq2wq->vdev, RES_TYPE_CQ);
	intr_count = vnicky_dev_get_res_count(rq2wq->vdev, RES_TYPE_INTR_CTRL);

	printk(KERN_INFO PFX "vNIC resources avail: "
		"wq %d rq %d cq %d intr %d\n",
		wq_count, rq_count,
		cq_count, intr_count);

	/* Need 1 WQ, 1 RQ, 1 CQ, and 1 INTR resources
	 */

	if ((wq_count < 1) ||
	    (rq_count < 1) ||
	    (cq_count < 1) ||
	    (intr_count < 1)) {
		printk(KERN_INFO PFX "Not enough vNIC resources avail\n");
		return -ENODEV;
	}

	printk(KERN_INFO PFX "vNIC resources used:  "
		"wq 1 rq 1 cq 1 intr 1 intr mode MSI\n");

	return 0;
}

static void rq2wq_init_vnicky_resources(struct rq2wq *rq2wq)
{
	unsigned int mask_on_assertion;
	unsigned int error_interrupt_enable;
	unsigned int error_interrupt_offset;
	unsigned int cq_index;

	/* Init RQ/WQ resources.
	 *
	 * RQ[0] points to CQ[0]
	 * WQ[0] points to CQ[0]
	 */

	error_interrupt_enable = 0;
	error_interrupt_offset = 0;

	cq_index = 0;
	vnicky_rq_init(&rq2wq->rq,
		cq_index,
		error_interrupt_enable,
		error_interrupt_offset);

	vnicky_wq_init(&rq2wq->wq,
		cq_index,
		error_interrupt_enable,
		error_interrupt_offset);

	/* Init CQ resources
	 *
	 * CQ[0] points to INTR[0]
	 */

	vnicky_cq_init(&rq2wq->cq,
		0 /* flow_control_enable */,
		1 /* color_enable */,
		0 /* cq_head */,
		0 /* cq_tail */,
		1 /* cq_tail_color */,
		1 /* interrupt_enable */,
		1 /* cq_entry_enable */,
		0 /* cq_message_enable */,
		0 /* interrupt_offset */,
		0 /* cq_message_addr */);

	/* Init INTR resources
	 */

	mask_on_assertion = 1;

	vnicky_intr_init(&rq2wq->intr,
		INTR_COALESCE_USEC_TO_HW(rq2wq->config.intr_timer_usec),
		rq2wq->config.intr_timer_type,
		mask_on_assertion);

	/* Clear LIF stats
	 */

	vnicky_dev_stats_clear(rq2wq->vdev);
}

static void rq2wq_free_vnicky_resources(struct rq2wq *rq2wq)
{
	vnicky_wq_free(&rq2wq->wq);
	vnicky_rq_free(&rq2wq->rq);
	vnicky_cq_free(&rq2wq->cq);
	vnicky_intr_free(&rq2wq->intr);
}

static int rq2wq_alloc_vnicky_resources(struct rq2wq *rq2wq)
{
	int err;

	/* Allocate queue resources
	 */

	err = vnicky_wq_alloc(rq2wq->vdev, &rq2wq->wq, 0,
		rq2wq->config.wq_desc_count,
		sizeof(struct wq_enet_desc));
	if (err)
		goto err_out_cleanup;

	err = vnicky_rq_alloc(rq2wq->vdev, &rq2wq->rq, 0,
		rq2wq->config.rq_desc_count,
		sizeof(struct rq_enet_desc));
	if (err)
		goto err_out_cleanup;

	err = vnicky_cq_alloc(rq2wq->vdev, &rq2wq->cq, 0,
		rq2wq->config.wq_desc_count + rq2wq->config.rq_desc_count,
		sizeof(struct cq_enet_rq_desc));
	if (err)
		goto err_out_cleanup;

	err = vnicky_intr_alloc(rq2wq->vdev, &rq2wq->intr, 0);
	if (err)
		goto err_out_cleanup;

	return 0;

err_out_cleanup:
	rq2wq_free_vnicky_resources(rq2wq);

	return err;
}

static void rq2wq_free_wq_buf(struct vnicky_wq *wq, struct vnicky_wq_buf *buf)
{
	__free_page(buf->os_buf);
}

static void rq2wq_free_rq_buf(struct vnicky_rq *rq, struct vnicky_rq_buf *buf)
{
	__free_page(buf->os_buf);
}

static int rq2wq_alloc_buf(struct vnicky_rq *rq)
{
	struct rq2wq *rq2wq = vnicky_dev_priv(rq->vdev);
	struct page *page = alloc_page(GFP_KERNEL);
	dma_addr_t dma_addr;

	if (!page)
		return -ENOMEM;

	/* One-time mapping we can re-use as the buf
	 * cycles between RQ and WQ.
	 */

	dma_addr = pci_map_page(rq2wq->pdev, page, 0,
		PAGE_SIZE, PCI_DMA_BIDIRECTIONAL);

	rq2wq_rq_queue(rq, page, dma_addr);

	return 0;
}

static int rq2wq_set_niccfg(struct rq2wq *rq2wq)
{
	u8 rss_default_cpu = 0;
	u8 rss_hash_type = 0;
	u8 rss_hash_bits = 0;
	u8 rss_base_cpu = 0;
	u8 rss_enable = 0;
	u8 tso_ipid_split_en = 0;
	u8 ig_vlan_strip_en = 0;
	u64 a0, a1;
	u32 nic_cfg;
	int wait = 1000;

	vnicky_set_nic_cfg(&nic_cfg, rss_default_cpu,
		rss_hash_type, rss_hash_bits, rss_base_cpu,
		rss_enable, tso_ipid_split_en, ig_vlan_strip_en);

	a0 = nic_cfg;
	a1 = 0;

	return vnicky_dev_cmd(rq2wq->vdev, CMD_NIC_CFG, &a0, &a1, wait);
}

static void rq2wq_iounmap(struct rq2wq *rq2wq)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(rq2wq->bar); i++)
		if (rq2wq->bar[i].vaddr)
			iounmap(rq2wq->bar[i].vaddr);
}

static int rq2wq_open(struct inode *inode, struct file *file)
{
	file->private_data = container_of(inode->i_cdev, struct rq2wq, cdev);

	return 0;
}

static long rq2wq_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    int err = 0;
    unsigned long flags;
	struct rq2wq *rq2wq = file->private_data;

	switch (cmd) {
	case SIOCPROTOPRIVATE+0:
		/* Enable I/O
		*/
		printk(KERN_INFO "rq2wq_ioctl Enable I/O\n");
        spin_lock_irqsave(&rq2wq->rq2wq_lock, flags);
		vnicky_wq_enable(&rq2wq->wq);
		vnicky_rq_enable(&rq2wq->rq);
		vnicky_dev_enable(rq2wq->vdev);
		vnicky_intr_unmask(&rq2wq->intr);
        spin_unlock_irqrestore(&rq2wq->rq2wq_lock, flags);
		break;

	case SIOCPROTOPRIVATE+1:
		/* Disable I/O
		*/
		printk(KERN_INFO "rq2wq_ioctl Disable I/O\n");
        spin_lock_irqsave(&rq2wq->rq2wq_lock, flags);
		vnicky_intr_mask(&rq2wq->intr);
		synchronize_irq(rq2wq->pdev->irq);
		vnicky_dev_disable(rq2wq->vdev);
		vnicky_wq_disable(&rq2wq->wq);
		vnicky_rq_disable(&rq2wq->rq);
        spin_unlock_irqrestore(&rq2wq->rq2wq_lock, flags);
		break;

	default:
		err = -ENOTTY;
	}

	return err;
}

static dev_t dev;

static struct file_operations rq2wq_fops = {
	.owner = THIS_MODULE,
	.open = rq2wq_open,
	.unlocked_ioctl = rq2wq_ioctl,
};

static int __devinit rq2wq_probe(struct pci_dev *pdev,
	const struct pci_device_id *ent)
{
	struct rq2wq *rq2wq;
	unsigned int i;
	static unsigned int instance = 0;
	int devno;
	int err;

	if (instance >= DRV_INSTANCE_MAX)
		return -ENODEV;

	rq2wq = kzalloc(sizeof(struct rq2wq), GFP_KERNEL);
	if (!rq2wq) {
		printk(KERN_ERR PFX "Alloc of rq2wq struct failed\n");
		return -ENOMEM;
	}

	pci_set_drvdata(pdev, rq2wq);
	rq2wq->pdev = pdev;

    spin_lock_init(&rq2wq->rq2wq_lock);

	err = pci_enable_device(pdev);
	if (err) {
		printk(KERN_ERR PFX
			"Cannot enable PCI device, aborting.\n");
		goto err_out_free_rq2wq;
	}

	err = adapter_pci_request_regions(pdev, DRV_NAME);
	if (err) {
		printk(KERN_ERR PFX
			"Cannot request PCI regions, aborting.\n");
		goto err_out_disable_device;
	}

	pci_set_master(pdev);

	err = pci_set_dma_mask(pdev, DMA_BIT_MASK(32));
	if (err) {
		printk(KERN_ERR PFX
			"No usable DMA configuration, aborting.\n");
		goto err_out_release_regions;
	}
	err = adapter_pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
	if (err) {
		printk(KERN_ERR PFX
			"Unable to obtain 32-bit DMA "
			"for consistent allocations, aborting.\n");
		goto err_out_release_regions;
	}

	/* Map vNIC resources from BAR0-5
	 */

	for (i = 0; i < ARRAY_SIZE(rq2wq->bar); i++) {
		if (!(pci_resource_flags(pdev, i) & IORESOURCE_MEM))
			continue;
		rq2wq->bar[i].len = pci_resource_len(pdev, i);
		rq2wq->bar[i].vaddr = pci_iomap(pdev, i, rq2wq->bar[i].len);
		if (!rq2wq->bar[i].vaddr) {
			printk(KERN_ERR PFX
				"Cannot memory-map BAR %d, aborting.\n", i);
			err = -ENODEV;
			goto err_out_iounmap;
	}
	rq2wq->bar[i].bus_addr = pci_resource_start(pdev, i);
	}

	/* Register vNIC device
	 */

	rq2wq->vdev = vnicky_dev_register(NULL, rq2wq, pdev, rq2wq->bar,
			ARRAY_SIZE(rq2wq->bar));
	if (!rq2wq->vdev) {
			printk(KERN_ERR PFX
					"vNIC registration failed, aborting.\n");
			err = -ENODEV;
			goto err_out_iounmap;
	}

	/* Issue device open to get device in known state
	 */

	err = rq2wq_dev_open(rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"vNIC dev open failed, aborting.\n");
		goto err_out_vnicky_unregister;
	}

	err = vnicky_dev_init(rq2wq->vdev, 0);
	if (err) {
		printk(KERN_ERR PFX
			"vNIC dev init failed, aborting.\n");
		goto err_out_dev_close;
	}

	err = rq2wq_get_vnicky_config(rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"Get vNIC configuration failed, aborting.\n");
		goto err_out_dev_close;
	}

	err = rq2wq_get_res_counts(rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"Get vNIC resource counts failed, aborting.\n");
		goto err_out_dev_close;
	}

	/* Allocate and configure vNIC resources
	 */

	err = rq2wq_alloc_vnicky_resources(rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"Failed to alloc vNIC resources, aborting.\n");
		goto err_out_free_vnicky_resources;
	}

	rq2wq_init_vnicky_resources(rq2wq);

	err = rq2wq_set_niccfg(rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"Failed to config nic, aborting.\n");
		goto err_out_free_vnicky_resources;
	}

	err = request_irq(pdev->irq, rq2wq_isr_msi,
		0, DRV_NAME, rq2wq);
	if (err) {
		printk(KERN_ERR PFX
			"Failed to request IRQ, aborting.\n");
		goto err_out_free_vnicky_resources;
	}

	/* Fill RQ ring, enable queues, and unmask intr
	 */

	err = vnicky_rq_fill(&rq2wq->rq, rq2wq_alloc_buf);
	if (err) {
		printk(KERN_ERR PFX
			"Failed to alloc RQ bufs, aborting.\n");
		goto err_out_rq_unfill;
	}

	cdev_init(&rq2wq->cdev, &rq2wq_fops);
	rq2wq->cdev.owner = THIS_MODULE;
	devno = MKDEV(MAJOR(dev), instance);
	err = cdev_add(&rq2wq->cdev, devno, 1);
	if (err) {
		printk(KERN_ERR PFX
			"Failed to create char device, aborting.\n");
		goto err_out_rq_unfill;
	}
	
	instance++;

	return 0;

err_out_rq_unfill:
	vnicky_rq_clean(&rq2wq->rq, rq2wq_free_rq_buf);
	free_irq(pdev->irq, rq2wq);
err_out_free_vnicky_resources:
	rq2wq_free_vnicky_resources(rq2wq);
err_out_dev_close:
	vnicky_dev_close(rq2wq->vdev);
err_out_vnicky_unregister:
	vnicky_dev_unregister(rq2wq->vdev);
err_out_iounmap:
	rq2wq_iounmap(rq2wq);
err_out_release_regions:
	pci_release_regions(pdev);
err_out_disable_device:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
err_out_free_rq2wq:
	kfree(rq2wq);

	return err;
}

static void __devexit rq2wq_remove(struct pci_dev *pdev)
{
	struct rq2wq *rq2wq = pci_get_drvdata(pdev);

	cdev_del(&rq2wq->cdev);
	vnicky_wq_clean(&rq2wq->wq, rq2wq_free_wq_buf);
	vnicky_rq_clean(&rq2wq->rq, rq2wq_free_rq_buf);
	free_irq(pdev->irq, rq2wq);
	rq2wq_free_vnicky_resources(rq2wq);
	vnicky_dev_close(rq2wq->vdev);
	vnicky_dev_unregister(rq2wq->vdev);
	rq2wq_iounmap(rq2wq);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	kfree(rq2wq);
}

/* Supported devices */
static struct pci_device_id rq2wq_id_table[] = {
	{ PCI_VDEVICE(CISCO, PCI_DEVICE_ID_CISCO_VIC_ENET) },
	{ 0, }	/* end of table */
};

MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_AUTHOR("Scott Feldman <scofeldm@cisco.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
MODULE_DEVICE_TABLE(pci, rq2wq_id_table);

static struct pci_driver rq2wq_driver = {
	.name = DRV_NAME,
	.id_table = rq2wq_id_table,
	.probe = rq2wq_probe,
	.remove = __devexit_p(rq2wq_remove),
};

static int __init rq2wq_init_module(void)
{
	int err;
	
	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);

	err = alloc_chrdev_region(&dev, 0, DRV_INSTANCE_MAX, DRV_NAME);
	if (err) {
		printk(KERN_ERR PFX "failed to allocate char driver region\n");
		return err;
	}
	
	printk(KERN_INFO PFX " major %d\n", MAJOR(dev));

	return adapter_pci_register_driver(&rq2wq_driver);
}

static void __exit rq2wq_cleanup_module(void)
{
	pci_unregister_driver(&rq2wq_driver);
	unregister_chrdev_region(dev, DRV_INSTANCE_MAX);
}

module_init(rq2wq_init_module);
module_exit(rq2wq_cleanup_module);
