/*
 * Copyright 2008 Cisco Systems, Inc.  All rights reserved.
 * Copyright 2007 Nuova Systems, Inc.  All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ident "$Id: vnic_dev.c 21 2016-12-02 21:25:59Z buildsa $"

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/if_ether.h>

#include "kcompat.h"
#include "vnic_resource.h"
#include "vnic_devcmd.h"
#include "vnic_dev.h"
#include "vnic_stats.h"

struct vnicky_res {
	void __iomem *vaddr;
	dma_addr_t bus_addr;
	unsigned int count;
};

#define VNIC_DEV_CAP_INIT	0x0001

struct vnicky_dev {
	void *priv;
	struct pci_dev *pdev;
	struct vnicky_res res[RES_TYPE_MAX];
	enum vnicky_dev_intr_mode intr_mode;
	struct vnicky_devcmd __iomem *devcmd;
	struct vnicky_devcmd_notify *notify;
	struct vnicky_devcmd_notify notify_copy;
	dma_addr_t notify_pa;
	u32 notify_sz;
	u32 *linkstatus;
	dma_addr_t linkstatus_pa;
	struct vnicky_stats *stats;
	dma_addr_t stats_pa;
	struct vnicky_devcmd_fw_info *fw_info;
	dma_addr_t fw_info_pa;
	u32 cap_flags;
};

#define VNIC_MAX_RES_HDR_SIZE \
	(sizeof(struct vnicky_resource_header) + \
	sizeof(struct vnicky_resource) * RES_TYPE_MAX)
#define VNIC_RES_STRIDE	128

void *vnicky_dev_priv(struct vnicky_dev *vdev)
{
	return vdev->priv;
}

int vnicky_dev_get_size(void)
{
	return sizeof(struct vnicky_dev);
}

static int vnicky_dev_discover_res(struct vnicky_dev *vdev,
	struct vnicky_dev_bar *bar, unsigned int num_bars)
{
	struct vnicky_resource_header __iomem *rh;
	struct vnicky_resource __iomem *r;
	u8 type;

	if (num_bars == 0)
		return -EINVAL;

	if (bar->len < VNIC_MAX_RES_HDR_SIZE) {
		printk(KERN_ERR "vNIC BAR0 res hdr length error\n");
		return -EINVAL;
	}

	rh = bar->vaddr;
	if (!rh) {
		printk(KERN_ERR "vNIC BAR0 res hdr not mem-mapped\n");
		return -EINVAL;
	}

	if (ioread32(&rh->magic) != VNIC_RES_MAGIC ||
	    ioread32(&rh->version) != VNIC_RES_VERSION) {
		printk(KERN_ERR "vNIC BAR0 res magic/version error "
			"exp (%lx/%lx) curr (%x/%x)\n",
			VNIC_RES_MAGIC, VNIC_RES_VERSION,
			ioread32(&rh->magic), ioread32(&rh->version));
		return -EINVAL;
	}

	r = (struct vnicky_resource __iomem *)(rh + 1);

	while ((type = ioread8(&r->type)) != RES_TYPE_EOL) {

		u8 bar_num = ioread8(&r->bar);
		u32 bar_offset = ioread32(&r->bar_offset);
		u32 count = ioread32(&r->count);
		u32 len;

		r++;

		if (bar_num >= num_bars)
			continue;

		if (!bar[bar_num].len || !bar[bar_num].vaddr)
			continue;

		switch (type) {
		case RES_TYPE_WQ:
		case RES_TYPE_RQ:
		case RES_TYPE_CQ:
		case RES_TYPE_INTR_CTRL:
			/* each count is stride bytes long */
			len = count * VNIC_RES_STRIDE;
			if (len + bar_offset > bar[bar_num].len) {
				printk(KERN_ERR "vNIC BAR0 resource %d "
					"out-of-bounds, offset 0x%x + "
					"size 0x%x > bar len 0x%lx\n",
					type, bar_offset,
					len,
					bar[bar_num].len);
				return -EINVAL;
			}
			break;
		case RES_TYPE_INTR_PBA_LEGACY:
		case RES_TYPE_DEV:
		case RES_TYPE_DEVCMD:
#ifdef ENIC_UPT
		case RES_TYPE_PASS_THRU_PAGE:
#endif
			len = count;
			break;
		default:
			continue;
		}

		vdev->res[type].count = count;
		vdev->res[type].vaddr = (char __iomem *)bar[bar_num].vaddr +
			bar_offset;
		vdev->res[type].bus_addr = bar[bar_num].bus_addr + bar_offset;
	}

	return 0;
}

unsigned int vnicky_dev_get_res_count(struct vnicky_dev *vdev,
	enum vnicky_res_type type)
{
	return vdev->res[type].count;
}

void __iomem *vnicky_dev_get_res(struct vnicky_dev *vdev, enum vnicky_res_type type,
	unsigned int index)
{
	if (!vdev->res[type].vaddr)
		return NULL;

	switch (type) {
	case RES_TYPE_WQ:
	case RES_TYPE_RQ:
	case RES_TYPE_CQ:
	case RES_TYPE_INTR_CTRL:
		return (char __iomem *)vdev->res[type].vaddr +
			index * VNIC_RES_STRIDE;
	default:
		return (char __iomem *)vdev->res[type].vaddr;
	}
}

dma_addr_t vnicky_dev_get_res_bus_addr(struct vnicky_dev *vdev,
	enum vnicky_res_type type, unsigned int index)
{
	switch (type) {
	case RES_TYPE_WQ:
	case RES_TYPE_RQ:
	case RES_TYPE_CQ:
	case RES_TYPE_INTR_CTRL:
		return vdev->res[type].bus_addr +
			index * VNIC_RES_STRIDE;
	default:
		return vdev->res[type].bus_addr;
	}
}

unsigned int vnicky_dev_desc_ring_size(struct vnicky_dev_ring *ring,
	unsigned int desc_count, unsigned int desc_size)
{
	/* The base address of the desc rings must be 512 byte aligned.
	 * Descriptor count is aligned to groups of 32 descriptors.  A
	 * count of 0 means the maximum 4096 descriptors.  Descriptor
	 * size is aligned to 16 bytes.
	 */

	unsigned int count_align = 32;
	unsigned int desc_align = 16;

	ring->base_align = 512;

	if (desc_count == 0)
		desc_count = 4096;

	ring->desc_count = ALIGN(desc_count, count_align);

	ring->desc_size = ALIGN(desc_size, desc_align);

	ring->size = ring->desc_count * ring->desc_size;
	ring->size_unaligned = ring->size + ring->base_align;

	return ring->size_unaligned;
}

void vnicky_dev_clear_desc_ring(struct vnicky_dev_ring *ring)
{
	memset(ring->descs, 0, ring->size);
}

int vnicky_dev_alloc_desc_ring(struct vnicky_dev *vdev, struct vnicky_dev_ring *ring,
	unsigned int desc_count, unsigned int desc_size)
{
	vnicky_dev_desc_ring_size(ring, desc_count, desc_size);

	ring->descs_unaligned = pci_alloc_consistent(vdev->pdev,
		ring->size_unaligned,
		&ring->base_addr_unaligned);

	if (!ring->descs_unaligned) {
		printk(KERN_ERR
		  "Failed to allocate ring (size=%d), aborting\n",
			(int)ring->size);
		return -ENOMEM;
	}

	ring->base_addr = ALIGN(ring->base_addr_unaligned,
		ring->base_align);
	ring->descs = (u8 *)ring->descs_unaligned +
		(ring->base_addr - ring->base_addr_unaligned);

	vnicky_dev_clear_desc_ring(ring);

	ring->desc_avail = ring->desc_count - 1;

	return 0;
}

void vnicky_dev_free_desc_ring(struct vnicky_dev *vdev, struct vnicky_dev_ring *ring)
{
	if (ring->descs) {
		pci_free_consistent(vdev->pdev,
			ring->size_unaligned,
			ring->descs_unaligned,
			ring->base_addr_unaligned);
		ring->descs = NULL;
	}
}

int vnicky_dev_cmd(struct vnicky_dev *vdev, enum vnicky_devcmd_cmd cmd,
	u64 *a0, u64 *a1, int wait)
{
	return 0;
}

static int vnicky_dev_capable(struct vnicky_dev *vdev, enum vnicky_devcmd_cmd cmd)
{
	u64 a0 = (u32)cmd, a1 = 0;
	int wait = 1000;
	int err;

	err = vnicky_dev_cmd(vdev, CMD_CAPABILITY, &a0, &a1, wait);

	return !(err || a0);
}

int vnicky_dev_fw_info(struct vnicky_dev *vdev,
	struct vnicky_devcmd_fw_info **fw_info)
{
	u64 a0, a1 = 0;
	int wait = 1000;
	int err = 0;

	if (!vdev->fw_info) {
		vdev->fw_info = pci_alloc_consistent(vdev->pdev,
			sizeof(struct vnicky_devcmd_fw_info),
			&vdev->fw_info_pa);
		if (!vdev->fw_info)
			return -ENOMEM;

		a0 = vdev->fw_info_pa;

		/* only get fw_info once and cache it */
		err = vnicky_dev_cmd(vdev, CMD_MCPU_FW_INFO, &a0, &a1, wait);
	}

	*fw_info = vdev->fw_info;

	return err;
}

int vnicky_dev_hw_version(struct vnicky_dev *vdev, enum vnicky_dev_hw_version *hw_ver)
{
	struct vnicky_devcmd_fw_info *fw_info;
	int err;

	err = vnicky_dev_fw_info(vdev, &fw_info);
	if (err)
		return err;

	if (strncmp(fw_info->hw_version, "A1", sizeof("A1")) == 0)
		*hw_ver = VNIC_DEV_HW_VER_A1;
	else if (strncmp(fw_info->hw_version, "A2", sizeof("A2")) == 0)
		*hw_ver = VNIC_DEV_HW_VER_A2;
	else
		*hw_ver = VNIC_DEV_HW_VER_UNKNOWN;

	return 0;
}

int vnicky_dev_spec(struct vnicky_dev *vdev, unsigned int offset, unsigned int size,
	void *value)
{
	u8 *v = vnicky_dev_get_res(vdev, RES_TYPE_DEV, 0);
	if (!v) {
		printk(KERN_ERR
			"vNIC device-specific region not found.\n");
		return -EINVAL;
	}

	switch (size) {
	case 1: *(u8 *)value = ioread8(v + offset); break;
	case 2: *(u16 *)value = ioread16(v + offset); break;
	case 4: *(u32 *)value = ioread32(v + offset); break;
	case 8: *(u64 *)value = readq(v + offset); break;
	default: BUG(); break;
	}

	return 0;
}

int vnicky_dev_stats_clear(struct vnicky_dev *vdev)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_STATS_CLEAR, &a0, &a1, wait);
}

int vnicky_dev_stats_dump(struct vnicky_dev *vdev, struct vnicky_stats **stats)
{
	u64 a0, a1;
	int wait = 1000;

	if (!vdev->stats) {
		vdev->stats = pci_alloc_consistent(vdev->pdev,
			sizeof(struct vnicky_stats), &vdev->stats_pa);
		if (!vdev->stats)
			return -ENOMEM;
	}

	*stats = vdev->stats;
	a0 = vdev->stats_pa;
	a1 = sizeof(struct vnicky_stats);

	return vnicky_dev_cmd(vdev, CMD_STATS_DUMP, &a0, &a1, wait);
}

int vnicky_dev_close(struct vnicky_dev *vdev)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_CLOSE, &a0, &a1, wait);
}

int vnicky_dev_enable(struct vnicky_dev *vdev)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_ENABLE, &a0, &a1, wait);
}

int vnicky_dev_disable(struct vnicky_dev *vdev)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_DISABLE, &a0, &a1, wait);
}

int vnicky_dev_open(struct vnicky_dev *vdev, int arg)
{
	u64 a0 = (u32)arg, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_OPEN, &a0, &a1, wait);
}

int vnicky_dev_open_done(struct vnicky_dev *vdev, int *done)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	int err;

	*done = 0;

	err = vnicky_dev_cmd(vdev, CMD_OPEN_STATUS, &a0, &a1, wait);
	if (err)
		return err;

	*done = (a0 == 0);

	return 0;
}

int vnicky_dev_soft_reset(struct vnicky_dev *vdev, int arg)
{
	u64 a0 = (u32)arg, a1 = 0;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_SOFT_RESET, &a0, &a1, wait);
}

int vnicky_dev_soft_reset_done(struct vnicky_dev *vdev, int *done)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	int err;

	*done = 0;

	err = vnicky_dev_cmd(vdev, CMD_SOFT_RESET_STATUS, &a0, &a1, wait);
	if (err)
		return err;

	*done = (a0 == 0);

	return 0;
}

int vnicky_dev_hang_notify(struct vnicky_dev *vdev)
{
	u64 a0, a1;
	int wait = 1000;
	return vnicky_dev_cmd(vdev, CMD_HANG_NOTIFY, &a0, &a1, wait);
}

int vnicky_dev_mac_addr(struct vnicky_dev *vdev, u8 *mac_addr)
{
	u64 laa = 0x02;
	memcpy(mac_addr, &laa, ETH_ALEN);
	return 0;
}

void vnicky_dev_packet_filter(struct vnicky_dev *vdev, int directed, int multicast,
	int broadcast, int promisc, int allmulti)
{
	u64 a0, a1 = 0;
	int wait = 1000;
	int err;

	a0 = (directed ? CMD_PFILTER_DIRECTED : 0) |
	     (multicast ? CMD_PFILTER_MULTICAST : 0) |
	     (broadcast ? CMD_PFILTER_BROADCAST : 0) |
	     (promisc ? CMD_PFILTER_PROMISCUOUS : 0) |
	     (allmulti ? CMD_PFILTER_ALL_MULTICAST : 0);

	err = vnicky_dev_cmd(vdev, CMD_PACKET_FILTER, &a0, &a1, wait);
	if (err)
		printk(KERN_ERR "Can't set packet filter\n");
}

void vnicky_dev_packet_filter_all(struct vnicky_dev *vdev, int directed,
	int multicast, int broadcast, int promisc, int allmulti)
{
	u64 a0, a1 = 0;
	int wait = 1000;
	int err;

	a0 = (directed ? CMD_PFILTER_DIRECTED : 0) |
	     (multicast ? CMD_PFILTER_MULTICAST : 0) |
	     (broadcast ? CMD_PFILTER_BROADCAST : 0) |
	     (promisc ? CMD_PFILTER_PROMISCUOUS : 0) |
	     (allmulti ? CMD_PFILTER_ALL_MULTICAST : 0);

	err = vnicky_dev_cmd(vdev, CMD_PACKET_FILTER_ALL, &a0, &a1, wait);
	if (err)
		printk(KERN_ERR "Can't set packet filter\n");
}

void vnicky_dev_add_addr(struct vnicky_dev *vdev, u8 *addr)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	int err;
	int i;

	for (i = 0; i < ETH_ALEN; i++)
		((u8 *)&a0)[i] = addr[i];

	err = vnicky_dev_cmd(vdev, CMD_ADDR_ADD, &a0, &a1, wait);
	if (err)
		printk(KERN_ERR
			"Can't add addr [%02x:%02x:%02x:%02x:%02x:%02x], %d\n",
			addr[0], addr[1], addr[2], addr[3], addr[4], addr[5],
			err);
}

void vnicky_dev_del_addr(struct vnicky_dev *vdev, u8 *addr)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	int err;
	int i;

	for (i = 0; i < ETH_ALEN; i++)
		((u8 *)&a0)[i] = addr[i];

	err = vnicky_dev_cmd(vdev, CMD_ADDR_DEL, &a0, &a1, wait);
	if (err)
		printk(KERN_ERR
			"Can't del addr [%02x:%02x:%02x:%02x:%02x:%02x], %d\n",
			addr[0], addr[1], addr[2], addr[3], addr[4], addr[5],
			err);
}

int vnicky_dev_raise_intr(struct vnicky_dev *vdev, u16 intr)
{
	u64 a0 = intr, a1 = 0;
	int wait = 1000;
	int err;

	err = vnicky_dev_cmd(vdev, CMD_IAR, &a0, &a1, wait);
	if (err)
		printk(KERN_ERR "Failed to raise INTR[%d], err %d\n",
			intr, err);

	return err;
}

int vnicky_dev_notify_setcmd(struct vnicky_dev *vdev, 
	void *notify_addr, dma_addr_t notify_pa, u16 intr)
{
	u64 a0, a1;
	int wait = 1000;
	int r;

        if (vdev->notify || vdev->notify_pa) {
		printk(KERN_ERR "notify block %p still allocated",
			vdev->notify);
		return -EINVAL;
        }
	memset(notify_addr, 0, sizeof(struct vnicky_devcmd_notify));
	vdev->notify = notify_addr;
	vdev->notify_pa = notify_pa;

	a0 = (u64)notify_pa;
	a1 = ((u64)intr << 32) & 0x0000ffff00000000ULL;
	a1 += sizeof(struct vnicky_devcmd_notify);

	r = vnicky_dev_cmd(vdev, CMD_NOTIFY, &a0, &a1, wait);
	vdev->notify_sz = (r == 0) ? (u32)a1 : 0;
	return r;
}

int vnicky_dev_notify_set(struct vnicky_dev *vdev, u16 intr)
{
	void *notify_addr;
	dma_addr_t notify_pa;

	notify_addr = pci_alloc_consistent(vdev->pdev,
			sizeof(struct vnicky_devcmd_notify),
			&notify_pa);
	if (!notify_addr)
		return -ENOMEM;

	return vnicky_dev_notify_setcmd(vdev, notify_addr, notify_pa, intr);
}

void vnicky_dev_notify_unsetcmd(struct vnicky_dev *vdev)
{
	u64 a0, a1;
	int wait = 1000;

	a0 = 0;  /* paddr = 0 to unset notify buffer */
	a1 = 0x0000ffff00000000ULL; /* intr num = -1 to unreg for intr */
	a1 += sizeof(struct vnicky_devcmd_notify);

	vnicky_dev_cmd(vdev, CMD_NOTIFY, &a0, &a1, wait);
	vdev->notify = NULL;
	vdev->notify_pa = 0;
	vdev->notify_sz = 0;
}

void vnicky_dev_notify_unset(struct vnicky_dev *vdev)
{
	if (vdev->notify) {
		pci_free_consistent(vdev->pdev,
			sizeof(struct vnicky_devcmd_notify),
			vdev->notify,
			vdev->notify_pa);
	}

	vnicky_dev_notify_unsetcmd(vdev);
}

static int vnicky_dev_notify_ready(struct vnicky_dev *vdev)
{
	u32 *words;
	unsigned int nwords = vdev->notify_sz / 4;
	unsigned int i;
	u32 csum;

	if (!vdev->notify || !vdev->notify_sz)
		return 0;

	do {
		csum = 0;
		memcpy(&vdev->notify_copy, vdev->notify, vdev->notify_sz);
		words = (u32 *)&vdev->notify_copy;
		for (i = 1; i < nwords; i++)
			csum += words[i];
	} while (csum != words[0]);

	return 1;
}

int vnicky_dev_init(struct vnicky_dev *vdev, int arg)
{
	u64 a0 = (u32)arg, a1 = 0;
	int wait = 1000;
	int r = 0;

	if (vdev->cap_flags & VNIC_DEV_CAP_INIT)
		r = vnicky_dev_cmd(vdev, CMD_INIT, &a0, &a1, wait);
	else {
		vnicky_dev_cmd(vdev, CMD_INIT_v1, &a0, &a1, wait);
		if (a0 & CMD_INITF_DEFAULT_MAC) {
			// Emulate these for old CMD_INIT_v1 which
			// didn't pass a0 so no CMD_INITF_*.
			vnicky_dev_cmd(vdev, CMD_MAC_ADDR, &a0, &a1, wait);
			vnicky_dev_cmd(vdev, CMD_ADDR_ADD, &a0, &a1, wait);
		}
	}
	return r;
}

int vnicky_dev_init_done(struct vnicky_dev *vdev, int *done)
{
	u64 a0 = 0, a1 = 0;
	int wait = 1000;
	int err;

	*done = 0;

	err = vnicky_dev_cmd(vdev, CMD_INIT_STATUS, &a0, &a1, wait);
	if (err)
		return err;

	*done = (a0 == 0);

	return 0;
}

int vnicky_dev_link_status(struct vnicky_dev *vdev)
{
	return 1;
}

u32 vnicky_dev_port_speed(struct vnicky_dev *vdev)
{
	if (!vnicky_dev_notify_ready(vdev))
		return 0;

	return vdev->notify_copy.port_speed;
}

u32 vnicky_dev_msg_lvl(struct vnicky_dev *vdev)
{
	if (!vnicky_dev_notify_ready(vdev))
		return 0;

	return vdev->notify_copy.msglvl;
}

u32 vnicky_dev_mtu(struct vnicky_dev *vdev)
{
	return 1500;
}

u32 vnicky_dev_link_down_cnt(struct vnicky_dev *vdev)
{
	if (!vnicky_dev_notify_ready(vdev))
		return 0;

	return vdev->notify_copy.link_down_cnt;
}

u32 vnicky_dev_notify_status(struct vnicky_dev *vdev)
{
	if (!vnicky_dev_notify_ready(vdev))
		return 0;

	return vdev->notify_copy.status;
}

u32 vnicky_dev_perbi_rebuild_cnt(struct vnicky_dev *vdev)
{
	if (!vnicky_dev_notify_ready(vdev))
		return 0;

	return vdev->notify_copy.perbi_rebuild_cnt;
}

void vnicky_dev_set_intr_mode(struct vnicky_dev *vdev,
	enum vnicky_dev_intr_mode intr_mode)
{
	vdev->intr_mode = intr_mode;
}

enum vnicky_dev_intr_mode vnicky_dev_get_intr_mode(
	struct vnicky_dev *vdev)
{
	return vdev->intr_mode;
}

void vnicky_dev_unregister(struct vnicky_dev *vdev)
{
	if (vdev) {
		if (vdev->notify)
			pci_free_consistent(vdev->pdev,
				sizeof(struct vnicky_devcmd_notify),
				vdev->notify,
				vdev->notify_pa);
		if (vdev->linkstatus)
			pci_free_consistent(vdev->pdev,
				sizeof(u32),
				vdev->linkstatus,
				vdev->linkstatus_pa);
		if (vdev->stats)
			pci_free_consistent(vdev->pdev,
				sizeof(struct vnicky_dev),
				vdev->stats, vdev->stats_pa);
		if (vdev->fw_info)
			pci_free_consistent(vdev->pdev,
				sizeof(struct vnicky_devcmd_fw_info),
				vdev->fw_info, vdev->fw_info_pa);
		kfree(vdev);
	}
}

struct vnicky_dev *vnicky_dev_register(struct vnicky_dev *vdev,
	void *priv, struct pci_dev *pdev, struct vnicky_dev_bar *bar,
	unsigned int num_bars)
{
	if (!vdev) {
		vdev = kzalloc(sizeof(struct vnicky_dev), GFP_ATOMIC);
		if (!vdev)
			return NULL;
	}

	vdev->priv = priv;
	vdev->pdev = pdev;

	if (vnicky_dev_discover_res(vdev, bar, num_bars))
		goto err_out;


	vdev->cap_flags = 0;

	if (vnicky_dev_capable(vdev, CMD_INIT))
		vdev->cap_flags |= VNIC_DEV_CAP_INIT;

	return vdev;

err_out:
	vnicky_dev_unregister(vdev);
	return NULL;
}

struct pci_dev *vnicky_dev_get_pdev(struct vnicky_dev *vdev)
{
	return vdev->pdev;
}

int vnicky_dev_int13(struct vnicky_dev *vdev, u64 arg, u32 op)
{
	u64 a0 = arg, a1 = op;
	int wait = 1000;
	int r = 0;

	r = vnicky_dev_cmd(vdev, CMD_INT13, &a0, &a1, wait);
	return r;
}

int vnicky_dev_perbi(struct vnicky_dev *vdev, u64 arg, u32 op)
{
	u64 a0 = arg, a1 = op;
	int wait = 5000;
	int r = 0;

	r = vnicky_dev_cmd(vdev, CMD_PERBI, &a0, &a1, wait);

	return r;
}


