/*
 *  Linux MegaRAID driver for SAS based RAID controllers
 *
 *  Copyright (c) 2009-2012  LSI Corporation.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  FILE: megaraid_sas_raptor.c
 *
 *  Authors: LSI Corporation
 *           Sumant Patro
 *           Adam Radford <linuxraid@lsi.com>
 *
 *  Send feedback to: <megaraidlinux@lsi.com>
 *
 *  Mail to: LSI Corporation, 1621 Barber Lane, Milpitas, CA 95035
 *     ATTN: Linuxraid
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/mutex.h>
#include <linux/interrupt.h>
#include <linux/uio.h>
#include <linux/compat.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>

#include "megaraid_sas_fusion.h"
#include "megaraid_sas.h"

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef RAPTOR_HYBRID_IO_MASK
#define RAPTOR_HYBRID_IO_MASK  0x80
#endif


struct
megasas_cmd *megasas_get_cmd(
		struct megasas_instance *instance
		);

void
megasas_return_cmd(
		struct megasas_instance *instance,
		struct megasas_cmd *cmd
		);

int
megasas_issue_polled(
		struct megasas_instance *instance,
		struct megasas_cmd *cmd
		);

u32
mega_mod64(
		u64 dividend,
		u32 divisor
		);



int
raptor_logic_qualify_for_cmd_failed(
		struct megasas_instance *instance,
		struct megasas_cmd_fusion *cmd,
		struct IO_REQUEST_INFO *io_info
		)
{
	struct fusion_context *fusion =
		(struct fusion_context *)instance->ctrl_context;
	MR_DRV_RAID_MAP_ALL *local_map_ptr;
	MR_LD_RAID  *raid;
	u32 ld;

	local_map_ptr = fusion->ld_drv_map[(instance->map_id & 1)];
	ld = MR_TargetIdToLdGet(io_info->ldTgtId, local_map_ptr);

	cmd->raptor_logic_enabled = FALSE;
	raid = MR_LdRaidGet(ld, local_map_ptr);

	if ((raid->writeMode == WRITE_BACK) &&
		(raid->level == 0 || raid->level == 5)) {
		cmd->raptor_logic_enabled = TRUE;

		if (list_empty(
				&fusion
				->raptor_hash_bucket_node_list
				->raptor_hash_bucket_node_free_list_head)
		   ) {
			printk(KERN_ERR "hash bucket node list is empty\n");
			cmd->raptor_logic_enabled = FALSE;
			cmd->raptor_unqualified = FALSE;
			return 1;
		}
	}
	return 0;
}

inline void
raptor_logic_unqualify_for_cmd(
		struct megasas_instance *instance,
		struct megasas_cmd_fusion *cmd,
		struct IO_REQUEST_INFO *io_info
		)
{
	cmd->raptor_logic_enabled = FALSE;
	if (!(io_info->isRead)) {
		atomic_inc(&instance->fw_write_outstanding);
		cmd->raptor_unqualified = TRUE;
	}
}


int
raptor_alloc_hash_bucket_list(struct megasas_instance *instance)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	RAPTOR_HASH_BUCKET_LIST *r_hash_bucket_list = NULL;
	RAPTOR_HASH_BUCKET *r_hash_bucket = NULL;
	int hash_bucket_index;

	spin_lock_init(&fusion->raptor_hash_bucket_list_lock);
	rwlock_init(&fusion->raptor_th_bucket_rwlock);

	fusion->raptor_hash_bucket_list =
			kzalloc(
				sizeof(RAPTOR_HASH_BUCKET_LIST),
				GFP_KERNEL
				);

	if (!fusion->raptor_hash_bucket_list) {
		printk(KERN_ERR "no memory for hash bucket list\n");
		return -ENOMEM;
	}

	r_hash_bucket_list =
			fusion->raptor_hash_bucket_list;
	r_hash_bucket_list->total_buckets =
			roundup_pow_of_two(instance->max_fw_cmds);
	r_hash_bucket_list->raptor_hash_bucket_start =
			kzalloc(
					sizeof(RAPTOR_HASH_BUCKET) *
						r_hash_bucket_list->total_buckets,
					GFP_KERNEL
					);

	if (!r_hash_bucket_list->raptor_hash_bucket_start) {
		printk(KERN_ERR "no memory for hash bucket\n");
		kfree(fusion->raptor_hash_bucket_list);
		return -ENOMEM;
	}

	r_hash_bucket = r_hash_bucket_list->raptor_hash_bucket_start;

	for (hash_bucket_index = 0;
		hash_bucket_index < r_hash_bucket_list->total_buckets;
		hash_bucket_index++, r_hash_bucket++) {
		INIT_LIST_HEAD(&r_hash_bucket->head);
		r_hash_bucket->index = hash_bucket_index;
	}

	return 0;
}

void
raptor_free_hash_bucket_list(struct megasas_instance *instance)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;

	kfree(fusion->raptor_hash_bucket_list->raptor_hash_bucket_start);
	fusion->raptor_hash_bucket_list->raptor_hash_bucket_start = NULL;
	fusion->raptor_hash_bucket_list->total_buckets = 0;

	kfree(fusion->raptor_hash_bucket_list);
	fusion->raptor_hash_bucket_list = NULL;
}

int
raptor_alloc_hash_bucket_node_list(struct megasas_instance *instance)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	RAPTOR_HASH_BUCKET_NODE_LIST *r_hash_bucket_node_list = NULL;
	RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node = NULL;
	int hash_bucket_node;

	fusion->raptor_hash_bucket_node_list =
				kzalloc(
					sizeof(RAPTOR_HASH_BUCKET_NODE_LIST),
					GFP_KERNEL
					);

	if (!fusion->raptor_hash_bucket_node_list) {
		printk(KERN_ERR "no memory for hash bucket node list\n");
		return -ENOMEM;
	}

	r_hash_bucket_node_list = fusion->raptor_hash_bucket_node_list;
	r_hash_bucket_node_list->total_nodes =
				roundup_pow_of_two(instance->max_fw_cmds);

	r_hash_bucket_node_list->raptor_hash_bucket_node_start =
				kzalloc(
						sizeof(RAPTOR_HASH_BUCKET_NODE) *
						r_hash_bucket_node_list->total_nodes,
						GFP_KERNEL
						);

	if (!r_hash_bucket_node_list->raptor_hash_bucket_node_start) {
		printk(KERN_ERR "no memory for hash bucket nodes\n");
		kfree(fusion->raptor_hash_bucket_node_list);
		return -ENOMEM;
	}

	INIT_LIST_HEAD(
		&r_hash_bucket_node_list->
			raptor_hash_bucket_node_free_list_head);

	r_hash_bucket_node = r_hash_bucket_node_list->
							raptor_hash_bucket_node_start;

	for (hash_bucket_node = 0;
		hash_bucket_node < r_hash_bucket_node_list->total_nodes;
		hash_bucket_node++, r_hash_bucket_node++) {
		INIT_LIST_HEAD(&r_hash_bucket_node->node);
		list_add_tail(
			&r_hash_bucket_node->node,
			&r_hash_bucket_node_list->
				raptor_hash_bucket_node_free_list_head
			);
	}

	return 0;
}

void
raptor_free_hash_bucket_node_list(struct megasas_instance *instance)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;

	kfree(fusion->raptor_hash_bucket_node_list
				->raptor_hash_bucket_node_start);
	fusion->raptor_hash_bucket_node_list
				->raptor_hash_bucket_node_start = NULL;

	kfree(fusion->raptor_hash_bucket_node_list);
	fusion->raptor_hash_bucket_node_list = NULL;
}

RAPTOR_HASH_BUCKET_NODE*
raptor_get_hash_bucket_node(struct megasas_instance *instance)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node = NULL;
	unsigned long lock_flags;

	spin_lock_irqsave(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);

	if (list_empty(
			&fusion->raptor_hash_bucket_node_list
			->raptor_hash_bucket_node_free_list_head
		)) {
		spin_unlock_irqrestore(
			&fusion->raptor_hash_bucket_list_lock,
			lock_flags
			);
		printk(KERN_ERR "hash bucket node is empty");
		return NULL;
	}

	r_hash_bucket_node =
		list_first_entry(
			&fusion->raptor_hash_bucket_node_list
					->raptor_hash_bucket_node_free_list_head,
			RAPTOR_HASH_BUCKET_NODE, node
			);
	list_del_init(&r_hash_bucket_node->node);

	spin_unlock_irqrestore(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);

	r_hash_bucket_node->ldTgtId = USHRT_MAX;
	r_hash_bucket_node->outstanding = 0;

	return r_hash_bucket_node;
}

/* Assumed that caller holds the lock to the list */
void
raptor_put_hash_bucket_node(
	struct megasas_instance *instance,
	RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node
	)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;

	r_hash_bucket_node->ldTgtId = USHRT_MAX;

	list_del_init(&r_hash_bucket_node->node);
	list_add_tail(
		&r_hash_bucket_node->node,
		&fusion->raptor_hash_bucket_node_list
				->raptor_hash_bucket_node_free_list_head
		);
}

void raptor_add_node_to_hash_bucket_list(
			struct megasas_instance *instance,
			RAPTOR_HASH_BUCKET *r_hash_bucket,
			RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node
			)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	unsigned long lock_flags;

	spin_lock_irqsave(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);

	list_add_tail(
				&r_hash_bucket_node->node,
				&r_hash_bucket->head
				);

	spin_unlock_irqrestore(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);
}


void
raptor_complete_cmd(
				struct megasas_instance *instance,
				struct megasas_cmd_fusion *cmd
				)
{
	RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node;
	struct fusion_context *fusion;
	unsigned long lock_flags;

	/* In case of non-Raptor Controller, we have nothing to do */
	if (!instance->fw_supportFastPathWB)
		return;

	fusion = (struct fusion_context *)instance->ctrl_context;


	if (cmd->raptor_logic_enabled) {
		/* r_hash_bucket_node needs to be protected since
		 * it is being searched in dispatch code. Hold the
		 * lock so that both r_hash_bucket_node and associated
		 * lists are protected
		 */
		spin_lock_irqsave(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);

		r_hash_bucket_node = cmd->r_hash_bucket_node;
		r_hash_bucket_node->outstanding--;
		if (r_hash_bucket_node->outstanding == 0) {
			raptor_put_hash_bucket_node(
					instance,
					r_hash_bucket_node
					);
			atomic_dec(&instance->fw_write_outstanding);
			cmd->r_hash_bucket_node = NULL;
			cmd->raptor_logic_enabled = FALSE;
			cmd->raptor_unqualified = FALSE;
		}

		spin_unlock_irqrestore(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);
	} else {
		if (instance->fw_supportFastPathWB &&
			(cmd->scmd->sc_data_direction == PCI_DMA_TODEVICE)
			&& cmd->raptor_unqualified) {
			atomic_dec(&instance->fw_write_outstanding);
			cmd->raptor_unqualified = FALSE;
		}
	}
}


/*
 * raptor_populate_threshold_buckets -  populate the driver data structures for
 *                                      bucket list by using the data from
 *                                      firmware
 * @instance                          - megasas instance
 * @mrBucketList                      - pointer to MR bucket list data got from
 *                                      firmware
 *
 * Return Value:
 *                                 0  - success
 *                                 <0 - error
 */
int
raptor_populate_threshold_buckets(
						struct megasas_instance *instance,
						MR_BUCKET_LIST *mrBucketList
						)
{
	u32 ld;
	MR_LD_DATA *ldData = NULL;
	struct fusion_context *fusion =
			(struct fusion_context *)(instance->ctrl_context);
	RAPTOR_THRESHOLD_BUCKET_LIST *raptorThresholdBucketList = NULL;
	RAPTOR_LD_DATA *raptorLdData = NULL;
	u8 bucketNum;
	RAPTOR_BUCKET *raptorBucket = NULL;
	MR_BUCKET *mrBucket = NULL;
	unsigned long lock_flags;

	if (!fusion) {
		printk(KERN_ERR "Fusion context is NULL\n");
		return -1;
	}

	write_lock_irqsave(
			&fusion->raptor_th_bucket_rwlock,
			lock_flags
			);

	raptorThresholdBucketList = fusion->raptor_threshold_bucket_list;

	if (raptorThresholdBucketList == NULL) {
		printk(KERN_ERR "raptor threshold bucket list is NULL\n");

		write_unlock_irqrestore(
			&fusion->raptor_th_bucket_rwlock,
			lock_flags
			);
		return -1;
	}

	memset(
		raptorThresholdBucketList,
		0,
		sizeof(RAPTOR_THRESHOLD_BUCKET_LIST)
		);

	raptorThresholdBucketList->ldCount = mrBucketList->ldCount;

	for (ld = 0; ld < mrBucketList->ldCount; ld++) {
		ldData = &mrBucketList->ldData[0];
		ldData =
			(MR_LD_DATA *)(
				(char *)ldData +
				ld *
					(offsetof(MR_LD_DATA, bucket) +
					mrBucketList->maxBuckets *
					sizeof(MR_BUCKET)
					)
			);
		if (ldData->ldTargetId != 0xFF) {
			/* Only populate driver structures for the LDs
			 * for which firmware has populated.
			 */
			raptorLdData =
				&(raptorThresholdBucketList->
					raptorLdData[ldData->ldTargetId]);

			raptorLdData->valid = TRUE;
			raptorLdData->isDefaultLdIo = ldData->isDefaultLdIo;
			raptorLdData->numBuckets = ldData->numBuckets;

			for (bucketNum = 0;
					bucketNum < raptorLdData->numBuckets;
					bucketNum++) {
				raptorBucket =
					&(raptorLdData->
						bucket[bucketNum]);
				mrBucket = ldData->bucket + bucketNum;
				raptorBucket->isLdIo = mrBucket->isLdIo;
				raptorBucket->isAlignReqd =
						mrBucket->isAlignReqd;
				raptorBucket->isBucketRangeAbsolute =
						mrBucket->isBucketRangeAbsolute;
				raptorBucket->threshold = mrBucket->threshold;

				/* Convert the KB into number of blocks */
				raptorBucket->bucketRange =
						mrBucket->bucketRange << 1;
			}
		}
	}

	write_unlock_irqrestore(&fusion->raptor_th_bucket_rwlock, lock_flags);
	return 0;
}

/*
 * raptor_get_ctrl_tunable  - This function is used to fire the DCMD to get
 *                              bucket list from firmware. It's a synchronous
 *                              call.
 *
 * @instance                  - megasas instance
 *
 * Return value
 *                          0  - success
 *                          <0 - error
 */
int
raptor_get_ctrl_tunable(struct megasas_instance *instance)
{
	int ret = 0;
	struct megasas_cmd *cmd;
	struct megasas_dcmd_frame *dcmd;
	MR_BUCKET_LIST *ci;
	dma_addr_t ci_h = 0;
	u32 size;
	struct fusion_context *fusion;

	cmd = megasas_get_cmd(instance);

	if (!cmd) {
		printk(KERN_DEBUG "megasas (%s): Failed to get cmd\n",
				__func__);
		return -ENOMEM;
	}

	fusion = instance->ctrl_context;

	if (!fusion) {
		megasas_return_cmd(instance, cmd);
		return 1;
	}

	dcmd = &cmd->frame->dcmd;

	size = offsetof(MR_BUCKET_LIST, ldData)
			+ instance->CurLdCount *
			(offsetof(MR_LD_DATA, bucket)
			+ RAPTOR_MAX_THRESHOLD_BUCKETS * sizeof(MR_BUCKET)
			);

	ci = pci_alloc_consistent(instance->pdev, size, &ci_h);

	if (!ci) {
		printk(
			KERN_DEBUG
			"megasas: Failed to alloc mem for ctrl tunable list\n"
			);
		megasas_return_cmd(instance, cmd);
		return -ENOMEM;
	}

	memset(ci, 0, sizeof(*ci));
	ci->ldCount = instance->CurLdCount;
	ci->maxBuckets = RAPTOR_MAX_THRESHOLD_BUCKETS;
	ci->bucketLdDataSize = sizeof(MR_LD_DATA);
	ci->size = size;

	memset(dcmd->mbox.b, 0, MFI_MBOX_SIZE);

	dcmd->cmd = MFI_CMD_DCMD;
	dcmd->cmd_status = 0xFF;
	dcmd->sge_count = 1;
	dcmd->flags = MFI_FRAME_DIR_READ;
	dcmd->timeout = 0;
	dcmd->pad_0 = 0;
	dcmd->data_xfer_len = size;
	dcmd->opcode = MR_DCMD_CTRL_TUNABLE_LIST_GET;
	dcmd->sgl.sge32[0].phys_addr = ci_h;
	dcmd->sgl.sge32[0].length = size;

	if (!megasas_issue_polled(instance, cmd))
		ret = 0;
	else {
		printk(KERN_ERR "megasas: Get Ctrl tunable Failed\n");
		ret = -1;
	}



	if ((ret == 0) && (dcmd->cmd_status == 0)) {
		instance->fw_io_optimization =
			(raptor_populate_threshold_buckets(instance, ci)) ?
				0 : 1;
	} else {
		printk(
			KERN_ERR
			"megasas: Ctrl tunable ret %x, dcmd status %x\n",
			ret, dcmd->cmd_status
			);
		instance->fw_io_optimization = 0;
	}

	pci_free_consistent(instance->pdev, size, ci, ci_h);

	megasas_return_cmd(instance, cmd);

	return ret;
}

/* The caller has to take care to call this function for raptor firmware only */
inline RAPTOR_HASH_BUCKET*
raptor_get_hash_bucket(struct megasas_instance *instance, u64 end_row)
{
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	return
		fusion->raptor_hash_bucket_list->raptor_hash_bucket_start
		+
		(end_row &
			(fusion->raptor_hash_bucket_list->total_buckets - 1)
		);
}


/*
 ******************************************************************************
 *
 * This routine idenfies if the write command is outstanding on the row where
 * write IO has come.
 *
 * Inputs :
 *
 *    instance           - megasas instance
 *    ldTgtId            - Ld target id
 *    end_row            - end row number of this write Io
 *    r_hash_bucket      - pointer to hash bucket
 *
 * Outputs :
 *
 *    r_hash_bucket_node - if the bucket node corresponding to end_row and
 *                         ldTgtId is found, pointer to that node is returned,
 *                         else NULL is returned.
 *    TRUE               - if node is found
 *    FALSE              - if node is not found
 */
u8
raptor_is_write_cmd_pending_on_this_row(
		struct megasas_instance *instance,
		u16 ldTgtId,
		u64 end_row,
		RAPTOR_HASH_BUCKET *r_hash_bucket,
		RAPTOR_HASH_BUCKET_NODE **r_hash_bucket_node
		)
{
	struct list_head *pos;
	RAPTOR_HASH_BUCKET_NODE *temp_node;
	struct fusion_context *fusion =
				(struct fusion_context *)instance->ctrl_context;
	unsigned long lock_flags;

	*r_hash_bucket_node = NULL;

	spin_lock_irqsave(&fusion->raptor_hash_bucket_list_lock, lock_flags);

	if (list_empty(&r_hash_bucket->head)) {
		spin_unlock_irqrestore(
				&fusion->raptor_hash_bucket_list_lock,
				lock_flags
				);
		return FALSE;
	}

	list_for_each(pos, &r_hash_bucket->head) {
		temp_node = container_of(pos, RAPTOR_HASH_BUCKET_NODE, node);

		if ((temp_node->ldTgtId == ldTgtId)
				&& (temp_node->row == end_row)) {
			temp_node->outstanding++;
			*r_hash_bucket_node = temp_node;
			spin_unlock_irqrestore(
					&fusion->raptor_hash_bucket_list_lock,
					lock_flags
					);
			return TRUE;
		}
	}

	spin_unlock_irqrestore(
			&fusion->raptor_hash_bucket_list_lock,
			lock_flags
			);
	return FALSE;
}


/*
 *  **********************************************************************
 *
 * This routine processes every write IO and put the IO in the right
 * hash bucket. If the IO is on Raid 0/ Raid 5 Raptor VD, it will put the
 * IO in the right bucket or create new bucket node.  Else, it will not
 * put it in bucket but increment the outstanding counters.
 *
 * Inputs :
 *
 *    instance           - megasas instance
 *    cmd                - megasas write command
 *    io_info            - information about this IO
 *    numBlocks          - IO size in blocks
 *    ldTgtId            - Ld target id
 *    end_row            - end row number of this write Io
 *    raid               - raid info for this LD
 *    pRAID_Context      - raid context
 *
 * Outputs :
 *          None
 */

void
raptor_process_writes_in_same_row(struct megasas_instance *instance,
				struct megasas_cmd_fusion *cmd,
				struct IO_REQUEST_INFO *io_info,
				u32 numBlocks,
				u32 ldTgtId,
				u64 endRow,
				MR_LD_RAID *raid,
				RAID_CONTEXT **pRAID_Context)
{
	struct fusion_context *fusion;
	unsigned long lock_flags;
	RAPTOR_HASH_BUCKET_NODE *r_hash_bucket_node;
	RAPTOR_HASH_BUCKET *r_hash_bucket;
	RAPTOR_LD_DATA *raptorLdData;
	u8 bucketNum;
	u8 gotBucket = FALSE;

	if (cmd->raptor_logic_enabled) {
		fusion = (struct fusion_context *)instance->ctrl_context;
		r_hash_bucket = raptor_get_hash_bucket(instance, endRow);

		if (raptor_is_write_cmd_pending_on_this_row(
					instance, ldTgtId,
					endRow, r_hash_bucket,
					&r_hash_bucket_node)
				) {
			io_info->fpOkForIo = FALSE;
		} else {
			/* We don't check for fpOkForIo for Raid 5, as it
			 * is always false for R5 writes. For R0, it can
			 * be true or false.
			 */
			if ((io_info->fpOkForIo && raid->level == 0) ||
				raid->level == 5
				) {
				read_lock_irqsave(
					&fusion->raptor_th_bucket_rwlock,
					lock_flags
					);

				raptorLdData =
					&(fusion->
						raptor_threshold_bucket_list->
						raptorLdData[io_info->ldTgtId]
					);

				if (!raptorLdData->valid) {
					read_unlock_irqrestore(
						&fusion->raptor_th_bucket_rwlock,
						lock_flags
						);

					cmd->raptor_logic_enabled = FALSE;
					atomic_inc(&instance->fw_write_outstanding);
					cmd->raptor_unqualified = TRUE;
					return;
				}
				for (bucketNum = 0; bucketNum < raptorLdData->numBuckets; bucketNum++) {
					if (raptorLdData->bucket[bucketNum].isBucketRangeAbsolute) {
						/* Bucket Range absolute */
						if (numBlocks == raptorLdData->bucket[bucketNum].bucketRange) {
							if (raptorLdData->bucket[bucketNum].isAlignReqd) {
								/* IO Alignment check required */
								if (!mega_mod64(io_info->ldStartBlock, numBlocks)) {
									/* IO is aligned */
									if (raid->level == 0) {
										if (atomic_read(&instance->fw_write_outstanding) <
												raptorLdData->bucket[bucketNum].threshold) {
											io_info->fpOkForIo = !(raptorLdData->bucket[bucketNum].isLdIo);
										}
									} else if (raid->level == 5) {
									/* R5 Writes are anyway not supported in fast path. We
									 * send only those IOs above bucket threshold to PD. Else
									 * we write them only to DRAM and flush them later
									 */
										if (atomic_read(&instance->fw_write_outstanding) >
												raptorLdData->bucket[bucketNum].threshold) {
											/* Set the hybrid bit to send
											 * this IO in WT mode
											 */
											(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
										}
									} else {
										printk(
										"megaraid_sas Raptor: Aligned IO received for \
										an unexpected raid level %u for FP or non-FP \
										determination\n", raid->level);
										read_unlock_irqrestore(
											&fusion->raptor_th_bucket_rwlock,
											lock_flags
											);
										return;
									}
								} else {
									/* IO is unaligned */
									if (raid->level == 0)
										io_info->fpOkForIo = !(raptorLdData->bucket[bucketNum].isLdIo);
									else if (raid->level == 5) {
										if (atomic_read(&instance->fw_write_outstanding) >
											raptorLdData->bucket[bucketNum].threshold) {
										/* Set the hybrid bit to send
										 * this IO in WT mode
										 */
											(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
										}
									} else {
										printk(
										"megaraid_sas Raptor: Unaligned IO received for \
										an unexpected raid level %u for FP or non-FP \
										determination\n", raid->level);

										read_unlock_irqrestore(
												&fusion->raptor_th_bucket_rwlock,
												lock_flags
												);
										return;
									}
								}
							} else {
								/* IO Alignment check NOT required */
								if (raid->level == 0) {
									if (atomic_read(&instance->fw_write_outstanding) <
											raptorLdData->bucket[bucketNum].threshold)
										io_info->fpOkForIo =
											!(raptorLdData->bucket[bucketNum].isLdIo);
								} else if (raid->level == 5) {
									if (atomic_read(&instance->fw_write_outstanding) >
											raptorLdData->bucket[bucketNum].threshold) {
										(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
									}
								} else {
									printk(
									"megaraid_sas Raptor: IO received for \
									an unexpected raid level %u for FP or \
									non-FP determination\n", raid->level);
									read_unlock_irqrestore(
											&fusion->raptor_th_bucket_rwlock,
											lock_flags
											);
									return;
								}
							}
							gotBucket = TRUE;
							break;
						}
					} else {
						/* Bucket Range not absolute */
						/* Need such a big if check to handle the case where the
						 * threshold bucket is not in ascending order
						 */
						if (((bucketNum != 0) &&
							(numBlocks	> raptorLdData->bucket[bucketNum - 1].bucketRange) &&
							(numBlocks <= raptorLdData->bucket[bucketNum].bucketRange)
							)
							||
							((bucketNum == 0) &&
							 (numBlocks <= raptorLdData->bucket[bucketNum].bucketRange)
							)
							) {
							if (raptorLdData->bucket[bucketNum].isAlignReqd) {
								/* IO Alignment check required */
								if (!mega_mod64(io_info->ldStartBlock, numBlocks)) {
									/* IO is aligned */
									if (raid->level == 0) {
										if (atomic_read(&instance->fw_write_outstanding) <
												raptorLdData->bucket[bucketNum].threshold)
											io_info->fpOkForIo =
												!(raptorLdData->bucket[bucketNum].isLdIo);
									} else if (raid->level == 5) {
										if (atomic_read(&instance->fw_write_outstanding) >
												raptorLdData->bucket[bucketNum].threshold) {
											(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
										}
									} else {
										printk(
										"megaraid_sas Raptor: Aligned IO received \
										for an unexpected raid level %u for FP or \
										non-FP determination\n", raid->level);
										read_unlock_irqrestore(
												&fusion->raptor_th_bucket_rwlock,
												lock_flags
												);
										return;
									}
								} else {
									/* IO is not aligned */
									if (raid->level == 0) {
										io_info->fpOkForIo =
											!(raptorLdData->bucket[bucketNum].isLdIo);
									} else if (raid->level == 5) {
										if (atomic_read(&instance->fw_write_outstanding) >
												raptorLdData->bucket[bucketNum].threshold) {
											/* Set the hybrid bit to send
											 * this IO in WT mode */
											(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
										}
									} else {
										printk(
										"megaraid_sas Raptor: Unaligned IO received \
										for an unexpected raid level %u for FP or \
										non-FP determination\n", raid->level);
										read_unlock_irqrestore(
												&fusion->raptor_th_bucket_rwlock,
												lock_flags
												);
										return;
									}
								}
							} else {
								/* IO Alignment check NOT required */
								if (raid->level == 0) {
									if (atomic_read(&instance->fw_write_outstanding) <
											raptorLdData->bucket[bucketNum].threshold) {
										io_info->fpOkForIo =
											!(raptorLdData->bucket[bucketNum].isLdIo);
									}
								} else if (raid->level == 5) {
									if (atomic_read(&instance->fw_write_outstanding) >
											raptorLdData->bucket[bucketNum].threshold) {
										(*pRAID_Context)->logArm |= RAPTOR_HYBRID_IO_MASK;
									}
								} else {
									printk(
									"megaraid_sas Raptor: IO received \
									for an unexpected raid level %u for \
									FP or non-FP determination\n", raid->level);
									read_unlock_irqrestore(
											&fusion->raptor_th_bucket_rwlock,
											lock_flags
											);
									return;
								}
							}
							gotBucket = TRUE;
							break;
						}
					} /* end BucketRangeCheckAbsolute */
				} /* end for */
				if (!gotBucket && raid->level == 0) {
					io_info->fpOkForIo = !(raptorLdData->isDefaultLdIo);
				}
				read_unlock_irqrestore(
						&fusion->raptor_th_bucket_rwlock,
						lock_flags
						);
			}
			r_hash_bucket_node = raptor_get_hash_bucket_node(instance);
			/* The node is now isolated.
			 * No need to hold a lock for modifying it
			 */
			r_hash_bucket_node->row = endRow;
			r_hash_bucket_node->ldTgtId = ldTgtId;
			r_hash_bucket_node->outstanding = 1;

			raptor_add_node_to_hash_bucket_list(
					instance,
					r_hash_bucket,
					r_hash_bucket_node
					);
			atomic_inc(&instance->fw_write_outstanding);
		}
		cmd->r_hash_bucket_node = r_hash_bucket_node;
		cmd->raptor_unqualified = FALSE;

	} else {
		if (instance->fw_supportFastPathWB && !(io_info->isRead)) {
			atomic_inc(&instance->fw_write_outstanding);
			cmd->raptor_unqualified = TRUE;
		}
	}
}
