/*
 * Copyright 2007 Nuova Systems, Inc.  All rights reserved.
 * Copyright 2009-2014 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: uio_paloints.c 109099 2012-07-26 20:28:42Z daclear $"

#include <linux/device.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/uio_driver.h>
#include <asm/mach-palo/paloint.h>

#define DRV_NAME "paloints"
#define DRV_DESCRIPTION "Driver for Palo Interrupts"
#define DRV_VERSION "1.0"
#define DRV_COPYRIGHT "Copyright 2007-2009 Cisco Systems"

MODULE_DESCRIPTION(DRV_DESCRIPTION);
MODULE_LICENSE("TBD");
MODULE_VERSION(DRV_VERSION);

#define PFX DRV_NAME ": "
/*
#define pi_printk(lvl, dev, fmt, arg...) \
	printk(lvl "%s: " dev->driver.name, fmt, ##arg)
#define pi_info(dev, fmt, arg...) \
	pi_printk(KERN_INFO, dev, fmt, ##arg)
#define pi_err(dev, fmt, arg...) \
	pi_printk(KERN_ERR, dev, fmt, ##arg)
*/
/*
 * We claim all these ints for user-level paloints.
 * Any bits NOT set are left for other kernel modules
 * to claim.  We'll still create a device for it to keep
 * the device minor number predictable, but open will fail
 * for device minor numbers that we don't claim (see paloints_open).
 */
static u32 paloints_claimmask[PALO_NUM_INTS] = {
	/* 0 */0x00000fff, /* [22-25]GPIOs */
	/* 1 */0x007fffff, /* [28-31]rnaxi0-3 INTs */
	/* 2 */0x0000ffff,
	/* 3 */0x4fff4000,
	/* 4 */0x000000ff,
};

static int npaloints;

struct paloints_data {
	int id;
	struct uio_info info[PALO_IRQS_PER_INT];
};

static int paloints_open(struct uio_info *info, struct inode *inode)
{
	if (info->irq == UIO_IRQ_NONE) {
		struct platform_device *dev = info->priv;
		dev_err(&dev->dev, "open: unclaimed intr\n");
		return -EACCES;
	}
	return 0;
}

static irqreturn_t paloints_intr(int irq, struct uio_info *info)
{
	return IRQ_HANDLED;
}

static int __devinit paloints_probe(struct platform_device *dev)
{
	struct paloints_data *pd;
	int i, irq_s, irq_e;

	if (npaloints >= PALO_NUM_INTS)
		return -ENODEV;

	pd = kzalloc(sizeof (struct paloints_data), GFP_KERNEL);
	if (!pd)
		return -ENOMEM;

	pd->id = npaloints++;

	irq_s = palo_int_to_irq(pd->id, 0);
	irq_e = irq_s + PALO_IRQS_PER_INT - 1;

	dev_info(&dev->dev, "claim ints %08x irqs %3d-%3d\n",
		paloints_claimmask[pd->id], irq_s, irq_e);

	for (i = 0; i < PALO_IRQS_PER_INT; i++) {
		struct uio_info *info = &pd->info[i];

		info->name = DRV_NAME;
		info->version = DRV_VERSION;
		info->priv = dev;
		info->open = paloints_open;

		if (paloints_claimmask[pd->id] & (1 << i)) {
			info->irq = palo_int_to_irq(pd->id, i);
			info->irq_flags = IRQF_DISABLED | IRQF_SHARED;
			info->handler = paloints_intr;
		} else {
			info->irq = UIO_IRQ_NONE;
		}

		if (uio_register_device(&dev->dev, info))
			goto out_unregister;
	}

        platform_set_drvdata(dev, pd);
	return 0;

out_unregister:
	while (--i >= 0)
		uio_unregister_device(&pd->info[i]);
	dev_err(&dev->dev, "probe failed\n");
	return -ENODEV;
}

static int __devinit paloints_remove(struct platform_device *dev)
{
	struct paloints_data *pd = platform_get_drvdata(dev);
	int i;

	for (i = 0; i < PALO_IRQS_PER_INT; i++) {
		uio_unregister_device(&pd->info[i]);
	}
	kfree(pd);
	return 0;
}


static struct platform_driver paloints_driver = {
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
	},
	.probe = paloints_probe,
	.remove = __devexit_p(paloints_remove),
};

static int __init paloints_init(void)
{
	printk(KERN_INFO PFX "%s, ver %s\n", DRV_DESCRIPTION, DRV_VERSION);
	printk(KERN_INFO PFX "%s\n", DRV_COPYRIGHT);

	return platform_driver_register(&paloints_driver);
}

static void __exit paloints_exit(void)
{
	platform_driver_unregister(&paloints_driver);
}

module_init(paloints_init);
module_exit(paloints_exit);
