/*
 * Copyright (c) 2013 Cisco Systems, Inc.  All rights reserved.
 * Hardware burst write support code.
 */
#if defined(CONFIG_BASE_SERENO) || defined(CONFIG_BASE_CRUZ)

#ifdef CONFIG_OS_TYPE_LINUXKERNEL
#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>

#include <palo/include/proc_mregs.h>
#include <common/include/hwecpu_addr.h>
#include "regio.h"

#include <palo/fw/common/libburst/include/burst_write.h>
#include <linux/string.h>

#else

#include <sys/types.h>
#include "linux_types.h" /* <sa/src/palo/include/linux_types.h> */

#include "sa_assert.h"   /* <sa/src/common/libsa/include/sa_assert.h> */
#include "proc_mregs.h"

#include "ecpu_types.h"
#include "ecpu_common.h"
#include "ecpu_cache.h"
#include "hwecpu_addr.h"
#include "regio.h"
#include "burst_write.h"
#include <string.h>

#endif /* CONFIG_OS_TYPE_LINUXKERNEL */

#define memory_barrier()        asm volatile("" ::: "memory")

/*
 * NOTE: On fcpu(s), FC uses first 3 contexts (0 - 2) of burst write and
 *       leaves context 3 free.  MCPU has all 4 free.
 *       Therefore, uses context 3 as default for all ECPUS.
 *       On MCPU, other contexts (0 - 2) can also be used as need
 */

struct bw_ctx {
    u_int32_t offset;
    u_int32_t addr;
    u_int32_t size;
    u_int32_t control;
    u_int32_t launch;
} bw_ctxs[BURST_CONTEXT_COUNT] = {
   { BURST_CONTEXT0_OFFSET, /* context 0 */
     PROC_MREGS_BURST_ADDR0, PROC_MREGS_BURST_SIZE0,
     PROC_MREGS_BURST_CONTROL0, PROC_MREGS_BURST_LAUNCH0},
   { BURST_CONTEXT1_OFFSET, /* context 1 */
     PROC_MREGS_BURST_ADDR1, PROC_MREGS_BURST_SIZE1,
     PROC_MREGS_BURST_CONTROL1, PROC_MREGS_BURST_LAUNCH1},
   { BURST_CONTEXT2_OFFSET, /* context 2 */
     PROC_MREGS_BURST_ADDR2, PROC_MREGS_BURST_SIZE2,
     PROC_MREGS_BURST_CONTROL2, PROC_MREGS_BURST_LAUNCH2},
   { BURST_CONTEXT3_OFFSET, /* context 3 */
     PROC_MREGS_BURST_ADDR3, PROC_MREGS_BURST_SIZE3,
     PROC_MREGS_BURST_CONTROL3, PROC_MREGS_BURST_LAUNCH3}
};


static inline void
_intr_enable(unsigned long flags)
{
#if  ! defined(__i386__)

#ifdef CONFIG_OS_TYPE_LINUXKERNEL
    local_irq_restore(flags);
#else

#if  defined(CONFIG_OS_TYPE_LINUXUSER)
    /* linux user space */
#else
	/* threadx */
    __asm__ __volatile__ ("ei %0; ehb" : "=d" (flags));
#endif

#endif /* CONFIG_OS_TYPE_LINUXKERNEL */

#endif /* !i386 */
}


static inline unsigned long
_intr_disable(void)
{
    register unsigned long flags = 0;

#if  ! defined(__i386__)

#ifdef CONFIG_OS_TYPE_LINUXKERNEL
    local_irq_save(flags);
#else

#if  defined(CONFIG_OS_TYPE_LINUXUSER)
    /* linux user space */
#else
	/* threadx */
    __asm__ __volatile__ ("di %0; ehb" : "=d" (flags));
    flags = (flags & SR_IE);
#endif
#endif /* CONFIG_OS_TYPE_LINUXKERNEL */

#endif /* !i386 */
    return flags;
}
/*
 * src_va => Virt src address of data to copy into the burst data registers
 * dst_pa => PHY dst address (MUST be 8-byte aligned) mips memory
 * offset => byte offset into the burst data registers AND destination memory
 * length => Length in bytes of data to move
 * ctx_no => context number
 *
 * Returns:
 *  0 => Success
 * BURST_WRITE_ERR_SIZE(-1) => if length exceeds maximum reserved burst area
 * BURST_WRITE_ERR_ALIGN(-2) => if dst addr not 8 byte aligned
 * BURST_WRITE_ERR_ALIGN_XFER(-3) =>
 *                          if (dst addr + length) crosses 256 byte boundary
 *
 * NOTE:: caller need to make sure no concurrently calls.
 */
int
burst_write_buffer(void  *src_va,
                   u_int64_t dst_pa,
                   u_int32_t length,
                   int ctx_no)
{
    u_int32_t dst_bottom = ((u_int32_t)dst_pa & 0xFF);
    u_int32_t reg_base;
    struct bw_ctx *ctx;
    unsigned long flags;

    if ((ctx_no < 0) || (ctx_no >= BURST_CONTEXT_COUNT)) {
        return BURST_WRITE_ERR_CTX_NO;
    }
    if (length & 3) { /* length must be multiple of 4 */
        return BURST_WRITE_ERR_SIZE;
    }
    if (length > BURST_CONTEXT_SIZE) { /* length too large */
        return BURST_WRITE_ERR_SIZE;
    }
    if (dst_bottom & 0x7) { /* bogus not 8 byte aligned */
        return BURST_WRITE_ERR_ALIGN;
    }
    if ((dst_bottom + length) > 256) { /* can't pass 256 byte boundary */
        return BURST_WRITE_ERR_ALIGN_XFER;
    }

    flags  = _intr_disable();
    ctx = &bw_ctxs[ctx_no];
    /*
     * setup context burst data registers
     */
    writereg64(ctx->addr, local_to_hwecpu(dst_pa));

    /* control is all zero */
    writereg(ctx->control, 0);

    writereg(ctx->size, length);

    //
    // move caller's data into burst data registers
    //
    reg_base = (u_int32_t)PROC_MREGS_BURST_DATA + ctx->offset;
    writeregs(reg_base, src_va, length / sizeof (int));

    memory_barrier();

    /* launch burst accelerator */
    writereg(ctx->launch + ctx->offset, 0);
    _intr_enable(flags);

    return 0;
}
#ifdef CONFIG_OS_TYPE_LINUXKERNEL
EXPORT_SYMBOL(burst_write_buffer);
#endif

int
burst_writeregs(u_int32_t dst_pa, void *src_va, u_int32_t ndw)
{
#if defined(CONFIG_BASE_PALO)
    return -1;
#endif
    int r;
    unsigned long flags;


    if (ndw == 1) {
        writereg(dst_pa, *(u_int32_t *)src_va);
        return 0;
    }

    flags  = _intr_disable();

    r = burst_write_buffer(src_va,
                              (u_int64_t)dst_pa,
                              ndw * sizeof (u_int32_t),
                              BURST_CONTEXT_DEFAULT);

    _intr_enable(flags);

    return (r);
}

int
burst_writereg64(u_int32_t dst_pa, u_int64_t v64)
{
    u_int32_t v[2];

    v[0] = v64 & 0xFFFFFFFF;
    v[1] = (v64 >> 32) & 0xFFFFFFFF;
    return  burst_writeregs(dst_pa, v, 2);
}

#endif /* defined(CONFIG_BASE_SERENO) || defined(CONFIG_BASE_CRUZ) */
