/*
 * Copyright (c) 2009 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: paloecpu.h 59 2017-01-21 21:37:10Z buildsa $"

#ifndef _PALOECPU_H_
#define _PALOECPU_H_

/*
 * IOCTLs
 */
#define PALOECPU_IOC            'E'
#define PALOECPU_IOC_RUN        _IO(PALOECPU_IOC, 0)
#define PALOECPU_IOC_STOP       _IO(PALOECPU_IOC, 1)
#define PALOECPU_IOC_NMI        _IO(PALOECPU_IOC, 2)
#define PALOECPU_IOC_GETDBPAGE  _IOR(PALOECPU_IOC, 3, u_int32_t)

/************************* Virtual Uart **************************************/

#ifndef __ASSEMBLY__

/*
 * Accesses to the virtual uart must be uncached.
 */
#define PALOECPU_VCONS_BUFSZ   64
typedef struct {
    u_int8_t    buf[PALOECPU_VCONS_BUFSZ];  // data buffer
    u_int32_t   head;                   // writer index
    u_int32_t   tail;                   // reader index
} paloecpu_vcons_fifo_t;

typedef struct {
    int connected;                      // set by mcpu. non-zero if connected
    paloecpu_vcons_fifo_t rx;           // mcpu->ecpu fifo
    paloecpu_vcons_fifo_t tx;           // ecpu->mcpu fifo
} paloecpu_vcons_t;

#ifndef memory_barrier
#define memory_barrier() do { asm volatile("sync" ::: "memory"); } while (0)
#endif

static inline int
paloecpu_vcons_fifo_isempty(paloecpu_vcons_fifo_t *p)
{
    return (p->head == p->tail);
}

static inline int
paloecpu_vcons_fifo_isfull(paloecpu_vcons_fifo_t *p)
{
    return (((p->head + 1) % PALOECPU_VCONS_BUFSZ) == p->tail);
}

static inline int
paloecpu_vcons_fifo_usedcount(paloecpu_vcons_fifo_t *p)
{
    if (p->head >= p->tail) {
        return p->head - p->tail;
    } else {
        return p->head + PALOECPU_VCONS_BUFSZ - p->tail;
    }
}

static inline int
paloecpu_vcons_fifo_freecount(paloecpu_vcons_fifo_t *p)
{
    return (PALOECPU_VCONS_BUFSZ - paloecpu_vcons_fifo_usedcount(p) - 1);
}

static inline int
paloecpu_vcons_fifo_write(paloecpu_vcons_fifo_t *p, const char *buf, int len)
{
    int n = paloecpu_vcons_fifo_freecount(p);
    int i, head;

    if (n > len) {
        n = len;
    }
    head = p->head;
    for (i = 0; i < n; i++) {
        p->buf[head] = *buf++;
        head = (head + 1) % PALOECPU_VCONS_BUFSZ;
    }
    memory_barrier();
    p->head = head;
    return n;
}

static inline int
paloecpu_vcons_fifo_read(paloecpu_vcons_fifo_t *p, char *buf, int len)
{
    int n = paloecpu_vcons_fifo_usedcount(p);
    int i, tail;

    if (n > len) {
        n = len;
    }
    tail = p->tail;
    for (i = 0; i < n; i++) {
        *buf++ = p->buf[tail];
        tail = (tail + 1) % PALOECPU_VCONS_BUFSZ;
    }
    memory_barrier();
    p->tail = tail;
    return n;
}

#endif /* __ASSEMBLY__ */

/**************************** Debugger ***************************************/

/*
 * The ectx block provides the KSEG1 address of a paloecpu_dbpage_t.  This
 * structure holds the infomation for the software EJTAG debugger.
 */

#define PALOECPU_DBPAGESIZE     4096

/*
 *            sp -> +-------------------+
 *                  |                   |
 *                  |    ejtag stack    |
 *                  |                   |
 *                  +-------------------+
 *                  |///////////////////|
 *                  +-------------------+
 *                  |                   |
 *                  | paloecpu_dbpage_t |
 *                  |                   |
 *   ectx.dbpage -> +-------------------+
 */

#ifndef __ASSEMBLY__

/*
 * The debug page is the software interface between the debugger running on
 * the ecpu and the application running on mcpu.
 */
typedef struct {
    u_int32_t   gpr[32];                // +0x00 - 0x7c
    u_int32_t   lo;                     // +0x80
    u_int32_t   hi;                     // +0x84
    u_int32_t   status;                 // +0x88
    u_int32_t   cause;                  // +0x8c
    u_int32_t   epc;                    // +0x90
    u_int32_t   badvaddr;               // +0x94
    u_int32_t   pc;                     // +0x98
    u_int32_t   debug;                  // +0x9c
} paloecpu_xcptregs_t; 

#endif

#define PALOECPU_XCPTREGS_SIZE  0xa0    // Must be sizeof (paloecpu_xcptregs_t)

#ifndef __ASSEMBLY__

/*
 * Statistics on debugger activity.
 */
typedef struct {
    u_int32_t   entry_count;            // entry into the handler count
    u_int32_t   dint_count;             // entry due to DINT count
    u_int32_t   disarmed_count;         // entry when disarmed count
    u_int32_t   reentry_count;          // re-entry into the handler count
} paloecpu_dbpage_stats_t;

/*
 * Debug info structure.
 */
typedef struct {
    int         stopped;                // program stopped
    int         armed;                  // non-zero if debugger armed
    u_int32_t   cmd_stat;               // command/status
    u_int32_t   argc;                   // arg count
    u_int32_t   arg[4];                 // args
    u_int32_t   data[256];              // command/response data (1KB)
    u_int32_t   cmdseqno;               // command sequence number
    u_int32_t   cmdackno;               // command ack number
    paloecpu_xcptregs_t regs;           // registers
    paloecpu_dbpage_stats_t stats;      // statistics
} paloecpu_dbpage_t;

enum {
    PALOECPU_DBG_CMD_NONE,
    PALOECPU_DBG_CMD_GO,
    PALOECPU_DBG_CMD_STEP,
    PALOECPU_DBG_CMD_READ_MEM,          // addr, size(1,2,4), count
    PALOECPU_DBG_CMD_WRITE_MEM,         // addr, size(1,2,4), count
    PALOECPU_DBG_CMD_READ_CP0,          // reg, sel
    PALOECPU_DBG_CMD_WRITE_CP0,         // reg, sel, val
    PALOECPU_DBG_CMD_CACHE_FLUSH,       // PALOECPU_DBG_CACHE_x
    PALOECPU_DBG_CMD_WRITE_INSTR,       // addr opcode
    PALOECPU_DBG_CMD_READ_TLB,          // index
    PALOECPU_DBG_CMD_WRITE_TLB,         // index
    PALOECPU_DBG_CMD_READ_DCACHET,      // index
    PALOECPU_DBG_CMD_READ_ICACHET,      // index
    PALOECPU_DBG_CMD_READ_DCACHE,       // offset
    PALOECPU_DBG_CMD_READ_ICACHE,       // offset
    PALOECPU_DBG_CMD_NUM
};

enum {
    PALOECPU_DBG_CACHE_D,
    PALOECPU_DBG_CACHE_I,
};

#define PALOECPU_DBG_STAT_LINE(x)        ((x) >> 16)
#define PALOECPU_DBG_STAT_CODE(x)        ((x) & 0xffff)
#define PALOECPU_DBG_STAT_MK(x, y)       (((x) << 16) | (y))

enum {
    PALOECPU_DBG_STAT_OK,
    PALOECPU_DBG_STAT_ARGCNT,           // bad argument count
    PALOECPU_DBG_STAT_INVAL,            // invalid argument
    PALOECPU_DBG_STAT_EXCEPTION,        // data[0] is exception code
    PALOECPU_DBG_STAT_CMDUNKNOWN,       // invalid command
};

#endif /* __ASSEMBLY__ */

#endif /* _PALOECPU_H_ */
