/*
 * Copyright (c) 2007 Nuova Systems, Inc.  All rights reserved.
 */
#ident "$Id: ppcie_dev.h 43 2016-12-03 22:59:19Z buildsa $"

#ifndef _PALOPCIE_H_
#define _PALOPCIE_H_

#ifdef __KERNEL__
#include <palo/include/palo_defs.h>
#else // __KERNEL__
#include "palo_defs.h"
#include "queue.h"
#endif // __KERNEL__

#define PPCIE_NBARS  6
#define NORES        0xffff

enum ppcie_event_type {
	PPCIE_EVT_NONE,
	PPCIE_EVT_DEVCMD,		/* incoming devcmd */
	PPCIE_EVT_PCIE_MSG,		/* incoming pcie message */
	PPCIE_EVT_BUS_CHG,		/* switch secondary bus change */
	PPCIE_EVT_DEV_RESET,		/* host-initiated device reset */
	PPCIE_EVT_HOSTUP,		/* host is up */
	PPCIE_EVT_HOSTDN,		/* host is down */
	PPCIE_EVT_LINKERR,		/* link error detected */

	PPCIE_EVT_MAX
};

#define PPCIE_EVT_DEV_RESET_BUS 1
#define PPCIE_EVT_DEV_RESET_FLR 2

struct ppcie_event_dev_reset {
	u_int32_t reset_type;
};

struct ppcie_event_pcie_msg {
	u_int64_t msg_data;
};

struct ppcie_event_bus_chg {
	u_int8_t busno;
};

struct ppcie_event_devcmd {
	u_int32_t status;
	u_int32_t cmd;
	u_int64_t args[15];
};

/* ppcie_event_linkerr.errtype */
enum {
	PPCIE_ERRTYPE_NONE,		/* no error */
	PPCIE_ERRTYPE_SOFT,		/* soft, correctable error */
	PPCIE_ERRTYPE_HARD,		/* hard, uncorrectable error */
};

/* ppcie_event_linkerr.errcode */
enum {
	PPCIE_ERRCODE_NONE,		/* no error */
	PPCIE_ERRCODE_LTSSM_THRESHOLD,	/* ltssm state changed > threshold */
	PPCIE_ERRCODE_FSM_FAULT,	/* link fsm error */
	PPCIE_ERRCODE_PLL_LOCK,		/* global pll lock failed at startup */
	PPCIE_ERRCODE_PPS_PLL,		/* pps pll locked failed at init */
};

struct ppcie_event_linkerr {
	u_int8_t errtype;
	u_int8_t errcode;
};

struct ppcie_event_host_iter {
	u_int32_t host_iter;
};

struct ppcie_event {
	enum ppcie_event_type type;
	int id;
	u_int16_t bdf;
	u_int16_t len;
	union {
		struct ppcie_event_devcmd devcmd;
		struct ppcie_event_pcie_msg msg;
		struct ppcie_event_dev_reset reset;
		struct ppcie_event_bus_chg bus_chg;
		struct ppcie_event_linkerr linkerr;
		struct ppcie_event_host_iter host_iter;
	} u;
	TAILQ_ENTRY(ppcie_event) list;
};

#define PPCIE_GET_EVENTS_MAX    16

#ifdef	XXX
struct ppcie_stats {
	u_int64_t bar_miss;
	u_int64_t bar_missfail;
	u_int64_t bars_evicted;
	u_int64_t bars_loaded;
	u_int64_t bars_loadov;
	u_int64_t bars_reset;
	u_int64_t ioctl;
	struct host_stats {
		u_int64_t cfg_rd;
		u_int64_t cfg_wr;
		u_int64_t cfg_unaligned_rd;
		u_int64_t cfg_unaligned_wr;
		u_int64_t cmt_miss;
		u_int64_t bar_rd;
		u_int64_t bar_wr;
		u_int64_t bmt_miss;
		u_int64_t bmt_miss_retries;
		u_int64_t bmt_oor;
		u_int64_t pmv;
		u_int64_t unhandled_cause;
		u_int64_t unexpected_reqidx;
	} host_stats[PALO_MAX_HOSTS];
};
#endif

struct ppcie_hostinfo {
	u_int32_t nhids;
	u_int32_t nhosts;
	struct ppcie_hidmap {
		u_int8_t hid;
		u_int8_t width;
	} hidmap[PALO_MAX_HIDS];
	struct ppcie_hostmap {
		u_int8_t hid;
		u_int8_t shid;
	} hostmap[PALO_MAX_HOSTS];
};

struct ppcie_host_ready {
	u_int32_t hostid;
	u_int32_t hostdn_iter;
};

/*
 * PPCIE_IOC_LOAD
 *     Load the device in ppcie_dev into ppcie hardware.
 *
 *     struct ppcie_dev ppcie_dev;
 *     ioctl(fd, PPCIE_IOC_LOAD, &ppcie_dev);
 *
 * PPCIE_IOC_GET
 *     Get the device with bus/dev/fnc, return entire ppcie_dev in buffer.
 *
 *     struct ppcie_dev ppcie_dev;
 *     ppcie_dev.bdf = bdf_of_interest;
 *     ioctl(fd, PPCIE_IOC_GET, &ppcie_dev);
 *
 * PPCIE_IOC_UNLOAD
 *     Unload the device with pci bus/dev/fnc from ppcie hardware,
 *     freeing all resources used by that device.
 *
 *     int bdf;
 *     ioctl(fd, PPCIE_IOC_UNLOAD, bdf);
 *
 * PPCIE_IOC_READY
 *     Turn off pcie config retries and allow the host to probe pcie bus.
 *
 *     struct ppcie_host_ready host_ready;
 *     ioctl(fd, PPCIE_IOC_READY, &host_ready);
 *
 * PPCIE_IOC_GET_EVENTS
 *     Get pending events.  Returns 0 if no events pending,
 *     n if n events returned, or <0 on error.
 *
 *     struct ppcie_event ev[PPCIE_GET_EVENTS_MAX];
 *     nevents = ioctl(fd, PPCIE_IOC_GET_EVENTS, ev);
 *
 * PPCIE_IOC_GET_STATS
 *     Get current stats.
 *
 *     struct ppcie_stats ppcie_stats;
 *     ioctl(fd, PPCIE_IOC_GET_STATS, &ppcie_stats);
 *
 * PPCIE_IOC_GET_HOSTINFO
 *     Get platform host info.
 *
 *     struct ppcie_hostinfo ppcie_hostinfo;
 *     ioctl(fd, PPCIE_IOC_GET_HOSTINFO, &ppcie_hostinfo);
 */

#define PPCIE_IOC               'P'
#define PPCIE_IOC_LOAD          _IOR(PPCIE_IOC, 0, struct ppcie_dev *)
// #define PPCIE_IOC_GET           _IOWR(PPCIE_IOC, 1, struct ppcie_dev *)
#define PPCIE_IOC_UNLOAD        _IOW(PPCIE_IOC, 2, int)
#define PPCIE_IOC_READY         _IOW(PPCIE_IOC, 3, struct ppcie_host_ready *)
#define PPCIE_IOC_GET_EVENTS	_IOW(PPCIE_IOC, 4, struct ppcie_event *)
// #define PPCIE_IOC_GET_STATS	_IOW(PPCIE_IOC, 5, struct ppcie_stats *)
#define PPCIE_IOC_GET_HOSTINFO  _IOW(PPCIE_IOC, 6, struct ppcie_hostinfo *)
#define PPCIE_IOC_ACTIVE_STANDBY _IOWR(PPCIE_IOC, 7, struct ppcie_dev *)
// #define PPCIE_IOC_BDFT_LOAD   _IOR(PPCIE_IOC, 7, struct ppcie_bdft *)
#define PPCIE_IOC_CHIP_RESET    _IO(PPCIE_IOC, 8)
#define PPCIE_IOC_CRT_RELOAD    _IOW(PPCIE_IOC, 9, int)
#define PPCIE_IOC_DELETE_DEVCMDS _IOW(PPCIE_IOC, 10, struct ppcie_dev *)
#define PPCIE_IOC_VNIC_STATE_CFG _IOW(PPCIE_IOC, 11, struct vnic_st_info *)

#ifndef __KERNEL__
const char *ppcie_event_name(const enum ppcie_event_type e);
int ppcie_event_encode(const char *s);
#endif /* __KENREL__ */


#endif /* _PALOPCIE_H_ */
