
/*
 * Copyright 2012 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id: trace.h 6 2016-12-02 03:26:55Z buildsa $"

#ifndef _TRACE_H_
#define _TRACE_H_

/*
 * Tracing.
 * The trace function copies the fmt pointer and TRACE_NWORDS words of
 * arguments to a trace buffer entry - so it does not incur much of a time
 * penalty, but it does come with caveats:
 *      - Do not construct a call with more than TRACE_NWORDS 32-bit words
 *        worth of arguments.  For example, if TRACE_NWORDS is 6 then the
 *        following might not work:
 *          trace(0, "Event %x %x %llx %llx", a, b, c, d);
 *        Disagree?  While it is true that the above call uses 6 words worth
 *        of arguments, MIPS alignment rules require that the 64-bit arguments
 *        c and d be push dword-aligned on the stack.  This might require one
 *        additional pad word between b and c, and thus require 7 words.
 *        Keep in mind that every 64-bit argument that is not preceeded by
 *        another 64-bit argument may require one extra pad word.
 *      - This will NOT be caught at either compile-time or run-time.  The
 *        tail arguments will be lost and if one is a "%s", it will likely
 *        crash when the trace buffer is dumped.
 *      - Do not reference a stack-based string buffer.  When the trace is
 *        dumped the original stack-based string will be long gone.
 */

/*
 * Define TRACE_NWORDS as 8.  This makes for a 48-byte entry and allows for the
 * format string: "%x %llx %x %llx" (8 word worst-case alignment).
 */
#define TRACE_NWORDS        8           // for a 48 byte entry size

typedef struct trace_entry {
    u_int64_t   timestamp;              // mips timestamp
    unsigned int genid;                 // matches ring genid, for verification
    u_int8_t    ev;                     // event id (0..31)
    u_int8_t    align;                  // stack alignment for original va_list
    const char  *fmt;                   // format string
    u_int32_t   word[TRACE_NWORDS];     // format arguments (va_list)
} trace_entry_t;

typedef struct trace_log {
    char name[32];                      // public name of the trace buffer
    trace_entry_t *buf;                 // trace buffer data
    int         nentries;               // capacity of the buffer in entries
    int         head;                   // head (write) index
    int         tail;                   // tail (read) index
    unsigned int genid;                 // ring generation id to detect wraps
    u_int32_t   evmask;                 // bitmap of tracable events
} trace_log_t;

#ifdef __KERNEL__

void _trace(trace_log_t *log, int ev, const char *fmt, ...)
        __attribute__((format(printf, 3, 4)));

/*
 * trace API
 *
 * trace(trace_log_t log, int ev, const char *fmt, ...)
 * If event ev has been enabled for tracing then a trace record is written.
 *
 * trace_if(expr, log, ev, const char *fmt, ...)
 * If the expression is true, and the event ev is enabled for tracing then
 * generate a trace record.
 */
#define trace(log, ev, fmt, ...) \
    do { \
        if ((1 << (ev)) & (log)->evmask) { \
            _trace((log), (ev), (fmt), ##__VA_ARGS__); \
        } \
    } while (0)

#define trace_if(expr, log, ev, fmt, ...) \
    do { \
        if ((expr) && ((1 << (ev)) & (log)->evmask)) { \
            _trace((log), (ev), (fmt), ##__VA_ARGS__); \
        } \
    } while (0)

/*
 * Set and get the event mask.
 */
#define trace_set_evmask(log, m) \
    do { \
        (log)->evmask = m; \
    } while (0)

#define trace_get_evmask(log)      ((log)->evmask)

extern void trace_init(trace_log_t *log, const char *name, int nentries);
extern void trace_free(trace_log_t *log);

#endif // __KERNEL__
#endif // _TRACE_H_
