/*
 * Copyright 2008-2010 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id$"

#ifndef _PALO_PALOMEM_H_
#define _PALO_PALOMEM_H_

/*
 * Layout of physical memory used by various components
 * of the ecpu firmware and associated applications.
 *
 * 0x08000000 +-----------------+ 128 MB
 *            |                 |
 *            |      ecpu2      |
 *            |                 |
 * 0x06300000 +-----------------+
 *            |                 |
 *            |      ecpu1      |
 *            |                 |
 * 0x04600000 +-----------------+
 *            |      extbl      |
 * 0x04000000 +-----------------+
 *            |    rq cache     |
 * 0x03f80000 +-----------------+
 *            |    wq cache     |
 * 0x03f00000 +-----------------+
 *            |      diag       |
 * 0x03eff000 +-----------------+
 *            |      hwio       |
 *            |                 |
 * 0x03c00000 +-----------------+
 *            |     memtun      |
 * 0x03bf0000 +-----------------+
 *            |                 |
 *            |                 |
 *            |                 |
 *            |      Linux      |
 *            |                 |
 *            |                 |
 *            |                 |
 * 0x00001000 +-----------------+
 *            |      ectx       |
 * 0x00000000 +-----------------+
 */

/*
 * Define the top of the memory map.
 * 128MB for all Palo cards.
 */
#define PALOMEM_MEMTOP          0x08000000

/*
 * Ecpu boot vectors and arguments.
 * See ecpuctx.h for details.
 */
#define PALOMEM_ECTX_NAME       "ectx"
#define PALOMEM_ECTX_BASE       0x00000000
#define PALOMEM_ECTX_SIZE       0x00001000

/*
 * Linux memory region for Linux kernel and user processes.
 * Linux RAMTOP points to the top of this region, any access
 * to memory from Linux above RAMTOP is uncached access.
 * See mcpu/firmware/linux/arch/mips/nuova/palo/palo_mem.c
 * for details.
 */
#define PALOMEM_LINUX_NAME      "linux"
#define PALOMEM_LINUX_BASE      0x00001000
#define PALOMEM_LINUX_SIZE      0x03bef000

/*
 * The memtun region contains the transmit/receive rings for
 * exchanging packets with the host through the memtun facility.
 * See sa/src/palo/fw/mcpu/memtun for details.
 */
#define PALOMEM_MEMTUN_NAME     "memtun"
#define PALOMEM_MEMTUN_BASE     0x03bf0000
#define PALOMEM_MEMTUN_SIZE     0x00010000

/*
 * The hwio region is reserved for ecpu physical memory access.
 * See sa/src/common/libhwio/include/hwio_mipslinux_impl.h for details.
 */
#define PALOMEM_HWIO_NAME       "hwio"
#define PALOMEM_HWIO_BASE       0x03c00000
#define PALOMEM_HWIO_SIZE       0x00200000

#define PALOMEM_PMPCIE_NAME     "pmpcie"
#define PALOMEM_PMPCIE_BASE     0x03e00000
#define PALOMEM_PMPCIE_SIZE     0x000ff000

/*
 * Diags uses this page to pass data from the host diags app to diags fw.
 * Only used by the diags firmware image.
 */
#define PALOMEM_DIAG_NAME       "diag"
#define PALOMEM_DIAG_BASE       0x03eff000
#define PALOMEM_DIAG_SIZE       0x00001000

/*
 * Palo local memory WQ descriptor cache.
 * Used by sa/src/palo/fw/mcpu/mcp/src/hwinit.c.
 * See Palo TOO for details.
 */
#define PALOMEM_WQCACHE_NAME    "wqcache"
#define PALOMEM_WQCACHE_BASE    0x03f00000
#define PALOMEM_WQCACHE_SIZE    0x00080000

/*
 * Palo local memory RQ descriptor cache.
 * Used by sa/src/palo/fw/mcpu/mcp/src/hwinit.c.
 * See Palo TOO for details.
 */
#define PALOMEM_RQCACHE_NAME    "rqcache"
#define PALOMEM_RQCACHE_BASE    0x03f80000
#define PALOMEM_RQCACHE_SIZE    0x00080000

/*
 * Region allocated for ecpu code/data/exchange table.
 * We give 64 MB total, extbl gets the first 6 MB, then
 * 29 MB for each ecpu code/data.
 * See palo/fw/build for details.
 */
#define PALOMEM_EXTBL_NAME      "extbl"
#define PALOMEM_EXTBL_BASE      0x04000000
#define PALOMEM_EXTBL_SIZE      0x00600000
/*
 * Part of the exchange table is used for cache in Palo but not in Sereno.
 * Need to know the amount that is cache in the palo_dfw.lds file.
 */
#define PALOMEM_EXTBL_CA_SIZE   0x00400000

/*
 * Exchange table param/stats size is 1MB each in Palo.
 */
#define PALOMEM_EXTBL_PM_SIZE   0x00100000
#define PALOMEM_EXTBL_SS_SIZE   0x00100000

#define PALOMEM_ECPU1_NAME      "ecpu1"
#define PALOMEM_ECPU1_BASE      0x04600000
#define PALOMEM_ECPU1_SIZE      0x01d00000

#define PALOMEM_ECPU2_NAME      "ecpu2"
#define PALOMEM_ECPU2_BASE      0x06300000
#define PALOMEM_ECPU2_SIZE      0x01d00000

#endif /* _PALO_PALOMEM_H_ */
