/*
 * Copyright 2007 Nuova Systems, Inc. All Rights Reserved
 */
#ident "$Id: pcie_tlp.h 6 2016-12-02 03:26:55Z buildsa $"

#ifndef _PCIE_TLP_H_
#define _PCIE_TLP_H_

//
// PCIe Transaction Layer Protocol, based on [PCIe Spec]:
//     PCI Express Base Specification
//     Revision 1.1
//     March 28, 2005
//

//
// Palo always separates header and data for us, and payload data
// always begins at byte 16 (dw 4).
//
//     DW0 - PCIe Request Header DW0
//     DW1 - PCIe Request Header DW1
//     DW2 - PCIe Request Header DW2
//     DW3 - PCIe Request Header DW3
//     DW4 - PCIe Request Data DW0
//     DW5 - PCIe Request Data DW1
//     DW6 - PCIe Request Data DW2
//     DW7 - PCIe Request Data DW3
//
#define PCIE_TLP_HEADER_DW      4       // up to 4 DW for header
#define PCIE_TLP_PAYLOAD_DW     4       // up to 4 DW for ind_req payload
#define PCIE_TLP_DW            (PCIE_TLP_HEADER_DW + PCIE_TLP_PAYLOAD_DW)

struct pcie_tlp {
    u_int8_t data[PCIE_TLP_DW * sizeof(u_int32_t)];
};

//
// [PCIe Spec]: Table 2-2
//
enum pcie_tlp_fmt {
    PCIE_TLP_FMT_3DW  = 0x0,
    PCIE_TLP_FMT_4DW  = 0x1,
    PCIE_TLP_FMT_3DWD = 0x2,
    PCIE_TLP_FMT_4DWD = 0x3,
};

#define _tlp_type(fmt, type)    ((PCIE_TLP_FMT_##fmt << 5) | ((type) & 0x1f))

//
// [PCIe Spec]: Table 2-3
//
enum pcie_tlp_type {
    PCIE_TLP_TYPE_MEMRD      = _tlp_type(3DW,  0x0),
    PCIE_TLP_TYPE_MEMRD64    = _tlp_type(4DW,  0x0),
    PCIE_TLP_TYPE_MEMRD_LK   = _tlp_type(3DW,  0x1),
    PCIE_TLP_TYPE_MEMRD64_LK = _tlp_type(4DW,  0x1),
    PCIE_TLP_TYPE_MEMWR      = _tlp_type(3DWD, 0x0),
    PCIE_TLP_TYPE_MEMWR64    = _tlp_type(4DWD, 0x0),
    PCIE_TLP_TYPE_IORD       = _tlp_type(3DW,  0x2),
    PCIE_TLP_TYPE_IOWR       = _tlp_type(3DWD, 0x2),
    PCIE_TLP_TYPE_CFGRD0     = _tlp_type(3DW,  0x4),
    PCIE_TLP_TYPE_CFGWR0     = _tlp_type(3DWD, 0x4),
    PCIE_TLP_TYPE_CFGRD1     = _tlp_type(3DW,  0x5),
    PCIE_TLP_TYPE_CFGWR1     = _tlp_type(3DWD, 0x5),
    PCIE_TLP_TYPE_CPL        = _tlp_type(3DW,  0xa),
    PCIE_TLP_TYPE_CPLD       = _tlp_type(3DWD, 0xa),
    PCIE_TLP_TYPE_CPL_LK     = _tlp_type(3DW,  0xb),
    PCIE_TLP_TYPE_CPLD_LK    = _tlp_type(3DWD, 0xb),
};


static inline int
bitcount(u_int32_t n)
{
    int count = 0;

    while (n) {
        count++;
        n &= ~(-n);     // clear low order 1 bit
    }
    return count;
}

static inline void
pcie_tlp_hdr_enc(struct pcie_tlp *tlp, u_int8_t type,
                 u_int64_t addr, u_int16_t len)
{
    u_int16_t len_dw;
    u_int8_t bef, bel, bemask;

    //assert(len <= 8);

    bemask = (1 << len) - 1;
    bemask <<= (addr & 0x3);

    bef = bemask & 0xf;
    bel = (bemask >> 4) & 0xf;

    len_dw = (bef > 0) + (bel > 0);

    if (len_dw == 0) {
        // 0-length allowed, use len==1 DW, BE==0
        len_dw = 1;
    } else if (len_dw == 0x400) {
        len_dw = 0;
    }

    tlp->data[0] = type;
    tlp->data[2] = (len_dw >> 8) & 0x3;
    tlp->data[3] = len_dw & 0xff;
    tlp->data[7] = (bel << 4) | bef;
}

static inline void
pcie_tlp_hdr_dec(struct pcie_tlp *tlp, u_int8_t *typep,
                 u_int8_t *addroffp, u_int16_t *lenp)
{
    u_int16_t len_dw;
    u_int8_t bemask, be_dw;

    if (typep) {
        *typep = tlp->data[0];
    }

    len_dw = ((((u_int16_t)tlp->data[2] && 0x3) << 8) | tlp->data[3]);
    if (len_dw == 0) {
        len_dw = 0x400;
    }

    bemask = tlp->data[7];  // both bel, bef

    be_dw = ((bemask & 0xf) != 0) + ((bemask & 0xf0) != 0);

    if (addroffp) {
        u_int8_t fbit = ffs(bemask);
        *addroffp = fbit ? fbit - 1 : 0;
    }

    if (lenp) {
        u_int8_t becount = bitcount(bemask);
        *lenp = bemask ? ((len_dw - be_dw) << 2) + becount : 0;
    }
}


static inline void
pcie_tlp_type_dec(struct pcie_tlp *tlp, u_int8_t *type)
{
    if (type) {
        *type = tlp->data[0];
    }
}

static inline u_int64_t
_val2bl(u_int64_t addr, u_int16_t len, u_int64_t val)
{
    switch (len) {
    case 0: val = 0; break;
    case 1: val &= 0x000000ff; break;
    case 2: val &= 0x0000ffff; break;
    case 4: val &= 0xffffffff; break;
    default: break;
    }
    // shift data val over to byte lanes based on addr offset
    return (val << ((addr & 0x3) * 8));
}

//
// Addresses are encoded in big-endian order.
//
static inline void
_addr32_enc(u_int8_t *dst, u_int32_t addr)
{
    dst[0] = (addr >> 24) & 0xff;
    dst[1] = (addr >> 16) & 0xff;
    dst[2] = (addr >>  8) & 0xff;
    dst[3] = (addr >>  0) & 0xff;
}
static inline void
_addr64_enc(u_int8_t *dst, u_int64_t addr)
{
    _addr32_enc(dst,   addr >> 32);
    _addr32_enc(dst+4, addr);
}

static inline void
_addr32_dec(u_int8_t *src, u_int32_t *addr)
{
    *addr = (((u_int32_t)src[0] << 24) |
             ((u_int32_t)src[1] << 16) |
             ((u_int32_t)src[2] <<  8) |
             ((u_int32_t)src[3] <<  0));
}
static inline void
_addr64_dec(u_int8_t *src, u_int64_t *addr)
{
    u_int32_t addrlo, addrhi;
    _addr32_dec(src  , &addrhi);
    _addr32_dec(src+4, &addrlo);
    *addr = (u_int64_t)addrhi << 32 | addrlo;
}

//
// Data/payload values are encoded in little-endian order.
//
static inline void
_val32_enc(u_int8_t *dst, u_int32_t val)
{
    dst[0] = (val >>  0) & 0xff;
    dst[1] = (val >>  8) & 0xff;
    dst[2] = (val >> 16) & 0xff;
    dst[3] = (val >> 24) & 0xff;
}
static inline void
_val64_enc(u_int8_t *dst, u_int64_t val)
{
    _val32_enc(dst,   val);
    _val32_enc(dst+4, val >> 32);
}

static inline void
_val32_dec(u_int8_t *src, u_int32_t *val)
{
    *val = (((u_int32_t)src[0] <<  0) |
            ((u_int32_t)src[1] <<  8) |
            ((u_int32_t)src[2] << 16) |
            ((u_int32_t)src[3] << 24));
}
static inline void
_val64_dec(u_int8_t *src, u_int64_t *val)
{
    u_int32_t vallo, valhi;
    _val32_dec(src  , &vallo);
    _val32_dec(src+4, &valhi);
    *val = (u_int64_t)valhi << 32 | vallo;
}


static inline void
pcie_tlp_mem32_enc(struct pcie_tlp *tlp, u_int8_t is_wr,
                   u_int64_t addr, u_int16_t len, u_int64_t val)
{
    u_int8_t type = is_wr ? PCIE_TLP_TYPE_MEMWR : PCIE_TLP_TYPE_MEMRD;

    //assert(len <= 8);

    memset(tlp, 0, sizeof (struct pcie_tlp));

    pcie_tlp_hdr_enc(tlp, type, addr, len);
    _addr32_enc(&tlp->data[8], addr & ~0x3); // DW aligned addr

    if (is_wr && len) {
        val = _val2bl(addr, len, val);
        _val64_enc(&tlp->data[16], val);
    }
}

static inline void
pcie_tlp_mem32_dec(struct pcie_tlp *tlp, u_int8_t *is_wrp,
                   u_int64_t *addrp, u_int16_t *lenp, u_int64_t *valp)
{
    u_int8_t is_wr, type, addroff;
    u_int16_t len;

    pcie_tlp_hdr_dec(tlp, &type, &addroff, &len);

    is_wr = (type == PCIE_TLP_TYPE_MEMWR);

    if (is_wrp) {
        *is_wrp = is_wr;
    }
    if (addrp) {
        u_int32_t addr;
        _addr32_dec(&tlp->data[8], &addr);
        *addrp = addr + addroff;
    }
    if (lenp) {
        *lenp = len;
    }
    if (valp) {
        if (is_wr) {
            u_int64_t val;
            _val64_dec(&tlp->data[16], &val);
            val >>= addroff * 8;
            *valp = val;
        } else {
            *valp = 0;
        }
    }
}

static inline void
pcie_tlp_mem64_enc(struct pcie_tlp *tlp, u_int8_t is_wr,
                   u_int64_t addr, u_int16_t len, u_int64_t val)
{
    u_int8_t type = is_wr ? PCIE_TLP_TYPE_MEMWR64 : PCIE_TLP_TYPE_MEMRD64;

    //assert(len <= 8);

    memset(tlp, 0, sizeof (struct pcie_tlp));

    pcie_tlp_hdr_enc(tlp, type, addr, len);
    _addr64_enc(&tlp->data[8], addr & ~0x3); // DW aligned addr

    if (is_wr && len) {
        val = _val2bl(addr, len, val);
        _val64_enc(&tlp->data[16], val);
    }
}

static inline void
pcie_tlp_mem64_dec(struct pcie_tlp *tlp, u_int8_t *is_wrp,
                   u_int64_t *addrp, u_int16_t *lenp, u_int64_t *valp)
{
    u_int8_t is_wr, type, addroff;
    u_int16_t len;

    pcie_tlp_hdr_dec(tlp, &type, &addroff, &len);

    is_wr = (type == PCIE_TLP_TYPE_MEMWR64);

    if (is_wrp) {
        *is_wrp = is_wr;
    }
    if (addrp) {
        u_int64_t addr;
        _addr64_dec(&tlp->data[8], &addr);
        *addrp = addr + addroff;
    }
    if (lenp) {
        *lenp = len;
    }
    if (valp) {
        if (is_wr) {
            u_int64_t val;
            _val64_dec(&tlp->data[16], &val);
            val >>= addroff * 8;
            *valp = val;
        } else {
            *valp = 0;
        }
    }
}

static inline void
pcie_tlp_mem_enc(struct pcie_tlp *tlp, u_int8_t is_wr,
                 u_int64_t addr, u_int16_t len, u_int64_t val)
{
    if (addr > 0xffffffff) {
        pcie_tlp_mem64_enc(tlp, is_wr, addr, len, val);
    } else {
        pcie_tlp_mem32_enc(tlp, is_wr, addr, len, val);
    }
}

static inline void
pcie_tlp_mem_dec(struct pcie_tlp *tlp, u_int8_t *is_wrp,
                 u_int64_t *addrp, u_int16_t *lenp, u_int64_t *valp)
{
    u_int8_t type;

    pcie_tlp_type_dec(tlp, &type);

    switch (type) {
    case PCIE_TLP_TYPE_MEMRD64:
    case PCIE_TLP_TYPE_MEMWR64:
        pcie_tlp_mem64_dec(tlp, is_wrp, addrp, lenp, valp);
        break;
    case PCIE_TLP_TYPE_MEMRD:
    case PCIE_TLP_TYPE_MEMWR:
        pcie_tlp_mem32_dec(tlp, is_wrp, addrp, lenp, valp);
        break;
    default:
        //assert(0);
        break;
    }
}

static inline void
pcie_tlp_io_enc(struct pcie_tlp *tlp, u_int8_t is_wr,
                u_int32_t addr, u_int16_t len, u_int32_t val)
{
    u_int8_t type = is_wr ? PCIE_TLP_TYPE_IOWR : PCIE_TLP_TYPE_IORD;

    //assert(len > 0 && len <= 4);
    memset(tlp, 0, sizeof (struct pcie_tlp));

    pcie_tlp_hdr_enc(tlp, type, addr, len);
    _addr32_enc(&tlp->data[8], addr & ~0x3); // DW aligned addr

    if (is_wr && len) {
        val = _val2bl(addr, len, val);
        _val32_enc(&tlp->data[16], val);
    }
}

static inline void
pcie_tlp_io_dec(struct pcie_tlp *tlp, u_int8_t *is_wrp,
                u_int32_t *addrp, u_int16_t *lenp, u_int32_t *valp)
{
    u_int8_t is_wr, type, addroff;
    u_int16_t len;

    pcie_tlp_hdr_dec(tlp, &type, &addroff, &len);

    is_wr = (type == PCIE_TLP_TYPE_IOWR);

    if (is_wrp) {
        *is_wrp = is_wr;
    }
    if (addrp) {
        u_int32_t addr;
        _addr32_dec(&tlp->data[8], &addr);
        *addrp = addr + addroff;
    }
    if (lenp) {
        *lenp = len;
    }
    if (valp) {
        if (is_wr) {
            u_int32_t val;
            _val32_dec(&tlp->data[16], &val);
            val >>= addroff * 8;
            *valp = val;
        } else {
            *valp = 0;
        }
    }
}

static inline void
pcie_tlp_cfg_enc(struct pcie_tlp *tlp, u_int8_t is_wr,
                 u_int8_t bus, u_int8_t dev, u_int8_t fnc,
                 u_int16_t addr, u_int16_t len, u_int32_t val)
{
    u_int8_t type = is_wr ? PCIE_TLP_TYPE_CFGWR0 : PCIE_TLP_TYPE_CFGRD0;

    //assert(len >= 1 && len <= 4);
    //assert((len & (len - 1)) == 0);     // cfg reg 1, 2, 4

    memset(tlp, 0, sizeof (struct pcie_tlp));

    pcie_tlp_hdr_enc(tlp, type, addr, len);

    tlp->data[8] = bus;
    tlp->data[9] = (dev << 3) | fnc;
    tlp->data[10] = (addr >> 8) & 0xf;
    tlp->data[11] = addr & ~0x3;        // DW aligned

    if (is_wr && len) {
        val = _val2bl(addr, len, val);
        _val32_enc(&tlp->data[16], val);
    }
}

static inline void
pcie_tlp_cfg_dec(struct pcie_tlp *tlp, u_int8_t *is_wrp,
                 u_int8_t *busp, u_int8_t *devp, u_int8_t *fncp,
                 u_int16_t *addrp, u_int16_t *lenp, u_int32_t *valp)
{
    u_int8_t is_wr, type, addroff;
    u_int16_t len;

    pcie_tlp_hdr_dec(tlp, &type, &addroff, &len);

    is_wr = (type == PCIE_TLP_TYPE_CFGWR0 || type == PCIE_TLP_TYPE_CFGWR1);

    if (is_wrp) {
        *is_wrp = is_wr;
    }
    if (busp) {
        *busp = tlp->data[8];
    }
    if (devp) {
        *devp = (tlp->data[9] >> 3) & 0x1f;
    }
    if (fncp) {
        *fncp = tlp->data[9] & 0x7;
    }
    if (addrp) {
        *addrp = (((((u_int16_t)tlp->data[10] & 0xf) << 8) | tlp->data[11]) +
                  addroff);
    }
    if (lenp) {
        *lenp = len;
    }
    if (valp) {
        if (is_wr) {
            u_int32_t val;
            _val32_dec(&tlp->data[16], &val);
            val >>= addroff * 8;
            // mask off unselected upper bytes of data
            if (len == 1) {
                val &= 0xff;
            } else if (len == 2) {
                val &= 0xffff;
            }
            *valp = val;
        } else {
            *valp = 0;
        }
    }
}

#endif /* _PCIE_TLP_H_ */
