/*
 * Copyright 2008-2010 Cisco Systems, Inc.  All rights reserved.
 */
#ident "$Id$"

#ifndef _PALO_PALOMEM_H_
#define _PALO_PALOMEM_H_

/*
 * Layout of physical memory used by various components
 * of the ecpu firmware and associated applications. The
 * 128MB and 256MB configurations are the same up through
 * ecpu1. Note that ecpu2 always loads at the same address.
 * (It's not loaded in the 128MB configuration.) hwio is the
 * only region who's address changes between 128 and 256MB.
 * It's last to allow it to use all of the remaining memory.
 * NOTE: 128MB is a lab (FPGA) version
 *
 *            256MB Configuration   128MB Configuration
 *
 * 0x10000000 +-----------------+
 *            |       ecc       |
 * 0x0e000000 +-----------------+
 *            |                 |
 *            |     flowtbl     |
 *            |                 |
 * 0x0a000000 +-----------------+
 *            |      genio      |
 * 0x09800000 +-----------------+
 *            |      pmpcie     |
 * 0x09600000 +-----------------+
 *            |                 |
 *            |                 |
 *            |      hwio       |
 *            |                 |   +-----------------+ 0x08000000
 * 0x07800000 +-----------------+   |       ecc       |
 *            |                 |   +-----------------+ 0x07000000
 *            |      ecpu2      |   |                 |
 *            |                 |   |      hwio       |
 * 0x05b00000 +-----------------+   +-----------------+ 0x05f00000
 *            |                 |
 *            |      ecpu1      |           ...
 *            |                 |
 * 0x03e00000 +-----------------+
 *            |      extbl      |           ...
 * 0x03c00000 +-----------------+
 *            |                 |
 *            |      Linux      |           ...
 *            |                 |
 * 0x00012000 +-----------------+
 *            |     memtun      |           ...
 * 0x00002000 +-----------------+
 *            |      diag       |           ...
 * 0x00001000 +-----------------+
 *            |      ectx       |           ...
 * 0x00000000 +-----------------+
 */

/*
 * Define the top of the memory map.
 * 128MB for Serenote, 256MB for Sereno.
 */
#define PALOMEM128_MEMTOP       0x08000000
#define PALOMEM256_MEMTOP       0x10000000

/*
 * Ecpu boot vectors and arguments.
 * See ecpuctx.h for details.
 */
#define PALOMEM_ECTX_NAME       "ectx"
#define PALOMEM_ECTX_BASE       0x00000000
#define PALOMEM_ECTX_SIZE       0x00001000

/*
 * Diags uses this page to pass data from the host diags app to diags fw.
 * Only used by the diags firmware image.
 */
#define PALOMEM_DIAG_NAME       "diag"
#define PALOMEM_DIAG_BASE       0x00001000
#define PALOMEM_DIAG_SIZE       0x00001000

/*
 * The memtun region contains the transmit/receive rings for
 * exchanging packets with the host through the memtun facility.
 * See sa/src/palo/fw/mcpu/memtun for details.
 */
#define PALOMEM_MEMTUN_NAME     "memtun"
#define PALOMEM_MEMTUN_BASE     0x00002000
#define PALOMEM_MEMTUN_SIZE     0x00010000

/*
 * Linux memory region for Linux kernel and user processes.
 * N.B. *** The Linux kernel load address load-$(CONFIG_BASE_SERENO)
 * from mcpu/firmware/linux/arch/mips/Makefile must be within this
 * region ***.
 */
#define PALOMEM_LINUX_NAME      "linux"
#define PALOMEM_LINUX_BASE      0x00012000
#define PALOMEM_LINUX_SIZE      0x03bee000

/*
 * Region allocated for ecpu code/data/exchange table.
 * We give 64 MB total, extbl gets the first 6 MB, then
 * 29 MB for each ecpu code/data. ecpu2 is not loaded in
 * 128MB configurations.
 * See palo/fw/build for details.
 */
#define PALOMEM_EXTBL_NAME      "extbl"
#define PALOMEM_EXTBL_BASE      0x03c00000
#define PALOMEM_EXTBL_SIZE      0x00200000
/*
 * Part of the exchange table is used for cache in Palo but not in Sereno.
 * Need to know the amount that is cache in the palo_dfw.lds file.
 */
#define PALOMEM_EXTBL_CA_SIZE   0x00000000

/*
 * Exchange table param/stats size is 1MB each in Sereno.
 */
#define PALOMEM_EXTBL_PM_SIZE   0x00100000
#define PALOMEM_EXTBL_SS_SIZE   0x00100000

#define PALOMEM_ECPU1_NAME      "ecpu1"
#define PALOMEM_ECPU1_BASE      0x03e00000
#define PALOMEM_ECPU1_SIZE      0x01d00000

#define PALOMEM_ECPU2_NAME      "ecpu2"
#define PALOMEM_ECPU2_BASE      0x05b00000
#define PALOMEM_ECPU2_SIZE      0x01d00000

#ifdef CONFIG_BASE_SERENO
#define PALOMEM_FLOW_TBL_NAME   "flowtbl"
#define PALOMEM_FLOW_TBL_BASE   0x0a000000
#define PALOMEM_FLOW_TBL_SIZE   0x04000000
#endif
/* ecpu2 is not mapped in 128MB configurations. */

/*
 * The hwio region is reserved for ecpu physical memory access.
 * See sa/src/common/libhwio/include/hwio_mipslinux_impl.h for details.
 * It loads in one of two different places depending on the available
 * memory size.
 */
#define PALOMEM_HWIO_NAME       "hwio"
#define PALOMEM_HWIO128_NAME    PALOMEM_HWIO_NAME
#define PALOMEM_HWIO128_BASE    0x05b00000
#define PALOMEM_HWIO128_SIZE    0x01500000

#define PALOMEM_HWIO256_NAME    PALOMEM_HWIO_NAME
#define PALOMEM_HWIO256_BASE    0x07800000
#define PALOMEM_HWIO256_SIZE    0x01e00000

#define PALOMEM_PMPCIE_NAME     "pmpcie"
#define PALOMEM_PMPCIE256_BASE  0x09600000
#define PALOMEM_PMPCIE_SIZE     0x00200000
/* 128M is only for lab (FPGA) verion, pciemgr assumes 256M one */
#define PALOMEM_PMPCIE_BASE     PALOMEM_PMPCIE256_BASE

#define PALOMEM_GENIO_NAME      "genio"
#define PALOMEM_GENIO_BASE      0x09800000
#define PALOMEM_GENIO_SIZE      0x04800000

#define PALOMEM_ECC_NAME        "ecc"
#define PALOMEM_ECC128_NAME     PALOMEM_ECC_NAME
#define PALOMEM_ECC128_BASE     0x07000000
#define PALOMEM_ECC128_SIZE     0x01000000
#define PALOMEM_ECC256_NAME     PALOMEM_ECC_NAME
#define PALOMEM_ECC256_BASE     0x0e000000
#define PALOMEM_ECC256_SIZE     0x02000000

#endif /* _PALO_PALOMEM_H_ */
